# Copyright (c) 2024-Present
# Author: Jiawei Zhang <jiawei@ifmlab.org>
# Affiliation: IFM Lab, UC Davis

import torch
from torch.utils.data import DataLoader

import warnings
import os.path
import numpy as np

from sklearn.model_selection import train_test_split, KFold

from tinybig.data.base_data import dataloader, dataset


# we preprocess data into a unified format
# headline row: physical meanings and data types of each column
# remaining rows: data instances (id, features, label)
class tabular_dataloader(dataloader):

    def __init__(self, name='tabular_dataloader', data_contents: list | tuple = None,
                 train_batch_size=64, test_batch_size=64, *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)
        self.train_batch_size = train_batch_size
        self.test_batch_size = test_batch_size
        self.data_contents = data_contents

    def load_file(self, cache_dir: str = None, filename: str = None):
        file_path = cache_dir + filename
        if not os.path.exists(file_path):
            raise ValueError('The provided file path doesn\'t exist...'.format(file_path))
        f = open(file_path, 'r')
        rows = f.readlines()
        f.close()
        return rows

    def load_raw_data(self, data_contents=None, cache_dir: str = None, filename: str = None):
        data_contents = data_contents if data_contents is not None else self.data_contents
        if data_contents is not None:
            return data_contents
        elif cache_dir is not None and filename is not None:
            return self.load_file(cache_dir=cache_dir, filename=filename)

    def process_data(self, data_contents=None, cache_dir: str = None, filename: str = None,
                     str_converted_to_numerical: bool = True, *args, **kwargs):

        rows = data_contents if data_contents is not None else self.load_file(cache_dir=cache_dir, filename=filename)

        data_dict = {
            'profile': {
                'name': self.name,
                'str_converted_to_numerical': str_converted_to_numerical,
                'column_names': [],  #[list of column names],
                'column_value_types': [],  #[list of column value types, e.g., float, int, str, etc.],
                'column_codings': [],  #[list of column index: {raw_values: coding_values}]
            },
            'contents': []  # [[vector of instance 1], [vector of instance 2], ... ]
        }


        headline = rows[0]
        column_names, column_value_types = [], []
        feature_name_value_type_list = headline.strip('\n').split(',')
        for feature_name_value_type in feature_name_value_type_list:
            name, value_type = feature_name_value_type.split('_')
            column_names.append(name)
            column_value_types.append(value_type)

        column_codings = [None]*len(column_names)
        contents = []
        nan_detect = False
        nan_columns = {}
        for row_index in range(1, len(rows)):
            row = rows[row_index]
            x = []
            features = row.strip('\n').split(',')
            for index in range(len(features)):
                feature = features[index]
                feature_value_type = column_value_types[index]
                if feature_value_type in ['int', 'integer', 'Int', 'Integer', 'INT', 'INTEGER']:
                    try:
                        feature = int(feature)
                    except:
                        nan_detect = True
                        nan_columns[index] = 1
                        feature = np.nan
                elif feature_value_type in ['float', 'Float', 'double', 'Double', 'FLOAT', 'DOUBLE']:
                    try:
                        feature = float(feature)
                    except:
                        nan_detect = True
                        nan_columns[index] = 1
                        feature = np.nan
                elif feature_value_type in ['str', 'string', 'Str', 'String', 'strings', 'Strings', 'STR', 'STRING', 'STRINGS']:
                    if str_converted_to_numerical:
                        if column_codings[index] is None:
                            column_codings[index] = {}
                        if feature not in column_codings[index]:
                            column_codings[index][feature] = len(column_codings[index])
                        feature = column_codings[index][feature]
                x.append(feature)
            contents.append(x)
        data_dict['profile']['feature_names'] = column_names
        data_dict['profile']['column_value_types'] = column_value_types
        data_dict['profile']['column_codings'] = column_codings
        data_dict['profile']['row_number'] = len(contents)
        data_dict['profile']['column_number'] = len(column_names)
        data_dict['contents'] = contents

        if nan_detect:
            warnings.warn('The loaded dataset may contain NaN elements, and the columns containing nan are listed as follows', UserWarning)
            print(nan_columns.keys())
        return data_dict

    def load_complete_data(self, data_contents: list | tuple = None, cache_dir: str = None,
                           filename: str = None, *args, **kwargs):
        data_contents = data_contents if data_contents is not None else self.data_contents
        raw_data = self.process_data(data_contents=data_contents, cache_dir=cache_dir, filename=filename,
                                     str_converted_to_numerical=True, *args, **kwargs)

        contents = np.array(raw_data['contents'])
        row_number = raw_data['profile']['row_number']
        column_number = raw_data['profile']['column_number']
        column_names = raw_data['profile']['feature_names']
        column_value_types = raw_data['profile']['column_value_types']
        ids = contents[:, 0]
        X = torch.Tensor(contents[:, 1:-1])
        y = torch.Tensor(contents[:, -1])

        return {
            'X': X,
            'y': y,
            'profile': {
                'name': raw_data['profile']['name'],
                'instance_number': row_number,
                'feature_number': column_number - 2,
                'feature_codings': raw_data['profile']['column_codings'][1:-1],
                'label_codings': raw_data['profile']['column_codings'][-1],
                'feature_names': column_names[1:-1],
                'label_names': column_names[-1],
                'feature_value_types': column_value_types[1:-1],
                'label_value_types': column_value_types[-1]
            }
        }

    @staticmethod
    def normalize(input: torch.Tensor, normalize_type=None, normalize_range: list | tuple=None):
        if normalize_type == 'min_max':
            if normalize_range is None:
                min_value, max_value = (0, 1)
            else:
                min_value, max_value = normalize_range
            X_std = (input - input.numpy().min(axis=0)) / (input.numpy().max(axis=0) - input.numpy().min(axis=0))
            X_scaled = X_std * (max_value - min_value) + min_value
        elif normalize_type == 'mean_std':
            X_mean = torch.mean(input, dim=0, keepdim=True)
            X_std = torch.std(input, dim=0, keepdim=True)
            X_scaled = (input - X_mean) / X_std
        else:
            X_scaled = input
        return X_scaled

    def split(self, X, y, split_type='train_test_split', train_percentage=0.9, fold=10, random_state=1234, shuffle=True):
        train_loader, test_loader = None, None
        if split_type == 'train_test_split':
            X_train, X_test, y_train, y_test = train_test_split(
                X, y,
                train_size=int(train_percentage * len(X)),
                random_state=random_state, shuffle=shuffle
            )
            train_dataset = dataset(X_train, torch.unsqueeze(y_train, 1))
            test_dataset = dataset(X_test, torch.unsqueeze(y_test, 1))
            train_loader = DataLoader(train_dataset, batch_size=self.train_batch_size, shuffle=True)
            test_loader = DataLoader(test_dataset, batch_size=self.test_batch_size, shuffle=False)
        elif split_type in ['KFold', 'cross_validation']:
            kf = KFold(n_splits=fold, random_state=random_state, shuffle=shuffle)
            train_loader, test_loader = {}, {}
            for i, (train_index, test_index) in enumerate(kf.split(X)):
                X_train, y_train = X[train_index], y[train_index]
                X_test, y_test = X[test_index], y[test_index]
                train_dataset = dataset(X_train, torch.unsqueeze(y_train, 1))
                test_dataset = dataset(X_test, torch.unsqueeze(y_test, 1))
                train_loader[i] = DataLoader(train_dataset, batch_size=self.train_batch_size, shuffle=True)
                test_loader[i] = DataLoader(test_dataset, batch_size=self.test_batch_size, shuffle=False)
        return train_loader, test_loader

    def load(self, cache_dir: str = './data/', filename: str = 'classic_dataset_filename',
             split_type: str = 'train_test_split', train_percentage: float = 0.9, fold: int = 10,
             random_state: int = 123, shuffle: bool = True,
             normalize_X: bool = False, normalize_y: bool = False,
             normalize_type: str = 'min_max', normalize_range: list | tuple = (0, 1), *args, **kwargs):

        complete_data = self.load_complete_data(data_contents=self.data_contents, cache_dir=cache_dir, filename=filename)
        X, y = complete_data['X'], complete_data['y']

        if normalize_X:
            X = self.normalize(input=X, normalize_type=normalize_type, normalize_range=normalize_range)
        if normalize_y:
            y = self.normalize(input=y, normalize_type=normalize_type, normalize_range=normalize_range)

        train_loader, test_loader = self.split(X=X, y=y, split_type=split_type, train_percentage=train_percentage,
                                               fold=fold, random_state=random_state, shuffle=shuffle)

        return {
            'train_loader': train_loader,
            'test_loader': test_loader,
            'profile': complete_data['profile']
        }

Iris_Dataset = (
    "Id_int,SepalLengthCm_float,SepalWidthCm_float,PetalLengthCm_float,PetalWidthCm_float,Species_str",
    "1,5.1,3.5,1.4,0.2,Iris-setosa",
    "2,4.9,3,1.4,0.2,Iris-setosa",
    "3,4.7,3.2,1.3,0.2,Iris-setosa",
    "4,4.6,3.1,1.5,0.2,Iris-setosa",
    "5,5,3.6,1.4,0.2,Iris-setosa",
    "6,5.4,3.9,1.7,0.4,Iris-setosa",
    "7,4.6,3.4,1.4,0.3,Iris-setosa",
    "8,5,3.4,1.5,0.2,Iris-setosa",
    "9,4.4,2.9,1.4,0.2,Iris-setosa",
    "10,4.9,3.1,1.5,0.1,Iris-setosa",
    "11,5.4,3.7,1.5,0.2,Iris-setosa",
    "12,4.8,3.4,1.6,0.2,Iris-setosa",
    "13,4.8,3,1.4,0.1,Iris-setosa",
    "14,4.3,3,1.1,0.1,Iris-setosa",
    "15,5.8,4,1.2,0.2,Iris-setosa",
    "16,5.7,4.4,1.5,0.4,Iris-setosa",
    "17,5.4,3.9,1.3,0.4,Iris-setosa",
    "18,5.1,3.5,1.4,0.3,Iris-setosa",
    "19,5.7,3.8,1.7,0.3,Iris-setosa",
    "20,5.1,3.8,1.5,0.3,Iris-setosa",
    "21,5.4,3.4,1.7,0.2,Iris-setosa",
    "22,5.1,3.7,1.5,0.4,Iris-setosa",
    "23,4.6,3.6,1,0.2,Iris-setosa",
    "24,5.1,3.3,1.7,0.5,Iris-setosa",
    "25,4.8,3.4,1.9,0.2,Iris-setosa",
    "26,5,3,1.6,0.2,Iris-setosa",
    "27,5,3.4,1.6,0.4,Iris-setosa",
    "28,5.2,3.5,1.5,0.2,Iris-setosa",
    "29,5.2,3.4,1.4,0.2,Iris-setosa",
    "30,4.7,3.2,1.6,0.2,Iris-setosa",
    "31,4.8,3.1,1.6,0.2,Iris-setosa",
    "32,5.4,3.4,1.5,0.4,Iris-setosa",
    "33,5.2,4.1,1.5,0.1,Iris-setosa",
    "34,5.5,4.2,1.4,0.2,Iris-setosa",
    "35,4.9,3.1,1.5,0.1,Iris-setosa",
    "36,5,3.2,1.2,0.2,Iris-setosa",
    "37,5.5,3.5,1.3,0.2,Iris-setosa",
    "38,4.9,3.1,1.5,0.1,Iris-setosa",
    "39,4.4,3,1.3,0.2,Iris-setosa",
    "40,5.1,3.4,1.5,0.2,Iris-setosa",
    "41,5,3.5,1.3,0.3,Iris-setosa",
    "42,4.5,2.3,1.3,0.3,Iris-setosa",
    "43,4.4,3.2,1.3,0.2,Iris-setosa",
    "44,5,3.5,1.6,0.6,Iris-setosa",
    "45,5.1,3.8,1.9,0.4,Iris-setosa",
    "46,4.8,3,1.4,0.3,Iris-setosa",
    "47,5.1,3.8,1.6,0.2,Iris-setosa",
    "48,4.6,3.2,1.4,0.2,Iris-setosa",
    "49,5.3,3.7,1.5,0.2,Iris-setosa",
    "50,5,3.3,1.4,0.2,Iris-setosa",
    "51,7,3.2,4.7,1.4,Iris-versicolor",
    "52,6.4,3.2,4.5,1.5,Iris-versicolor",
    "53,6.9,3.1,4.9,1.5,Iris-versicolor",
    "54,5.5,2.3,4,1.3,Iris-versicolor",
    "55,6.5,2.8,4.6,1.5,Iris-versicolor",
    "56,5.7,2.8,4.5,1.3,Iris-versicolor",
    "57,6.3,3.3,4.7,1.6,Iris-versicolor",
    "58,4.9,2.4,3.3,1,Iris-versicolor",
    "59,6.6,2.9,4.6,1.3,Iris-versicolor",
    "60,5.2,2.7,3.9,1.4,Iris-versicolor",
    "61,5,2,3.5,1,Iris-versicolor",
    "62,5.9,3,4.2,1.5,Iris-versicolor",
    "63,6,2.2,4,1,Iris-versicolor",
    "64,6.1,2.9,4.7,1.4,Iris-versicolor",
    "65,5.6,2.9,3.6,1.3,Iris-versicolor",
    "66,6.7,3.1,4.4,1.4,Iris-versicolor",
    "67,5.6,3,4.5,1.5,Iris-versicolor",
    "68,5.8,2.7,4.1,1,Iris-versicolor",
    "69,6.2,2.2,4.5,1.5,Iris-versicolor",
    "70,5.6,2.5,3.9,1.1,Iris-versicolor",
    "71,5.9,3.2,4.8,1.8,Iris-versicolor",
    "72,6.1,2.8,4,1.3,Iris-versicolor",
    "73,6.3,2.5,4.9,1.5,Iris-versicolor",
    "74,6.1,2.8,4.7,1.2,Iris-versicolor",
    "75,6.4,2.9,4.3,1.3,Iris-versicolor",
    "76,6.6,3,4.4,1.4,Iris-versicolor",
    "77,6.8,2.8,4.8,1.4,Iris-versicolor",
    "78,6.7,3,5,1.7,Iris-versicolor",
    "79,6,2.9,4.5,1.5,Iris-versicolor",
    "80,5.7,2.6,3.5,1,Iris-versicolor",
    "81,5.5,2.4,3.8,1.1,Iris-versicolor",
    "82,5.5,2.4,3.7,1,Iris-versicolor",
    "83,5.8,2.7,3.9,1.2,Iris-versicolor",
    "84,6,2.7,5.1,1.6,Iris-versicolor",
    "85,5.4,3,4.5,1.5,Iris-versicolor",
    "86,6,3.4,4.5,1.6,Iris-versicolor",
    "87,6.7,3.1,4.7,1.5,Iris-versicolor",
    "88,6.3,2.3,4.4,1.3,Iris-versicolor",
    "89,5.6,3,4.1,1.3,Iris-versicolor",
    "90,5.5,2.5,4,1.3,Iris-versicolor",
    "91,5.5,2.6,4.4,1.2,Iris-versicolor",
    "92,6.1,3,4.6,1.4,Iris-versicolor",
    "93,5.8,2.6,4,1.2,Iris-versicolor",
    "94,5,2.3,3.3,1,Iris-versicolor",
    "95,5.6,2.7,4.2,1.3,Iris-versicolor",
    "96,5.7,3,4.2,1.2,Iris-versicolor",
    "97,5.7,2.9,4.2,1.3,Iris-versicolor",
    "98,6.2,2.9,4.3,1.3,Iris-versicolor",
    "99,5.1,2.5,3,1.1,Iris-versicolor",
    "100,5.7,2.8,4.1,1.3,Iris-versicolor",
    "101,6.3,3.3,6,2.5,Iris-virginica",
    "102,5.8,2.7,5.1,1.9,Iris-virginica",
    "103,7.1,3,5.9,2.1,Iris-virginica",
    "104,6.3,2.9,5.6,1.8,Iris-virginica",
    "105,6.5,3,5.8,2.2,Iris-virginica",
    "106,7.6,3,6.6,2.1,Iris-virginica",
    "107,4.9,2.5,4.5,1.7,Iris-virginica",
    "108,7.3,2.9,6.3,1.8,Iris-virginica",
    "109,6.7,2.5,5.8,1.8,Iris-virginica",
    "110,7.2,3.6,6.1,2.5,Iris-virginica",
    "111,6.5,3.2,5.1,2,Iris-virginica",
    "112,6.4,2.7,5.3,1.9,Iris-virginica",
    "113,6.8,3,5.5,2.1,Iris-virginica",
    "114,5.7,2.5,5,2,Iris-virginica",
    "115,5.8,2.8,5.1,2.4,Iris-virginica",
    "116,6.4,3.2,5.3,2.3,Iris-virginica",
    "117,6.5,3,5.5,1.8,Iris-virginica",
    "118,7.7,3.8,6.7,2.2,Iris-virginica",
    "119,7.7,2.6,6.9,2.3,Iris-virginica",
    "120,6,2.2,5,1.5,Iris-virginica",
    "121,6.9,3.2,5.7,2.3,Iris-virginica",
    "122,5.6,2.8,4.9,2,Iris-virginica",
    "123,7.7,2.8,6.7,2,Iris-virginica",
    "124,6.3,2.7,4.9,1.8,Iris-virginica",
    "125,6.7,3.3,5.7,2.1,Iris-virginica",
    "126,7.2,3.2,6,1.8,Iris-virginica",
    "127,6.2,2.8,4.8,1.8,Iris-virginica",
    "128,6.1,3,4.9,1.8,Iris-virginica",
    "129,6.4,2.8,5.6,2.1,Iris-virginica",
    "130,7.2,3,5.8,1.6,Iris-virginica",
    "131,7.4,2.8,6.1,1.9,Iris-virginica",
    "132,7.9,3.8,6.4,2,Iris-virginica",
    "133,6.4,2.8,5.6,2.2,Iris-virginica",
    "134,6.3,2.8,5.1,1.5,Iris-virginica",
    "135,6.1,2.6,5.6,1.4,Iris-virginica",
    "136,7.7,3,6.1,2.3,Iris-virginica",
    "137,6.3,3.4,5.6,2.4,Iris-virginica",
    "138,6.4,3.1,5.5,1.8,Iris-virginica",
    "139,6,3,4.8,1.8,Iris-virginica",
    "140,6.9,3.1,5.4,2.1,Iris-virginica",
    "141,6.7,3.1,5.6,2.4,Iris-virginica",
    "142,6.9,3.1,5.1,2.3,Iris-virginica",
    "143,5.8,2.7,5.1,1.9,Iris-virginica",
    "144,6.8,3.2,5.9,2.3,Iris-virginica",
    "145,6.7,3.3,5.7,2.5,Iris-virginica",
    "146,6.7,3,5.2,2.3,Iris-virginica",
    "147,6.3,2.5,5,1.9,Iris-virginica",
    "148,6.5,3,5.2,2,Iris-virginica",
    "149,6.2,3.4,5.4,2.3,Iris-virginica",
    "150,5.9,3,5.1,1.8,Iris-virginica",
)

Titanic_Dataset = (
    "id_int,pclass_int,name_str,sex_str,age_int,sibsp_int,parch_int,ticket_str,fare_float,cabin_str,embarked_str,boat_str,body_int,home.dest_str,survived_int",
    "1,1,Allen. Miss. Elisabeth Walton,female,29,0,0,24160,211.3375,B5,S,2,?,St Louis. MO,1",
    "2,1,Allison. Master. Hudson Trevor,male,0.9167,1,2,113781,151.55,C22 C26,S,11,?,Montreal. PQ / Chesterville. ON,1",
    "3,1,Allison. Miss. Helen Loraine,female,2,1,2,113781,151.55,C22 C26,S,?,?,Montreal. PQ / Chesterville. ON,0",
    "4,1,Allison. Mr. Hudson Joshua Creighton,male,30,1,2,113781,151.55,C22 C26,S,?,135,Montreal. PQ / Chesterville. ON,0",
    "5,1,Allison. Mrs. Hudson J C (Bessie Waldo Daniels),female,25,1,2,113781,151.55,C22 C26,S,?,?,Montreal. PQ / Chesterville. ON,0",
    "6,1,Anderson. Mr. Harry,male,48,0,0,19952,26.55,E12,S,3,?,New York. NY,1",
    "7,1,Andrews. Miss. Kornelia Theodosia,female,63,1,0,13502,77.9583,D7,S,10,?,Hudson. NY,1",
    "8,1,Andrews. Mr. Thomas Jr,male,39,0,0,112050,0,A36,S,?,?,Belfast. NI,0",
    "9,1,Appleton. Mrs. Edward Dale (Charlotte Lamson),female,53,2,0,11769,51.4792,C101,S,D,?,Bayside. Queens. NY,1",
    "10,1,Artagaveytia. Mr. Ramon,male,71,0,0,PC 17609,49.5042,?,C,?,22,Montevideo. Uruguay,0",
    "11,1,Astor. Col. John Jacob,male,47,1,0,PC 17757,227.525,C62 C64,C,?,124,New York. NY,0",
    "12,1,Astor. Mrs. John Jacob (Madeleine Talmadge Force),female,18,1,0,PC 17757,227.525,C62 C64,C,4,?,New York. NY,1",
    "13,1,Aubart. Mme. Leontine Pauline,female,24,0,0,PC 17477,69.3,B35,C,9,?,Paris. France,1",
    "14,1,Barber. Miss. Ellen 'Nellie',female,26,0,0,19877,78.85,?,S,6,?,?,1",
    "15,1,Barkworth. Mr. Algernon Henry Wilson,male,80,0,0,27042,30,A23,S,B,?,Hessle. Yorks,1",
    "16,1,Baumann. Mr. John D,male,?,0,0,PC 17318,25.925,?,S,?,?,New York. NY,0",
    "17,1,Baxter. Mr. Quigg Edmond,male,24,0,1,PC 17558,247.5208,B58 B60,C,?,?,Montreal. PQ,0",
    "18,1,Baxter. Mrs. James (Helene DeLaudeniere Chaput),female,50,0,1,PC 17558,247.5208,B58 B60,C,6,?,Montreal. PQ,1",
    "19,1,Bazzani. Miss. Albina,female,32,0,0,11813,76.2917,D15,C,8,?,?,1",
    "20,1,Beattie. Mr. Thomson,male,36,0,0,13050,75.2417,C6,C,A,?,Winnipeg. MN,0",
    "21,1,Beckwith. Mr. Richard Leonard,male,37,1,1,11751,52.5542,D35,S,5,?,New York. NY,1",
    "22,1,Beckwith. Mrs. Richard Leonard (Sallie Monypeny),female,47,1,1,11751,52.5542,D35,S,5,?,New York. NY,1",
    "23,1,Behr. Mr. Karl Howell,male,26,0,0,111369,30,C148,C,5,?,New York. NY,1",
    "24,1,Bidois. Miss. Rosalie,female,42,0,0,PC 17757,227.525,?,C,4,?,?,1",
    "25,1,Bird. Miss. Ellen,female,29,0,0,PC 17483,221.7792,C97,S,8,?,?,1",
    "26,1,Birnbaum. Mr. Jakob,male,25,0,0,13905,26,?,C,?,148,San Francisco. CA,0",
    "27,1,Bishop. Mr. Dickinson H,male,25,1,0,11967,91.0792,B49,C,7,?,Dowagiac. MI,1",
    "28,1,Bishop. Mrs. Dickinson H (Helen Walton),female,19,1,0,11967,91.0792,B49,C,7,?,Dowagiac. MI,1",
    "29,1,Bissette. Miss. Amelia,female,35,0,0,PC 17760,135.6333,C99,S,8,?,?,1",
    "30,1,Bjornstrom-Steffansson. Mr. Mauritz Hakan,male,28,0,0,110564,26.55,C52,S,D,?,Stockholm. Sweden / Washington. DC,1",
    "31,1,Blackwell. Mr. Stephen Weart,male,45,0,0,113784,35.5,T,S,?,?,Trenton. NJ,0",
    "32,1,Blank. Mr. Henry,male,40,0,0,112277,31,A31,C,7,?,Glen Ridge. NJ,1",
    "33,1,Bonnell. Miss. Caroline,female,30,0,0,36928,164.8667,C7,S,8,?,Youngstown. OH,1",
    "34,1,Bonnell. Miss. Elizabeth,female,58,0,0,113783,26.55,C103,S,8,?,Birkdale. England Cleveland. Ohio,1",
    "35,1,Borebank. Mr. John James,male,42,0,0,110489,26.55,D22,S,?,?,London / Winnipeg. MB,0",
    "36,1,Bowen. Miss. Grace Scott,female,45,0,0,PC 17608,262.375,?,C,4,?,Cooperstown. NY,1",
    "37,1,Bowerman. Miss. Elsie Edith,female,22,0,1,113505,55,E33,S,6,?,St Leonards-on-Sea. England Ohio,1",
    "38,1,Bradley. Mr. George ('George Arthur Brayton'),male,?,0,0,111427,26.55,?,S,9,?,Los Angeles. CA,1",
    "39,1,Brady. Mr. John Bertram,male,41,0,0,113054,30.5,A21,S,?,?,Pomeroy. WA,0",
    "40,1,Brandeis. Mr. Emil,male,48,0,0,PC 17591,50.4958,B10,C,?,208,Omaha. NE,0",
    "41,1,Brewe. Dr. Arthur Jackson,male,?,0,0,112379,39.6,?,C,?,?,Philadelphia. PA,0",
    "42,1,Brown. Mrs. James Joseph (Margaret Tobin),female,44,0,0,PC 17610,27.7208,B4,C,6,?,Denver. CO,1",
    "43,1,Brown. Mrs. John Murray (Caroline Lane Lamson),female,59,2,0,11769,51.4792,C101,S,D,?,Belmont. MA,1",
    "44,1,Bucknell. Mrs. William Robert (Emma Eliza Ward),female,60,0,0,11813,76.2917,D15,C,8,?,Philadelphia. PA,1",
    "45,1,Burns. Miss. Elizabeth Margaret,female,41,0,0,16966,134.5,E40,C,3,?,?,1",
    "46,1,Butt. Major. Archibald Willingham,male,45,0,0,113050,26.55,B38,S,?,?,Washington. DC,0",
    "47,1,Cairns. Mr. Alexander,male,?,0,0,113798,31,?,S,?,?,?,0",
    "48,1,Calderhead. Mr. Edward Pennington,male,42,0,0,PC 17476,26.2875,E24,S,5,?,New York. NY,1",
    "49,1,Candee. Mrs. Edward (Helen Churchill Hungerford),female,53,0,0,PC 17606,27.4458,?,C,6,?,Washington. DC,1",
    "50,1,Cardeza. Mr. Thomas Drake Martinez,male,36,0,1,PC 17755,512.3292,B51 B53 B55,C,3,?,Austria-Hungary / Germantown. Philadelphia. PA,1",
    "51,1,Cardeza. Mrs. James Warburton Martinez (Charlotte Wardle Drake),female,58,0,1,PC 17755,512.3292,B51 B53 B55,C,3,?,Germantown. Philadelphia. PA,1",
    "52,1,Carlsson. Mr. Frans Olof,male,33,0,0,695,5,B51 B53 B55,S,?,?,New York. NY,0",
    "53,1,Carrau. Mr. Francisco M,male,28,0,0,113059,47.1,?,S,?,?,Montevideo. Uruguay,0",
    "54,1,Carrau. Mr. Jose Pedro,male,17,0,0,113059,47.1,?,S,?,?,Montevideo. Uruguay,0",
    "55,1,Carter. Master. William Thornton II,male,11,1,2,113760,120,B96 B98,S,4,?,Bryn Mawr. PA,1",
    "56,1,Carter. Miss. Lucile Polk,female,14,1,2,113760,120,B96 B98,S,4,?,Bryn Mawr. PA,1",
    "57,1,Carter. Mr. William Ernest,male,36,1,2,113760,120,B96 B98,S,C,?,Bryn Mawr. PA,1",
    "58,1,Carter. Mrs. William Ernest (Lucile Polk),female,36,1,2,113760,120,B96 B98,S,4,?,Bryn Mawr. PA,1",
    "59,1,Case. Mr. Howard Brown,male,49,0,0,19924,26,?,S,?,?,Ascot. Berkshire / Rochester. NY,0",
    "60,1,Cassebeer. Mrs. Henry Arthur Jr (Eleanor Genevieve Fosdick),female,?,0,0,17770,27.7208,?,C,5,?,New York. NY,1",
    "61,1,Cavendish. Mr. Tyrell William,male,36,1,0,19877,78.85,C46,S,?,172,Little Onn Hall. Staffs,0",
    "62,1,Cavendish. Mrs. Tyrell William (Julia Florence Siegel),female,76,1,0,19877,78.85,C46,S,6,?,Little Onn Hall. Staffs,1",
    "63,1,Chaffee. Mr. Herbert Fuller,male,46,1,0,W.E.P. 5734,61.175,E31,S,?,?,Amenia. ND,0",
    "64,1,Chaffee. Mrs. Herbert Fuller (Carrie Constance Toogood),female,47,1,0,W.E.P. 5734,61.175,E31,S,4,?,Amenia. ND,1",
    "65,1,Chambers. Mr. Norman Campbell,male,27,1,0,113806,53.1,E8,S,5,?,New York. NY / Ithaca. NY,1",
    "66,1,Chambers. Mrs. Norman Campbell (Bertha Griggs),female,33,1,0,113806,53.1,E8,S,5,?,New York. NY / Ithaca. NY,1",
    "67,1,Chaudanson. Miss. Victorine,female,36,0,0,PC 17608,262.375,B61,C,4,?,?,1",
    "68,1,Cherry. Miss. Gladys,female,30,0,0,110152,86.5,B77,S,8,?,London. England,1",
    "69,1,Chevre. Mr. Paul Romaine,male,45,0,0,PC 17594,29.7,A9,C,7,?,Paris. France,1",
    "70,1,Chibnall. Mrs. (Edith Martha Bowerman),female,?,0,1,113505,55,E33,S,6,?,St Leonards-on-Sea. England Ohio,1",
    "71,1,Chisholm. Mr. Roderick Robert Crispin,male,?,0,0,112051,0,?,S,?,?,Liverpool. England / Belfast,0",
    "72,1,Clark. Mr. Walter Miller,male,27,1,0,13508,136.7792,C89,C,?,?,Los Angeles. CA,0",
    "73,1,Clark. Mrs. Walter Miller (Virginia McDowell),female,26,1,0,13508,136.7792,C89,C,4,?,Los Angeles. CA,1",
    "74,1,Cleaver. Miss. Alice,female,22,0,0,113781,151.55,?,S,11,?,?,1",
    "75,1,Clifford. Mr. George Quincy,male,?,0,0,110465,52,A14,S,?,?,Stoughton. MA,0",
    "76,1,Colley. Mr. Edward Pomeroy,male,47,0,0,5727,25.5875,E58,S,?,?,Victoria. BC,0",
    "77,1,Compton. Miss. Sara Rebecca,female,39,1,1,PC 17756,83.1583,E49,C,14,?,Lakewood. NJ,1",
    "78,1,Compton. Mr. Alexander Taylor Jr,male,37,1,1,PC 17756,83.1583,E52,C,?,?,Lakewood. NJ,0",
    "79,1,Compton. Mrs. Alexander Taylor (Mary Eliza Ingersoll),female,64,0,2,PC 17756,83.1583,E45,C,14,?,Lakewood. NJ,1",
    "80,1,Cornell. Mrs. Robert Clifford (Malvina Helen Lamson),female,55,2,0,11770,25.7,C101,S,2,?,New York. NY,1",
    "81,1,Crafton. Mr. John Bertram,male,?,0,0,113791,26.55,?,S,?,?,Roachdale. IN,0",
    "82,1,Crosby. Capt. Edward Gifford,male,70,1,1,WE/P 5735,71,B22,S,?,269,Milwaukee. WI,0",
    "83,1,Crosby. Miss. Harriet R,female,36,0,2,WE/P 5735,71,B22,S,7,?,Milwaukee. WI,1",
    "84,1,Crosby. Mrs. Edward Gifford (Catherine Elizabeth Halstead),female,64,1,1,112901,26.55,B26,S,7,?,Milwaukee. WI,1",
    "85,1,Cumings. Mr. John Bradley,male,39,1,0,PC 17599,71.2833,C85,C,?,?,New York. NY,0",
    "86,1,Cumings. Mrs. John Bradley (Florence Briggs Thayer),female,38,1,0,PC 17599,71.2833,C85,C,4,?,New York. NY,1",
    "87,1,Daly. Mr. Peter Denis,male,51,0,0,113055,26.55,E17,S,5 9,?,Lima. Peru,1",
    "88,1,Daniel. Mr. Robert Williams,male,27,0,0,113804,30.5,?,S,3,?,Philadelphia. PA,1",
    "89,1,Daniels. Miss. Sarah,female,33,0,0,113781,151.55,?,S,8,?,?,1",
    "90,1,Davidson. Mr. Thornton,male,31,1,0,F.C. 12750,52,B71,S,?,?,Montreal. PQ,0",
    "91,1,Davidson. Mrs. Thornton (Orian Hays),female,27,1,2,F.C. 12750,52,B71,S,3,?,Montreal. PQ,1",
    "92,1,Dick. Mr. Albert Adrian,male,31,1,0,17474,57,B20,S,3,?,Calgary. AB,1",
    "93,1,Dick. Mrs. Albert Adrian (Vera Gillespie),female,17,1,0,17474,57,B20,S,3,?,Calgary. AB,1",
    "94,1,Dodge. Dr. Washington,male,53,1,1,33638,81.8583,A34,S,13,?,San Francisco. CA,1",
    "95,1,Dodge. Master. Washington,male,4,0,2,33638,81.8583,A34,S,5,?,San Francisco. CA,1",
    "96,1,Dodge. Mrs. Washington (Ruth Vidaver),female,54,1,1,33638,81.8583,A34,S,5,?,San Francisco. CA,1",
    "97,1,Douglas. Mr. Walter Donald,male,50,1,0,PC 17761,106.425,C86,C,?,62,Deephaven. MN / Cedar Rapids. IA,0",
    "98,1,Douglas. Mrs. Frederick Charles (Mary Helene Baxter),female,27,1,1,PC 17558,247.5208,B58 B60,C,6,?,Montreal. PQ,1",
    "99,1,Douglas. Mrs. Walter Donald (Mahala Dutton),female,48,1,0,PC 17761,106.425,C86,C,2,?,Deephaven. MN / Cedar Rapids. IA,1",
    "100,1,Duff Gordon. Lady. (Lucille Christiana Sutherland) ('Mrs Morgan'),female,48,1,0,11755,39.6,A16,C,1,?,London / Paris,1",
    "101,1,Duff Gordon. Sir. Cosmo Edmund ('Mr Morgan'),male,49,1,0,PC 17485,56.9292,A20,C,1,?,London / Paris,1",
    "102,1,Dulles. Mr. William Crothers,male,39,0,0,PC 17580,29.7,A18,C,?,133,Philadelphia. PA,0",
    "103,1,Earnshaw. Mrs. Boulton (Olive Potter),female,23,0,1,11767,83.1583,C54,C,7,?,Mt Airy. Philadelphia. PA,1",
    "104,1,Endres. Miss. Caroline Louise,female,38,0,0,PC 17757,227.525,C45,C,4,?,New York. NY,1",
    "105,1,Eustis. Miss. Elizabeth Mussey,female,54,1,0,36947,78.2667,D20,C,4,?,Brookline. MA,1",
    "106,1,Evans. Miss. Edith Corse,female,36,0,0,PC 17531,31.6792,A29,C,?,?,New York. NY,0",
    "107,1,Farthing. Mr. John,male,?,0,0,PC 17483,221.7792,C95,S,?,?,?,0",
    "108,1,Flegenheim. Mrs. Alfred (Antoinette),female,?,0,0,PC 17598,31.6833,?,S,7,?,New York. NY,1",
    "109,1,Fleming. Miss. Margaret,female,?,0,0,17421,110.8833,?,C,4,?,?,1",
    "110,1,Flynn. Mr. John Irwin ('Irving'),male,36,0,0,PC 17474,26.3875,E25,S,5,?,Brooklyn. NY,1",
    "111,1,Foreman. Mr. Benjamin Laventall,male,30,0,0,113051,27.75,C111,C,?,?,New York. NY,0",
    "112,1,Fortune. Miss. Alice Elizabeth,female,24,3,2,19950,263,C23 C25 C27,S,10,?,Winnipeg. MB,1",
    "113,1,Fortune. Miss. Ethel Flora,female,28,3,2,19950,263,C23 C25 C27,S,10,?,Winnipeg. MB,1",
    "114,1,Fortune. Miss. Mabel Helen,female,23,3,2,19950,263,C23 C25 C27,S,10,?,Winnipeg. MB,1",
    "115,1,Fortune. Mr. Charles Alexander,male,19,3,2,19950,263,C23 C25 C27,S,?,?,Winnipeg. MB,0",
    "116,1,Fortune. Mr. Mark,male,64,1,4,19950,263,C23 C25 C27,S,?,?,Winnipeg. MB,0",
    "117,1,Fortune. Mrs. Mark (Mary McDougald),female,60,1,4,19950,263,C23 C25 C27,S,10,?,Winnipeg. MB,1",
    "118,1,Francatelli. Miss. Laura Mabel,female,30,0,0,PC 17485,56.9292,E36,C,1,?,?,1",
    "119,1,Franklin. Mr. Thomas Parham,male,?,0,0,113778,26.55,D34,S,?,?,Westcliff-on-Sea. Essex,0",
    "120,1,Frauenthal. Dr. Henry William,male,50,2,0,PC 17611,133.65,?,S,5,?,New York. NY,1",
    "121,1,Frauenthal. Mr. Isaac Gerald,male,43,1,0,17765,27.7208,D40,C,5,?,New York. NY,1",
    "122,1,Frauenthal. Mrs. Henry William (Clara Heinsheimer),female,?,1,0,PC 17611,133.65,?,S,5,?,New York. NY,1",
    "123,1,Frolicher. Miss. Hedwig Margaritha,female,22,0,2,13568,49.5,B39,C,5,?,Zurich. Switzerland,1",
    "124,1,Frolicher-Stehli. Mr. Maxmillian,male,60,1,1,13567,79.2,B41,C,5,?,Zurich. Switzerland,1",
    "125,1,Frolicher-Stehli. Mrs. Maxmillian (Margaretha Emerentia Stehli),female,48,1,1,13567,79.2,B41,C,5,?,Zurich. Switzerland,1",
    "126,1,Fry. Mr. Richard,male,?,0,0,112058,0,B102,S,?,?,?,0",
    "127,1,Futrelle. Mr. Jacques Heath,male,37,1,0,113803,53.1,C123,S,?,?,Scituate. MA,0",
    "128,1,Futrelle. Mrs. Jacques Heath (Lily May Peel),female,35,1,0,113803,53.1,C123,S,D,?,Scituate. MA,1",
    "129,1,Gee. Mr. Arthur H,male,47,0,0,111320,38.5,E63,S,?,275,St Anne's-on-Sea. Lancashire,0",
    "130,1,Geiger. Miss. Amalie,female,35,0,0,113503,211.5,C130,C,4,?,?,1",
    "131,1,Gibson. Miss. Dorothy Winifred,female,22,0,1,112378,59.4,?,C,7,?,New York. NY,1",
    "132,1,Gibson. Mrs. Leonard (Pauline C Boeson),female,45,0,1,112378,59.4,?,C,7,?,New York. NY,1",
    "133,1,Giglio. Mr. Victor,male,24,0,0,PC 17593,79.2,B86,C,?,?,?,0",
    "134,1,Goldenberg. Mr. Samuel L,male,49,1,0,17453,89.1042,C92,C,5,?,Paris. France / New York. NY,1",
    "135,1,Goldenberg. Mrs. Samuel L (Edwiga Grabowska),female,?,1,0,17453,89.1042,C92,C,5,?,Paris. France / New York. NY,1",
    "136,1,Goldschmidt. Mr. George B,male,71,0,0,PC 17754,34.6542,A5,C,?,?,New York. NY,0",
    "137,1,Gracie. Col. Archibald IV,male,53,0,0,113780,28.5,C51,C,B,?,Washington. DC,1",
    "138,1,Graham. Miss. Margaret Edith,female,19,0,0,112053,30,B42,S,3,?,Greenwich. CT,1",
    "139,1,Graham. Mr. George Edward,male,38,0,1,PC 17582,153.4625,C91,S,?,147,Winnipeg. MB,0",
    "140,1,Graham. Mrs. William Thompson (Edith Junkins),female,58,0,1,PC 17582,153.4625,C125,S,3,?,Greenwich. CT,1",
    "141,1,Greenfield. Mr. William Bertram,male,23,0,1,PC 17759,63.3583,D10 D12,C,7,?,New York. NY,1",
    "142,1,Greenfield. Mrs. Leo David (Blanche Strouse),female,45,0,1,PC 17759,63.3583,D10 D12,C,7,?,New York. NY,1",
    "143,1,Guggenheim. Mr. Benjamin,male,46,0,0,PC 17593,79.2,B82 B84,C,?,?,New York. NY,0",
    "144,1,Harder. Mr. George Achilles,male,25,1,0,11765,55.4417,E50,C,5,?,Brooklyn. NY,1",
    "145,1,Harder. Mrs. George Achilles (Dorothy Annan),female,25,1,0,11765,55.4417,E50,C,5,?,Brooklyn. NY,1",
    "146,1,Harper. Mr. Henry Sleeper,male,48,1,0,PC 17572,76.7292,D33,C,3,?,New York. NY,1",
    "147,1,Harper. Mrs. Henry Sleeper (Myna Haxtun),female,49,1,0,PC 17572,76.7292,D33,C,3,?,New York. NY,1",
    "148,1,Harrington. Mr. Charles H,male,?,0,0,113796,42.4,?,S,?,?,?,0",
    "149,1,Harris. Mr. Henry Birkhardt,male,45,1,0,36973,83.475,C83,S,?,?,New York. NY,0",
    "150,1,Harris. Mrs. Henry Birkhardt (Irene Wallach),female,35,1,0,36973,83.475,C83,S,D,?,New York. NY,1",
    "151,1,Harrison. Mr. William,male,40,0,0,112059,0,B94,S,?,110,?,0",
    "152,1,Hassab. Mr. Hammad,male,27,0,0,PC 17572,76.7292,D49,C,3,?,?,1",
    "153,1,Hawksford. Mr. Walter James,male,?,0,0,16988,30,D45,S,3,?,Kingston. Surrey,1",
    "154,1,Hays. Miss. Margaret Bechstein,female,24,0,0,11767,83.1583,C54,C,7,?,New York. NY,1",
    "155,1,Hays. Mr. Charles Melville,male,55,1,1,12749,93.5,B69,S,?,307,Montreal. PQ,0",
    "156,1,Hays. Mrs. Charles Melville (Clara Jennings Gregg),female,52,1,1,12749,93.5,B69,S,3,?,Montreal. PQ,1",
    "157,1,Head. Mr. Christopher,male,42,0,0,113038,42.5,B11,S,?,?,London / Middlesex,0",
    "158,1,Hilliard. Mr. Herbert Henry,male,?,0,0,17463,51.8625,E46,S,?,?,Brighton. MA,0",
    "159,1,Hipkins. Mr. William Edward,male,55,0,0,680,50,C39,S,?,?,London / Birmingham,0",
    "160,1,Hippach. Miss. Jean Gertrude,female,16,0,1,111361,57.9792,B18,C,4,?,Chicago. IL,1",
    "161,1,Hippach. Mrs. Louis Albert (Ida Sophia Fischer),female,44,0,1,111361,57.9792,B18,C,4,?,Chicago. IL,1",
    "162,1,Hogeboom. Mrs. John C (Anna Andrews),female,51,1,0,13502,77.9583,D11,S,10,?,Hudson. NY,1",
    "163,1,Holverson. Mr. Alexander Oskar,male,42,1,0,113789,52,?,S,?,38,New York. NY,0",
    "164,1,Holverson. Mrs. Alexander Oskar (Mary Aline Towner),female,35,1,0,113789,52,?,S,8,?,New York. NY,1",
    "165,1,Homer. Mr. Harry ('Mr E Haven'),male,35,0,0,111426,26.55,?,C,15,?,Indianapolis. IN,1",
    "166,1,Hoyt. Mr. Frederick Maxfield,male,38,1,0,19943,90,C93,S,D,?,New York. NY /  Stamford CT,1",
    "167,1,Hoyt. Mr. William Fisher,male,?,0,0,PC 17600,30.6958,?,C,14,?,New York. NY,0",
    "168,1,Hoyt. Mrs. Frederick Maxfield (Jane Anne Forby),female,35,1,0,19943,90,C93,S,D,?,New York. NY /  Stamford CT,1",
    "169,1,Icard. Miss. Amelie,female,38,0,0,113572,80,B28,?,6,?,?,1",
    "170,1,Isham. Miss. Ann Elizabeth,female,50,0,0,PC 17595,28.7125,C49,C,?,?,Paris. France New York. NY,0",
    "171,1,Ismay. Mr. Joseph Bruce,male,49,0,0,112058,0,B52 B54 B56,S,C,?,Liverpool,1",
    "172,1,Jones. Mr. Charles Cresson,male,46,0,0,694,26,?,S,?,80,Bennington. VT,0",
    "173,1,Julian. Mr. Henry Forbes,male,50,0,0,113044,26,E60,S,?,?,London,0",
    "174,1,Keeping. Mr. Edwin,male,32.5,0,0,113503,211.5,C132,C,?,45,?,0",
    "175,1,Kent. Mr. Edward Austin,male,58,0,0,11771,29.7,B37,C,?,258,Buffalo. NY,0",
    "176,1,Kenyon. Mr. Frederick R,male,41,1,0,17464,51.8625,D21,S,?,?,Southington / Noank. CT,0",
    "177,1,Kenyon. Mrs. Frederick R (Marion),female,?,1,0,17464,51.8625,D21,S,8,?,Southington / Noank. CT,1",
    "178,1,Kimball. Mr. Edwin Nelson Jr,male,42,1,0,11753,52.5542,D19,S,5,?,Boston. MA,1",
    "179,1,Kimball. Mrs. Edwin Nelson Jr (Gertrude Parsons),female,45,1,0,11753,52.5542,D19,S,5,?,Boston. MA,1",
    "180,1,Klaber. Mr. Herman,male,?,0,0,113028,26.55,C124,S,?,?,Portland. OR,0",
    "181,1,Kreuchen. Miss. Emilie,female,39,0,0,24160,211.3375,?,S,2,?,?,1",
    "182,1,Leader. Dr. Alice (Farnham),female,49,0,0,17465,25.9292,D17,S,8,?,New York. NY,1",
    "183,1,LeRoy. Miss. Bertha,female,30,0,0,PC 17761,106.425,?,C,2,?,?,1",
    "184,1,Lesurer. Mr. Gustave J,male,35,0,0,PC 17755,512.3292,B101,C,3,?,?,1",
    "185,1,Lewy. Mr. Ervin G,male,?,0,0,PC 17612,27.7208,?,C,?,?,Chicago. IL,0",
    "186,1,Lindeberg-Lind. Mr. Erik Gustaf ('Mr Edward Lingrey'),male,42,0,0,17475,26.55,?,S,?,?,Stockholm. Sweden,0",
    "187,1,Lindstrom. Mrs. Carl Johan (Sigrid Posse),female,55,0,0,112377,27.7208,?,C,6,?,Stockholm. Sweden,1",
    "188,1,Lines. Miss. Mary Conover,female,16,0,1,PC 17592,39.4,D28,S,9,?,Paris. France,1",
    "189,1,Lines. Mrs. Ernest H (Elizabeth Lindsey James),female,51,0,1,PC 17592,39.4,D28,S,9,?,Paris. France,1",
    "190,1,Long. Mr. Milton Clyde,male,29,0,0,113501,30,D6,S,?,126,Springfield. MA,0",
    "191,1,Longley. Miss. Gretchen Fiske,female,21,0,0,13502,77.9583,D9,S,10,?,Hudson. NY,1",
    "192,1,Loring. Mr. Joseph Holland,male,30,0,0,113801,45.5,?,S,?,?,London / New York. NY,0",
    "193,1,Lurette. Miss. Elise,female,58,0,0,PC 17569,146.5208,B80,C,?,?,?,1",
    "194,1,Madill. Miss. Georgette Alexandra,female,15,0,1,24160,211.3375,B5,S,2,?,St Louis. MO,1",
    "195,1,Maguire. Mr. John Edward,male,30,0,0,110469,26,C106,S,?,?,Brockton. MA,0",
    "196,1,Maioni. Miss. Roberta,female,16,0,0,110152,86.5,B79,S,8,?,?,1",
    "197,1,Marechal. Mr. Pierre,male,?,0,0,11774,29.7,C47,C,7,?,Paris. France,1",
    "198,1,Marvin. Mr. Daniel Warner,male,19,1,0,113773,53.1,D30,S,?,?,New York. NY,0",
    "199,1,Marvin. Mrs. Daniel Warner (Mary Graham Carmichael Farquarson),female,18,1,0,113773,53.1,D30,S,10,?,New York. NY,1",
    "200,1,Mayne. Mlle. Berthe Antonine ('Mrs de Villiers'),female,24,0,0,PC 17482,49.5042,C90,C,6,?,Belgium  Montreal. PQ,1",
    "201,1,McCaffry. Mr. Thomas Francis,male,46,0,0,13050,75.2417,C6,C,?,292,Vancouver. BC,0",
    "202,1,McCarthy. Mr. Timothy J,male,54,0,0,17463,51.8625,E46,S,?,175,Dorchester. MA,0",
    "203,1,McGough. Mr. James Robert,male,36,0,0,PC 17473,26.2875,E25,S,7,?,Philadelphia. PA,1",
    "204,1,Meyer. Mr. Edgar Joseph,male,28,1,0,PC 17604,82.1708,?,C,?,?,New York. NY,0",
    "205,1,Meyer. Mrs. Edgar Joseph (Leila Saks),female,?,1,0,PC 17604,82.1708,?,C,6,?,New York. NY,1",
    "206,1,Millet. Mr. Francis Davis,male,65,0,0,13509,26.55,E38,S,?,249,East Bridgewater. MA,0",
    "207,1,Minahan. Dr. William Edward,male,44,2,0,19928,90,C78,Q,?,230,Fond du Lac. WI,0",
    "208,1,Minahan. Miss. Daisy E,female,33,1,0,19928,90,C78,Q,14,?,Green Bay. WI,1",
    "209,1,Minahan. Mrs. William Edward (Lillian E Thorpe),female,37,1,0,19928,90,C78,Q,14,?,Fond du Lac. WI,1",
    "210,1,Mock. Mr. Philipp Edmund,male,30,1,0,13236,57.75,C78,C,11,?,New York. NY,1",
    "211,1,Molson. Mr. Harry Markland,male,55,0,0,113787,30.5,C30,S,?,?,Montreal. PQ,0",
    "212,1,Moore. Mr. Clarence Bloomfield,male,47,0,0,113796,42.4,?,S,?,?,Washington. DC,0",
    "213,1,Natsch. Mr. Charles H,male,37,0,1,PC 17596,29.7,C118,C,?,?,Brooklyn. NY,0",
    "214,1,Newell. Miss. Madeleine,female,31,1,0,35273,113.275,D36,C,6,?,Lexington. MA,1",
    "215,1,Newell. Miss. Marjorie,female,23,1,0,35273,113.275,D36,C,6,?,Lexington. MA,1",
    "216,1,Newell. Mr. Arthur Webster,male,58,0,2,35273,113.275,D48,C,?,122,Lexington. MA,0",
    "217,1,Newsom. Miss. Helen Monypeny,female,19,0,2,11752,26.2833,D47,S,5,?,New York. NY,1",
    "218,1,Nicholson. Mr. Arthur Ernest,male,64,0,0,693,26,?,S,?,263,Isle of Wight. England,0",
    "219,1,Oliva y Ocana. Dona. Fermina,female,39,0,0,PC 17758,108.9,C105,C,8,?,?,1",
    "220,1,Omont. Mr. Alfred Fernand,male,?,0,0,F.C. 12998,25.7417,?,C,7,?,Paris. France,1",
    "221,1,Ostby. Miss. Helene Ragnhild,female,22,0,1,113509,61.9792,B36,C,5,?,Providence. RI,1",
    "222,1,Ostby. Mr. Engelhart Cornelius,male,65,0,1,113509,61.9792,B30,C,?,234,Providence. RI,0",
    "223,1,Ovies y Rodriguez. Mr. Servando,male,28.5,0,0,PC 17562,27.7208,D43,C,?,189,?Havana. Cuba,0",
    "224,1,Parr. Mr. William Henry Marsh,male,?,0,0,112052,0,?,S,?,?,Belfast,0",
    "225,1,Partner. Mr. Austen,male,45.5,0,0,113043,28.5,C124,S,?,166,Surbiton Hill. Surrey,0",
    "226,1,Payne. Mr. Vivian Ponsonby,male,23,0,0,12749,93.5,B24,S,?,?,Montreal. PQ,0",
    "227,1,Pears. Mr. Thomas Clinton,male,29,1,0,113776,66.6,C2,S,?,?,Isleworth. England,0",
    "228,1,Pears. Mrs. Thomas (Edith Wearne),female,22,1,0,113776,66.6,C2,S,8,?,Isleworth. England,1",
    "229,1,Penasco y Castellana. Mr. Victor de Satode,male,18,1,0,PC 17758,108.9,C65,C,?,?,Madrid. Spain,0",
    "230,1,Penasco y Castellana. Mrs. Victor de Satode (Maria Josefa Perez de Soto y Vallejo),female,17,1,0,PC 17758,108.9,C65,C,8,?,Madrid. Spain,1",
    "231,1,Perreault. Miss. Anne,female,30,0,0,12749,93.5,B73,S,3,?,?,1",
    "232,1,Peuchen. Major. Arthur Godfrey,male,52,0,0,113786,30.5,C104,S,6,?,Toronto. ON,1",
    "233,1,Porter. Mr. Walter Chamberlain,male,47,0,0,110465,52,C110,S,?,207,Worcester. MA,0",
    "234,1,Potter. Mrs. Thomas Jr (Lily Alexenia Wilson),female,56,0,1,11767,83.1583,C50,C,7,?,Mt Airy. Philadelphia. PA,1",
    "235,1,Reuchlin. Jonkheer. John George,male,38,0,0,19972,0,?,S,?,?,Rotterdam. Netherlands,0",
    "236,1,Rheims. Mr. George Alexander Lucien,male,?,0,0,PC 17607,39.6,?,S,A,?,Paris /  New York. NY,1",
    "237,1,Ringhini. Mr. Sante,male,22,0,0,PC 17760,135.6333,?,C,?,232,?,0",
    "238,1,Robbins. Mr. Victor,male,?,0,0,PC 17757,227.525,?,C,?,?,?,0",
    "239,1,Robert. Mrs. Edward Scott (Elisabeth Walton McMillan),female,43,0,1,24160,211.3375,B3,S,2,?,St Louis. MO,1",
    "240,1,Roebling. Mr. Washington Augustus II,male,31,0,0,PC 17590,50.4958,A24,S,?,?,Trenton. NJ,0",
    "241,1,Romaine. Mr. Charles Hallace ('Mr C Rolmane'),male,45,0,0,111428,26.55,?,S,9,?,New York. NY,1",
    "242,1,Rood. Mr. Hugh Roscoe,male,?,0,0,113767,50,A32,S,?,?,Seattle. WA,0",
    "243,1,Rosenbaum. Miss. Edith Louise,female,33,0,0,PC 17613,27.7208,A11,C,11,?,Paris. France,1",
    "244,1,Rosenshine. Mr. George ('Mr George Thorne'),male,46,0,0,PC 17585,79.2,?,C,?,16,New York. NY,0",
    "245,1,Ross. Mr. John Hugo,male,36,0,0,13049,40.125,A10,C,?,?,Winnipeg. MB,0",
    "246,1,Rothes. the Countess. of (Lucy Noel Martha Dyer-Edwards),female,33,0,0,110152,86.5,B77,S,8,?,London  Vancouver. BC,1",
    "247,1,Rothschild. Mr. Martin,male,55,1,0,PC 17603,59.4,?,C,?,?,New York. NY,0",
    "248,1,Rothschild. Mrs. Martin (Elizabeth L. Barrett),female,54,1,0,PC 17603,59.4,?,C,6,?,New York. NY,1",
    "249,1,Rowe. Mr. Alfred G,male,33,0,0,113790,26.55,?,S,?,109,London,0",
    "250,1,Ryerson. Master. John Borie,male,13,2,2,PC 17608,262.375,B57 B59 B63 B66,C,4,?,Haverford. PA / Cooperstown. NY,1",
    "251,1,Ryerson. Miss. Emily Borie,female,18,2,2,PC 17608,262.375,B57 B59 B63 B66,C,4,?,Haverford. PA / Cooperstown. NY,1",
    "252,1,Ryerson. Miss. Susan Parker 'Suzette',female,21,2,2,PC 17608,262.375,B57 B59 B63 B66,C,4,?,Haverford. PA / Cooperstown. NY,1",
    "253,1,Ryerson. Mr. Arthur Larned,male,61,1,3,PC 17608,262.375,B57 B59 B63 B66,C,?,?,Haverford. PA / Cooperstown. NY,0",
    "254,1,Ryerson. Mrs. Arthur Larned (Emily Maria Borie),female,48,1,3,PC 17608,262.375,B57 B59 B63 B66,C,4,?,Haverford. PA / Cooperstown. NY,1",
    "255,1,Saalfeld. Mr. Adolphe,male,?,0,0,19988,30.5,C106,S,3,?,Manchester. England,1",
    "256,1,Sagesser. Mlle. Emma,female,24,0,0,PC 17477,69.3,B35,C,9,?,?,1",
    "257,1,Salomon. Mr. Abraham L,male,?,0,0,111163,26,?,S,1,?,New York. NY,1",
    "258,1,Schabert. Mrs. Paul (Emma Mock),female,35,1,0,13236,57.75,C28,C,11,?,New York. NY,1",
    "259,1,Serepeca. Miss. Augusta,female,30,0,0,113798,31,?,C,4,?,?,1",
    "260,1,Seward. Mr. Frederic Kimber,male,34,0,0,113794,26.55,?,S,7,?,New York. NY,1",
    "261,1,Shutes. Miss. Elizabeth W,female,40,0,0,PC 17582,153.4625,C125,S,3,?,New York. NY / Greenwich CT,1",
    "262,1,Silverthorne. Mr. Spencer Victor,male,35,0,0,PC 17475,26.2875,E24,S,5,?,St Louis. MO,1",
    "263,1,Silvey. Mr. William Baird,male,50,1,0,13507,55.9,E44,S,?,?,Duluth. MN,0",
    "264,1,Silvey. Mrs. William Baird (Alice Munger),female,39,1,0,13507,55.9,E44,S,11,?,Duluth. MN,1",
    "265,1,Simonius-Blumer. Col. Oberst Alfons,male,56,0,0,13213,35.5,A26,C,3,?,Basel. Switzerland,1",
    "266,1,Sloper. Mr. William Thompson,male,28,0,0,113788,35.5,A6,S,7,?,New Britain. CT,1",
    "267,1,Smart. Mr. John Montgomery,male,56,0,0,113792,26.55,?,S,?,?,New York. NY,0",
    "268,1,Smith. Mr. James Clinch,male,56,0,0,17764,30.6958,A7,C,?,?,St James. Long Island. NY,0",
    "269,1,Smith. Mr. Lucien Philip,male,24,1,0,13695,60,C31,S,?,?,Huntington. WV,0",
    "270,1,Smith. Mr. Richard William,male,?,0,0,113056,26,A19,S,?,?,Streatham. Surrey,0",
    "271,1,Smith. Mrs. Lucien Philip (Mary Eloise Hughes),female,18,1,0,13695,60,C31,S,6,?,Huntington. WV,1",
    "272,1,Snyder. Mr. John Pillsbury,male,24,1,0,21228,82.2667,B45,S,7,?,Minneapolis. MN,1",
    "273,1,Snyder. Mrs. John Pillsbury (Nelle Stevenson),female,23,1,0,21228,82.2667,B45,S,7,?,Minneapolis. MN,1",
    "274,1,Spedden. Master. Robert Douglas,male,6,0,2,16966,134.5,E34,C,3,?,Tuxedo Park. NY,1",
    "275,1,Spedden. Mr. Frederic Oakley,male,45,1,1,16966,134.5,E34,C,3,?,Tuxedo Park. NY,1",
    "276,1,Spedden. Mrs. Frederic Oakley (Margaretta Corning Stone),female,40,1,1,16966,134.5,E34,C,3,?,Tuxedo Park. NY,1",
    "277,1,Spencer. Mr. William Augustus,male,57,1,0,PC 17569,146.5208,B78,C,?,?,Paris. France,0",
    "278,1,Spencer. Mrs. William Augustus (Marie Eugenie),female,?,1,0,PC 17569,146.5208,B78,C,6,?,Paris. France,1",
    "279,1,Stahelin-Maeglin. Dr. Max,male,32,0,0,13214,30.5,B50,C,3,?,Basel. Switzerland,1",
    "280,1,Stead. Mr. William Thomas,male,62,0,0,113514,26.55,C87,S,?,?,Wimbledon Park. London / Hayling Island. Hants,0",
    "281,1,Stengel. Mr. Charles Emil Henry,male,54,1,0,11778,55.4417,C116,C,1,?,Newark. NJ,1",
    "282,1,Stengel. Mrs. Charles Emil Henry (Annie May Morris),female,43,1,0,11778,55.4417,C116,C,5,?,Newark. NJ,1",
    "283,1,Stephenson. Mrs. Walter Bertram (Martha Eustis),female,52,1,0,36947,78.2667,D20,C,4,?,Haverford. PA,1",
    "284,1,Stewart. Mr. Albert A,male,?,0,0,PC 17605,27.7208,?,C,?,?,Gallipolis. Ohio / ? Paris / New York,0",
    "285,1,Stone. Mrs. George Nelson (Martha Evelyn),female,62,0,0,113572,80,B28,?,6,?,Cincinatti. OH,1",
    "286,1,Straus. Mr. Isidor,male,67,1,0,PC 17483,221.7792,C55 C57,S,?,96,New York. NY,0",
    "287,1,Straus. Mrs. Isidor (Rosalie Ida Blun),female,63,1,0,PC 17483,221.7792,C55 C57,S,?,?,New York. NY,0",
    "288,1,Sutton. Mr. Frederick,male,61,0,0,36963,32.3208,D50,S,?,46,Haddenfield. NJ,0",
    "289,1,Swift. Mrs. Frederick Joel (Margaret Welles Barron),female,48,0,0,17466,25.9292,D17,S,8,?,Brooklyn. NY,1",
    "290,1,Taussig. Miss. Ruth,female,18,0,2,110413,79.65,E68,S,8,?,New York. NY,1",
    "291,1,Taussig. Mr. Emil,male,52,1,1,110413,79.65,E67,S,?,?,New York. NY,0",
    "292,1,Taussig. Mrs. Emil (Tillie Mandelbaum),female,39,1,1,110413,79.65,E67,S,8,?,New York. NY,1",
    "293,1,Taylor. Mr. Elmer Zebley,male,48,1,0,19996,52,C126,S,5 7,?,London /  East Orange. NJ,1",
    "294,1,Taylor. Mrs. Elmer Zebley (Juliet Cummins Wright),female,?,1,0,19996,52,C126,S,5 7,?,London /  East Orange. NJ,1",
    "295,1,Thayer. Mr. John Borland,male,49,1,1,17421,110.8833,C68,C,?,?,Haverford. PA,0",
    "296,1,Thayer. Mr. John Borland Jr,male,17,0,2,17421,110.8833,C70,C,B,?,Haverford. PA,1",
    "297,1,Thayer. Mrs. John Borland (Marian Longstreth Morris),female,39,1,1,17421,110.8833,C68,C,4,?,Haverford. PA,1",
    "298,1,Thorne. Mrs. Gertrude Maybelle,female,?,0,0,PC 17585,79.2,?,C,D,?,New York. NY,1",
    "299,1,Tucker. Mr. Gilbert Milligan Jr,male,31,0,0,2543,28.5375,C53,C,7,?,Albany. NY,1",
    "300,1,Uruchurtu. Don. Manuel E,male,40,0,0,PC 17601,27.7208,?,C,?,?,Mexico City. Mexico,0",
    "301,1,Van der hoef. Mr. Wyckoff,male,61,0,0,111240,33.5,B19,S,?,245,Brooklyn. NY,0",
    "302,1,Walker. Mr. William Anderson,male,47,0,0,36967,34.0208,D46,S,?,?,East Orange. NJ,0",
    "303,1,Ward. Miss. Anna,female,35,0,0,PC 17755,512.3292,?,C,3,?,?,1",
    "304,1,Warren. Mr. Frank Manley,male,64,1,0,110813,75.25,D37,C,?,?,Portland. OR,0",
    "305,1,Warren. Mrs. Frank Manley (Anna Sophia Atkinson),female,60,1,0,110813,75.25,D37,C,5,?,Portland. OR,1",
    "306,1,Weir. Col. John,male,60,0,0,113800,26.55,?,S,?,?,England Salt Lake City. Utah,0",
    "307,1,White. Mr. Percival Wayland,male,54,0,1,35281,77.2875,D26,S,?,?,Brunswick. ME,0",
    "308,1,White. Mr. Richard Frasar,male,21,0,1,35281,77.2875,D26,S,?,169,Brunswick. ME,0",
    "309,1,White. Mrs. John Stuart (Ella Holmes),female,55,0,0,PC 17760,135.6333,C32,C,8,?,New York. NY / Briarcliff Manor NY,1",
    "310,1,Wick. Miss. Mary Natalie,female,31,0,2,36928,164.8667,C7,S,8,?,Youngstown. OH,1",
    "311,1,Wick. Mr. George Dennick,male,57,1,1,36928,164.8667,?,S,?,?,Youngstown. OH,0",
    "312,1,Wick. Mrs. George Dennick (Mary Hitchcock),female,45,1,1,36928,164.8667,?,S,8,?,Youngstown. OH,1",
    "313,1,Widener. Mr. George Dunton,male,50,1,1,113503,211.5,C80,C,?,?,Elkins Park. PA,0",
    "314,1,Widener. Mr. Harry Elkins,male,27,0,2,113503,211.5,C82,C,?,?,Elkins Park. PA,0",
    "315,1,Widener. Mrs. George Dunton (Eleanor Elkins),female,50,1,1,113503,211.5,C80,C,4,?,Elkins Park. PA,1",
    "316,1,Willard. Miss. Constance,female,21,0,0,113795,26.55,?,S,8 10,?,Duluth. MN,1",
    "317,1,Williams. Mr. Charles Duane,male,51,0,1,PC 17597,61.3792,?,C,?,?,Geneva. Switzerland / Radnor. PA,0",
    "318,1,Williams. Mr. Richard Norris II,male,21,0,1,PC 17597,61.3792,?,C,A,?,Geneva. Switzerland / Radnor. PA,1",
    "319,1,Williams-Lambert. Mr. Fletcher Fellows,male,?,0,0,113510,35,C128,S,?,?,London. England,0",
    "320,1,Wilson. Miss. Helen Alice,female,31,0,0,16966,134.5,E39 E41,C,3,?,?,1",
    "321,1,Woolner. Mr. Hugh,male,?,0,0,19947,35.5,C52,S,D,?,London. England,1",
    "322,1,Wright. Mr. George,male,62,0,0,113807,26.55,?,S,?,?,Halifax. NS,0",
    "323,1,Young. Miss. Marie Grice,female,36,0,0,PC 17760,135.6333,C32,C,8,?,New York. NY / Washington. DC,1",
    "324,2,Abelson. Mr. Samuel,male,30,1,0,P/PP 3381,24,?,C,?,?,Russia New York. NY,0",
    "325,2,Abelson. Mrs. Samuel (Hannah Wizosky),female,28,1,0,P/PP 3381,24,?,C,10,?,Russia New York. NY,1",
    "326,2,Aldworth. Mr. Charles Augustus,male,30,0,0,248744,13,?,S,?,?,Bryn Mawr. PA. USA,0",
    "327,2,Andrew. Mr. Edgardo Samuel,male,18,0,0,231945,11.5,?,S,?,?,Buenos Aires. Argentina / New Jersey. NJ,0",
    "328,2,Andrew. Mr. Frank Thomas,male,25,0,0,C.A. 34050,10.5,?,S,?,?,Cornwall. England Houghton. MI,0",
    "329,2,Angle. Mr. William A,male,34,1,0,226875,26,?,S,?,?,Warwick. England,0",
    "330,2,Angle. Mrs. William A (Florence 'Mary' Agnes Hughes),female,36,1,0,226875,26,?,S,11,?,Warwick. England,1",
    "331,2,Ashby. Mr. John,male,57,0,0,244346,13,?,S,?,?,West Hoboken. NJ,0",
    "332,2,Bailey. Mr. Percy Andrew,male,18,0,0,29108,11.5,?,S,?,?,Penzance. Cornwall / Akron. OH,0",
    "333,2,Baimbrigge. Mr. Charles Robert,male,23,0,0,C.A. 31030,10.5,?,S,?,?,Guernsey,0",
    "334,2,Ball. Mrs. (Ada E Hall),female,36,0,0,28551,13,D,S,10,?,Bristol. Avon / Jacksonville. FL,1",
    "335,2,Banfield. Mr. Frederick James,male,28,0,0,C.A./SOTON 34068,10.5,?,S,?,?,Plymouth. Dorset / Houghton. MI,0",
    "336,2,Bateman. Rev. Robert James,male,51,0,0,S.O.P. 1166,12.525,?,S,?,174,Jacksonville. FL,0",
    "337,2,Beane. Mr. Edward,male,32,1,0,2908,26,?,S,13,?,Norwich / New York. NY,1",
    "338,2,Beane. Mrs. Edward (Ethel Clarke),female,19,1,0,2908,26,?,S,13,?,Norwich / New York. NY,1",
    "339,2,Beauchamp. Mr. Henry James,male,28,0,0,244358,26,?,S,?,?,England,0",
    "340,2,Becker. Master. Richard F,male,1,2,1,230136,39,F4,S,11,?,Guntur. India / Benton Harbour. MI,1",
    "341,2,Becker. Miss. Marion Louise,female,4,2,1,230136,39,F4,S,11,?,Guntur. India / Benton Harbour. MI,1",
    "342,2,Becker. Miss. Ruth Elizabeth,female,12,2,1,230136,39,F4,S,13,?,Guntur. India / Benton Harbour. MI,1",
    "343,2,Becker. Mrs. Allen Oliver (Nellie E Baumgardner),female,36,0,3,230136,39,F4,S,11,?,Guntur. India / Benton Harbour. MI,1",
    "344,2,Beesley. Mr. Lawrence,male,34,0,0,248698,13,D56,S,13,?,London,1",
    "345,2,Bentham. Miss. Lilian W,female,19,0,0,28404,13,?,S,12,?,Rochester. NY,1",
    "346,2,Berriman. Mr. William John,male,23,0,0,28425,13,?,S,?,?,St Ives. Cornwall / Calumet. MI,0",
    "347,2,Botsford. Mr. William Hull,male,26,0,0,237670,13,?,S,?,?,Elmira. NY / Orange. NJ,0",
    "348,2,Bowenur. Mr. Solomon,male,42,0,0,211535,13,?,S,?,?,London,0",
    "349,2,Bracken. Mr. James H,male,27,0,0,220367,13,?,S,?,?,Lake Arthur. Chavez County. NM,0",
    "350,2,Brown. Miss. Amelia 'Mildred',female,24,0,0,248733,13,F33,S,11,?,London / Montreal. PQ,1",
    "351,2,Brown. Miss. Edith Eileen,female,15,0,2,29750,39,?,S,14,?,Cape Town. South Africa / Seattle. WA,1",
    "352,2,Brown. Mr. Thomas William Solomon,male,60,1,1,29750,39,?,S,?,?,Cape Town. South Africa / Seattle. WA,0",
    "353,2,Brown. Mrs. Thomas William Solomon (Elizabeth Catherine Ford),female,40,1,1,29750,39,?,S,14,?,Cape Town. South Africa / Seattle. WA,1",
    "354,2,Bryhl. Miss. Dagmar Jenny Ingeborg,female,20,1,0,236853,26,?,S,12,?,Skara. Sweden / Rockford. IL,1",
    "355,2,Bryhl. Mr. Kurt Arnold Gottfrid,male,25,1,0,236853,26,?,S,?,?,Skara. Sweden / Rockford. IL,0",
    "356,2,Buss. Miss. Kate,female,36,0,0,27849,13,?,S,9,?,Sittingbourne. England / San Diego. CA,1",
    "357,2,Butler. Mr. Reginald Fenton,male,25,0,0,234686,13,?,S,?,97,Southsea. Hants,0",
    "358,2,Byles. Rev. Thomas Roussel Davids,male,42,0,0,244310,13,?,S,?,?,London,0",
    "359,2,Bystrom. Mrs. (Karolina),female,42,0,0,236852,13,?,S,?,?,New York. NY,1",
    "360,2,Caldwell. Master. Alden Gates,male,0.8333,0,2,248738,29,?,S,13,?,Bangkok. Thailand / Roseville. IL,1",
    "361,2,Caldwell. Mr. Albert Francis,male,26,1,1,248738,29,?,S,13,?,Bangkok. Thailand / Roseville. IL,1",
    "362,2,Caldwell. Mrs. Albert Francis (Sylvia Mae Harbaugh),female,22,1,1,248738,29,?,S,13,?,Bangkok. Thailand / Roseville. IL,1",
    "363,2,Cameron. Miss. Clear Annie,female,35,0,0,F.C.C. 13528,21,?,S,14,?,Mamaroneck. NY,1",
    "364,2,Campbell. Mr. William,male,?,0,0,239853,0,?,S,?,?,Belfast,0",
    "365,2,Carbines. Mr. William,male,19,0,0,28424,13,?,S,?,18,St Ives. Cornwall / Calumet. MI,0",
    "366,2,Carter. Mrs. Ernest Courtenay (Lilian Hughes),female,44,1,0,244252,26,?,S,?,?,London,0",
    "367,2,Carter. Rev. Ernest Courtenay,male,54,1,0,244252,26,?,S,?,?,London,0",
    "368,2,Chapman. Mr. Charles Henry,male,52,0,0,248731,13.5,?,S,?,130,Bronx. NY,0",
    "369,2,Chapman. Mr. John Henry,male,37,1,0,SC/AH 29037,26,?,S,?,17,Cornwall / Spokane. WA,0",
    "370,2,Chapman. Mrs. John Henry (Sara Elizabeth Lawry),female,29,1,0,SC/AH 29037,26,?,S,?,?,Cornwall / Spokane. WA,0",
    "371,2,Christy. Miss. Julie Rachel,female,25,1,1,237789,30,?,S,12,?,London,1",
    "372,2,Christy. Mrs. (Alice Frances),female,45,0,2,237789,30,?,S,12,?,London,1",
    "373,2,Clarke. Mr. Charles Valentine,male,29,1,0,2003,26,?,S,?,?,England / San Francisco. CA,0",
    "374,2,Clarke. Mrs. Charles V (Ada Maria Winfield),female,28,1,0,2003,26,?,S,14,?,England / San Francisco. CA,1",
    "375,2,Coleridge. Mr. Reginald Charles,male,29,0,0,W./C. 14263,10.5,?,S,?,?,Hartford. Huntingdonshire,0",
    "376,2,Collander. Mr. Erik Gustaf,male,28,0,0,248740,13,?,S,?,?,Helsinki. Finland Ashtabula. Ohio,0",
    "377,2,Collett. Mr. Sidney C Stuart,male,24,0,0,28034,10.5,?,S,9,?,London / Fort Byron. NY,1",
    "378,2,Collyer. Miss. Marjorie 'Lottie',female,8,0,2,C.A. 31921,26.25,?,S,14,?,Bishopstoke. Hants / Fayette Valley. ID,1",
    "379,2,Collyer. Mr. Harvey,male,31,1,1,C.A. 31921,26.25,?,S,?,?,Bishopstoke. Hants / Fayette Valley. ID,0",
    "380,2,Collyer. Mrs. Harvey (Charlotte Annie Tate),female,31,1,1,C.A. 31921,26.25,?,S,14,?,Bishopstoke. Hants / Fayette Valley. ID,1",
    "381,2,Cook. Mrs. (Selena Rogers),female,22,0,0,W./C. 14266,10.5,F33,S,14,?,Pennsylvania,1",
    "382,2,Corbett. Mrs. Walter H (Irene Colvin),female,30,0,0,237249,13,?,S,?,?,Provo. UT,0",
    "383,2,Corey. Mrs. Percy C (Mary Phyllis Elizabeth Miller),female,?,0,0,F.C.C. 13534,21,?,S,?,?,Upper Burma. India Pittsburgh. PA,0",
    "384,2,Cotterill. Mr. Henry 'Harry',male,21,0,0,29107,11.5,?,S,?,?,Penzance. Cornwall / Akron. OH,0",
    "385,2,Cunningham. Mr. Alfred Fleming,male,?,0,0,239853,0,?,S,?,?,Belfast,0",
    "386,2,Davies. Master. John Morgan Jr,male,8,1,1,C.A. 33112,36.75,?,S,14,?,St Ives. Cornwall / Hancock. MI,1",
    "387,2,Davies. Mr. Charles Henry,male,18,0,0,S.O.C. 14879,73.5,?,S,?,?,Lyndhurst. England,0",
    "388,2,Davies. Mrs. John Morgan (Elizabeth Agnes Mary White),female,48,0,2,C.A. 33112,36.75,?,S,14,?,St Ives. Cornwall / Hancock. MI,1",
    "389,2,Davis. Miss. Mary,female,28,0,0,237668,13,?,S,13,?,London / Staten Island. NY,1",
    "390,2,de Brito. Mr. Jose Joaquim,male,32,0,0,244360,13,?,S,?,?,Portugal / Sau Paulo. Brazil,0",
    "391,2,Deacon. Mr. Percy William,male,17,0,0,S.O.C. 14879,73.5,?,S,?,?,?,0",
    "392,2,del Carlo. Mr. Sebastiano,male,29,1,0,SC/PARIS 2167,27.7208,?,C,?,295,Lucca. Italy / California,0",
    "393,2,del Carlo. Mrs. Sebastiano (Argenia Genovesi),female,24,1,0,SC/PARIS 2167,27.7208,?,C,12,?,Lucca. Italy / California,1",
    "394,2,Denbury. Mr. Herbert,male,25,0,0,C.A. 31029,31.5,?,S,?,?,Guernsey / Elizabeth. NJ,0",
    "395,2,Dibden. Mr. William,male,18,0,0,S.O.C. 14879,73.5,?,S,?,?,New Forest. England,0",
    "396,2,Doling. Miss. Elsie,female,18,0,1,231919,23,?,S,?,?,Southampton,1",
    "397,2,Doling. Mrs. John T (Ada Julia Bone),female,34,0,1,231919,23,?,S,?,?,Southampton,1",
    "398,2,Downton. Mr. William James,male,54,0,0,28403,26,?,S,?,?,Holley. NY,0",
    "399,2,Drew. Master. Marshall Brines,male,8,0,2,28220,32.5,?,S,10,?,Greenport. NY,1",
    "400,2,Drew. Mr. James Vivian,male,42,1,1,28220,32.5,?,S,?,?,Greenport. NY,0",
    "401,2,Drew. Mrs. James Vivian (Lulu Thorne Christian),female,34,1,1,28220,32.5,?,S,10,?,Greenport. NY,1",
    "402,2,Duran y More. Miss. Asuncion,female,27,1,0,SC/PARIS 2149,13.8583,?,C,12,?,Barcelona. Spain / Havana. Cuba,1",
    "403,2,Duran y More. Miss. Florentina,female,30,1,0,SC/PARIS 2148,13.8583,?,C,12,?,Barcelona. Spain / Havana. Cuba,1",
    "404,2,Eitemiller. Mr. George Floyd,male,23,0,0,29751,13,?,S,?,?,England / Detroit. MI,0",
    "405,2,Enander. Mr. Ingvar,male,21,0,0,236854,13,?,S,?,?,Goteborg. Sweden / Rockford. IL,0",
    "406,2,Fahlstrom. Mr. Arne Jonas,male,18,0,0,236171,13,?,S,?,?,Oslo. Norway Bayonne. NJ,0",
    "407,2,Faunthorpe. Mr. Harry,male,40,1,0,2926,26,?,S,?,286,England / Philadelphia. PA,0",
    "408,2,Faunthorpe. Mrs. Lizzie (Elizabeth Anne Wilkinson),female,29,1,0,2926,26,?,S,16,?,?,1",
    "409,2,Fillbrook. Mr. Joseph Charles,male,18,0,0,C.A. 15185,10.5,?,S,?,?,Cornwall / Houghton. MI,0",
    "410,2,Fox. Mr. Stanley Hubert,male,36,0,0,229236,13,?,S,?,236,Rochester. NY,0",
    "411,2,Frost. Mr. Anthony Wood 'Archie',male,?,0,0,239854,0,?,S,?,?,Belfast,0",
    "412,2,Funk. Miss. Annie Clemmer,female,38,0,0,237671,13,?,S,?,?,Janjgir. India / Pennsylvania,0",
    "413,2,Fynney. Mr. Joseph J,male,35,0,0,239865,26,?,S,?,322,Liverpool / Montreal. PQ,0",
    "414,2,Gale. Mr. Harry,male,38,1,0,28664,21,?,S,?,?,Cornwall / Clear Creek. CO,0",
    "415,2,Gale. Mr. Shadrach,male,34,1,0,28664,21,?,S,?,?,Cornwall / Clear Creek. CO,0",
    "416,2,Garside. Miss. Ethel,female,34,0,0,243880,13,?,S,12,?,Brooklyn. NY,1",
    "417,2,Gaskell. Mr. Alfred,male,16,0,0,239865,26,?,S,?,?,Liverpool / Montreal. PQ,0",
    "418,2,Gavey. Mr. Lawrence,male,26,0,0,31028,10.5,?,S,?,?,Guernsey / Elizabeth. NJ,0",
    "419,2,Gilbert. Mr. William,male,47,0,0,C.A. 30769,10.5,?,S,?,?,Cornwall,0",
    "420,2,Giles. Mr. Edgar,male,21,1,0,28133,11.5,?,S,?,?,Cornwall / Camden. NJ,0",
    "421,2,Giles. Mr. Frederick Edward,male,21,1,0,28134,11.5,?,S,?,?,Cornwall / Camden. NJ,0",
    "422,2,Giles. Mr. Ralph,male,24,0,0,248726,13.5,?,S,?,297,West Kensington. London,0",
    "423,2,Gill. Mr. John William,male,24,0,0,233866,13,?,S,?,155,Clevedon. England,0",
    "424,2,Gillespie. Mr. William Henry,male,34,0,0,12233,13,?,S,?,?,Vancouver. BC,0",
    "425,2,Givard. Mr. Hans Kristensen,male,30,0,0,250646,13,?,S,?,305,?,0",
    "426,2,Greenberg. Mr. Samuel,male,52,0,0,250647,13,?,S,?,19,Bronx. NY,0",
    "427,2,Hale. Mr. Reginald,male,30,0,0,250653,13,?,S,?,75,Auburn. NY,0",
    "428,2,Hamalainen. Master. Viljo,male,0.6667,1,1,250649,14.5,?,S,4,?,Detroit. MI,1",
    "429,2,Hamalainen. Mrs. William (Anna),female,24,0,2,250649,14.5,?,S,4,?,Detroit. MI,1",
    "430,2,Harbeck. Mr. William H,male,44,0,0,248746,13,?,S,?,35,Seattle. WA / Toledo. OH,0",
    "431,2,Harper. Miss. Annie Jessie 'Nina',female,6,0,1,248727,33,?,S,11,?,Denmark Hill. Surrey / Chicago,1",
    "432,2,Harper. Rev. John,male,28,0,1,248727,33,?,S,?,?,Denmark Hill. Surrey / Chicago,0",
    "433,2,Harris. Mr. George,male,62,0,0,S.W./PP 752,10.5,?,S,15,?,London,1",
    "434,2,Harris. Mr. Walter,male,30,0,0,W/C 14208,10.5,?,S,?,?,Walthamstow. England,0",
    "435,2,Hart. Miss. Eva Miriam,female,7,0,2,F.C.C. 13529,26.25,?,S,14,?,Ilford. Essex / Winnipeg. MB,1",
    "436,2,Hart. Mr. Benjamin,male,43,1,1,F.C.C. 13529,26.25,?,S,?,?,Ilford. Essex / Winnipeg. MB,0",
    "437,2,Hart. Mrs. Benjamin (Esther Ada Bloomfield),female,45,1,1,F.C.C. 13529,26.25,?,S,14,?,Ilford. Essex / Winnipeg. MB,1",
    "438,2,Herman. Miss. Alice,female,24,1,2,220845,65,?,S,9,?,Somerset / Bernardsville. NJ,1",
    "439,2,Herman. Miss. Kate,female,24,1,2,220845,65,?,S,9,?,Somerset / Bernardsville. NJ,1",
    "440,2,Herman. Mr. Samuel,male,49,1,2,220845,65,?,S,?,?,Somerset / Bernardsville. NJ,0",
    "441,2,Herman. Mrs. Samuel (Jane Laver),female,48,1,2,220845,65,?,S,9,?,Somerset / Bernardsville. NJ,1",
    "442,2,Hewlett. Mrs. (Mary D Kingcome),female,55,0,0,248706,16,?,S,13,?,India / Rapid City. SD,1",
    "443,2,Hickman. Mr. Leonard Mark,male,24,2,0,S.O.C. 14879,73.5,?,S,?,?,West Hampstead. London / Neepawa. MB,0",
    "444,2,Hickman. Mr. Lewis,male,32,2,0,S.O.C. 14879,73.5,?,S,?,256,West Hampstead. London / Neepawa. MB,0",
    "445,2,Hickman. Mr. Stanley George,male,21,2,0,S.O.C. 14879,73.5,?,S,?,?,West Hampstead. London / Neepawa. MB,0",
    "446,2,Hiltunen. Miss. Marta,female,18,1,1,250650,13,?,S,?,?,Kontiolahti. Finland / Detroit. MI,0",
    "447,2,Hocking. Miss. Ellen 'Nellie',female,20,2,1,29105,23,?,S,4,?,Cornwall / Akron. OH,1",
    "448,2,Hocking. Mr. Richard George,male,23,2,1,29104,11.5,?,S,?,?,Cornwall / Akron. OH,0",
    "449,2,Hocking. Mr. Samuel James Metcalfe,male,36,0,0,242963,13,?,S,?,?,Devonport. England,0",
    "450,2,Hocking. Mrs. Elizabeth (Eliza Needs),female,54,1,3,29105,23,?,S,4,?,Cornwall / Akron. OH,1",
    "451,2,Hodges. Mr. Henry Price,male,50,0,0,250643,13,?,S,?,149,Southampton,0",
    "452,2,Hold. Mr. Stephen,male,44,1,0,26707,26,?,S,?,?,England / Sacramento. CA,0",
    "453,2,Hold. Mrs. Stephen (Annie Margaret Hill),female,29,1,0,26707,26,?,S,10,?,England / Sacramento. CA,1",
    "454,2,Hood. Mr. Ambrose Jr,male,21,0,0,S.O.C. 14879,73.5,?,S,?,?,New Forest. England,0",
    "455,2,Hosono. Mr. Masabumi,male,42,0,0,237798,13,?,S,10,?,Tokyo. Japan,1",
    "456,2,Howard. Mr. Benjamin,male,63,1,0,24065,26,?,S,?,?,Swindon. England,0",
    "457,2,Howard. Mrs. Benjamin (Ellen Truelove Arman),female,60,1,0,24065,26,?,S,?,?,Swindon. England,0",
    "458,2,Hunt. Mr. George Henry,male,33,0,0,SCO/W 1585,12.275,?,S,?,?,Philadelphia. PA,0",
    "459,2,Ilett. Miss. Bertha,female,17,0,0,SO/C 14885,10.5,?,S,?,?,Guernsey,1",
    "460,2,Jacobsohn. Mr. Sidney Samuel,male,42,1,0,243847,27,?,S,?,?,London,0",
    "461,2,Jacobsohn. Mrs. Sidney Samuel (Amy Frances Christy),female,24,2,1,243847,27,?,S,12,?,London,1",
    "462,2,Jarvis. Mr. John Denzil,male,47,0,0,237565,15,?,S,?,?,North Evington. England,0",
    "463,2,Jefferys. Mr. Clifford Thomas,male,24,2,0,C.A. 31029,31.5,?,S,?,?,Guernsey / Elizabeth. NJ,0",
    "464,2,Jefferys. Mr. Ernest Wilfred,male,22,2,0,C.A. 31029,31.5,?,S,?,?,Guernsey / Elizabeth. NJ,0",
    "465,2,Jenkin. Mr. Stephen Curnow,male,32,0,0,C.A. 33111,10.5,?,S,?,?,St Ives. Cornwall / Houghton. MI,0",
    "466,2,Jerwan. Mrs. Amin S (Marie Marthe Thuillard),female,23,0,0,SC/AH Basle 541,13.7917,D,C,11,?,New York. NY,1",
    "467,2,Kantor. Mr. Sinai,male,34,1,0,244367,26,?,S,?,283,Moscow / Bronx. NY,0",
    "468,2,Kantor. Mrs. Sinai (Miriam Sternin),female,24,1,0,244367,26,?,S,12,?,Moscow / Bronx. NY,1",
    "469,2,Karnes. Mrs. J Frank (Claire Bennett),female,22,0,0,F.C.C. 13534,21,?,S,?,?,India / Pittsburgh. PA,0",
    "470,2,Keane. Miss. Nora A,female,?,0,0,226593,12.35,E101,Q,10,?,Harrisburg. PA,1",
    "471,2,Keane. Mr. Daniel,male,35,0,0,233734,12.35,?,Q,?,?,?,0",
    "472,2,Kelly. Mrs. Florence 'Fannie',female,45,0,0,223596,13.5,?,S,9,?,London / New York. NY,1",
    "473,2,Kirkland. Rev. Charles Leonard,male,57,0,0,219533,12.35,?,Q,?,?,Glasgow / Bangor. ME,0",
    "474,2,Knight. Mr. Robert J,male,?,0,0,239855,0,?,S,?,?,Belfast,0",
    "475,2,Kvillner. Mr. Johan Henrik Johannesson,male,31,0,0,C.A. 18723,10.5,?,S,?,165,Sweden / Arlington. NJ,0",
    "476,2,Lahtinen. Mrs. William (Anna Sylfven),female,26,1,1,250651,26,?,S,?,?,Minneapolis. MN,0",
    "477,2,Lahtinen. Rev. William,male,30,1,1,250651,26,?,S,?,?,Minneapolis. MN,0",
    "478,2,Lamb. Mr. John Joseph,male,?,0,0,240261,10.7083,?,Q,?,?,?,0",
    "479,2,Laroche. Miss. Louise,female,1,1,2,SC/Paris 2123,41.5792,?,C,14,?,Paris / Haiti,1",
    "480,2,Laroche. Miss. Simonne Marie Anne Andree,female,3,1,2,SC/Paris 2123,41.5792,?,C,14,?,Paris / Haiti,1",
    "481,2,Laroche. Mr. Joseph Philippe Lemercier,male,25,1,2,SC/Paris 2123,41.5792,?,C,?,?,Paris / Haiti,0",
    "482,2,Laroche. Mrs. Joseph (Juliette Marie Louise Lafargue),female,22,1,2,SC/Paris 2123,41.5792,?,C,14,?,Paris / Haiti,1",
    "483,2,Lehmann. Miss. Bertha,female,17,0,0,SC 1748,12,?,C,12,?,Berne. Switzerland / Central City. IA,1",
    "484,2,Leitch. Miss. Jessie Wills,female,?,0,0,248727,33,?,S,11,?,London / Chicago. IL,1",
    "485,2,Lemore. Mrs. (Amelia Milley),female,34,0,0,C.A. 34260,10.5,F33,S,14,?,Chicago. IL,1",
    "486,2,Levy. Mr. Rene Jacques,male,36,0,0,SC/Paris 2163,12.875,D,C,?,?,Montreal. PQ,0",
    "487,2,Leyson. Mr. Robert William Norman,male,24,0,0,C.A. 29566,10.5,?,S,?,108,?,0",
    "488,2,Lingane. Mr. John,male,61,0,0,235509,12.35,?,Q,?,?,?,0",
    "489,2,Louch. Mr. Charles Alexander,male,50,1,0,SC/AH 3085,26,?,S,?,121,Weston-Super-Mare. Somerset,0",
    "490,2,Louch. Mrs. Charles Alexander (Alice Adelaide Slow),female,42,1,0,SC/AH 3085,26,?,S,?,?,Weston-Super-Mare. Somerset,1",
    "491,2,Mack. Mrs. (Mary),female,57,0,0,S.O./P.P. 3,10.5,E77,S,?,52,Southampton / New York. NY,0",
    "492,2,Malachard. Mr. Noel,male,?,0,0,237735,15.0458,D,C,?,?,Paris,0",
    "493,2,Mallet. Master. Andre,male,1,0,2,S.C./PARIS 2079,37.0042,?,C,10,?,Paris / Montreal. PQ,1",
    "494,2,Mallet. Mr. Albert,male,31,1,1,S.C./PARIS 2079,37.0042,?,C,?,?,Paris / Montreal. PQ,0",
    "495,2,Mallet. Mrs. Albert (Antoinette Magnin),female,24,1,1,S.C./PARIS 2079,37.0042,?,C,10,?,Paris / Montreal. PQ,1",
    "496,2,Mangiavacchi. Mr. Serafino Emilio,male,?,0,0,SC/A.3 2861,15.5792,?,C,?,?,New York. NY,0",
    "497,2,Matthews. Mr. William John,male,30,0,0,28228,13,?,S,?,?,St Austall. Cornwall,0",
    "498,2,Maybery. Mr. Frank Hubert,male,40,0,0,239059,16,?,S,?,?,Weston-Super-Mare / Moose Jaw. SK,0",
    "499,2,McCrae. Mr. Arthur Gordon,male,32,0,0,237216,13.5,?,S,?,209,Sydney. Australia,0",
    "500,2,McCrie. Mr. James Matthew,male,30,0,0,233478,13,?,S,?,?,Sarnia. ON,0",
    "501,2,McKane. Mr. Peter David,male,46,0,0,28403,26,?,S,?,?,Rochester. NY,0",
    "502,2,Mellinger. Miss. Madeleine Violet,female,13,0,1,250644,19.5,?,S,14,?,England / Bennington. VT,1",
    "503,2,Mellinger. Mrs. (Elizabeth Anne Maidment),female,41,0,1,250644,19.5,?,S,14,?,England / Bennington. VT,1",
    "504,2,Mellors. Mr. William John,male,19,0,0,SW/PP 751,10.5,?,S,B,?,Chelsea. London,1",
    "505,2,Meyer. Mr. August,male,39,0,0,248723,13,?,S,?,?,Harrow-on-the-Hill. Middlesex,0",
    "506,2,Milling. Mr. Jacob Christian,male,48,0,0,234360,13,?,S,?,271,Copenhagen. Denmark,0",
    "507,2,Mitchell. Mr. Henry Michael,male,70,0,0,C.A. 24580,10.5,?,S,?,?,Guernsey / Montclair. NJ and/or Toledo. Ohio,0",
    "508,2,Montvila. Rev. Juozas,male,27,0,0,211536,13,?,S,?,?,Worcester. MA,0",
    "509,2,Moraweck. Dr. Ernest,male,54,0,0,29011,14,?,S,?,?,Frankfort. KY,0",
    "510,2,Morley. Mr. Henry Samuel ('Mr Henry Marshall'),male,39,0,0,250655,26,?,S,?,?,?,0",
    "511,2,Mudd. Mr. Thomas Charles,male,16,0,0,S.O./P.P. 3,10.5,?,S,?,?,Halesworth. England,0",
    "512,2,Myles. Mr. Thomas Francis,male,62,0,0,240276,9.6875,?,Q,?,?,Cambridge. MA,0",
    "513,2,Nasser. Mr. Nicholas,male,32.5,1,0,237736,30.0708,?,C,?,43,New York. NY,0",
    "514,2,Nasser. Mrs. Nicholas (Adele Achem),female,14,1,0,237736,30.0708,?,C,?,?,New York. NY,1",
    "515,2,Navratil. Master. Edmond Roger,male,2,1,1,230080,26,F2,S,D,?,Nice. France,1",
    "516,2,Navratil. Master. Michel M,male,3,1,1,230080,26,F2,S,D,?,Nice. France,1",
    "517,2,Navratil. Mr. Michel ('Louis M Hoffman'),male,36.5,0,2,230080,26,F2,S,?,15,Nice. France,0",
    "518,2,Nesson. Mr. Israel,male,26,0,0,244368,13,F2,S,?,?,Boston. MA,0",
    "519,2,Nicholls. Mr. Joseph Charles,male,19,1,1,C.A. 33112,36.75,?,S,?,101,Cornwall / Hancock. MI,0",
    "520,2,Norman. Mr. Robert Douglas,male,28,0,0,218629,13.5,?,S,?,287,Glasgow,0",
    "521,2,Nourney. Mr. Alfred ('Baron von Drachstedt'),male,20,0,0,SC/PARIS 2166,13.8625,D38,C,7,?,Cologne. Germany,1",
    "522,2,Nye. Mrs. (Elizabeth Ramell),female,29,0,0,C.A. 29395,10.5,F33,S,11,?,Folkstone. Kent / New York. NY,1",
    "523,2,Otter. Mr. Richard,male,39,0,0,28213,13,?,S,?,?,Middleburg Heights. OH,0",
    "524,2,Oxenham. Mr. Percy Thomas,male,22,0,0,W./C. 14260,10.5,?,S,13,?,Pondersend. England / New Durham. NJ,1",
    "525,2,Padro y Manent. Mr. Julian,male,?,0,0,SC/PARIS 2146,13.8625,?,C,9,?,Spain / Havana. Cuba,1",
    "526,2,Pain. Dr. Alfred,male,23,0,0,244278,10.5,?,S,?,?,Hamilton. ON,0",
    "527,2,Pallas y Castello. Mr. Emilio,male,29,0,0,SC/PARIS 2147,13.8583,?,C,9,?,Spain / Havana. Cuba,1",
    "528,2,Parker. Mr. Clifford Richard,male,28,0,0,SC 14888,10.5,?,S,?,?,St Andrews. Guernsey,0",
    "529,2,Parkes. Mr. Francis 'Frank',male,?,0,0,239853,0,?,S,?,?,Belfast,0",
    "530,2,Parrish. Mrs. (Lutie Davis),female,50,0,1,230433,26,?,S,12,?,Woodford County. KY,1",
    "531,2,Pengelly. Mr. Frederick William,male,19,0,0,28665,10.5,?,S,?,?,Gunnislake. England / Butte. MT,0",
    "532,2,Pernot. Mr. Rene,male,?,0,0,SC/PARIS 2131,15.05,?,C,?,?,?,0",
    "533,2,Peruschitz. Rev. Joseph Maria,male,41,0,0,237393,13,?,S,?,?,?,0",
    "534,2,Phillips. Miss. Alice Frances Louisa,female,21,0,1,S.O./P.P. 2,21,?,S,12,?,Ilfracombe. Devon,1",
    "535,2,Phillips. Miss. Kate Florence ('Mrs Kate Louise Phillips Marshall'),female,19,0,0,250655,26,?,S,11,?,Worcester. England,1",
    "536,2,Phillips. Mr. Escott Robert,male,43,0,1,S.O./P.P. 2,21,?,S,?,?,Ilfracombe. Devon,0",
    "537,2,Pinsky. Mrs. (Rosa),female,32,0,0,234604,13,?,S,9,?,Russia,1",
    "538,2,Ponesell. Mr. Martin,male,34,0,0,250647,13,?,S,?,?,Denmark / New York. NY,0",
    "539,2,Portaluppi. Mr. Emilio Ilario Giuseppe,male,30,0,0,C.A. 34644,12.7375,?,C,14,?,Milford. NH,1",
    "540,2,Pulbaum. Mr. Franz,male,27,0,0,SC/PARIS 2168,15.0333,?,C,?,?,Paris,0",
    "541,2,Quick. Miss. Phyllis May,female,2,1,1,26360,26,?,S,11,?,Plymouth. Devon / Detroit. MI,1",
    "542,2,Quick. Miss. Winifred Vera,female,8,1,1,26360,26,?,S,11,?,Plymouth. Devon / Detroit. MI,1",
    "543,2,Quick. Mrs. Frederick Charles (Jane Richards),female,33,0,2,26360,26,?,S,11,?,Plymouth. Devon / Detroit. MI,1",
    "544,2,Reeves. Mr. David,male,36,0,0,C.A. 17248,10.5,?,S,?,?,Brighton. Sussex,0",
    "545,2,Renouf. Mr. Peter Henry,male,34,1,0,31027,21,?,S,12,?,Elizabeth. NJ,0",
    "546,2,Renouf. Mrs. Peter Henry (Lillian Jefferys),female,30,3,0,31027,21,?,S,?,?,Elizabeth. NJ,1",
    "547,2,Reynaldo. Ms. Encarnacion,female,28,0,0,230434,13,?,S,9,?,Spain,1",
    "548,2,Richard. Mr. Emile,male,23,0,0,SC/PARIS 2133,15.0458,?,C,?,?,Paris / Montreal. PQ,0",
    "549,2,Richards. Master. George Sibley,male,0.8333,1,1,29106,18.75,?,S,4,?,Cornwall / Akron. OH,1",
    "550,2,Richards. Master. William Rowe,male,3,1,1,29106,18.75,?,S,4,?,Cornwall / Akron. OH,1",
    "551,2,Richards. Mrs. Sidney (Emily Hocking),female,24,2,3,29106,18.75,?,S,4,?,Cornwall / Akron. OH,1",
    "552,2,Ridsdale. Miss. Lucy,female,50,0,0,W./C. 14258,10.5,?,S,13,?,London. England / Marietta. Ohio and Milwaukee. WI,1",
    "553,2,Rogers. Mr. Reginald Harry,male,19,0,0,28004,10.5,?,S,?,?,?,0",
    "554,2,Rugg. Miss. Emily,female,21,0,0,C.A. 31026,10.5,?,S,12,?,Guernsey / Wilmington. DE,1",
    "555,2,Schmidt. Mr. August,male,26,0,0,248659,13,?,S,?,?,Newark. NJ,0",
    "556,2,Sedgwick. Mr. Charles Frederick Waddington,male,25,0,0,244361,13,?,S,?,?,Liverpool,0",
    "557,2,Sharp. Mr. Percival James R,male,27,0,0,244358,26,?,S,?,?,Hornsey. England,0",
    "558,2,Shelley. Mrs. William (Imanita Parrish Hall),female,25,0,1,230433,26,?,S,12,?,Deer Lodge. MT,1",
    "559,2,Silven. Miss. Lyyli Karoliina,female,18,0,2,250652,13,?,S,16,?,Finland / Minneapolis. MN,1",
    "560,2,Sincock. Miss. Maude,female,20,0,0,C.A. 33112,36.75,?,S,11,?,Cornwall / Hancock. MI,1",
    "561,2,Sinkkonen. Miss. Anna,female,30,0,0,250648,13,?,S,10,?,Finland / Washington. DC,1",
    "562,2,Sjostedt. Mr. Ernst Adolf,male,59,0,0,237442,13.5,?,S,?,?,Sault St Marie. ON,0",
    "563,2,Slayter. Miss. Hilda Mary,female,30,0,0,234818,12.35,?,Q,13,?,Halifax. NS,1",
    "564,2,Slemen. Mr. Richard James,male,35,0,0,28206,10.5,?,S,?,?,Cornwall,0",
    "565,2,Smith. Miss. Marion Elsie,female,40,0,0,31418,13,?,S,9,?,?,1",
    "566,2,Sobey. Mr. Samuel James Hayden,male,25,0,0,C.A. 29178,13,?,S,?,?,Cornwall / Houghton. MI,0",
    "567,2,Stanton. Mr. Samuel Ward,male,41,0,0,237734,15.0458,?,C,?,?,New York. NY,0",
    "568,2,Stokes. Mr. Philip Joseph,male,25,0,0,F.C.C. 13540,10.5,?,S,?,81,Catford. Kent / Detroit. MI,0",
    "569,2,Swane. Mr. George,male,18.5,0,0,248734,13,F,S,?,294,?,0",
    "570,2,Sweet. Mr. George Frederick,male,14,0,0,220845,65,?,S,?,?,Somerset / Bernardsville. NJ,0",
    "571,2,Toomey. Miss. Ellen,female,50,0,0,F.C.C. 13531,10.5,?,S,9,?,Indianapolis. IN,1",
    "572,2,Troupiansky. Mr. Moses Aaron,male,23,0,0,233639,13,?,S,?,?,?,0",
    "573,2,Trout. Mrs. William H (Jessie L),female,28,0,0,240929,12.65,?,S,?,?,Columbus. OH,1",
    "574,2,Troutt. Miss. Edwina Celia 'Winnie',female,27,0,0,34218,10.5,E101,S,16,?,Bath. England / Massachusetts,1",
    "575,2,Turpin. Mr. William John Robert,male,29,1,0,11668,21,?,S,?,?,Plymouth. England,0",
    "576,2,Turpin. Mrs. William John Robert (Dorothy Ann Wonnacott),female,27,1,0,11668,21,?,S,?,?,Plymouth. England,0",
    "577,2,Veal. Mr. James,male,40,0,0,28221,13,?,S,?,?,Barre. Co Washington. VT,0",
    "578,2,Walcroft. Miss. Nellie,female,31,0,0,F.C.C. 13528,21,?,S,14,?,Mamaroneck. NY,1",
    "579,2,Ware. Mr. John James,male,30,1,0,CA 31352,21,?,S,?,?,Bristol. England / New Britain. CT,0",
    "580,2,Ware. Mr. William Jeffery,male,23,1,0,28666,10.5,?,S,?,?,?,0",
    "581,2,Ware. Mrs. John James (Florence Louise Long),female,31,0,0,CA 31352,21,?,S,10,?,Bristol. England / New Britain. CT,1",
    "582,2,Watson. Mr. Ennis Hastings,male,?,0,0,239856,0,?,S,?,?,Belfast,0",
    "583,2,Watt. Miss. Bertha J,female,12,0,0,C.A. 33595,15.75,?,S,9,?,Aberdeen / Portland. OR,1",
    "584,2,Watt. Mrs. James (Elizabeth 'Bessie' Inglis Milne),female,40,0,0,C.A. 33595,15.75,?,S,9,?,Aberdeen / Portland. OR,1",
    "585,2,Webber. Miss. Susan,female,32.5,0,0,27267,13,E101,S,12,?,England / Hartford. CT,1",
    "586,2,Weisz. Mr. Leopold,male,27,1,0,228414,26,?,S,?,293,Bromsgrove. England / Montreal. PQ,0",
    "587,2,Weisz. Mrs. Leopold (Mathilde Francoise Pede),female,29,1,0,228414,26,?,S,10,?,Bromsgrove. England / Montreal. PQ,1",
    "588,2,Wells. Master. Ralph Lester,male,2,1,1,29103,23,?,S,14,?,Cornwall / Akron. OH,1",
    "589,2,Wells. Miss. Joan,female,4,1,1,29103,23,?,S,14,?,Cornwall / Akron. OH,1",
    "590,2,Wells. Mrs. Arthur Henry ('Addie' Dart Trevaskis),female,29,0,2,29103,23,?,S,14,?,Cornwall / Akron. OH,1",
    "591,2,West. Miss. Barbara J,female,0.9167,1,2,C.A. 34651,27.75,?,S,10,?,Bournmouth. England,1",
    "592,2,West. Miss. Constance Mirium,female,5,1,2,C.A. 34651,27.75,?,S,10,?,Bournmouth. England,1",
    "593,2,West. Mr. Edwy Arthur,male,36,1,2,C.A. 34651,27.75,?,S,?,?,Bournmouth. England,0",
    "594,2,West. Mrs. Edwy Arthur (Ada Mary Worth),female,33,1,2,C.A. 34651,27.75,?,S,10,?,Bournmouth. England,1",
    "595,2,Wheadon. Mr. Edward H,male,66,0,0,C.A. 24579,10.5,?,S,?,?,Guernsey. England / Edgewood. RI,0",
    "596,2,Wheeler. Mr. Edwin 'Frederick',male,?,0,0,SC/PARIS 2159,12.875,?,S,?,?,?,0",
    "597,2,Wilhelms. Mr. Charles,male,31,0,0,244270,13,?,S,9,?,London. England,1",
    "598,2,Williams. Mr. Charles Eugene,male,?,0,0,244373,13,?,S,14,?,Harrow. England,1",
    "599,2,Wright. Miss. Marion,female,26,0,0,220844,13.5,?,S,9,?,Yoevil. England / Cottage Grove. OR,1",
    "600,2,Yrois. Miss. Henriette ('Mrs Harbeck'),female,24,0,0,248747,13,?,S,?,?,Paris,0",
    "601,3,Abbing. Mr. Anthony,male,42,0,0,C.A. 5547,7.55,?,S,?,?,?,0",
    "602,3,Abbott. Master. Eugene Joseph,male,13,0,2,C.A. 2673,20.25,?,S,?,?,East Providence. RI,0",
    "603,3,Abbott. Mr. Rossmore Edward,male,16,1,1,C.A. 2673,20.25,?,S,?,190,East Providence. RI,0",
    "604,3,Abbott. Mrs. Stanton (Rosa Hunt),female,35,1,1,C.A. 2673,20.25,?,S,A,?,East Providence. RI,1",
    "605,3,Abelseth. Miss. Karen Marie,female,16,0,0,348125,7.65,?,S,16,?,Norway Los Angeles. CA,1",
    "606,3,Abelseth. Mr. Olaus Jorgensen,male,25,0,0,348122,7.65,F G63,S,A,?,Perkins County. SD,1",
    "607,3,Abrahamsson. Mr. Abraham August Johannes,male,20,0,0,SOTON/O2 3101284,7.925,?,S,15,?,Taalintehdas. Finland Hoboken. NJ,1",
    "608,3,Abrahim. Mrs. Joseph (Sophie Halaut Easu),female,18,0,0,2657,7.2292,?,C,C,?,Greensburg. PA,1",
    "609,3,Adahl. Mr. Mauritz Nils Martin,male,30,0,0,C 7076,7.25,?,S,?,72,Asarum. Sweden Brooklyn. NY,0",
    "610,3,Adams. Mr. John,male,26,0,0,341826,8.05,?,S,?,103,Bournemouth. England,0",
    "611,3,Ahlin. Mrs. Johan (Johanna Persdotter Larsson),female,40,1,0,7546,9.475,?,S,?,?,Sweden Akeley. MN,0",
    "612,3,Aks. Master. Philip Frank,male,0.8333,0,1,392091,9.35,?,S,11,?,London. England Norfolk. VA,1",
    "613,3,Aks. Mrs. Sam (Leah Rosen),female,18,0,1,392091,9.35,?,S,13,?,London. England Norfolk. VA,1",
    "614,3,Albimona. Mr. Nassef Cassem,male,26,0,0,2699,18.7875,?,C,15,?,Syria Fredericksburg. VA,1",
    "615,3,Alexander. Mr. William,male,26,0,0,3474,7.8875,?,S,?,?,England Albion. NY,0",
    "616,3,Alhomaki. Mr. Ilmari Rudolf,male,20,0,0,SOTON/O2 3101287,7.925,?,S,?,?,Salo. Finland Astoria. OR,0",
    "617,3,Ali. Mr. Ahmed,male,24,0,0,SOTON/O.Q. 3101311,7.05,?,S,?,?,?,0",
    "618,3,Ali. Mr. William,male,25,0,0,SOTON/O.Q. 3101312,7.05,?,S,?,79,Argentina,0",
    "619,3,Allen. Mr. William Henry,male,35,0,0,373450,8.05,?,S,?,?,Lower Clapton. Middlesex or Erdington. Birmingham,0",
    "620,3,Allum. Mr. Owen George,male,18,0,0,2223,8.3,?,S,?,259,Windsor. England New York. NY,0",
    "621,3,Andersen. Mr. Albert Karvin,male,32,0,0,C 4001,22.525,?,S,?,260,Bergen. Norway,0",
    "622,3,Andersen-Jensen. Miss. Carla Christine Nielsine,female,19,1,0,350046,7.8542,?,S,16,?,?,1",
    "623,3,Andersson. Master. Sigvard Harald Elias,male,4,4,2,347082,31.275,?,S,?,?,Sweden Winnipeg. MN,0",
    "624,3,Andersson. Miss. Ebba Iris Alfrida,female,6,4,2,347082,31.275,?,S,?,?,Sweden Winnipeg. MN,0",
    "625,3,Andersson. Miss. Ellis Anna Maria,female,2,4,2,347082,31.275,?,S,?,?,Sweden Winnipeg. MN,0",
    "626,3,Andersson. Miss. Erna Alexandra,female,17,4,2,3101281,7.925,?,S,D,?,Ruotsinphyhtaa. Finland New York. NY,1",
    "627,3,Andersson. Miss. Ida Augusta Margareta,female,38,4,2,347091,7.775,?,S,?,?,Vadsbro. Sweden Ministee. MI,0",
    "628,3,Andersson. Miss. Ingeborg Constanzia,female,9,4,2,347082,31.275,?,S,?,?,Sweden Winnipeg. MN,0",
    "629,3,Andersson. Miss. Sigrid Elisabeth,female,11,4,2,347082,31.275,?,S,?,?,Sweden Winnipeg. MN,0",
    "630,3,Andersson. Mr. Anders Johan,male,39,1,5,347082,31.275,?,S,?,?,Sweden Winnipeg. MN,0",
    "631,3,Andersson. Mr. August Edvard ('Wennerstrom'),male,27,0,0,350043,7.7958,?,S,A,?,?,1",
    "632,3,Andersson. Mr. Johan Samuel,male,26,0,0,347075,7.775,?,S,?,?,Hartford. CT,0",
    "633,3,Andersson. Mrs. Anders Johan (Alfrida Konstantia Brogren),female,39,1,5,347082,31.275,?,S,?,?,Sweden Winnipeg. MN,0",
    "634,3,Andreasson. Mr. Paul Edvin,male,20,0,0,347466,7.8542,?,S,?,?,Sweden Chicago. IL,0",
    "635,3,Angheloff. Mr. Minko,male,26,0,0,349202,7.8958,?,S,?,?,Bulgaria Chicago. IL,0",
    "636,3,Arnold-Franchi. Mr. Josef,male,25,1,0,349237,17.8,?,S,?,?,Altdorf. Switzerland,0",
    "637,3,Arnold-Franchi. Mrs. Josef (Josefine Franchi),female,18,1,0,349237,17.8,?,S,?,?,Altdorf. Switzerland,0",
    "638,3,Aronsson. Mr. Ernst Axel Algot,male,24,0,0,349911,7.775,?,S,?,?,Sweden Joliet. IL,0",
    "639,3,Asim. Mr. Adola,male,35,0,0,SOTON/O.Q. 3101310,7.05,?,S,?,?,?,0",
    "640,3,Asplund. Master. Carl Edgar,male,5,4,2,347077,31.3875,?,S,?,?,Sweden  Worcester. MA,0",
    "641,3,Asplund. Master. Clarence Gustaf Hugo,male,9,4,2,347077,31.3875,?,S,?,?,Sweden Worcester. MA,0",
    "642,3,Asplund. Master. Edvin Rojj Felix,male,3,4,2,347077,31.3875,?,S,15,?,Sweden Worcester. MA,1",
    "643,3,Asplund. Master. Filip Oscar,male,13,4,2,347077,31.3875,?,S,?,?,Sweden Worcester. MA,0",
    "644,3,Asplund. Miss. Lillian Gertrud,female,5,4,2,347077,31.3875,?,S,15,?,Sweden Worcester. MA,1",
    "645,3,Asplund. Mr. Carl Oscar Vilhelm Gustafsson,male,40,1,5,347077,31.3875,?,S,?,142,Sweden  Worcester. MA,0",
    "646,3,Asplund. Mr. Johan Charles,male,23,0,0,350054,7.7958,?,S,13,?,Oskarshamn. Sweden Minneapolis. MN,1",
    "647,3,Asplund. Mrs. Carl Oscar (Selma Augusta Emilia Johansson),female,38,1,5,347077,31.3875,?,S,15,?,Sweden  Worcester. MA,1",
    "648,3,Assaf Khalil. Mrs. Mariana ('Miriam'),female,45,0,0,2696,7.225,?,C,C,?,Ottawa. ON,1",
    "649,3,Assaf. Mr. Gerios,male,21,0,0,2692,7.225,?,C,?,?,Ottawa. ON,0",
    "650,3,Assam. Mr. Ali,male,23,0,0,SOTON/O.Q. 3101309,7.05,?,S,?,?,?,0",
    "651,3,Attalah. Miss. Malake,female,17,0,0,2627,14.4583,?,C,?,?,?,0",
    "652,3,Attalah. Mr. Sleiman,male,30,0,0,2694,7.225,?,C,?,?,Ottawa. ON,0",
    "653,3,Augustsson. Mr. Albert,male,23,0,0,347468,7.8542,?,S,?,?,Krakoryd. Sweden Bloomington. IL,0",
    "654,3,Ayoub. Miss. Banoura,female,13,0,0,2687,7.2292,?,C,C,?,Syria Youngstown. OH,1",
    "655,3,Baccos. Mr. Raffull,male,20,0,0,2679,7.225,?,C,?,?,?,0",
    "656,3,Backstrom. Mr. Karl Alfred,male,32,1,0,3101278,15.85,?,S,D,?,Ruotsinphytaa. Finland New York. NY,0",
    "657,3,Backstrom. Mrs. Karl Alfred (Maria Mathilda Gustafsson),female,33,3,0,3101278,15.85,?,S,?,?,Ruotsinphytaa. Finland New York. NY,1",
    "658,3,Baclini. Miss. Eugenie,female,0.75,2,1,2666,19.2583,?,C,C,?,Syria New York. NY,1",
    "659,3,Baclini. Miss. Helene Barbara,female,0.75,2,1,2666,19.2583,?,C,C,?,Syria New York. NY,1",
    "660,3,Baclini. Miss. Marie Catherine,female,5,2,1,2666,19.2583,?,C,C,?,Syria New York. NY,1",
    "661,3,Baclini. Mrs. Solomon (Latifa Qurban),female,24,0,3,2666,19.2583,?,C,C,?,Syria New York. NY,1",
    "662,3,Badman. Miss. Emily Louisa,female,18,0,0,A/4 31416,8.05,?,S,C,?,London Skanteales. NY,1",
    "663,3,Badt. Mr. Mohamed,male,40,0,0,2623,7.225,?,C,?,?,?,0",
    "664,3,Balkic. Mr. Cerin,male,26,0,0,349248,7.8958,?,S,?,?,?,0",
    "665,3,Barah. Mr. Hanna Assi,male,20,0,0,2663,7.2292,?,C,15,?,?,1",
    "666,3,Barbara. Miss. Saiide,female,18,0,1,2691,14.4542,?,C,?,?,Syria Ottawa. ON,0",
    "667,3,Barbara. Mrs. (Catherine David),female,45,0,1,2691,14.4542,?,C,?,?,Syria Ottawa. ON,0",
    "668,3,Barry. Miss. Julia,female,27,0,0,330844,7.8792,?,Q,?,?,New York. NY,0",
    "669,3,Barton. Mr. David John,male,22,0,0,324669,8.05,?,S,?,?,England New York. NY,0",
    "670,3,Beavan. Mr. William Thomas,male,19,0,0,323951,8.05,?,S,?,?,England,0",
    "671,3,Bengtsson. Mr. John Viktor,male,26,0,0,347068,7.775,?,S,?,?,Krakudden. Sweden Moune. IL,0",
    "672,3,Berglund. Mr. Karl Ivar Sven,male,22,0,0,PP 4348,9.35,?,S,?,?,Tranvik. Finland New York,0",
    "673,3,Betros. Master. Seman,male,?,0,0,2622,7.2292,?,C,?,?,?,0",
    "674,3,Betros. Mr. Tannous,male,20,0,0,2648,4.0125,?,C,?,?,Syria,0",
    "675,3,Bing. Mr. Lee,male,32,0,0,1601,56.4958,?,S,C,?,Hong Kong New York. NY,1",
    "676,3,Birkeland. Mr. Hans Martin Monsen,male,21,0,0,312992,7.775,?,S,?,?,Brennes. Norway New York,0",
    "677,3,Bjorklund. Mr. Ernst Herbert,male,18,0,0,347090,7.75,?,S,?,?,Stockholm. Sweden New York,0",
    "678,3,Bostandyeff. Mr. Guentcho,male,26,0,0,349224,7.8958,?,S,?,?,Bulgaria Chicago. IL,0",
    "679,3,Boulos. Master. Akar,male,6,1,1,2678,15.2458,?,C,?,?,Syria Kent. ON,0",
    "680,3,Boulos. Miss. Nourelain,female,9,1,1,2678,15.2458,?,C,?,?,Syria Kent. ON,0",
    "681,3,Boulos. Mr. Hanna,male,?,0,0,2664,7.225,?,C,?,?,Syria,0",
    "682,3,Boulos. Mrs. Joseph (Sultana),female,?,0,2,2678,15.2458,?,C,?,?,Syria Kent. ON,0",
    "683,3,Bourke. Miss. Mary,female,?,0,2,364848,7.75,?,Q,?,?,Ireland Chicago. IL,0",
    "684,3,Bourke. Mr. John,male,40,1,1,364849,15.5,?,Q,?,?,Ireland Chicago. IL,0",
    "685,3,Bourke. Mrs. John (Catherine),female,32,1,1,364849,15.5,?,Q,?,?,Ireland Chicago. IL,0",
    "686,3,Bowen. Mr. David John 'Dai',male,21,0,0,54636,16.1,?,S,?,?,Treherbert. Cardiff. Wales,0",
    "687,3,Bradley. Miss. Bridget Delia,female,22,0,0,334914,7.725,?,Q,13,?,Kingwilliamstown. Co Cork. Ireland Glens Falls. NY,1",
    "688,3,Braf. Miss. Elin Ester Maria,female,20,0,0,347471,7.8542,?,S,?,?,Medeltorp. Sweden Chicago. IL,0",
    "689,3,Braund. Mr. Lewis Richard,male,29,1,0,3460,7.0458,?,S,?,?,Bridgerule. Devon,0",
    "690,3,Braund. Mr. Owen Harris,male,22,1,0,A/5 21171,7.25,?,S,?,?,Bridgerule. Devon,0",
    "691,3,Brobeck. Mr. Karl Rudolf,male,22,0,0,350045,7.7958,?,S,?,?,Sweden Worcester. MA,0",
    "692,3,Brocklebank. Mr. William Alfred,male,35,0,0,364512,8.05,?,S,?,?,Broomfield. Chelmsford. England,0",
    "693,3,Buckley. Miss. Katherine,female,18.5,0,0,329944,7.2833,?,Q,?,299,Co Cork. Ireland Roxbury. MA,0",
    "694,3,Buckley. Mr. Daniel,male,21,0,0,330920,7.8208,?,Q,13,?,Kingwilliamstown. Co Cork. Ireland New York. NY,1",
    "695,3,Burke. Mr. Jeremiah,male,19,0,0,365222,6.75,?,Q,?,?,Co Cork. Ireland Charlestown. MA,0",
    "696,3,Burns. Miss. Mary Delia,female,18,0,0,330963,7.8792,?,Q,?,?,Co Sligo. Ireland New York. NY,0",
    "697,3,Cacic. Miss. Manda,female,21,0,0,315087,8.6625,?,S,?,?,?,0",
    "698,3,Cacic. Miss. Marija,female,30,0,0,315084,8.6625,?,S,?,?,?,0",
    "699,3,Cacic. Mr. Jego Grga,male,18,0,0,315091,8.6625,?,S,?,?,?,0",
    "700,3,Cacic. Mr. Luka,male,38,0,0,315089,8.6625,?,S,?,?,Croatia,0",
    "701,3,Calic. Mr. Jovo,male,17,0,0,315093,8.6625,?,S,?,?,?,0",
    "702,3,Calic. Mr. Petar,male,17,0,0,315086,8.6625,?,S,?,?,?,0",
    "703,3,Canavan. Miss. Mary,female,21,0,0,364846,7.75,?,Q,?,?,?,0",
    "704,3,Canavan. Mr. Patrick,male,21,0,0,364858,7.75,?,Q,?,?,Ireland Philadelphia. PA,0",
    "705,3,Cann. Mr. Ernest Charles,male,21,0,0,A./5. 2152,8.05,?,S,?,?,?,0",
    "706,3,Caram. Mr. Joseph,male,?,1,0,2689,14.4583,?,C,?,?,Ottawa. ON,0",
    "707,3,Caram. Mrs. Joseph (Maria Elias),female,?,1,0,2689,14.4583,?,C,?,?,Ottawa. ON,0",
    "708,3,Carlsson. Mr. August Sigfrid,male,28,0,0,350042,7.7958,?,S,?,?,Dagsas. Sweden Fower. MN,0",
    "709,3,Carlsson. Mr. Carl Robert,male,24,0,0,350409,7.8542,?,S,?,?,Goteborg. Sweden Huntley. IL,0",
    "710,3,Carr. Miss. Helen 'Ellen',female,16,0,0,367231,7.75,?,Q,16,?,Co Longford. Ireland New York. NY,1",
    "711,3,Carr. Miss. Jeannie,female,37,0,0,368364,7.75,?,Q,?,?,Co Sligo. Ireland Hartford. CT,0",
    "712,3,Carver. Mr. Alfred John,male,28,0,0,392095,7.25,?,S,?,?,St Denys. Southampton. Hants,0",
    "713,3,Celotti. Mr. Francesco,male,24,0,0,343275,8.05,?,S,?,?,London,0",
    "714,3,Charters. Mr. David,male,21,0,0,A/5. 13032,7.7333,?,Q,?,?,Ireland New York. NY,0",
    "715,3,Chip. Mr. Chang,male,32,0,0,1601,56.4958,?,S,C,?,Hong Kong New York. NY,1",
    "716,3,Christmann. Mr. Emil,male,29,0,0,343276,8.05,?,S,?,?,?,0",
    "717,3,Chronopoulos. Mr. Apostolos,male,26,1,0,2680,14.4542,?,C,?,?,Greece,0",
    "718,3,Chronopoulos. Mr. Demetrios,male,18,1,0,2680,14.4542,?,C,?,?,Greece,0",
    "719,3,Coelho. Mr. Domingos Fernandeo,male,20,0,0,SOTON/O.Q. 3101307,7.05,?,S,?,?,Portugal,0",
    "720,3,Cohen. Mr. Gurshon 'Gus',male,18,0,0,A/5 3540,8.05,?,S,12,?,London Brooklyn. NY,1",
    "721,3,Colbert. Mr. Patrick,male,24,0,0,371109,7.25,?,Q,?,?,Co Limerick. Ireland Sherbrooke. PQ,0",
    "722,3,Coleff. Mr. Peju,male,36,0,0,349210,7.4958,?,S,?,?,Bulgaria Chicago. IL,0",
    "723,3,Coleff. Mr. Satio,male,24,0,0,349209,7.4958,?,S,?,?,?,0",
    "724,3,Conlon. Mr. Thomas Henry,male,31,0,0,21332,7.7333,?,Q,?,?,Philadelphia. PA,0",
    "725,3,Connaghton. Mr. Michael,male,31,0,0,335097,7.75,?,Q,?,?,Ireland Brooklyn. NY,0",
    "726,3,Connolly. Miss. Kate,female,22,0,0,370373,7.75,?,Q,13,?,Ireland,1",
    "727,3,Connolly. Miss. Kate,female,30,0,0,330972,7.6292,?,Q,?,?,Ireland,0",
    "728,3,Connors. Mr. Patrick,male,70.5,0,0,370369,7.75,?,Q,?,171,?,0",
    "729,3,Cook. Mr. Jacob,male,43,0,0,A/5 3536,8.05,?,S,?,?,?,0",
    "730,3,Cor. Mr. Bartol,male,35,0,0,349230,7.8958,?,S,?,?,Austria,0",
    "731,3,Cor. Mr. Ivan,male,27,0,0,349229,7.8958,?,S,?,?,Austria,0",
    "732,3,Cor. Mr. Liudevit,male,19,0,0,349231,7.8958,?,S,?,?,Austria,0",
    "733,3,Corn. Mr. Harry,male,30,0,0,SOTON/OQ 392090,8.05,?,S,?,?,London,0",
    "734,3,Coutts. Master. Eden Leslie 'Neville',male,9,1,1,C.A. 37671,15.9,?,S,2,?,England Brooklyn. NY,1",
    "735,3,Coutts. Master. William Loch 'William',male,3,1,1,C.A. 37671,15.9,?,S,2,?,England Brooklyn. NY,1",
    "736,3,Coutts. Mrs. William (Winnie 'Minnie' Treanor),female,36,0,2,C.A. 37671,15.9,?,S,2,?,England Brooklyn. NY,1",
    "737,3,Coxon. Mr. Daniel,male,59,0,0,364500,7.25,?,S,?,?,Merrill. WI,0",
    "738,3,Crease. Mr. Ernest James,male,19,0,0,S.P. 3464,8.1583,?,S,?,?,Bristol. England Cleveland. OH,0",
    "739,3,Cribb. Miss. Laura Alice,female,17,0,1,371362,16.1,?,S,12,?,Bournemouth. England Newark. NJ,1",
    "740,3,Cribb. Mr. John Hatfield,male,44,0,1,371362,16.1,?,S,?,?,Bournemouth. England Newark. NJ,0",
    "741,3,Culumovic. Mr. Jeso,male,17,0,0,315090,8.6625,?,S,?,?,Austria-Hungary,0",
    "742,3,Daher. Mr. Shedid,male,22.5,0,0,2698,7.225,?,C,?,9,?,0",
    "743,3,Dahl. Mr. Karl Edwart,male,45,0,0,7598,8.05,?,S,15,?,Australia Fingal. ND,1",
    "744,3,Dahlberg. Miss. Gerda Ulrika,female,22,0,0,7552,10.5167,?,S,?,?,Norrlot. Sweden Chicago. IL,0",
    "745,3,Dakic. Mr. Branko,male,19,0,0,349228,10.1708,?,S,?,?,Austria,0",
    "746,3,Daly. Miss. Margaret Marcella 'Maggie',female,30,0,0,382650,6.95,?,Q,15,?,Co Athlone. Ireland New York. NY,1",
    "747,3,Daly. Mr. Eugene Patrick,male,29,0,0,382651,7.75,?,Q,13 15 B,?,Co Athlone. Ireland New York. NY,1",
    "748,3,Danbom. Master. Gilbert Sigvard Emanuel,male,0.3333,0,2,347080,14.4,?,S,?,?,Stanton. IA,0",
    "749,3,Danbom. Mr. Ernst Gilbert,male,34,1,1,347080,14.4,?,S,?,197,Stanton. IA,0",
    "750,3,Danbom. Mrs. Ernst Gilbert (Anna Sigrid Maria Brogren),female,28,1,1,347080,14.4,?,S,?,?,Stanton. IA,0",
    "751,3,Danoff. Mr. Yoto,male,27,0,0,349219,7.8958,?,S,?,?,Bulgaria Chicago. IL,0",
    "752,3,Dantcheff. Mr. Ristiu,male,25,0,0,349203,7.8958,?,S,?,?,Bulgaria Chicago. IL,0",
    "753,3,Davies. Mr. Alfred J,male,24,2,0,A/4 48871,24.15,?,S,?,?,West Bromwich. England Pontiac. MI,0",
    "754,3,Davies. Mr. Evan,male,22,0,0,SC/A4 23568,8.05,?,S,?,?,?,0",
    "755,3,Davies. Mr. John Samuel,male,21,2,0,A/4 48871,24.15,?,S,?,?,West Bromwich. England Pontiac. MI,0",
    "756,3,Davies. Mr. Joseph,male,17,2,0,A/4 48873,8.05,?,S,?,?,West Bromwich. England Pontiac. MI,0",
    "757,3,Davison. Mr. Thomas Henry,male,?,1,0,386525,16.1,?,S,?,?,Liverpool. England Bedford. OH,0",
    "758,3,Davison. Mrs. Thomas Henry (Mary E Finck),female,?,1,0,386525,16.1,?,S,16,?,Liverpool. England Bedford. OH,1",
    "759,3,de Messemaeker. Mr. Guillaume Joseph,male,36.5,1,0,345572,17.4,?,S,15,?,Tampico. MT,1",
    "760,3,de Messemaeker. Mrs. Guillaume Joseph (Emma),female,36,1,0,345572,17.4,?,S,13,?,Tampico. MT,1",
    "761,3,de Mulder. Mr. Theodore,male,30,0,0,345774,9.5,?,S,11,?,Belgium Detroit. MI,1",
    "762,3,de Pelsmaeker. Mr. Alfons,male,16,0,0,345778,9.5,?,S,?,?,?,0",
    "763,3,Dean. Master. Bertram Vere,male,1,1,2,C.A. 2315,20.575,?,S,10,?,Devon. England Wichita. KS,1",
    "764,3,Dean. Miss. Elizabeth Gladys 'Millvina',female,0.1667,1,2,C.A. 2315,20.575,?,S,10,?,Devon. England Wichita. KS,1",
    "765,3,Dean. Mr. Bertram Frank,male,26,1,2,C.A. 2315,20.575,?,S,?,?,Devon. England Wichita. KS,0",
    "766,3,Dean. Mrs. Bertram (Eva Georgetta Light),female,33,1,2,C.A. 2315,20.575,?,S,10,?,Devon. England Wichita. KS,1",
    "767,3,Delalic. Mr. Redjo,male,25,0,0,349250,7.8958,?,S,?,?,?,0",
    "768,3,Demetri. Mr. Marinko,male,?,0,0,349238,7.8958,?,S,?,?,?,0",
    "769,3,Denkoff. Mr. Mitto,male,?,0,0,349225,7.8958,?,S,?,?,Bulgaria Coon Rapids. IA,0",
    "770,3,Dennis. Mr. Samuel,male,22,0,0,A/5 21172,7.25,?,S,?,?,?,0",
    "771,3,Dennis. Mr. William,male,36,0,0,A/5 21175,7.25,?,S,?,?,?,0",
    "772,3,Devaney. Miss. Margaret Delia,female,19,0,0,330958,7.8792,?,Q,C,?,Kilmacowen. Co Sligo. Ireland New York. NY,1",
    "773,3,Dika. Mr. Mirko,male,17,0,0,349232,7.8958,?,S,?,?,?,0",
    "774,3,Dimic. Mr. Jovan,male,42,0,0,315088,8.6625,?,S,?,?,?,0",
    "775,3,Dintcheff. Mr. Valtcho,male,43,0,0,349226,7.8958,?,S,?,?,?,0",
    "776,3,Doharr. Mr. Tannous,male,?,0,0,2686,7.2292,?,C,?,?,?,0",
    "777,3,Dooley. Mr. Patrick,male,32,0,0,370376,7.75,?,Q,?,?,Ireland New York. NY,0",
    "778,3,Dorking. Mr. Edward Arthur,male,19,0,0,A/5. 10482,8.05,?,S,B,?,England Oglesby. IL,1",
    "779,3,Dowdell. Miss. Elizabeth,female,30,0,0,364516,12.475,?,S,13,?,Union Hill. NJ,1",
    "780,3,Doyle. Miss. Elizabeth,female,24,0,0,368702,7.75,?,Q,?,?,Ireland New York. NY,0",
    "781,3,Drapkin. Miss. Jennie,female,23,0,0,SOTON/OQ 392083,8.05,?,S,?,?,London New York. NY,1",
    "782,3,Drazenoic. Mr. Jozef,male,33,0,0,349241,7.8958,?,C,?,51,Austria Niagara Falls. NY,0",
    "783,3,Duane. Mr. Frank,male,65,0,0,336439,7.75,?,Q,?,?,?,0",
    "784,3,Duquemin. Mr. Joseph,male,24,0,0,S.O./P.P. 752,7.55,?,S,D,?,England Albion. NY,1",
    "785,3,Dyker. Mr. Adolf Fredrik,male,23,1,0,347072,13.9,?,S,?,?,West Haven. CT,0",
    "786,3,Dyker. Mrs. Adolf Fredrik (Anna Elisabeth Judith Andersson),female,22,1,0,347072,13.9,?,S,16,?,West Haven. CT,1",
    "787,3,Edvardsson. Mr. Gustaf Hjalmar,male,18,0,0,349912,7.775,?,S,?,?,Tofta. Sweden Joliet. IL,0",
    "788,3,Eklund. Mr. Hans Linus,male,16,0,0,347074,7.775,?,S,?,?,Karberg. Sweden Jerome Junction. AZ,0",
    "789,3,Ekstrom. Mr. Johan,male,45,0,0,347061,6.975,?,S,?,?,Effington Rut. SD,0",
    "790,3,Elias. Mr. Dibo,male,?,0,0,2674,7.225,?,C,?,?,?,0",
    "791,3,Elias. Mr. Joseph,male,39,0,2,2675,7.2292,?,C,?,?,Syria Ottawa. ON,0",
    "792,3,Elias. Mr. Joseph Jr,male,17,1,1,2690,7.2292,?,C,?,?,?,0",
    "793,3,Elias. Mr. Tannous,male,15,1,1,2695,7.2292,?,C,?,?,Syria,0",
    "794,3,Elsbury. Mr. William James,male,47,0,0,A/5 3902,7.25,?,S,?,?,Illinois. USA,0",
    "795,3,Emanuel. Miss. Virginia Ethel,female,5,0,0,364516,12.475,?,S,13,?,New York. NY,1",
    "796,3,Emir. Mr. Farred Chehab,male,?,0,0,2631,7.225,?,C,?,?,?,0",
    "797,3,Everett. Mr. Thomas James,male,40.5,0,0,C.A. 6212,15.1,?,S,?,187,?,0",
    "798,3,Farrell. Mr. James,male,40.5,0,0,367232,7.75,?,Q,?,68,Aughnacliff. Co Longford. Ireland New York. NY,0",
    "799,3,Finoli. Mr. Luigi,male,?,0,0,SOTON/O.Q. 3101308,7.05,?,S,15,?,Italy Philadelphia. PA,1",
    "800,3,Fischer. Mr. Eberhard Thelander,male,18,0,0,350036,7.7958,?,S,?,?,?,0",
    "801,3,Fleming. Miss. Honora,female,?,0,0,364859,7.75,?,Q,?,?,?,0",
    "802,3,Flynn. Mr. James,male,?,0,0,364851,7.75,?,Q,?,?,?,0",
    "803,3,Flynn. Mr. John,male,?,0,0,368323,6.95,?,Q,?,?,?,0",
    "804,3,Foley. Mr. Joseph,male,26,0,0,330910,7.8792,?,Q,?,?,Ireland Chicago. IL,0",
    "805,3,Foley. Mr. William,male,?,0,0,365235,7.75,?,Q,?,?,Ireland,0",
    "806,3,Foo. Mr. Choong,male,?,0,0,1601,56.4958,?,S,13,?,Hong Kong New York. NY,1",
    "807,3,Ford. Miss. Doolina Margaret 'Daisy',female,21,2,2,W./C. 6608,34.375,?,S,?,?,Rotherfield. Sussex. England Essex Co. MA,0",
    "808,3,Ford. Miss. Robina Maggie 'Ruby',female,9,2,2,W./C. 6608,34.375,?,S,?,?,Rotherfield. Sussex. England Essex Co. MA,0",
    "809,3,Ford. Mr. Arthur,male,?,0,0,A/5 1478,8.05,?,S,?,?,Bridgwater. Somerset. England,0",
    "810,3,Ford. Mr. Edward Watson,male,18,2,2,W./C. 6608,34.375,?,S,?,?,Rotherfield. Sussex. England Essex Co. MA,0",
    "811,3,Ford. Mr. William Neal,male,16,1,3,W./C. 6608,34.375,?,S,?,?,Rotherfield. Sussex. England Essex Co. MA,0",
    "812,3,Ford. Mrs. Edward (Margaret Ann Watson),female,48,1,3,W./C. 6608,34.375,?,S,?,?,Rotherfield. Sussex. England Essex Co. MA,0",
    "813,3,Fox. Mr. Patrick,male,?,0,0,368573,7.75,?,Q,?,?,Ireland New York. NY,0",
    "814,3,Franklin. Mr. Charles (Charles Fardon),male,?,0,0,SOTON/O.Q. 3101314,7.25,?,S,?,?,?,0",
    "815,3,Gallagher. Mr. Martin,male,25,0,0,36864,7.7417,?,Q,?,?,New York. NY,0",
    "816,3,Garfirth. Mr. John,male,?,0,0,358585,14.5,?,S,?,?,?,0",
    "817,3,Gheorgheff. Mr. Stanio,male,?,0,0,349254,7.8958,?,C,?,?,?,0",
    "818,3,Gilinski. Mr. Eliezer,male,22,0,0,14973,8.05,?,S,?,47,?,0",
    "819,3,Gilnagh. Miss. Katherine 'Katie',female,16,0,0,35851,7.7333,?,Q,16,?,Co Longford. Ireland New York. NY,1",
    "820,3,Glynn. Miss. Mary Agatha,female,?,0,0,335677,7.75,?,Q,13,?,Co Clare. Ireland Washington. DC,1",
    "821,3,Goldsmith. Master. Frank John William 'Frankie',male,9,0,2,363291,20.525,?,S,C D,?,Strood. Kent. England Detroit. MI,1",
    "822,3,Goldsmith. Mr. Frank John,male,33,1,1,363291,20.525,?,S,?,?,Strood. Kent. England Detroit. MI,0",
    "823,3,Goldsmith. Mr. Nathan,male,41,0,0,SOTON/O.Q. 3101263,7.85,?,S,?,?,Philadelphia. PA,0",
    "824,3,Goldsmith. Mrs. Frank John (Emily Alice Brown),female,31,1,1,363291,20.525,?,S,C D,?,Strood. Kent. England Detroit. MI,1",
    "825,3,Goncalves. Mr. Manuel Estanslas,male,38,0,0,SOTON/O.Q. 3101306,7.05,?,S,?,?,Portugal,0",
    "826,3,Goodwin. Master. Harold Victor,male,9,5,2,CA 2144,46.9,?,S,?,?,Wiltshire. England Niagara Falls. NY,0",
    "827,3,Goodwin. Master. Sidney Leonard,male,1,5,2,CA 2144,46.9,?,S,?,?,Wiltshire. England Niagara Falls. NY,0",
    "828,3,Goodwin. Master. William Frederick,male,11,5,2,CA 2144,46.9,?,S,?,?,Wiltshire. England Niagara Falls. NY,0",
    "829,3,Goodwin. Miss. Jessie Allis,female,10,5,2,CA 2144,46.9,?,S,?,?,Wiltshire. England Niagara Falls. NY,0",
    "830,3,Goodwin. Miss. Lillian Amy,female,16,5,2,CA 2144,46.9,?,S,?,?,Wiltshire. England Niagara Falls. NY,0",
    "831,3,Goodwin. Mr. Charles Edward,male,14,5,2,CA 2144,46.9,?,S,?,?,Wiltshire. England Niagara Falls. NY,0",
    "832,3,Goodwin. Mr. Charles Frederick,male,40,1,6,CA 2144,46.9,?,S,?,?,Wiltshire. England Niagara Falls. NY,0",
    "833,3,Goodwin. Mrs. Frederick (Augusta Tyler),female,43,1,6,CA 2144,46.9,?,S,?,?,Wiltshire. England Niagara Falls. NY,0",
    "834,3,Green. Mr. George Henry,male,51,0,0,21440,8.05,?,S,?,?,Dorking. Surrey. England,0",
    "835,3,Gronnestad. Mr. Daniel Danielsen,male,32,0,0,8471,8.3625,?,S,?,?,Foresvik. Norway Portland. ND,0",
    "836,3,Guest. Mr. Robert,male,?,0,0,376563,8.05,?,S,?,?,?,0",
    "837,3,Gustafsson. Mr. Alfred Ossian,male,20,0,0,7534,9.8458,?,S,?,?,Waukegan. Chicago. IL,0",
    "838,3,Gustafsson. Mr. Anders Vilhelm,male,37,2,0,3101276,7.925,?,S,?,98,Ruotsinphytaa. Finland New York. NY,0",
    "839,3,Gustafsson. Mr. Johan Birger,male,28,2,0,3101277,7.925,?,S,?,?,Ruotsinphytaa. Finland New York. NY,0",
    "840,3,Gustafsson. Mr. Karl Gideon,male,19,0,0,347069,7.775,?,S,?,?,Myren. Sweden New York. NY,0",
    "841,3,Haas. Miss. Aloisia,female,24,0,0,349236,8.85,?,S,?,?,?,0",
    "842,3,Hagardon. Miss. Kate,female,17,0,0,AQ/3. 30631,7.7333,?,Q,?,?,?,0",
    "843,3,Hagland. Mr. Ingvald Olai Olsen,male,?,1,0,65303,19.9667,?,S,?,?,?,0",
    "844,3,Hagland. Mr. Konrad Mathias Reiersen,male,?,1,0,65304,19.9667,?,S,?,?,?,0",
    "845,3,Hakkarainen. Mr. Pekka Pietari,male,28,1,0,STON/O2. 3101279,15.85,?,S,?,?,?,0",
    "846,3,Hakkarainen. Mrs. Pekka Pietari (Elin Matilda Dolck),female,24,1,0,STON/O2. 3101279,15.85,?,S,15,?,?,1",
    "847,3,Hampe. Mr. Leon,male,20,0,0,345769,9.5,?,S,?,?,?,0",
    "848,3,Hanna. Mr. Mansour,male,23.5,0,0,2693,7.2292,?,C,?,188,?,0",
    "849,3,Hansen. Mr. Claus Peter,male,41,2,0,350026,14.1083,?,S,?,?,?,0",
    "850,3,Hansen. Mr. Henrik Juul,male,26,1,0,350025,7.8542,?,S,?,?,?,0",
    "851,3,Hansen. Mr. Henry Damsgaard,male,21,0,0,350029,7.8542,?,S,?,69,?,0",
    "852,3,Hansen. Mrs. Claus Peter (Jennie L Howard),female,45,1,0,350026,14.1083,?,S,11,?,?,1",
    "853,3,Harknett. Miss. Alice Phoebe,female,?,0,0,W./C. 6609,7.55,?,S,?,?,?,0",
    "854,3,Harmer. Mr. Abraham (David Lishin),male,25,0,0,374887,7.25,?,S,B,?,?,0",
    "855,3,Hart. Mr. Henry,male,?,0,0,394140,6.8583,?,Q,?,?,?,0",
    "856,3,Hassan. Mr. Houssein G N,male,11,0,0,2699,18.7875,?,C,?,?,?,0",
    "857,3,Healy. Miss. Hanora 'Nora',female,?,0,0,370375,7.75,?,Q,16,?,?,1",
    "858,3,Hedman. Mr. Oskar Arvid,male,27,0,0,347089,6.975,?,S,15,?,?,1",
    "859,3,Hee. Mr. Ling,male,?,0,0,1601,56.4958,?,S,C,?,?,1",
    "860,3,Hegarty. Miss. Hanora 'Nora',female,18,0,0,365226,6.75,?,Q,?,?,?,0",
    "861,3,Heikkinen. Miss. Laina,female,26,0,0,STON/O2. 3101282,7.925,?,S,?,?,?,1",
    "862,3,Heininen. Miss. Wendla Maria,female,23,0,0,STON/O2. 3101290,7.925,?,S,?,?,?,0",
    "863,3,Hellstrom. Miss. Hilda Maria,female,22,0,0,7548,8.9625,?,S,C,?,?,1",
    "864,3,Hendekovic. Mr. Ignjac,male,28,0,0,349243,7.8958,?,S,?,306,?,0",
    "865,3,Henriksson. Miss. Jenny Lovisa,female,28,0,0,347086,7.775,?,S,?,?,?,0",
    "866,3,Henry. Miss. Delia,female,?,0,0,382649,7.75,?,Q,?,?,?,0",
    "867,3,Hirvonen. Miss. Hildur E,female,2,0,1,3101298,12.2875,?,S,15,?,?,1",
    "868,3,Hirvonen. Mrs. Alexander (Helga E Lindqvist),female,22,1,1,3101298,12.2875,?,S,15,?,?,1",
    "869,3,Holm. Mr. John Fredrik Alexander,male,43,0,0,C 7075,6.45,?,S,?,?,?,0",
    "870,3,Holthen. Mr. Johan Martin,male,28,0,0,C 4001,22.525,?,S,?,?,?,0",
    "871,3,Honkanen. Miss. Eliina,female,27,0,0,STON/O2. 3101283,7.925,?,S,?,?,?,1",
    "872,3,Horgan. Mr. John,male,?,0,0,370377,7.75,?,Q,?,?,?,0",
    "873,3,Howard. Miss. May Elizabeth,female,?,0,0,A. 2. 39186,8.05,?,S,C,?,?,1",
    "874,3,Humblen. Mr. Adolf Mathias Nicolai Olsen,male,42,0,0,348121,7.65,F G63,S,?,120,?,0",
    "875,3,Hyman. Mr. Abraham,male,?,0,0,3470,7.8875,?,S,C,?,?,1",
    "876,3,Ibrahim Shawah. Mr. Yousseff,male,30,0,0,2685,7.2292,?,C,?,?,?,0",
    "877,3,Ilieff. Mr. Ylio,male,?,0,0,349220,7.8958,?,S,?,?,?,0",
    "878,3,Ilmakangas. Miss. Ida Livija,female,27,1,0,STON/O2. 3101270,7.925,?,S,?,?,?,0",
    "879,3,Ilmakangas. Miss. Pieta Sofia,female,25,1,0,STON/O2. 3101271,7.925,?,S,?,?,?,0",
    "880,3,Ivanoff. Mr. Kanio,male,?,0,0,349201,7.8958,?,S,?,?,?,0",
    "881,3,Jalsevac. Mr. Ivan,male,29,0,0,349240,7.8958,?,C,15,?,?,1",
    "882,3,Jansson. Mr. Carl Olof,male,21,0,0,350034,7.7958,?,S,A,?,?,1",
    "883,3,Jardin. Mr. Jose Neto,male,?,0,0,SOTON/O.Q. 3101305,7.05,?,S,?,?,?,0",
    "884,3,Jensen. Mr. Hans Peder,male,20,0,0,350050,7.8542,?,S,?,?,?,0",
    "885,3,Jensen. Mr. Niels Peder,male,48,0,0,350047,7.8542,?,S,?,?,?,0",
    "886,3,Jensen. Mr. Svend Lauritz,male,17,1,0,350048,7.0542,?,S,?,?,?,0",
    "887,3,Jermyn. Miss. Annie,female,?,0,0,14313,7.75,?,Q,D,?,?,1",
    "888,3,Johannesen-Bratthammer. Mr. Bernt,male,?,0,0,65306,8.1125,?,S,13,?,?,1",
    "889,3,Johanson. Mr. Jakob Alfred,male,34,0,0,3101264,6.4958,?,S,?,143,?,0",
    "890,3,Johansson Palmquist. Mr. Oskar Leander,male,26,0,0,347070,7.775,?,S,15,?,?,1",
    "891,3,Johansson. Mr. Erik,male,22,0,0,350052,7.7958,?,S,?,156,?,0",
    "892,3,Johansson. Mr. Gustaf Joel,male,33,0,0,7540,8.6542,?,S,?,285,?,0",
    "893,3,Johansson. Mr. Karl Johan,male,31,0,0,347063,7.775,?,S,?,?,?,0",
    "894,3,Johansson. Mr. Nils,male,29,0,0,347467,7.8542,?,S,?,?,?,0",
    "895,3,Johnson. Master. Harold Theodor,male,4,1,1,347742,11.1333,?,S,15,?,?,1",
    "896,3,Johnson. Miss. Eleanor Ileen,female,1,1,1,347742,11.1333,?,S,15,?,?,1",
    "897,3,Johnson. Mr. Alfred,male,49,0,0,LINE,0,?,S,?,?,?,0",
    "898,3,Johnson. Mr. Malkolm Joackim,male,33,0,0,347062,7.775,?,S,?,37,?,0",
    "899,3,Johnson. Mr. William Cahoone Jr,male,19,0,0,LINE,0,?,S,?,?,?,0",
    "900,3,Johnson. Mrs. Oscar W (Elisabeth Vilhelmina Berg),female,27,0,2,347742,11.1333,?,S,15,?,?,1",
    "901,3,Johnston. Master. William Arthur 'Willie',male,?,1,2,W./C. 6607,23.45,?,S,?,?,?,0",
    "902,3,Johnston. Miss. Catherine Helen 'Carrie',female,?,1,2,W./C. 6607,23.45,?,S,?,?,?,0",
    "903,3,Johnston. Mr. Andrew G,male,?,1,2,W./C. 6607,23.45,?,S,?,?,?,0",
    "904,3,Johnston. Mrs. Andrew G (Elizabeth 'Lily' Watson),female,?,1,2,W./C. 6607,23.45,?,S,?,?,?,0",
    "905,3,Jonkoff. Mr. Lalio,male,23,0,0,349204,7.8958,?,S,?,?,?,0",
    "906,3,Jonsson. Mr. Carl,male,32,0,0,350417,7.8542,?,S,15,?,?,1",
    "907,3,Jonsson. Mr. Nils Hilding,male,27,0,0,350408,7.8542,?,S,?,?,?,0",
    "908,3,Jussila. Miss. Katriina,female,20,1,0,4136,9.825,?,S,?,?,?,0",
    "909,3,Jussila. Miss. Mari Aina,female,21,1,0,4137,9.825,?,S,?,?,?,0",
    "910,3,Jussila. Mr. Eiriik,male,32,0,0,STON/O 2. 3101286,7.925,?,S,15,?,?,1",
    "911,3,Kallio. Mr. Nikolai Erland,male,17,0,0,STON/O 2. 3101274,7.125,?,S,?,?,?,0",
    "912,3,Kalvik. Mr. Johannes Halvorsen,male,21,0,0,8475,8.4333,?,S,?,?,?,0",
    "913,3,Karaic. Mr. Milan,male,30,0,0,349246,7.8958,?,S,?,?,?,0",
    "914,3,Karlsson. Mr. Einar Gervasius,male,21,0,0,350053,7.7958,?,S,13,?,?,1",
    "915,3,Karlsson. Mr. Julius Konrad Eugen,male,33,0,0,347465,7.8542,?,S,?,?,?,0",
    "916,3,Karlsson. Mr. Nils August,male,22,0,0,350060,7.5208,?,S,?,?,?,0",
    "917,3,Karun. Miss. Manca,female,4,0,1,349256,13.4167,?,C,15,?,?,1",
    "918,3,Karun. Mr. Franz,male,39,0,1,349256,13.4167,?,C,15,?,?,1",
    "919,3,Kassem. Mr. Fared,male,?,0,0,2700,7.2292,?,C,?,?,?,0",
    "920,3,Katavelas. Mr. Vassilios ('Catavelas Vassilios'),male,18.5,0,0,2682,7.2292,?,C,?,58,?,0",
    "921,3,Keane. Mr. Andrew 'Andy',male,?,0,0,12460,7.75,?,Q,?,?,?,0",
    "922,3,Keefe. Mr. Arthur,male,?,0,0,323592,7.25,?,S,A,?,?,0",
    "923,3,Kelly. Miss. Anna Katherine 'Annie Kate',female,?,0,0,9234,7.75,?,Q,16,?,?,1",
    "924,3,Kelly. Miss. Mary,female,?,0,0,14312,7.75,?,Q,D,?,?,1",
    "925,3,Kelly. Mr. James,male,34.5,0,0,330911,7.8292,?,Q,?,70,?,0",
    "926,3,Kelly. Mr. James,male,44,0,0,363592,8.05,?,S,?,?,?,0",
    "927,3,Kennedy. Mr. John,male,?,0,0,368783,7.75,?,Q,?,?,?,1",
    "928,3,Khalil. Mr. Betros,male,?,1,0,2660,14.4542,?,C,?,?,?,0",
    "929,3,Khalil. Mrs. Betros (Zahie 'Maria' Elias),female,?,1,0,2660,14.4542,?,C,?,?,?,0",
    "930,3,Kiernan. Mr. John,male,?,1,0,367227,7.75,?,Q,?,?,?,0",
    "931,3,Kiernan. Mr. Philip,male,?,1,0,367229,7.75,?,Q,?,?,?,0",
    "932,3,Kilgannon. Mr. Thomas J,male,?,0,0,36865,7.7375,?,Q,?,?,?,0",
    "933,3,Kink. Miss. Maria,female,22,2,0,315152,8.6625,?,S,?,?,?,0",
    "934,3,Kink. Mr. Vincenz,male,26,2,0,315151,8.6625,?,S,?,?,?,0",
    "935,3,Kink-Heilmann. Miss. Luise Gretchen,female,4,0,2,315153,22.025,?,S,2,?,?,1",
    "936,3,Kink-Heilmann. Mr. Anton,male,29,3,1,315153,22.025,?,S,2,?,?,1",
    "937,3,Kink-Heilmann. Mrs. Anton (Luise Heilmann),female,26,1,1,315153,22.025,?,S,2,?,?,1",
    "938,3,Klasen. Miss. Gertrud Emilia,female,1,1,1,350405,12.1833,?,S,?,?,?,0",
    "939,3,Klasen. Mr. Klas Albin,male,18,1,1,350404,7.8542,?,S,?,?,?,0",
    "940,3,Klasen. Mrs. (Hulda Kristina Eugenia Lofqvist),female,36,0,2,350405,12.1833,?,S,?,?,?,0",
    "941,3,Kraeff. Mr. Theodor,male,?,0,0,349253,7.8958,?,C,?,?,?,0",
    "942,3,Krekorian. Mr. Neshan,male,25,0,0,2654,7.2292,F E57,C,10,?,?,1",
    "943,3,Lahoud. Mr. Sarkis,male,?,0,0,2624,7.225,?,C,?,?,?,0",
    "944,3,Laitinen. Miss. Kristina Sofia,female,37,0,0,4135,9.5875,?,S,?,?,?,0",
    "945,3,Laleff. Mr. Kristo,male,?,0,0,349217,7.8958,?,S,?,?,?,0",
    "946,3,Lam. Mr. Ali,male,?,0,0,1601,56.4958,?,S,C,?,?,1",
    "947,3,Lam. Mr. Len,male,?,0,0,1601,56.4958,?,S,?,?,?,0",
    "948,3,Landergren. Miss. Aurora Adelia,female,22,0,0,C 7077,7.25,?,S,13,?,?,1",
    "949,3,Lane. Mr. Patrick,male,?,0,0,7935,7.75,?,Q,?,?,?,0",
    "950,3,Lang. Mr. Fang,male,26,0,0,1601,56.4958,?,S,14,?,?,1",
    "951,3,Larsson. Mr. August Viktor,male,29,0,0,7545,9.4833,?,S,?,?,?,0",
    "952,3,Larsson. Mr. Bengt Edvin,male,29,0,0,347067,7.775,?,S,?,?,?,0",
    "953,3,Larsson-Rondberg. Mr. Edvard A,male,22,0,0,347065,7.775,?,S,?,?,?,0",
    "954,3,Leeni. Mr. Fahim ('Philip Zenni'),male,22,0,0,2620,7.225,?,C,6,?,?,1",
    "955,3,Lefebre. Master. Henry Forbes,male,?,3,1,4133,25.4667,?,S,?,?,?,0",
    "956,3,Lefebre. Miss. Ida,female,?,3,1,4133,25.4667,?,S,?,?,?,0",
    "957,3,Lefebre. Miss. Jeannie,female,?,3,1,4133,25.4667,?,S,?,?,?,0",
    "958,3,Lefebre. Miss. Mathilde,female,?,3,1,4133,25.4667,?,S,?,?,?,0",
    "959,3,Lefebre. Mrs. Frank (Frances),female,?,0,4,4133,25.4667,?,S,?,?,?,0",
    "960,3,Leinonen. Mr. Antti Gustaf,male,32,0,0,STON/O 2. 3101292,7.925,?,S,?,?,?,0",
    "961,3,Lemberopolous. Mr. Peter L,male,34.5,0,0,2683,6.4375,?,C,?,196,?,0",
    "962,3,Lennon. Miss. Mary,female,?,1,0,370371,15.5,?,Q,?,?,?,0",
    "963,3,Lennon. Mr. Denis,male,?,1,0,370371,15.5,?,Q,?,?,?,0",
    "964,3,Leonard. Mr. Lionel,male,36,0,0,LINE,0,?,S,?,?,?,0",
    "965,3,Lester. Mr. James,male,39,0,0,A/4 48871,24.15,?,S,?,?,?,0",
    "966,3,Lievens. Mr. Rene Aime,male,24,0,0,345781,9.5,?,S,?,?,?,0",
    "967,3,Lindahl. Miss. Agda Thorilda Viktoria,female,25,0,0,347071,7.775,?,S,?,?,?,0",
    "968,3,Lindblom. Miss. Augusta Charlotta,female,45,0,0,347073,7.75,?,S,?,?,?,0",
    "969,3,Lindell. Mr. Edvard Bengtsson,male,36,1,0,349910,15.55,?,S,A,?,?,0",
    "970,3,Lindell. Mrs. Edvard Bengtsson (Elin Gerda Persson),female,30,1,0,349910,15.55,?,S,A,?,?,0",
    "971,3,Lindqvist. Mr. Eino William,male,20,1,0,STON/O 2. 3101285,7.925,?,S,15,?,?,1",
    "972,3,Linehan. Mr. Michael,male,?,0,0,330971,7.8792,?,Q,?,?,?,0",
    "973,3,Ling. Mr. Lee,male,28,0,0,1601,56.4958,?,S,?,?,?,0",
    "974,3,Lithman. Mr. Simon,male,?,0,0,S.O./P.P. 251,7.55,?,S,?,?,?,0",
    "975,3,Lobb. Mr. William Arthur,male,30,1,0,A/5. 3336,16.1,?,S,?,?,?,0",
    "976,3,Lobb. Mrs. William Arthur (Cordelia K Stanlick),female,26,1,0,A/5. 3336,16.1,?,S,?,?,?,0",
    "977,3,Lockyer. Mr. Edward,male,?,0,0,1222,7.8792,?,S,?,153,?,0",
    "978,3,Lovell. Mr. John Hall ('Henry'),male,20.5,0,0,A/5 21173,7.25,?,S,?,?,?,0",
    "979,3,Lulic. Mr. Nikola,male,27,0,0,315098,8.6625,?,S,15,?,?,1",
    "980,3,Lundahl. Mr. Johan Svensson,male,51,0,0,347743,7.0542,?,S,?,?,?,0",
    "981,3,Lundin. Miss. Olga Elida,female,23,0,0,347469,7.8542,?,S,10,?,?,1",
    "982,3,Lundstrom. Mr. Thure Edvin,male,32,0,0,350403,7.5792,?,S,15,?,?,1",
    "983,3,Lyntakoff. Mr. Stanko,male,?,0,0,349235,7.8958,?,S,?,?,?,0",
    "984,3,MacKay. Mr. George William,male,?,0,0,C.A. 42795,7.55,?,S,?,?,?,0",
    "985,3,Madigan. Miss. Margaret 'Maggie',female,?,0,0,370370,7.75,?,Q,15,?,?,1",
    "986,3,Madsen. Mr. Fridtjof Arne,male,24,0,0,C 17369,7.1417,?,S,13,?,?,1",
    "987,3,Maenpaa. Mr. Matti Alexanteri,male,22,0,0,STON/O 2. 3101275,7.125,?,S,?,?,?,0",
    "988,3,Mahon. Miss. Bridget Delia,female,?,0,0,330924,7.8792,?,Q,?,?,?,0",
    "989,3,Mahon. Mr. John,male,?,0,0,AQ/4 3130,7.75,?,Q,?,?,?,0",
    "990,3,Maisner. Mr. Simon,male,?,0,0,A/S 2816,8.05,?,S,?,?,?,0",
    "991,3,Makinen. Mr. Kalle Edvard,male,29,0,0,STON/O 2. 3101268,7.925,?,S,?,?,?,0",
    "992,3,Mamee. Mr. Hanna,male,?,0,0,2677,7.2292,?,C,15,?,?,1",
    "993,3,Mangan. Miss. Mary,female,30.5,0,0,364850,7.75,?,Q,?,61,?,0",
    "994,3,Mannion. Miss. Margareth,female,?,0,0,36866,7.7375,?,Q,16,?,?,1",
    "995,3,Mardirosian. Mr. Sarkis,male,?,0,0,2655,7.2292,F E46,C,?,?,?,0",
    "996,3,Markoff. Mr. Marin,male,35,0,0,349213,7.8958,?,C,?,?,?,0",
    "997,3,Markun. Mr. Johann,male,33,0,0,349257,7.8958,?,S,?,?,?,0",
    "998,3,Masselmani. Mrs. Fatima,female,?,0,0,2649,7.225,?,C,C,?,?,1",
    "999,3,Matinoff. Mr. Nicola,male,?,0,0,349255,7.8958,?,C,?,?,?,0",
    "1000,3,McCarthy. Miss. Catherine 'Katie',female,?,0,0,383123,7.75,?,Q,15 16,?,?,1",
    "1001,3,McCormack. Mr. Thomas Joseph,male,?,0,0,367228,7.75,?,Q,?,?,?,1",
    "1002,3,McCoy. Miss. Agnes,female,?,2,0,367226,23.25,?,Q,16,?,?,1",
    "1003,3,McCoy. Miss. Alicia,female,?,2,0,367226,23.25,?,Q,16,?,?,1",
    "1004,3,McCoy. Mr. Bernard,male,?,2,0,367226,23.25,?,Q,16,?,?,1",
    "1005,3,McDermott. Miss. Brigdet Delia,female,?,0,0,330932,7.7875,?,Q,13,?,?,1",
    "1006,3,McEvoy. Mr. Michael,male,?,0,0,36568,15.5,?,Q,?,?,?,0",
    "1007,3,McGovern. Miss. Mary,female,?,0,0,330931,7.8792,?,Q,13,?,?,1",
    "1008,3,McGowan. Miss. Anna 'Annie',female,15,0,0,330923,8.0292,?,Q,?,?,?,1",
    "1009,3,McGowan. Miss. Katherine,female,35,0,0,9232,7.75,?,Q,?,?,?,0",
    "1010,3,McMahon. Mr. Martin,male,?,0,0,370372,7.75,?,Q,?,?,?,0",
    "1011,3,McNamee. Mr. Neal,male,24,1,0,376566,16.1,?,S,?,?,?,0",
    "1012,3,McNamee. Mrs. Neal (Eileen O'Leary),female,19,1,0,376566,16.1,?,S,?,53,?,0",
    "1013,3,McNeill. Miss. Bridget,female,?,0,0,370368,7.75,?,Q,?,?,?,0",
    "1014,3,Meanwell. Miss. (Marion Ogden),female,?,0,0,SOTON/O.Q. 392087,8.05,?,S,?,?,?,0",
    "1015,3,Meek. Mrs. Thomas (Annie Louise Rowley),female,?,0,0,343095,8.05,?,S,?,?,?,0",
    "1016,3,Meo. Mr. Alfonzo,male,55.5,0,0,A.5. 11206,8.05,?,S,?,201,?,0",
    "1017,3,Mernagh. Mr. Robert,male,?,0,0,368703,7.75,?,Q,?,?,?,0",
    "1018,3,Midtsjo. Mr. Karl Albert,male,21,0,0,345501,7.775,?,S,15,?,?,1",
    "1019,3,Miles. Mr. Frank,male,?,0,0,359306,8.05,?,S,?,?,?,0",
    "1020,3,Mineff. Mr. Ivan,male,24,0,0,349233,7.8958,?,S,?,?,?,0",
    "1021,3,Minkoff. Mr. Lazar,male,21,0,0,349211,7.8958,?,S,?,?,?,0",
    "1022,3,Mionoff. Mr. Stoytcho,male,28,0,0,349207,7.8958,?,S,?,?,?,0",
    "1023,3,Mitkoff. Mr. Mito,male,?,0,0,349221,7.8958,?,S,?,?,?,0",
    "1024,3,Mockler. Miss. Helen Mary 'Ellie',female,?,0,0,330980,7.8792,?,Q,16,?,?,1",
    "1025,3,Moen. Mr. Sigurd Hansen,male,25,0,0,348123,7.65,F G73,S,?,309,?,0",
    "1026,3,Moor. Master. Meier,male,6,0,1,392096,12.475,E121,S,14,?,?,1",
    "1027,3,Moor. Mrs. (Beila),female,27,0,1,392096,12.475,E121,S,14,?,?,1",
    "1028,3,Moore. Mr. Leonard Charles,male,?,0,0,A4. 54510,8.05,?,S,?,?,?,0",
    "1029,3,Moran. Miss. Bertha,female,?,1,0,371110,24.15,?,Q,16,?,?,1",
    "1030,3,Moran. Mr. Daniel J,male,?,1,0,371110,24.15,?,Q,?,?,?,0",
    "1031,3,Moran. Mr. James,male,?,0,0,330877,8.4583,?,Q,?,?,?,0",
    "1032,3,Morley. Mr. William,male,34,0,0,364506,8.05,?,S,?,?,?,0",
    "1033,3,Morrow. Mr. Thomas Rowan,male,?,0,0,372622,7.75,?,Q,?,?,?,0",
    "1034,3,Moss. Mr. Albert Johan,male,?,0,0,312991,7.775,?,S,B,?,?,1",
    "1035,3,Moubarek. Master. Gerios,male,?,1,1,2661,15.2458,?,C,C,?,?,1",
    "1036,3,Moubarek. Master. Halim Gonios ('William George'),male,?,1,1,2661,15.2458,?,C,C,?,?,1",
    "1037,3,Moubarek. Mrs. George (Omine 'Amenia' Alexander),female,?,0,2,2661,15.2458,?,C,C,?,?,1",
    "1038,3,Moussa. Mrs. (Mantoura Boulos),female,?,0,0,2626,7.2292,?,C,?,?,?,1",
    "1039,3,Moutal. Mr. Rahamin Haim,male,?,0,0,374746,8.05,?,S,?,?,?,0",
    "1040,3,Mullens. Miss. Katherine 'Katie',female,?,0,0,35852,7.7333,?,Q,16,?,?,1",
    "1041,3,Mulvihill. Miss. Bertha E,female,24,0,0,382653,7.75,?,Q,15,?,?,1",
    "1042,3,Murdlin. Mr. Joseph,male,?,0,0,A./5. 3235,8.05,?,S,?,?,?,0",
    "1043,3,Murphy. Miss. Katherine 'Kate',female,?,1,0,367230,15.5,?,Q,16,?,?,1",
    "1044,3,Murphy. Miss. Margaret Jane,female,?,1,0,367230,15.5,?,Q,16,?,?,1",
    "1045,3,Murphy. Miss. Nora,female,?,0,0,36568,15.5,?,Q,16,?,?,1",
    "1046,3,Myhrman. Mr. Pehr Fabian Oliver Malkolm,male,18,0,0,347078,7.75,?,S,?,?,?,0",
    "1047,3,Naidenoff. Mr. Penko,male,22,0,0,349206,7.8958,?,S,?,?,?,0",
    "1048,3,Najib. Miss. Adele Kiamie 'Jane',female,15,0,0,2667,7.225,?,C,C,?,?,1",
    "1049,3,Nakid. Miss. Maria ('Mary'),female,1,0,2,2653,15.7417,?,C,C,?,?,1",
    "1050,3,Nakid. Mr. Sahid,male,20,1,1,2653,15.7417,?,C,C,?,?,1",
    "1051,3,Nakid. Mrs. Said (Waika 'Mary' Mowad),female,19,1,1,2653,15.7417,?,C,C,?,?,1",
    "1052,3,Nancarrow. Mr. William Henry,male,33,0,0,A./5. 3338,8.05,?,S,?,?,?,0",
    "1053,3,Nankoff. Mr. Minko,male,?,0,0,349218,7.8958,?,S,?,?,?,0",
    "1054,3,Nasr. Mr. Mustafa,male,?,0,0,2652,7.2292,?,C,?,?,?,0",
    "1055,3,Naughton. Miss. Hannah,female,?,0,0,365237,7.75,?,Q,?,?,?,0",
    "1056,3,Nenkoff. Mr. Christo,male,?,0,0,349234,7.8958,?,S,?,?,?,0",
    "1057,3,Nicola-Yarred. Master. Elias,male,12,1,0,2651,11.2417,?,C,C,?,?,1",
    "1058,3,Nicola-Yarred. Miss. Jamila,female,14,1,0,2651,11.2417,?,C,C,?,?,1",
    "1059,3,Nieminen. Miss. Manta Josefina,female,29,0,0,3101297,7.925,?,S,?,?,?,0",
    "1060,3,Niklasson. Mr. Samuel,male,28,0,0,363611,8.05,?,S,?,?,?,0",
    "1061,3,Nilsson. Miss. Berta Olivia,female,18,0,0,347066,7.775,?,S,D,?,?,1",
    "1062,3,Nilsson. Miss. Helmina Josefina,female,26,0,0,347470,7.8542,?,S,13,?,?,1",
    "1063,3,Nilsson. Mr. August Ferdinand,male,21,0,0,350410,7.8542,?,S,?,?,?,0",
    "1064,3,Nirva. Mr. Iisakki Antino Aijo,male,41,0,0,SOTON/O2 3101272,7.125,?,S,?,?,Finland Sudbury. ON,0",
    "1065,3,Niskanen. Mr. Juha,male,39,0,0,STON/O 2. 3101289,7.925,?,S,9,?,?,1",
    "1066,3,Nosworthy. Mr. Richard Cater,male,21,0,0,A/4. 39886,7.8,?,S,?,?,?,0",
    "1067,3,Novel. Mr. Mansouer,male,28.5,0,0,2697,7.2292,?,C,?,181,?,0",
    "1068,3,Nysten. Miss. Anna Sofia,female,22,0,0,347081,7.75,?,S,13,?,?,1",
    "1069,3,Nysveen. Mr. Johan Hansen,male,61,0,0,345364,6.2375,?,S,?,?,?,0",
    "1070,3,O'Brien. Mr. Thomas,male,?,1,0,370365,15.5,?,Q,?,?,?,0",
    "1071,3,O'Brien. Mr. Timothy,male,?,0,0,330979,7.8292,?,Q,?,?,?,0",
    "1072,3,O'Brien. Mrs. Thomas (Johanna 'Hannah' Godfrey),female,?,1,0,370365,15.5,?,Q,?,?,?,1",
    "1073,3,O'Connell. Mr. Patrick D,male,?,0,0,334912,7.7333,?,Q,?,?,?,0",
    "1074,3,O'Connor. Mr. Maurice,male,?,0,0,371060,7.75,?,Q,?,?,?,0",
    "1075,3,O'Connor. Mr. Patrick,male,?,0,0,366713,7.75,?,Q,?,?,?,0",
    "1076,3,Odahl. Mr. Nils Martin,male,23,0,0,7267,9.225,?,S,?,?,?,0",
    "1077,3,O'Donoghue. Ms. Bridget,female,?,0,0,364856,7.75,?,Q,?,?,?,0",
    "1078,3,O'Driscoll. Miss. Bridget,female,?,0,0,14311,7.75,?,Q,D,?,?,1",
    "1079,3,O'Dwyer. Miss. Ellen 'Nellie',female,?,0,0,330959,7.8792,?,Q,?,?,?,1",
    "1080,3,Ohman. Miss. Velin,female,22,0,0,347085,7.775,?,S,C,?,?,1",
    "1081,3,O'Keefe. Mr. Patrick,male,?,0,0,368402,7.75,?,Q,B,?,?,1",
    "1082,3,O'Leary. Miss. Hanora 'Norah',female,?,0,0,330919,7.8292,?,Q,13,?,?,1",
    "1083,3,Olsen. Master. Artur Karl,male,9,0,1,C 17368,3.1708,?,S,13,?,?,1",
    "1084,3,Olsen. Mr. Henry Margido,male,28,0,0,C 4001,22.525,?,S,?,173,?,0",
    "1085,3,Olsen. Mr. Karl Siegwart Andreas,male,42,0,1,4579,8.4042,?,S,?,?,?,0",
    "1086,3,Olsen. Mr. Ole Martin,male,?,0,0,Fa 265302,7.3125,?,S,?,?,?,0",
    "1087,3,Olsson. Miss. Elina,female,31,0,0,350407,7.8542,?,S,?,?,?,0",
    "1088,3,Olsson. Mr. Nils Johan Goransson,male,28,0,0,347464,7.8542,?,S,?,?,?,0",
    "1089,3,Olsson. Mr. Oscar Wilhelm,male,32,0,0,347079,7.775,?,S,A,?,?,1",
    "1090,3,Olsvigen. Mr. Thor Anderson,male,20,0,0,6563,9.225,?,S,?,89,Oslo. Norway Cameron. WI,0",
    "1091,3,Oreskovic. Miss. Jelka,female,23,0,0,315085,8.6625,?,S,?,?,?,0",
    "1092,3,Oreskovic. Miss. Marija,female,20,0,0,315096,8.6625,?,S,?,?,?,0",
    "1093,3,Oreskovic. Mr. Luka,male,20,0,0,315094,8.6625,?,S,?,?,?,0",
    "1094,3,Osen. Mr. Olaf Elon,male,16,0,0,7534,9.2167,?,S,?,?,?,0",
    "1095,3,Osman. Mrs. Mara,female,31,0,0,349244,8.6833,?,S,?,?,?,1",
    "1096,3,O'Sullivan. Miss. Bridget Mary,female,?,0,0,330909,7.6292,?,Q,?,?,?,0",
    "1097,3,Palsson. Master. Gosta Leonard,male,2,3,1,349909,21.075,?,S,?,4,?,0",
    "1098,3,Palsson. Master. Paul Folke,male,6,3,1,349909,21.075,?,S,?,?,?,0",
    "1099,3,Palsson. Miss. Stina Viola,female,3,3,1,349909,21.075,?,S,?,?,?,0",
    "1100,3,Palsson. Miss. Torborg Danira,female,8,3,1,349909,21.075,?,S,?,?,?,0",
    "1101,3,Palsson. Mrs. Nils (Alma Cornelia Berglund),female,29,0,4,349909,21.075,?,S,?,206,?,0",
    "1102,3,Panula. Master. Eino Viljami,male,1,4,1,3101295,39.6875,?,S,?,?,?,0",
    "1103,3,Panula. Master. Juha Niilo,male,7,4,1,3101295,39.6875,?,S,?,?,?,0",
    "1104,3,Panula. Master. Urho Abraham,male,2,4,1,3101295,39.6875,?,S,?,?,?,0",
    "1105,3,Panula. Mr. Ernesti Arvid,male,16,4,1,3101295,39.6875,?,S,?,?,?,0",
    "1106,3,Panula. Mr. Jaako Arnold,male,14,4,1,3101295,39.6875,?,S,?,?,?,0",
    "1107,3,Panula. Mrs. Juha (Maria Emilia Ojala),female,41,0,5,3101295,39.6875,?,S,?,?,?,0",
    "1108,3,Pasic. Mr. Jakob,male,21,0,0,315097,8.6625,?,S,?,?,?,0",
    "1109,3,Patchett. Mr. George,male,19,0,0,358585,14.5,?,S,?,?,?,0",
    "1110,3,Paulner. Mr. Uscher,male,?,0,0,3411,8.7125,?,C,?,?,?,0",
    "1111,3,Pavlovic. Mr. Stefo,male,32,0,0,349242,7.8958,?,S,?,?,?,0",
    "1112,3,Peacock. Master. Alfred Edward,male,0.75,1,1,SOTON/O.Q. 3101315,13.775,?,S,?,?,?,0",
    "1113,3,Peacock. Miss. Treasteall,female,3,1,1,SOTON/O.Q. 3101315,13.775,?,S,?,?,?,0",
    "1114,3,Peacock. Mrs. Benjamin (Edith Nile),female,26,0,2,SOTON/O.Q. 3101315,13.775,?,S,?,?,?,0",
    "1115,3,Pearce. Mr. Ernest,male,?,0,0,343271,7,?,S,?,?,?,0",
    "1116,3,Pedersen. Mr. Olaf,male,?,0,0,345498,7.775,?,S,?,?,?,0",
    "1117,3,Peduzzi. Mr. Joseph,male,?,0,0,A/5 2817,8.05,?,S,?,?,?,0",
    "1118,3,Pekoniemi. Mr. Edvard,male,21,0,0,STON/O 2. 3101294,7.925,?,S,?,?,?,0",
    "1119,3,Peltomaki. Mr. Nikolai Johannes,male,25,0,0,STON/O 2. 3101291,7.925,?,S,?,?,?,0",
    "1120,3,Perkin. Mr. John Henry,male,22,0,0,A/5 21174,7.25,?,S,?,?,?,0",
    "1121,3,Persson. Mr. Ernst Ulrik,male,25,1,0,347083,7.775,?,S,15,?,?,1",
    "1122,3,Peter. Master. Michael J,male,?,1,1,2668,22.3583,?,C,C,?,?,1",
    "1123,3,Peter. Miss. Anna,female,?,1,1,2668,22.3583,F E69,C,D,?,?,1",
    "1124,3,Peter. Mrs. Catherine (Catherine Rizk),female,?,0,2,2668,22.3583,?,C,D,?,?,1",
    "1125,3,Peters. Miss. Katie,female,?,0,0,330935,8.1375,?,Q,?,?,?,0",
    "1126,3,Petersen. Mr. Marius,male,24,0,0,342441,8.05,?,S,?,?,?,0",
    "1127,3,Petranec. Miss. Matilda,female,28,0,0,349245,7.8958,?,S,?,?,?,0",
    "1128,3,Petroff. Mr. Nedelio,male,19,0,0,349212,7.8958,?,S,?,?,?,0",
    "1129,3,Petroff. Mr. Pastcho ('Pentcho'),male,?,0,0,349215,7.8958,?,S,?,?,?,0",
    "1130,3,Petterson. Mr. Johan Emil,male,25,1,0,347076,7.775,?,S,?,?,?,0",
    "1131,3,Pettersson. Miss. Ellen Natalia,female,18,0,0,347087,7.775,?,S,?,?,?,0",
    "1132,3,Pickard. Mr. Berk (Berk Trembisky),male,32,0,0,SOTON/O.Q. 392078,8.05,E10,S,9,?,?,1",
    "1133,3,Plotcharsky. Mr. Vasil,male,?,0,0,349227,7.8958,?,S,?,?,?,0",
    "1134,3,Pokrnic. Mr. Mate,male,17,0,0,315095,8.6625,?,S,?,?,?,0",
    "1135,3,Pokrnic. Mr. Tome,male,24,0,0,315092,8.6625,?,S,?,?,?,0",
    "1136,3,Radeff. Mr. Alexander,male,?,0,0,349223,7.8958,?,S,?,?,?,0",
    "1137,3,Rasmussen. Mrs. (Lena Jacobsen Solvang),female,?,0,0,65305,8.1125,?,S,?,?,?,0",
    "1138,3,Razi. Mr. Raihed,male,?,0,0,2629,7.2292,?,C,?,?,?,0",
    "1139,3,Reed. Mr. James George,male,?,0,0,362316,7.25,?,S,?,?,?,0",
    "1140,3,Rekic. Mr. Tido,male,38,0,0,349249,7.8958,?,S,?,?,?,0",
    "1141,3,Reynolds. Mr. Harold J,male,21,0,0,342684,8.05,?,S,?,?,?,0",
    "1142,3,Rice. Master. Albert,male,10,4,1,382652,29.125,?,Q,?,?,?,0",
    "1143,3,Rice. Master. Arthur,male,4,4,1,382652,29.125,?,Q,?,?,?,0",
    "1144,3,Rice. Master. Eric,male,7,4,1,382652,29.125,?,Q,?,?,?,0",
    "1145,3,Rice. Master. Eugene,male,2,4,1,382652,29.125,?,Q,?,?,?,0",
    "1146,3,Rice. Master. George Hugh,male,8,4,1,382652,29.125,?,Q,?,?,?,0",
    "1147,3,Rice. Mrs. William (Margaret Norton),female,39,0,5,382652,29.125,?,Q,?,327,?,0",
    "1148,3,Riihivouri. Miss. Susanna Juhantytar 'Sanni',female,22,0,0,3101295,39.6875,?,S,?,?,?,0",
    "1149,3,Rintamaki. Mr. Matti,male,35,0,0,STON/O 2. 3101273,7.125,?,S,?,?,?,0",
    "1150,3,Riordan. Miss. Johanna 'Hannah',female,?,0,0,334915,7.7208,?,Q,13,?,?,1",
    "1151,3,Risien. Mr. Samuel Beard,male,?,0,0,364498,14.5,?,S,?,?,?,0",
    "1152,3,Risien. Mrs. Samuel (Emma),female,?,0,0,364498,14.5,?,S,?,?,?,0",
    "1153,3,Robins. Mr. Alexander A,male,50,1,0,A/5. 3337,14.5,?,S,?,119,?,0",
    "1154,3,Robins. Mrs. Alexander A (Grace Charity Laury),female,47,1,0,A/5. 3337,14.5,?,S,?,7,?,0",
    "1155,3,Rogers. Mr. William John,male,?,0,0,S.C./A.4. 23567,8.05,?,S,?,?,?,0",
    "1156,3,Rommetvedt. Mr. Knud Paust,male,?,0,0,312993,7.775,?,S,?,?,?,0",
    "1157,3,Rosblom. Miss. Salli Helena,female,2,1,1,370129,20.2125,?,S,?,?,?,0",
    "1158,3,Rosblom. Mr. Viktor Richard,male,18,1,1,370129,20.2125,?,S,?,?,?,0",
    "1159,3,Rosblom. Mrs. Viktor (Helena Wilhelmina),female,41,0,2,370129,20.2125,?,S,?,?,?,0",
    "1160,3,Roth. Miss. Sarah A,female,?,0,0,342712,8.05,?,S,C,?,?,1",
    "1161,3,Rouse. Mr. Richard Henry,male,50,0,0,A/5 3594,8.05,?,S,?,?,?,0",
    "1162,3,Rush. Mr. Alfred George John,male,16,0,0,A/4. 20589,8.05,?,S,?,?,?,0",
    "1163,3,Ryan. Mr. Edward,male,?,0,0,383162,7.75,?,Q,14,?,?,1",
    "1164,3,Ryan. Mr. Patrick,male,?,0,0,371110,24.15,?,Q,?,?,?,0",
    "1165,3,Saad. Mr. Amin,male,?,0,0,2671,7.2292,?,C,?,?,?,0",
    "1166,3,Saad. Mr. Khalil,male,25,0,0,2672,7.225,?,C,?,?,?,0",
    "1167,3,Saade. Mr. Jean Nassr,male,?,0,0,2676,7.225,?,C,?,?,?,0",
    "1168,3,Sadlier. Mr. Matthew,male,?,0,0,367655,7.7292,?,Q,?,?,?,0",
    "1169,3,Sadowitz. Mr. Harry,male,?,0,0,LP 1588,7.575,?,S,?,?,?,0",
    "1170,3,Saether. Mr. Simon Sivertsen,male,38.5,0,0,SOTON/O.Q. 3101262,7.25,?,S,?,32,?,0",
    "1171,3,Sage. Master. Thomas Henry,male,?,8,2,CA. 2343,69.55,?,S,?,?,?,0",
    "1172,3,Sage. Master. William Henry,male,14.5,8,2,CA. 2343,69.55,?,S,?,67,?,0",
    "1173,3,Sage. Miss. Ada,female,?,8,2,CA. 2343,69.55,?,S,?,?,?,0",
    "1174,3,Sage. Miss. Constance Gladys,female,?,8,2,CA. 2343,69.55,?,S,?,?,?,0",
    "1175,3,Sage. Miss. Dorothy Edith 'Dolly',female,?,8,2,CA. 2343,69.55,?,S,?,?,?,0",
    "1176,3,Sage. Miss. Stella Anna,female,?,8,2,CA. 2343,69.55,?,S,?,?,?,0",
    "1177,3,Sage. Mr. Douglas Bullen,male,?,8,2,CA. 2343,69.55,?,S,?,?,?,0",
    "1178,3,Sage. Mr. Frederick,male,?,8,2,CA. 2343,69.55,?,S,?,?,?,0",
    "1179,3,Sage. Mr. George John Jr,male,?,8,2,CA. 2343,69.55,?,S,?,?,?,0",
    "1180,3,Sage. Mr. John George,male,?,1,9,CA. 2343,69.55,?,S,?,?,?,0",
    "1181,3,Sage. Mrs. John (Annie Bullen),female,?,1,9,CA. 2343,69.55,?,S,?,?,?,0",
    "1182,3,Salander. Mr. Karl Johan,male,24,0,0,7266,9.325,?,S,?,?,?,0",
    "1183,3,Salkjelsvik. Miss. Anna Kristine,female,21,0,0,343120,7.65,?,S,C,?,?,1",
    "1184,3,Salonen. Mr. Johan Werner,male,39,0,0,3101296,7.925,?,S,?,?,?,0",
    "1185,3,Samaan. Mr. Elias,male,?,2,0,2662,21.6792,?,C,?,?,?,0",
    "1186,3,Samaan. Mr. Hanna,male,?,2,0,2662,21.6792,?,C,?,?,?,0",
    "1187,3,Samaan. Mr. Youssef,male,?,2,0,2662,21.6792,?,C,?,?,?,0",
    "1188,3,Sandstrom. Miss. Beatrice Irene,female,1,1,1,PP 9549,16.7,G6,S,13,?,?,1",
    "1189,3,Sandstrom. Mrs. Hjalmar (Agnes Charlotta Bengtsson),female,24,0,2,PP 9549,16.7,G6,S,13,?,?,1",
    "1190,3,Sandstrom. Miss. Marguerite Rut,female,4,1,1,PP 9549,16.7,G6,S,13,?,?,1",
    "1191,3,Sap. Mr. Julius,male,25,0,0,345768,9.5,?,S,11,?,?,1",
    "1192,3,Saundercock. Mr. William Henry,male,20,0,0,A/5. 2151,8.05,?,S,?,?,?,0",
    "1193,3,Sawyer. Mr. Frederick Charles,male,24.5,0,0,342826,8.05,?,S,?,284,?,0",
    "1194,3,Scanlan. Mr. James,male,?,0,0,36209,7.725,?,Q,?,?,?,0",
    "1195,3,Sdycoff. Mr. Todor,male,?,0,0,349222,7.8958,?,S,?,?,?,0",
    "1196,3,Shaughnessy. Mr. Patrick,male,?,0,0,370374,7.75,?,Q,?,?,?,0",
    "1197,3,Sheerlinck. Mr. Jan Baptist,male,29,0,0,345779,9.5,?,S,11,?,?,1",
    "1198,3,Shellard. Mr. Frederick William,male,?,0,0,C.A. 6212,15.1,?,S,?,?,?,0",
    "1199,3,Shine. Miss. Ellen Natalia,female,?,0,0,330968,7.7792,?,Q,?,?,?,1",
    "1200,3,Shorney. Mr. Charles Joseph,male,?,0,0,374910,8.05,?,S,?,?,?,0",
    "1201,3,Simmons. Mr. John,male,?,0,0,SOTON/OQ 392082,8.05,?,S,?,?,?,0",
    "1202,3,Sirayanian. Mr. Orsen,male,22,0,0,2669,7.2292,?,C,?,?,?,0",
    "1203,3,Sirota. Mr. Maurice,male,?,0,0,392092,8.05,?,S,?,?,?,0",
    "1204,3,Sivic. Mr. Husein,male,40,0,0,349251,7.8958,?,S,?,?,?,0",
    "1205,3,Sivola. Mr. Antti Wilhelm,male,21,0,0,STON/O 2. 3101280,7.925,?,S,?,?,?,0",
    "1206,3,Sjoblom. Miss. Anna Sofia,female,18,0,0,3101265,7.4958,?,S,16,?,?,1",
    "1207,3,Skoog. Master. Harald,male,4,3,2,347088,27.9,?,S,?,?,?,0",
    "1208,3,Skoog. Master. Karl Thorsten,male,10,3,2,347088,27.9,?,S,?,?,?,0",
    "1209,3,Skoog. Miss. Mabel,female,9,3,2,347088,27.9,?,S,?,?,?,0",
    "1210,3,Skoog. Miss. Margit Elizabeth,female,2,3,2,347088,27.9,?,S,?,?,?,0",
    "1211,3,Skoog. Mr. Wilhelm,male,40,1,4,347088,27.9,?,S,?,?,?,0",
    "1212,3,Skoog. Mrs. William (Anna Bernhardina Karlsson),female,45,1,4,347088,27.9,?,S,?,?,?,0",
    "1213,3,Slabenoff. Mr. Petco,male,?,0,0,349214,7.8958,?,S,?,?,?,0",
    "1214,3,Slocovski. Mr. Selman Francis,male,?,0,0,SOTON/OQ 392086,8.05,?,S,?,?,?,0",
    "1215,3,Smiljanic. Mr. Mile,male,?,0,0,315037,8.6625,?,S,?,?,?,0",
    "1216,3,Smith. Mr. Thomas,male,?,0,0,384461,7.75,?,Q,?,?,?,0",
    "1217,3,Smyth. Miss. Julia,female,?,0,0,335432,7.7333,?,Q,13,?,?,1",
    "1218,3,Soholt. Mr. Peter Andreas Lauritz Andersen,male,19,0,0,348124,7.65,F G73,S,?,?,?,0",
    "1219,3,Somerton. Mr. Francis William,male,30,0,0,A.5. 18509,8.05,?,S,?,?,?,0",
    "1220,3,Spector. Mr. Woolf,male,?,0,0,A.5. 3236,8.05,?,S,?,?,?,0",
    "1221,3,Spinner. Mr. Henry John,male,32,0,0,STON/OQ. 369943,8.05,?,S,?,?,?,0",
    "1222,3,Staneff. Mr. Ivan,male,?,0,0,349208,7.8958,?,S,?,?,?,0",
    "1223,3,Stankovic. Mr. Ivan,male,33,0,0,349239,8.6625,?,C,?,?,?,0",
    "1224,3,Stanley. Miss. Amy Zillah Elsie,female,23,0,0,CA. 2314,7.55,?,S,C,?,?,1",
    "1225,3,Stanley. Mr. Edward Roland,male,21,0,0,A/4 45380,8.05,?,S,?,?,?,0",
    "1226,3,Storey. Mr. Thomas,male,60.5,0,0,3701,?,?,S,?,261,?,0",
    "1227,3,Stoytcheff. Mr. Ilia,male,19,0,0,349205,7.8958,?,S,?,?,?,0",
    "1228,3,Strandberg. Miss. Ida Sofia,female,22,0,0,7553,9.8375,?,S,?,?,?,0",
    "1229,3,Stranden. Mr. Juho,male,31,0,0,STON/O 2. 3101288,7.925,?,S,9,?,?,1",
    "1230,3,Strilic. Mr. Ivan,male,27,0,0,315083,8.6625,?,S,?,?,?,0",
    "1231,3,Strom. Miss. Telma Matilda,female,2,0,1,347054,10.4625,G6,S,?,?,?,0",
    "1232,3,Strom. Mrs. Wilhelm (Elna Matilda Persson),female,29,1,1,347054,10.4625,G6,S,?,?,?,0",
    "1233,3,Sunderland. Mr. Victor Francis,male,16,0,0,SOTON/OQ 392089,8.05,?,S,B,?,?,1",
    "1234,3,Sundman. Mr. Johan Julian,male,44,0,0,STON/O 2. 3101269,7.925,?,S,15,?,?,1",
    "1235,3,Sutehall. Mr. Henry Jr,male,25,0,0,SOTON/OQ 392076,7.05,?,S,?,?,?,0",
    "1236,3,Svensson. Mr. Johan,male,74,0,0,347060,7.775,?,S,?,?,?,0",
    "1237,3,Svensson. Mr. Johan Cervin,male,14,0,0,7538,9.225,?,S,13,?,?,1",
    "1238,3,Svensson. Mr. Olof,male,24,0,0,350035,7.7958,?,S,?,?,?,0",
    "1239,3,Tenglin. Mr. Gunnar Isidor,male,25,0,0,350033,7.7958,?,S,13 15,?,?,1",
    "1240,3,Theobald. Mr. Thomas Leonard,male,34,0,0,363294,8.05,?,S,?,176,?,0",
    "1241,3,Thomas. Master. Assad Alexander,male,0.4167,0,1,2625,8.5167,?,C,16,?,?,1",
    "1242,3,Thomas. Mr. Charles P,male,?,1,0,2621,6.4375,?,C,?,?,?,0",
    "1243,3,Thomas. Mr. John,male,?,0,0,2681,6.4375,?,C,?,?,?,0",
    "1244,3,Thomas. Mr. Tannous,male,?,0,0,2684,7.225,?,C,?,?,?,0",
    "1245,3,Thomas. Mrs. Alexander (Thamine 'Thelma'),female,16,1,1,2625,8.5167,?,C,14,?,?,1",
    "1246,3,Thomson. Mr. Alexander Morrison,male,?,0,0,32302,8.05,?,S,?,?,?,0",
    "1247,3,Thorneycroft. Mr. Percival,male,?,1,0,376564,16.1,?,S,?,?,?,0",
    "1248,3,Thorneycroft. Mrs. Percival (Florence Kate White),female,?,1,0,376564,16.1,?,S,10,?,?,1",
    "1249,3,Tikkanen. Mr. Juho,male,32,0,0,STON/O 2. 3101293,7.925,?,S,?,?,?,0",
    "1250,3,Tobin. Mr. Roger,male,?,0,0,383121,7.75,F38,Q,?,?,?,0",
    "1251,3,Todoroff. Mr. Lalio,male,?,0,0,349216,7.8958,?,S,?,?,?,0",
    "1252,3,Tomlin. Mr. Ernest Portage,male,30.5,0,0,364499,8.05,?,S,?,50,?,0",
    "1253,3,Torber. Mr. Ernst William,male,44,0,0,364511,8.05,?,S,?,?,?,0",
    "1254,3,Torfa. Mr. Assad,male,?,0,0,2673,7.2292,?,C,?,?,?,0",
    "1255,3,Tornquist. Mr. William Henry,male,25,0,0,LINE,0,?,S,15,?,?,1",
    "1256,3,Toufik. Mr. Nakli,male,?,0,0,2641,7.2292,?,C,?,?,?,0",
    "1257,3,Touma. Master. Georges Youssef,male,7,1,1,2650,15.2458,?,C,C,?,?,1",
    "1258,3,Touma. Miss. Maria Youssef,female,9,1,1,2650,15.2458,?,C,C,?,?,1",
    "1259,3,Touma. Mrs. Darwis (Hanne Youssef Razi),female,29,0,2,2650,15.2458,?,C,C,?,?,1",
    "1260,3,Turcin. Mr. Stjepan,male,36,0,0,349247,7.8958,?,S,?,?,?,0",
    "1261,3,Turja. Miss. Anna Sofia,female,18,0,0,4138,9.8417,?,S,15,?,?,1",
    "1262,3,Turkula. Mrs. (Hedwig),female,63,0,0,4134,9.5875,?,S,15,?,?,1",
    "1263,3,van Billiard. Master. James William,male,?,1,1,A/5. 851,14.5,?,S,?,?,?,0",
    "1264,3,van Billiard. Master. Walter John,male,11.5,1,1,A/5. 851,14.5,?,S,?,1,?,0",
    "1265,3,van Billiard. Mr. Austin Blyler,male,40.5,0,2,A/5. 851,14.5,?,S,?,255,?,0",
    "1266,3,Van Impe. Miss. Catharina,female,10,0,2,345773,24.15,?,S,?,?,?,0",
    "1267,3,Van Impe. Mr. Jean Baptiste,male,36,1,1,345773,24.15,?,S,?,?,?,0",
    "1268,3,Van Impe. Mrs. Jean Baptiste (Rosalie Paula Govaert),female,30,1,1,345773,24.15,?,S,?,?,?,0",
    "1269,3,van Melkebeke. Mr. Philemon,male,?,0,0,345777,9.5,?,S,?,?,?,0",
    "1270,3,Vande Velde. Mr. Johannes Joseph,male,33,0,0,345780,9.5,?,S,?,?,?,0",
    "1271,3,Vande Walle. Mr. Nestor Cyriel,male,28,0,0,345770,9.5,?,S,?,?,?,0",
    "1272,3,Vanden Steen. Mr. Leo Peter,male,28,0,0,345783,9.5,?,S,?,?,?,0",
    "1273,3,Vander Cruyssen. Mr. Victor,male,47,0,0,345765,9,?,S,?,?,?,0",
    "1274,3,Vander Planke. Miss. Augusta Maria,female,18,2,0,345764,18,?,S,?,?,?,0",
    "1275,3,Vander Planke. Mr. Julius,male,31,3,0,345763,18,?,S,?,?,?,0",
    "1276,3,Vander Planke. Mr. Leo Edmondus,male,16,2,0,345764,18,?,S,?,?,?,0",
    "1277,3,Vander Planke. Mrs. Julius (Emelia Maria Vandemoortele),female,31,1,0,345763,18,?,S,?,?,?,0",
    "1278,3,Vartanian. Mr. David,male,22,0,0,2658,7.225,?,C,13 15,?,?,1",
    "1279,3,Vendel. Mr. Olof Edvin,male,20,0,0,350416,7.8542,?,S,?,?,?,0",
    "1280,3,Vestrom. Miss. Hulda Amanda Adolfina,female,14,0,0,350406,7.8542,?,S,?,?,?,0",
    "1281,3,Vovk. Mr. Janko,male,22,0,0,349252,7.8958,?,S,?,?,?,0",
    "1282,3,Waelens. Mr. Achille,male,22,0,0,345767,9,?,S,?,?,Antwerp. Belgium / Stanton. OH,0",
    "1283,3,Ware. Mr. Frederick,male,?,0,0,359309,8.05,?,S,?,?,?,0",
    "1284,3,Warren. Mr. Charles William,male,?,0,0,C.A. 49867,7.55,?,S,?,?,?,0",
    "1285,3,Webber. Mr. James,male,?,0,0,SOTON/OQ 3101316,8.05,?,S,?,?,?,0",
    "1286,3,Wenzel. Mr. Linhart,male,32.5,0,0,345775,9.5,?,S,?,298,?,0",
    "1287,3,Whabee. Mrs. George Joseph (Shawneene Abi-Saab),female,38,0,0,2688,7.2292,?,C,C,?,?,1",
    "1288,3,Widegren. Mr. Carl/Charles Peter,male,51,0,0,347064,7.75,?,S,?,?,?,0",
    "1289,3,Wiklund. Mr. Jakob Alfred,male,18,1,0,3101267,6.4958,?,S,?,314,?,0",
    "1290,3,Wiklund. Mr. Karl Johan,male,21,1,0,3101266,6.4958,?,S,?,?,?,0",
    "1291,3,Wilkes. Mrs. James (Ellen Needs),female,47,1,0,363272,7,?,S,?,?,?,1",
    "1292,3,Willer. Mr. Aaron ('Abi Weller'),male,?,0,0,3410,8.7125,?,S,?,?,?,0",
    "1293,3,Willey. Mr. Edward,male,?,0,0,S.O./P.P. 751,7.55,?,S,?,?,?,0",
    "1294,3,Williams. Mr. Howard Hugh 'Harry',male,?,0,0,A/5 2466,8.05,?,S,?,?,?,0",
    "1295,3,Williams. Mr. Leslie,male,28.5,0,0,54636,16.1,?,S,?,14,?,0",
    "1296,3,Windelov. Mr. Einar,male,21,0,0,SOTON/OQ 3101317,7.25,?,S,?,?,?,0",
    "1297,3,Wirz. Mr. Albert,male,27,0,0,315154,8.6625,?,S,?,131,?,0",
    "1298,3,Wiseman. Mr. Phillippe,male,?,0,0,A/4. 34244,7.25,?,S,?,?,?,0",
    "1299,3,Wittevrongel. Mr. Camille,male,36,0,0,345771,9.5,?,S,?,?,?,0",
    "1300,3,Yasbeck. Mr. Antoni,male,27,1,0,2659,14.4542,?,C,C,?,?,0",
    "1301,3,Yasbeck. Mrs. Antoni (Selini Alexander),female,15,1,0,2659,14.4542,?,C,?,?,?,1",
    "1302,3,Youseff. Mr. Gerious,male,45.5,0,0,2628,7.225,?,C,?,312,?,0",
    "1303,3,Yousif. Mr. Wazli,male,?,0,0,2647,7.225,?,C,?,?,?,0",
    "1304,3,Yousseff. Mr. Gerious,male,?,0,0,2627,14.4583,?,C,?,?,?,0",
    "1305,3,Zabour. Miss. Hileni,female,14.5,1,0,2665,14.4542,?,C,?,328,?,0",
    "1306,3,Zabour. Miss. Thamine,female,?,1,0,2665,14.4542,?,C,?,?,?,0",
    "1307,3,Zakarian. Mr. Mapriededer,male,26.5,0,0,2656,7.225,?,C,?,304,?,0",
    "1308,3,Zakarian. Mr. Ortin,male,27,0,0,2670,7.225,?,C,?,?,?,0",
    "1309,3,Zimmerman. Mr. Leo,male,29,0,0,315082,7.875,?,S,?,?,?,0",
)

Wine_Dataset = (
    "id_int,fixed acidity_float,volatile acidity_float,citric acid_float,residual sugar_float,chlorides_float,free sulfur dioxide_float,total sulfur dioxide_float,density_float,pH_float,sulphates_float,alcohol_float,quality_int",
    "1,7.4,0.7,0,1.9,0.076,11,34,0.9978,3.51,0.56,9.4,5",
    "2,7.8,0.88,0,2.6,0.098,25,67,0.9968,3.2,0.68,9.8,5",
    "3,7.8,0.76,0.04,2.3,0.092,15,54,0.997,3.26,0.65,9.8,5",
    "4,11.2,0.28,0.56,1.9,0.075,17,60,0.998,3.16,0.58,9.8,6",
    "5,7.4,0.7,0,1.9,0.076,11,34,0.9978,3.51,0.56,9.4,5",
    "6,7.4,0.66,0,1.8,0.075,13,40,0.9978,3.51,0.56,9.4,5",
    "7,7.9,0.6,0.06,1.6,0.069,15,59,0.9964,3.3,0.46,9.4,5",
    "8,7.3,0.65,0,1.2,0.065,15,21,0.9946,3.39,0.47,10,7",
    "9,7.8,0.58,0.02,2,0.073,9,18,0.9968,3.36,0.57,9.5,7",
    "10,7.5,0.5,0.36,6.1,0.071,17,102,0.9978,3.35,0.8,10.5,5",
    "11,6.7,0.58,0.08,1.8,0.097,15,65,0.9959,3.28,0.54,9.2,5",
    "12,7.5,0.5,0.36,6.1,0.071,17,102,0.9978,3.35,0.8,10.5,5",
    "13,5.6,0.615,0,1.6,0.089,16,59,0.9943,3.58,0.52,9.9,5",
    "14,7.8,0.61,0.29,1.6,0.114,9,29,0.9974,3.26,1.56,9.1,5",
    "15,8.9,0.62,0.18,3.8,0.176,52,145,0.9986,3.16,0.88,9.2,5",
    "16,8.9,0.62,0.19,3.9,0.17,51,148,0.9986,3.17,0.93,9.2,5",
    "17,8.5,0.28,0.56,1.8,0.092,35,103,0.9969,3.3,0.75,10.5,7",
    "18,8.1,0.56,0.28,1.7,0.368,16,56,0.9968,3.11,1.28,9.3,5",
    "19,7.4,0.59,0.08,4.4,0.086,6,29,0.9974,3.38,0.5,9,4",
    "20,7.9,0.32,0.51,1.8,0.341,17,56,0.9969,3.04,1.08,9.2,6",
    "21,8.9,0.22,0.48,1.8,0.077,29,60,0.9968,3.39,0.53,9.4,6",
    "22,7.6,0.39,0.31,2.3,0.082,23,71,0.9982,3.52,0.65,9.7,5",
    "23,7.9,0.43,0.21,1.6,0.106,10,37,0.9966,3.17,0.91,9.5,5",
    "24,8.5,0.49,0.11,2.3,0.084,9,67,0.9968,3.17,0.53,9.4,5",
    "25,6.9,0.4,0.14,2.4,0.085,21,40,0.9968,3.43,0.63,9.7,6",
    "26,6.3,0.39,0.16,1.4,0.08,11,23,0.9955,3.34,0.56,9.3,5",
    "27,7.6,0.41,0.24,1.8,0.08,4,11,0.9962,3.28,0.59,9.5,5",
    "28,7.9,0.43,0.21,1.6,0.106,10,37,0.9966,3.17,0.91,9.5,5",
    "29,7.1,0.71,0,1.9,0.08,14,35,0.9972,3.47,0.55,9.4,5",
    "30,7.8,0.645,0,2,0.082,8,16,0.9964,3.38,0.59,9.8,6",
    "31,6.7,0.675,0.07,2.4,0.089,17,82,0.9958,3.35,0.54,10.1,5",
    "32,6.9,0.685,0,2.5,0.105,22,37,0.9966,3.46,0.57,10.6,6",
    "33,8.3,0.655,0.12,2.3,0.083,15,113,0.9966,3.17,0.66,9.8,5",
    "34,6.9,0.605,0.12,10.7,0.073,40,83,0.9993,3.45,0.52,9.4,6",
    "35,5.2,0.32,0.25,1.8,0.103,13,50,0.9957,3.38,0.55,9.2,5",
    "36,7.8,0.645,0,5.5,0.086,5,18,0.9986,3.4,0.55,9.6,6",
    "37,7.8,0.6,0.14,2.4,0.086,3,15,0.9975,3.42,0.6,10.8,6",
    "38,8.1,0.38,0.28,2.1,0.066,13,30,0.9968,3.23,0.73,9.7,7",
    "39,5.7,1.13,0.09,1.5,0.172,7,19,0.994,3.5,0.48,9.8,4",
    "40,7.3,0.45,0.36,5.9,0.074,12,87,0.9978,3.33,0.83,10.5,5",
    "41,7.3,0.45,0.36,5.9,0.074,12,87,0.9978,3.33,0.83,10.5,5",
    "42,8.8,0.61,0.3,2.8,0.088,17,46,0.9976,3.26,0.51,9.3,4",
    "43,7.5,0.49,0.2,2.6,0.332,8,14,0.9968,3.21,0.9,10.5,6",
    "44,8.1,0.66,0.22,2.2,0.069,9,23,0.9968,3.3,1.2,10.3,5",
    "45,6.8,0.67,0.02,1.8,0.05,5,11,0.9962,3.48,0.52,9.5,5",
    "46,4.6,0.52,0.15,2.1,0.054,8,65,0.9934,3.9,0.56,13.1,4",
    "47,7.7,0.935,0.43,2.2,0.114,22,114,0.997,3.25,0.73,9.2,5",
    "48,8.7,0.29,0.52,1.6,0.113,12,37,0.9969,3.25,0.58,9.5,5",
    "49,6.4,0.4,0.23,1.6,0.066,5,12,0.9958,3.34,0.56,9.2,5",
    "50,5.6,0.31,0.37,1.4,0.074,12,96,0.9954,3.32,0.58,9.2,5",
    "51,8.8,0.66,0.26,1.7,0.074,4,23,0.9971,3.15,0.74,9.2,5",
    "52,6.6,0.52,0.04,2.2,0.069,8,15,0.9956,3.4,0.63,9.4,6",
    "53,6.6,0.5,0.04,2.1,0.068,6,14,0.9955,3.39,0.64,9.4,6",
    "54,8.6,0.38,0.36,3,0.081,30,119,0.997,3.2,0.56,9.4,5",
    "55,7.6,0.51,0.15,2.8,0.11,33,73,0.9955,3.17,0.63,10.2,6",
    "56,7.7,0.62,0.04,3.8,0.084,25,45,0.9978,3.34,0.53,9.5,5",
    "57,10.2,0.42,0.57,3.4,0.07,4,10,0.9971,3.04,0.63,9.6,5",
    "58,7.5,0.63,0.12,5.1,0.111,50,110,0.9983,3.26,0.77,9.4,5",
    "59,7.8,0.59,0.18,2.3,0.076,17,54,0.9975,3.43,0.59,10,5",
    "60,7.3,0.39,0.31,2.4,0.074,9,46,0.9962,3.41,0.54,9.4,6",
    "61,8.8,0.4,0.4,2.2,0.079,19,52,0.998,3.44,0.64,9.2,5",
    "62,7.7,0.69,0.49,1.8,0.115,20,112,0.9968,3.21,0.71,9.3,5",
    "63,7.5,0.52,0.16,1.9,0.085,12,35,0.9968,3.38,0.62,9.5,7",
    "64,7,0.735,0.05,2,0.081,13,54,0.9966,3.39,0.57,9.8,5",
    "65,7.2,0.725,0.05,4.65,0.086,4,11,0.9962,3.41,0.39,10.9,5",
    "66,7.2,0.725,0.05,4.65,0.086,4,11,0.9962,3.41,0.39,10.9,5",
    "67,7.5,0.52,0.11,1.5,0.079,11,39,0.9968,3.42,0.58,9.6,5",
    "68,6.6,0.705,0.07,1.6,0.076,6,15,0.9962,3.44,0.58,10.7,5",
    "69,9.3,0.32,0.57,2,0.074,27,65,0.9969,3.28,0.79,10.7,5",
    "70,8,0.705,0.05,1.9,0.074,8,19,0.9962,3.34,0.95,10.5,6",
    "71,7.7,0.63,0.08,1.9,0.076,15,27,0.9967,3.32,0.54,9.5,6",
    "72,7.7,0.67,0.23,2.1,0.088,17,96,0.9962,3.32,0.48,9.5,5",
    "73,7.7,0.69,0.22,1.9,0.084,18,94,0.9961,3.31,0.48,9.5,5",
    "74,8.3,0.675,0.26,2.1,0.084,11,43,0.9976,3.31,0.53,9.2,4",
    "75,9.7,0.32,0.54,2.5,0.094,28,83,0.9984,3.28,0.82,9.6,5",
    "76,8.8,0.41,0.64,2.2,0.093,9,42,0.9986,3.54,0.66,10.5,5",
    "77,8.8,0.41,0.64,2.2,0.093,9,42,0.9986,3.54,0.66,10.5,5",
    "78,6.8,0.785,0,2.4,0.104,14,30,0.9966,3.52,0.55,10.7,6",
    "79,6.7,0.75,0.12,2,0.086,12,80,0.9958,3.38,0.52,10.1,5",
    "80,8.3,0.625,0.2,1.5,0.08,27,119,0.9972,3.16,1.12,9.1,4",
    "81,6.2,0.45,0.2,1.6,0.069,3,15,0.9958,3.41,0.56,9.2,5",
    "82,7.8,0.43,0.7,1.9,0.464,22,67,0.9974,3.13,1.28,9.4,5",
    "83,7.4,0.5,0.47,2,0.086,21,73,0.997,3.36,0.57,9.1,5",
    "84,7.3,0.67,0.26,1.8,0.401,16,51,0.9969,3.16,1.14,9.4,5",
    "85,6.3,0.3,0.48,1.8,0.069,18,61,0.9959,3.44,0.78,10.3,6",
    "86,6.9,0.55,0.15,2.2,0.076,19,40,0.9961,3.41,0.59,10.1,5",
    "87,8.6,0.49,0.28,1.9,0.11,20,136,0.9972,2.93,1.95,9.9,6",
    "88,7.7,0.49,0.26,1.9,0.062,9,31,0.9966,3.39,0.64,9.6,5",
    "89,9.3,0.39,0.44,2.1,0.107,34,125,0.9978,3.14,1.22,9.5,5",
    "90,7,0.62,0.08,1.8,0.076,8,24,0.9978,3.48,0.53,9,5",
    "91,7.9,0.52,0.26,1.9,0.079,42,140,0.9964,3.23,0.54,9.5,5",
    "92,8.6,0.49,0.28,1.9,0.11,20,136,0.9972,2.93,1.95,9.9,6",
    "93,8.6,0.49,0.29,2,0.11,19,133,0.9972,2.93,1.98,9.8,5",
    "94,7.7,0.49,0.26,1.9,0.062,9,31,0.9966,3.39,0.64,9.6,5",
    "95,5,1.02,0.04,1.4,0.045,41,85,0.9938,3.75,0.48,10.5,4",
    "96,4.7,0.6,0.17,2.3,0.058,17,106,0.9932,3.85,0.6,12.9,6",
    "97,6.8,0.775,0,3,0.102,8,23,0.9965,3.45,0.56,10.7,5",
    "98,7,0.5,0.25,2,0.07,3,22,0.9963,3.25,0.63,9.2,5",
    "99,7.6,0.9,0.06,2.5,0.079,5,10,0.9967,3.39,0.56,9.8,5",
    "100,8.1,0.545,0.18,1.9,0.08,13,35,0.9972,3.3,0.59,9,6",
    "101,8.3,0.61,0.3,2.1,0.084,11,50,0.9972,3.4,0.61,10.2,6",
    "102,7.8,0.5,0.3,1.9,0.075,8,22,0.9959,3.31,0.56,10.4,6",
    "103,8.1,0.545,0.18,1.9,0.08,13,35,0.9972,3.3,0.59,9,6",
    "104,8.1,0.575,0.22,2.1,0.077,12,65,0.9967,3.29,0.51,9.2,5",
    "105,7.2,0.49,0.24,2.2,0.07,5,36,0.996,3.33,0.48,9.4,5",
    "106,8.1,0.575,0.22,2.1,0.077,12,65,0.9967,3.29,0.51,9.2,5",
    "107,7.8,0.41,0.68,1.7,0.467,18,69,0.9973,3.08,1.31,9.3,5",
    "108,6.2,0.63,0.31,1.7,0.088,15,64,0.9969,3.46,0.79,9.3,5",
    "109,8,0.33,0.53,2.5,0.091,18,80,0.9976,3.37,0.8,9.6,6",
    "110,8.1,0.785,0.52,2,0.122,37,153,0.9969,3.21,0.69,9.3,5",
    "111,7.8,0.56,0.19,1.8,0.104,12,47,0.9964,3.19,0.93,9.5,5",
    "112,8.4,0.62,0.09,2.2,0.084,11,108,0.9964,3.15,0.66,9.8,5",
    "113,8.4,0.6,0.1,2.2,0.085,14,111,0.9964,3.15,0.66,9.8,5",
    "114,10.1,0.31,0.44,2.3,0.08,22,46,0.9988,3.32,0.67,9.7,6",
    "115,7.8,0.56,0.19,1.8,0.104,12,47,0.9964,3.19,0.93,9.5,5",
    "116,9.4,0.4,0.31,2.2,0.09,13,62,0.9966,3.07,0.63,10.5,6",
    "117,8.3,0.54,0.28,1.9,0.077,11,40,0.9978,3.39,0.61,10,6",
    "118,7.8,0.56,0.12,2,0.082,7,28,0.997,3.37,0.5,9.4,6",
    "119,8.8,0.55,0.04,2.2,0.119,14,56,0.9962,3.21,0.6,10.9,6",
    "120,7,0.69,0.08,1.8,0.097,22,89,0.9959,3.34,0.54,9.2,6",
    "121,7.3,1.07,0.09,1.7,0.178,10,89,0.9962,3.3,0.57,9,5",
    "122,8.8,0.55,0.04,2.2,0.119,14,56,0.9962,3.21,0.6,10.9,6",
    "123,7.3,0.695,0,2.5,0.075,3,13,0.998,3.49,0.52,9.2,5",
    "124,8,0.71,0,2.6,0.08,11,34,0.9976,3.44,0.53,9.5,5",
    "125,7.8,0.5,0.17,1.6,0.082,21,102,0.996,3.39,0.48,9.5,5",
    "126,9,0.62,0.04,1.9,0.146,27,90,0.9984,3.16,0.7,9.4,5",
    "127,8.2,1.33,0,1.7,0.081,3,12,0.9964,3.53,0.49,10.9,5",
    "128,8.1,1.33,0,1.8,0.082,3,12,0.9964,3.54,0.48,10.9,5",
    "129,8,0.59,0.16,1.8,0.065,3,16,0.9962,3.42,0.92,10.5,7",
    "130,6.1,0.38,0.15,1.8,0.072,6,19,0.9955,3.42,0.57,9.4,5",
    "131,8,0.745,0.56,2,0.118,30,134,0.9968,3.24,0.66,9.4,5",
    "132,5.6,0.5,0.09,2.3,0.049,17,99,0.9937,3.63,0.63,13,5",
    "133,5.6,0.5,0.09,2.3,0.049,17,99,0.9937,3.63,0.63,13,5",
    "134,6.6,0.5,0.01,1.5,0.06,17,26,0.9952,3.4,0.58,9.8,6",
    "135,7.9,1.04,0.05,2.2,0.084,13,29,0.9959,3.22,0.55,9.9,6",
    "136,8.4,0.745,0.11,1.9,0.09,16,63,0.9965,3.19,0.82,9.6,5",
    "137,8.3,0.715,0.15,1.8,0.089,10,52,0.9968,3.23,0.77,9.5,5",
    "138,7.2,0.415,0.36,2,0.081,13,45,0.9972,3.48,0.64,9.2,5",
    "139,7.8,0.56,0.19,2.1,0.081,15,105,0.9962,3.33,0.54,9.5,5",
    "140,7.8,0.56,0.19,2,0.081,17,108,0.9962,3.32,0.54,9.5,5",
    "141,8.4,0.745,0.11,1.9,0.09,16,63,0.9965,3.19,0.82,9.6,5",
    "142,8.3,0.715,0.15,1.8,0.089,10,52,0.9968,3.23,0.77,9.5,5",
    "143,5.2,0.34,0,1.8,0.05,27,63,0.9916,3.68,0.79,14,6",
    "144,6.3,0.39,0.08,1.7,0.066,3,20,0.9954,3.34,0.58,9.4,5",
    "145,5.2,0.34,0,1.8,0.05,27,63,0.9916,3.68,0.79,14,6",
    "146,8.1,0.67,0.55,1.8,0.117,32,141,0.9968,3.17,0.62,9.4,5",
    "147,5.8,0.68,0.02,1.8,0.087,21,94,0.9944,3.54,0.52,10,5",
    "148,7.6,0.49,0.26,1.6,0.236,10,88,0.9968,3.11,0.8,9.3,5",
    "149,6.9,0.49,0.1,2.3,0.074,12,30,0.9959,3.42,0.58,10.2,6",
    "150,8.2,0.4,0.44,2.8,0.089,11,43,0.9975,3.53,0.61,10.5,6",
    "151,7.3,0.33,0.47,2.1,0.077,5,11,0.9958,3.33,0.53,10.3,6",
    "152,9.2,0.52,1,3.4,0.61,32,69,0.9996,2.74,2,9.4,4",
    "153,7.5,0.6,0.03,1.8,0.095,25,99,0.995,3.35,0.54,10.1,5",
    "154,7.5,0.6,0.03,1.8,0.095,25,99,0.995,3.35,0.54,10.1,5",
    "155,7.1,0.43,0.42,5.5,0.07,29,129,0.9973,3.42,0.72,10.5,5",
    "156,7.1,0.43,0.42,5.5,0.071,28,128,0.9973,3.42,0.71,10.5,5",
    "157,7.1,0.43,0.42,5.5,0.07,29,129,0.9973,3.42,0.72,10.5,5",
    "158,7.1,0.43,0.42,5.5,0.071,28,128,0.9973,3.42,0.71,10.5,5",
    "159,7.1,0.68,0,2.2,0.073,12,22,0.9969,3.48,0.5,9.3,5",
    "160,6.8,0.6,0.18,1.9,0.079,18,86,0.9968,3.59,0.57,9.3,6",
    "161,7.6,0.95,0.03,2,0.09,7,20,0.9959,3.2,0.56,9.6,5",
    "162,7.6,0.68,0.02,1.3,0.072,9,20,0.9965,3.17,1.08,9.2,4",
    "163,7.8,0.53,0.04,1.7,0.076,17,31,0.9964,3.33,0.56,10,6",
    "164,7.4,0.6,0.26,7.3,0.07,36,121,0.9982,3.37,0.49,9.4,5",
    "165,7.3,0.59,0.26,7.2,0.07,35,121,0.9981,3.37,0.49,9.4,5",
    "166,7.8,0.63,0.48,1.7,0.1,14,96,0.9961,3.19,0.62,9.5,5",
    "167,6.8,0.64,0.1,2.1,0.085,18,101,0.9956,3.34,0.52,10.2,5",
    "168,7.3,0.55,0.03,1.6,0.072,17,42,0.9956,3.37,0.48,9,4",
    "169,6.8,0.63,0.07,2.1,0.089,11,44,0.9953,3.47,0.55,10.4,6",
    "170,7.5,0.705,0.24,1.8,0.36,15,63,0.9964,3,1.59,9.5,5",
    "171,7.9,0.885,0.03,1.8,0.058,4,8,0.9972,3.36,0.33,9.1,4",
    "172,8,0.42,0.17,2,0.073,6,18,0.9972,3.29,0.61,9.2,6",
    "173,8,0.42,0.17,2,0.073,6,18,0.9972,3.29,0.61,9.2,6",
    "174,7.4,0.62,0.05,1.9,0.068,24,42,0.9961,3.42,0.57,11.5,6",
    "175,7.3,0.38,0.21,2,0.08,7,35,0.9961,3.33,0.47,9.5,5",
    "176,6.9,0.5,0.04,1.5,0.085,19,49,0.9958,3.35,0.78,9.5,5",
    "177,7.3,0.38,0.21,2,0.08,7,35,0.9961,3.33,0.47,9.5,5",
    "178,7.5,0.52,0.42,2.3,0.087,8,38,0.9972,3.58,0.61,10.5,6",
    "179,7,0.805,0,2.5,0.068,7,20,0.9969,3.48,0.56,9.6,5",
    "180,8.8,0.61,0.14,2.4,0.067,10,42,0.9969,3.19,0.59,9.5,5",
    "181,8.8,0.61,0.14,2.4,0.067,10,42,0.9969,3.19,0.59,9.5,5",
    "182,8.9,0.61,0.49,2,0.27,23,110,0.9972,3.12,1.02,9.3,5",
    "183,7.2,0.73,0.02,2.5,0.076,16,42,0.9972,3.44,0.52,9.3,5",
    "184,6.8,0.61,0.2,1.8,0.077,11,65,0.9971,3.54,0.58,9.3,5",
    "185,6.7,0.62,0.21,1.9,0.079,8,62,0.997,3.52,0.58,9.3,6",
    "186,8.9,0.31,0.57,2,0.111,26,85,0.9971,3.26,0.53,9.7,5",
    "187,7.4,0.39,0.48,2,0.082,14,67,0.9972,3.34,0.55,9.2,5",
    "188,7.7,0.705,0.1,2.6,0.084,9,26,0.9976,3.39,0.49,9.7,5",
    "189,7.9,0.5,0.33,2,0.084,15,143,0.9968,3.2,0.55,9.5,5",
    "190,7.9,0.49,0.32,1.9,0.082,17,144,0.9968,3.2,0.55,9.5,5",
    "191,8.2,0.5,0.35,2.9,0.077,21,127,0.9976,3.23,0.62,9.4,5",
    "192,6.4,0.37,0.25,1.9,0.074,21,49,0.9974,3.57,0.62,9.8,6",
    "193,6.8,0.63,0.12,3.8,0.099,16,126,0.9969,3.28,0.61,9.5,5",
    "194,7.6,0.55,0.21,2.2,0.071,7,28,0.9964,3.28,0.55,9.7,5",
    "195,7.6,0.55,0.21,2.2,0.071,7,28,0.9964,3.28,0.55,9.7,5",
    "196,7.8,0.59,0.33,2,0.074,24,120,0.9968,3.25,0.54,9.4,5",
    "197,7.3,0.58,0.3,2.4,0.074,15,55,0.9968,3.46,0.59,10.2,5",
    "198,11.5,0.3,0.6,2,0.067,12,27,0.9981,3.11,0.97,10.1,6",
    "199,5.4,0.835,0.08,1.2,0.046,13,93,0.9924,3.57,0.85,13,7",
    "200,6.9,1.09,0.06,2.1,0.061,12,31,0.9948,3.51,0.43,11.4,4",
    "201,9.6,0.32,0.47,1.4,0.056,9,24,0.99695,3.22,0.82,10.3,7",
    "202,8.8,0.37,0.48,2.1,0.097,39,145,0.9975,3.04,1.03,9.3,5",
    "203,6.8,0.5,0.11,1.5,0.075,16,49,0.99545,3.36,0.79,9.5,5",
    "204,7,0.42,0.35,1.6,0.088,16,39,0.9961,3.34,0.55,9.2,5",
    "205,7,0.43,0.36,1.6,0.089,14,37,0.99615,3.34,0.56,9.2,6",
    "206,12.8,0.3,0.74,2.6,0.095,9,28,0.9994,3.2,0.77,10.8,7",
    "207,12.8,0.3,0.74,2.6,0.095,9,28,0.9994,3.2,0.77,10.8,7",
    "208,7.8,0.57,0.31,1.8,0.069,26,120,0.99625,3.29,0.53,9.3,5",
    "209,7.8,0.44,0.28,2.7,0.1,18,95,0.9966,3.22,0.67,9.4,5",
    "210,11,0.3,0.58,2.1,0.054,7,19,0.998,3.31,0.88,10.5,7",
    "211,9.7,0.53,0.6,2,0.039,5,19,0.99585,3.3,0.86,12.4,6",
    "212,8,0.725,0.24,2.8,0.083,10,62,0.99685,3.35,0.56,10,6",
    "213,11.6,0.44,0.64,2.1,0.059,5,15,0.998,3.21,0.67,10.2,6",
    "214,8.2,0.57,0.26,2.2,0.06,28,65,0.9959,3.3,0.43,10.1,5",
    "215,7.8,0.735,0.08,2.4,0.092,10,41,0.9974,3.24,0.71,9.8,6",
    "216,7,0.49,0.49,5.6,0.06,26,121,0.9974,3.34,0.76,10.5,5",
    "217,8.7,0.625,0.16,2,0.101,13,49,0.9962,3.14,0.57,11,5",
    "218,8.1,0.725,0.22,2.2,0.072,11,41,0.9967,3.36,0.55,9.1,5",
    "219,7.5,0.49,0.19,1.9,0.076,10,44,0.9957,3.39,0.54,9.7,5",
    "220,7.8,0.53,0.33,2.4,0.08,24,144,0.99655,3.3,0.6,9.5,5",
    "221,7.8,0.34,0.37,2,0.082,24,58,0.9964,3.34,0.59,9.4,6",
    "222,7.4,0.53,0.26,2,0.101,16,72,0.9957,3.15,0.57,9.4,5",
    "223,6.8,0.61,0.04,1.5,0.057,5,10,0.99525,3.42,0.6,9.5,5",
    "224,8.6,0.645,0.25,2,0.083,8,28,0.99815,3.28,0.6,10,6",
    "225,8.4,0.635,0.36,2,0.089,15,55,0.99745,3.31,0.57,10.4,4",
    "226,7.7,0.43,0.25,2.6,0.073,29,63,0.99615,3.37,0.58,10.5,6",
    "227,8.9,0.59,0.5,2,0.337,27,81,0.9964,3.04,1.61,9.5,6",
    "228,9,0.82,0.14,2.6,0.089,9,23,0.9984,3.39,0.63,9.8,5",
    "229,7.7,0.43,0.25,2.6,0.073,29,63,0.99615,3.37,0.58,10.5,6",
    "230,6.9,0.52,0.25,2.6,0.081,10,37,0.99685,3.46,0.5,11,5",
    "231,5.2,0.48,0.04,1.6,0.054,19,106,0.9927,3.54,0.62,12.2,7",
    "232,8,0.38,0.06,1.8,0.078,12,49,0.99625,3.37,0.52,9.9,6",
    "233,8.5,0.37,0.2,2.8,0.09,18,58,0.998,3.34,0.7,9.6,6",
    "234,6.9,0.52,0.25,2.6,0.081,10,37,0.99685,3.46,0.5,11,5",
    "235,8.2,1,0.09,2.3,0.065,7,37,0.99685,3.32,0.55,9,6",
    "236,7.2,0.63,0,1.9,0.097,14,38,0.99675,3.37,0.58,9,6",
    "237,7.2,0.63,0,1.9,0.097,14,38,0.99675,3.37,0.58,9,6",
    "238,7.2,0.645,0,1.9,0.097,15,39,0.99675,3.37,0.58,9.2,6",
    "239,7.2,0.63,0,1.9,0.097,14,38,0.99675,3.37,0.58,9,6",
    "240,8.2,1,0.09,2.3,0.065,7,37,0.99685,3.32,0.55,9,6",
    "241,8.9,0.635,0.37,1.7,0.263,5,62,0.9971,3,1.09,9.3,5",
    "242,12,0.38,0.56,2.1,0.093,6,24,0.99925,3.14,0.71,10.9,6",
    "243,7.7,0.58,0.1,1.8,0.102,28,109,0.99565,3.08,0.49,9.8,6",
    "244,15,0.21,0.44,2.2,0.075,10,24,1.00005,3.07,0.84,9.2,7",
    "245,15,0.21,0.44,2.2,0.075,10,24,1.00005,3.07,0.84,9.2,7",
    "246,7.3,0.66,0,2,0.084,6,23,0.9983,3.61,0.96,9.9,6",
    "247,7.1,0.68,0.07,1.9,0.075,16,51,0.99685,3.38,0.52,9.5,5",
    "248,8.2,0.6,0.17,2.3,0.072,11,73,0.9963,3.2,0.45,9.3,5",
    "249,7.7,0.53,0.06,1.7,0.074,9,39,0.99615,3.35,0.48,9.8,6",
    "250,7.3,0.66,0,2,0.084,6,23,0.9983,3.61,0.96,9.9,6",
    "251,10.8,0.32,0.44,1.6,0.063,16,37,0.9985,3.22,0.78,10,6",
    "252,7.1,0.6,0,1.8,0.074,16,34,0.9972,3.47,0.7,9.9,6",
    "253,11.1,0.35,0.48,3.1,0.09,5,21,0.9986,3.17,0.53,10.5,5",
    "254,7.7,0.775,0.42,1.9,0.092,8,86,0.9959,3.23,0.59,9.5,5",
    "255,7.1,0.6,0,1.8,0.074,16,34,0.9972,3.47,0.7,9.9,6",
    "256,8,0.57,0.23,3.2,0.073,17,119,0.99675,3.26,0.57,9.3,5",
    "257,9.4,0.34,0.37,2.2,0.075,5,13,0.998,3.22,0.62,9.2,5",
    "258,6.6,0.695,0,2.1,0.075,12,56,0.9968,3.49,0.67,9.2,5",
    "259,7.7,0.41,0.76,1.8,0.611,8,45,0.9968,3.06,1.26,9.4,5",
    "260,10,0.31,0.47,2.6,0.085,14,33,0.99965,3.36,0.8,10.5,7",
    "261,7.9,0.33,0.23,1.7,0.077,18,45,0.99625,3.29,0.65,9.3,5",
    "262,7,0.975,0.04,2,0.087,12,67,0.99565,3.35,0.6,9.4,4",
    "263,8,0.52,0.03,1.7,0.07,10,35,0.99575,3.34,0.57,10,5",
    "264,7.9,0.37,0.23,1.8,0.077,23,49,0.9963,3.28,0.67,9.3,5",
    "265,12.5,0.56,0.49,2.4,0.064,5,27,0.9999,3.08,0.87,10.9,5",
    "266,11.8,0.26,0.52,1.8,0.071,6,10,0.9968,3.2,0.72,10.2,7",
    "267,8.1,0.87,0,3.3,0.096,26,61,1.00025,3.6,0.72,9.8,4",
    "268,7.9,0.35,0.46,3.6,0.078,15,37,0.9973,3.35,0.86,12.8,8",
    "269,6.9,0.54,0.04,3,0.077,7,27,0.9987,3.69,0.91,9.4,6",
    "270,11.5,0.18,0.51,4,0.104,4,23,0.9996,3.28,0.97,10.1,6",
    "271,7.9,0.545,0.06,4,0.087,27,61,0.9965,3.36,0.67,10.7,6",
    "272,11.5,0.18,0.51,4,0.104,4,23,0.9996,3.28,0.97,10.1,6",
    "273,10.9,0.37,0.58,4,0.071,17,65,0.99935,3.22,0.78,10.1,5",
    "274,8.4,0.715,0.2,2.4,0.076,10,38,0.99735,3.31,0.64,9.4,5",
    "275,7.5,0.65,0.18,7,0.088,27,94,0.99915,3.38,0.77,9.4,5",
    "276,7.9,0.545,0.06,4,0.087,27,61,0.9965,3.36,0.67,10.7,6",
    "277,6.9,0.54,0.04,3,0.077,7,27,0.9987,3.69,0.91,9.4,6",
    "278,11.5,0.18,0.51,4,0.104,4,23,0.9996,3.28,0.97,10.1,6",
    "279,10.3,0.32,0.45,6.4,0.073,5,13,0.9976,3.23,0.82,12.6,8",
    "280,8.9,0.4,0.32,5.6,0.087,10,47,0.9991,3.38,0.77,10.5,7",
    "281,11.4,0.26,0.44,3.6,0.071,6,19,0.9986,3.12,0.82,9.3,6",
    "282,7.7,0.27,0.68,3.5,0.358,5,10,0.9972,3.25,1.08,9.9,7",
    "283,7.6,0.52,0.12,3,0.067,12,53,0.9971,3.36,0.57,9.1,5",
    "284,8.9,0.4,0.32,5.6,0.087,10,47,0.9991,3.38,0.77,10.5,7",
    "285,9.9,0.59,0.07,3.4,0.102,32,71,1.00015,3.31,0.71,9.8,5",
    "286,9.9,0.59,0.07,3.4,0.102,32,71,1.00015,3.31,0.71,9.8,5",
    "287,12,0.45,0.55,2,0.073,25,49,0.9997,3.1,0.76,10.3,6",
    "288,7.5,0.4,0.12,3,0.092,29,53,0.9967,3.37,0.7,10.3,6",
    "289,8.7,0.52,0.09,2.5,0.091,20,49,0.9976,3.34,0.86,10.6,7",
    "290,11.6,0.42,0.53,3.3,0.105,33,98,1.001,3.2,0.95,9.2,5",
    "291,8.7,0.52,0.09,2.5,0.091,20,49,0.9976,3.34,0.86,10.6,7",
    "292,11,0.2,0.48,2,0.343,6,18,0.9979,3.3,0.71,10.5,5",
    "293,10.4,0.55,0.23,2.7,0.091,18,48,0.9994,3.22,0.64,10.3,6",
    "294,6.9,0.36,0.25,2.4,0.098,5,16,0.9964,3.41,0.6,10.1,6",
    "295,13.3,0.34,0.52,3.2,0.094,17,53,1.0014,3.05,0.81,9.5,6",
    "296,10.8,0.5,0.46,2.5,0.073,5,27,1.0001,3.05,0.64,9.5,5",
    "297,10.6,0.83,0.37,2.6,0.086,26,70,0.9981,3.16,0.52,9.9,5",
    "298,7.1,0.63,0.06,2,0.083,8,29,0.99855,3.67,0.73,9.6,5",
    "299,7.2,0.65,0.02,2.3,0.094,5,31,0.9993,3.67,0.8,9.7,5",
    "300,6.9,0.67,0.06,2.1,0.08,8,33,0.99845,3.68,0.71,9.6,5",
    "301,7.5,0.53,0.06,2.6,0.086,20,44,0.9965,3.38,0.59,10.7,6",
    "302,11.1,0.18,0.48,1.5,0.068,7,15,0.9973,3.22,0.64,10.1,6",
    "303,8.3,0.705,0.12,2.6,0.092,12,28,0.9994,3.51,0.72,10,5",
    "304,7.4,0.67,0.12,1.6,0.186,5,21,0.996,3.39,0.54,9.5,5",
    "305,8.4,0.65,0.6,2.1,0.112,12,90,0.9973,3.2,0.52,9.2,5",
    "306,10.3,0.53,0.48,2.5,0.063,6,25,0.9998,3.12,0.59,9.3,6",
    "307,7.6,0.62,0.32,2.2,0.082,7,54,0.9966,3.36,0.52,9.4,5",
    "308,10.3,0.41,0.42,2.4,0.213,6,14,0.9994,3.19,0.62,9.5,6",
    "309,10.3,0.43,0.44,2.4,0.214,5,12,0.9994,3.19,0.63,9.5,6",
    "310,7.4,0.29,0.38,1.7,0.062,9,30,0.9968,3.41,0.53,9.5,6",
    "311,10.3,0.53,0.48,2.5,0.063,6,25,0.9998,3.12,0.59,9.3,6",
    "312,7.9,0.53,0.24,2,0.072,15,105,0.996,3.27,0.54,9.4,6",
    "313,9,0.46,0.31,2.8,0.093,19,98,0.99815,3.32,0.63,9.5,6",
    "314,8.6,0.47,0.3,3,0.076,30,135,0.9976,3.3,0.53,9.4,5",
    "315,7.4,0.36,0.29,2.6,0.087,26,72,0.99645,3.39,0.68,11,5",
    "316,7.1,0.35,0.29,2.5,0.096,20,53,0.9962,3.42,0.65,11,6",
    "317,9.6,0.56,0.23,3.4,0.102,37,92,0.9996,3.3,0.65,10.1,5",
    "318,9.6,0.77,0.12,2.9,0.082,30,74,0.99865,3.3,0.64,10.4,6",
    "319,9.8,0.66,0.39,3.2,0.083,21,59,0.9989,3.37,0.71,11.5,7",
    "320,9.6,0.77,0.12,2.9,0.082,30,74,0.99865,3.3,0.64,10.4,6",
    "321,9.8,0.66,0.39,3.2,0.083,21,59,0.9989,3.37,0.71,11.5,7",
    "322,9.3,0.61,0.26,3.4,0.09,25,87,0.99975,3.24,0.62,9.7,5",
    "323,7.8,0.62,0.05,2.3,0.079,6,18,0.99735,3.29,0.63,9.3,5",
    "324,10.3,0.59,0.42,2.8,0.09,35,73,0.999,3.28,0.7,9.5,6",
    "325,10,0.49,0.2,11,0.071,13,50,1.0015,3.16,0.69,9.2,6",
    "326,10,0.49,0.2,11,0.071,13,50,1.0015,3.16,0.69,9.2,6",
    "327,11.6,0.53,0.66,3.65,0.121,6,14,0.9978,3.05,0.74,11.5,7",
    "328,10.3,0.44,0.5,4.5,0.107,5,13,0.998,3.28,0.83,11.5,5",
    "329,13.4,0.27,0.62,2.6,0.082,6,21,1.0002,3.16,0.67,9.7,6",
    "330,10.7,0.46,0.39,2,0.061,7,15,0.9981,3.18,0.62,9.5,5",
    "331,10.2,0.36,0.64,2.9,0.122,10,41,0.998,3.23,0.66,12.5,6",
    "332,10.2,0.36,0.64,2.9,0.122,10,41,0.998,3.23,0.66,12.5,6",
    "333,8,0.58,0.28,3.2,0.066,21,114,0.9973,3.22,0.54,9.4,6",
    "334,8.4,0.56,0.08,2.1,0.105,16,44,0.9958,3.13,0.52,11,5",
    "335,7.9,0.65,0.01,2.5,0.078,17,38,0.9963,3.34,0.74,11.7,7",
    "336,11.9,0.695,0.53,3.4,0.128,7,21,0.9992,3.17,0.84,12.2,7",
    "337,8.9,0.43,0.45,1.9,0.052,6,16,0.9948,3.35,0.7,12.5,6",
    "338,7.8,0.43,0.32,2.8,0.08,29,58,0.9974,3.31,0.64,10.3,5",
    "339,12.4,0.49,0.58,3,0.103,28,99,1.0008,3.16,1,11.5,6",
    "340,12.5,0.28,0.54,2.3,0.082,12,29,0.9997,3.11,1.36,9.8,7",
    "341,12.2,0.34,0.5,2.4,0.066,10,21,1,3.12,1.18,9.2,6",
    "342,10.6,0.42,0.48,2.7,0.065,5,18,0.9972,3.21,0.87,11.3,6",
    "343,10.9,0.39,0.47,1.8,0.118,6,14,0.9982,3.3,0.75,9.8,6",
    "344,10.9,0.39,0.47,1.8,0.118,6,14,0.9982,3.3,0.75,9.8,6",
    "345,11.9,0.57,0.5,2.6,0.082,6,32,1.0006,3.12,0.78,10.7,6",
    "346,7,0.685,0,1.9,0.067,40,63,0.9979,3.6,0.81,9.9,5",
    "347,6.6,0.815,0.02,2.7,0.072,17,34,0.9955,3.58,0.89,12.3,7",
    "348,13.8,0.49,0.67,3,0.093,6,15,0.9986,3.02,0.93,12,6",
    "349,9.6,0.56,0.31,2.8,0.089,15,46,0.9979,3.11,0.92,10,6",
    "350,9.1,0.785,0,2.6,0.093,11,28,0.9994,3.36,0.86,9.4,6",
    "351,10.7,0.67,0.22,2.7,0.107,17,34,1.0004,3.28,0.98,9.9,6",
    "352,9.1,0.795,0,2.6,0.096,11,26,0.9994,3.35,0.83,9.4,6",
    "353,7.7,0.665,0,2.4,0.09,8,19,0.9974,3.27,0.73,9.3,5",
    "354,13.5,0.53,0.79,4.8,0.12,23,77,1.0018,3.18,0.77,13,5",
    "355,6.1,0.21,0.4,1.4,0.066,40.5,165,0.9912,3.25,0.59,11.9,6",
    "356,6.7,0.75,0.01,2.4,0.078,17,32,0.9955,3.55,0.61,12.8,6",
    "357,11.5,0.41,0.52,3,0.08,29,55,1.0001,3.26,0.88,11,5",
    "358,10.5,0.42,0.66,2.95,0.116,12,29,0.997,3.24,0.75,11.7,7",
    "359,11.9,0.43,0.66,3.1,0.109,10,23,1,3.15,0.85,10.4,7",
    "360,12.6,0.38,0.66,2.6,0.088,10,41,1.001,3.17,0.68,9.8,6",
    "361,8.2,0.7,0.23,2,0.099,14,81,0.9973,3.19,0.7,9.4,5",
    "362,8.6,0.45,0.31,2.6,0.086,21,50,0.9982,3.37,0.91,9.9,6",
    "363,11.9,0.58,0.66,2.5,0.072,6,37,0.9992,3.05,0.56,10,5",
    "364,12.5,0.46,0.63,2,0.071,6,15,0.9988,2.99,0.87,10.2,5",
    "365,12.8,0.615,0.66,5.8,0.083,7,42,1.0022,3.07,0.73,10,7",
    "366,10,0.42,0.5,3.4,0.107,7,21,0.9979,3.26,0.93,11.8,6",
    "367,12.8,0.615,0.66,5.8,0.083,7,42,1.0022,3.07,0.73,10,7",
    "368,10.4,0.575,0.61,2.6,0.076,11,24,1,3.16,0.69,9,5",
    "369,10.3,0.34,0.52,2.8,0.159,15,75,0.9998,3.18,0.64,9.4,5",
    "370,9.4,0.27,0.53,2.4,0.074,6,18,0.9962,3.2,1.13,12,7",
    "371,6.9,0.765,0.02,2.3,0.063,35,63,0.9975,3.57,0.78,9.9,5",
    "372,7.9,0.24,0.4,1.6,0.056,11,25,0.9967,3.32,0.87,8.7,6",
    "373,9.1,0.28,0.48,1.8,0.067,26,46,0.9967,3.32,1.04,10.6,6",
    "374,7.4,0.55,0.22,2.2,0.106,12,72,0.9959,3.05,0.63,9.2,5",
    "375,14,0.41,0.63,3.8,0.089,6,47,1.0014,3.01,0.81,10.8,6",
    "376,11.5,0.54,0.71,4.4,0.124,6,15,0.9984,3.01,0.83,11.8,7",
    "377,11.5,0.45,0.5,3,0.078,19,47,1.0003,3.26,1.11,11,6",
    "378,9.4,0.27,0.53,2.4,0.074,6,18,0.9962,3.2,1.13,12,7",
    "379,11.4,0.625,0.66,6.2,0.088,6,24,0.9988,3.11,0.99,13.3,6",
    "380,8.3,0.42,0.38,2.5,0.094,24,60,0.9979,3.31,0.7,10.8,6",
    "381,8.3,0.26,0.42,2,0.08,11,27,0.9974,3.21,0.8,9.4,6",
    "382,13.7,0.415,0.68,2.9,0.085,17,43,1.0014,3.06,0.8,10,6",
    "383,8.3,0.26,0.42,2,0.08,11,27,0.9974,3.21,0.8,9.4,6",
    "384,8.3,0.26,0.42,2,0.08,11,27,0.9974,3.21,0.8,9.4,6",
    "385,7.7,0.51,0.28,2.1,0.087,23,54,0.998,3.42,0.74,9.2,5",
    "386,7.4,0.63,0.07,2.4,0.09,11,37,0.9979,3.43,0.76,9.7,6",
    "387,7.8,0.54,0.26,2,0.088,23,48,0.9981,3.41,0.74,9.2,6",
    "388,8.3,0.66,0.15,1.9,0.079,17,42,0.9972,3.31,0.54,9.6,6",
    "389,7.8,0.46,0.26,1.9,0.088,23,53,0.9981,3.43,0.74,9.2,6",
    "390,9.6,0.38,0.31,2.5,0.096,16,49,0.9982,3.19,0.7,10,7",
    "391,5.6,0.85,0.05,1.4,0.045,12,88,0.9924,3.56,0.82,12.9,8",
    "392,13.7,0.415,0.68,2.9,0.085,17,43,1.0014,3.06,0.8,10,6",
    "393,9.5,0.37,0.52,2,0.082,6,26,0.998,3.18,0.51,9.5,5",
    "394,8.4,0.665,0.61,2,0.112,13,95,0.997,3.16,0.54,9.1,5",
    "395,12.7,0.6,0.65,2.3,0.063,6,25,0.9997,3.03,0.57,9.9,5",
    "396,12,0.37,0.76,4.2,0.066,7,38,1.0004,3.22,0.6,13,7",
    "397,6.6,0.735,0.02,7.9,0.122,68,124,0.9994,3.47,0.53,9.9,5",
    "398,11.5,0.59,0.59,2.6,0.087,13,49,0.9988,3.18,0.65,11,6",
    "399,11.5,0.59,0.59,2.6,0.087,13,49,0.9988,3.18,0.65,11,6",
    "400,8.7,0.765,0.22,2.3,0.064,9,42,0.9963,3.1,0.55,9.4,5",
    "401,6.6,0.735,0.02,7.9,0.122,68,124,0.9994,3.47,0.53,9.9,5",
    "402,7.7,0.26,0.3,1.7,0.059,20,38,0.9949,3.29,0.47,10.8,6",
    "403,12.2,0.48,0.54,2.6,0.085,19,64,1,3.1,0.61,10.5,6",
    "404,11.4,0.6,0.49,2.7,0.085,10,41,0.9994,3.15,0.63,10.5,6",
    "405,7.7,0.69,0.05,2.7,0.075,15,27,0.9974,3.26,0.61,9.1,5",
    "406,8.7,0.31,0.46,1.4,0.059,11,25,0.9966,3.36,0.76,10.1,6",
    "407,9.8,0.44,0.47,2.5,0.063,9,28,0.9981,3.24,0.65,10.8,6",
    "408,12,0.39,0.66,3,0.093,12,30,0.9996,3.18,0.63,10.8,7",
    "409,10.4,0.34,0.58,3.7,0.174,6,16,0.997,3.19,0.7,11.3,6",
    "410,12.5,0.46,0.49,4.5,0.07,26,49,0.9981,3.05,0.57,9.6,4",
    "411,9,0.43,0.34,2.5,0.08,26,86,0.9987,3.38,0.62,9.5,6",
    "412,9.1,0.45,0.35,2.4,0.08,23,78,0.9987,3.38,0.62,9.5,5",
    "413,7.1,0.735,0.16,1.9,0.1,15,77,0.9966,3.27,0.64,9.3,5",
    "414,9.9,0.4,0.53,6.7,0.097,6,19,0.9986,3.27,0.82,11.7,7",
    "415,8.8,0.52,0.34,2.7,0.087,24,122,0.9982,3.26,0.61,9.5,5",
    "416,8.6,0.725,0.24,6.6,0.117,31,134,1.0014,3.32,1.07,9.3,5",
    "417,10.6,0.48,0.64,2.2,0.111,6,20,0.997,3.26,0.66,11.7,6",
    "418,7,0.58,0.12,1.9,0.091,34,124,0.9956,3.44,0.48,10.5,5",
    "419,11.9,0.38,0.51,2,0.121,7,20,0.9996,3.24,0.76,10.4,6",
    "420,6.8,0.77,0,1.8,0.066,34,52,0.9976,3.62,0.68,9.9,5",
    "421,9.5,0.56,0.33,2.4,0.089,35,67,0.9972,3.28,0.73,11.8,7",
    "422,6.6,0.84,0.03,2.3,0.059,32,48,0.9952,3.52,0.56,12.3,7",
    "423,7.7,0.96,0.2,2,0.047,15,60,0.9955,3.36,0.44,10.9,5",
    "424,10.5,0.24,0.47,2.1,0.066,6,24,0.9978,3.15,0.9,11,7",
    "425,7.7,0.96,0.2,2,0.047,15,60,0.9955,3.36,0.44,10.9,5",
    "426,6.6,0.84,0.03,2.3,0.059,32,48,0.9952,3.52,0.56,12.3,7",
    "427,6.4,0.67,0.08,2.1,0.045,19,48,0.9949,3.49,0.49,11.4,6",
    "428,9.5,0.78,0.22,1.9,0.077,6,32,0.9988,3.26,0.56,10.6,6",
    "429,9.1,0.52,0.33,1.3,0.07,9,30,0.9978,3.24,0.6,9.3,5",
    "430,12.8,0.84,0.63,2.4,0.088,13,35,0.9997,3.1,0.6,10.4,6",
    "431,10.5,0.24,0.47,2.1,0.066,6,24,0.9978,3.15,0.9,11,7",
    "432,7.8,0.55,0.35,2.2,0.074,21,66,0.9974,3.25,0.56,9.2,5",
    "433,11.9,0.37,0.69,2.3,0.078,12,24,0.9958,3,0.65,12.8,6",
    "434,12.3,0.39,0.63,2.3,0.091,6,18,1.0004,3.16,0.49,9.5,5",
    "435,10.4,0.41,0.55,3.2,0.076,22,54,0.9996,3.15,0.89,9.9,6",
    "436,12.3,0.39,0.63,2.3,0.091,6,18,1.0004,3.16,0.49,9.5,5",
    "437,8,0.67,0.3,2,0.06,38,62,0.9958,3.26,0.56,10.2,6",
    "438,11.1,0.45,0.73,3.2,0.066,6,22,0.9986,3.17,0.66,11.2,6",
    "439,10.4,0.41,0.55,3.2,0.076,22,54,0.9996,3.15,0.89,9.9,6",
    "440,7,0.62,0.18,1.5,0.062,7,50,0.9951,3.08,0.6,9.3,5",
    "441,12.6,0.31,0.72,2.2,0.072,6,29,0.9987,2.88,0.82,9.8,8",
    "442,11.9,0.4,0.65,2.15,0.068,7,27,0.9988,3.06,0.68,11.3,6",
    "443,15.6,0.685,0.76,3.7,0.1,6,43,1.0032,2.95,0.68,11.2,7",
    "444,10,0.44,0.49,2.7,0.077,11,19,0.9963,3.23,0.63,11.6,7",
    "445,5.3,0.57,0.01,1.7,0.054,5,27,0.9934,3.57,0.84,12.5,7",
    "446,9.5,0.735,0.1,2.1,0.079,6,31,0.9986,3.23,0.56,10.1,6",
    "447,12.5,0.38,0.6,2.6,0.081,31,72,0.9996,3.1,0.73,10.5,5",
    "448,9.3,0.48,0.29,2.1,0.127,6,16,0.9968,3.22,0.72,11.2,5",
    "449,8.6,0.53,0.22,2,0.1,7,27,0.9967,3.2,0.56,10.2,6",
    "450,11.9,0.39,0.69,2.8,0.095,17,35,0.9994,3.1,0.61,10.8,6",
    "451,11.9,0.39,0.69,2.8,0.095,17,35,0.9994,3.1,0.61,10.8,6",
    "452,8.4,0.37,0.53,1.8,0.413,9,26,0.9979,3.06,1.06,9.1,6",
    "453,6.8,0.56,0.03,1.7,0.084,18,35,0.9968,3.44,0.63,10,6",
    "454,10.4,0.33,0.63,2.8,0.084,5,22,0.9998,3.26,0.74,11.2,7",
    "455,7,0.23,0.4,1.6,0.063,21,67,0.9952,3.5,0.63,11.1,5",
    "456,11.3,0.62,0.67,5.2,0.086,6,19,0.9988,3.22,0.69,13.4,8",
    "457,8.9,0.59,0.39,2.3,0.095,5,22,0.9986,3.37,0.58,10.3,5",
    "458,9.2,0.63,0.21,2.7,0.097,29,65,0.9988,3.28,0.58,9.6,5",
    "459,10.4,0.33,0.63,2.8,0.084,5,22,0.9998,3.26,0.74,11.2,7",
    "460,11.6,0.58,0.66,2.2,0.074,10,47,1.0008,3.25,0.57,9,3",
    "461,9.2,0.43,0.52,2.3,0.083,14,23,0.9976,3.35,0.61,11.3,6",
    "462,8.3,0.615,0.22,2.6,0.087,6,19,0.9982,3.26,0.61,9.3,5",
    "463,11,0.26,0.68,2.55,0.085,10,25,0.997,3.18,0.61,11.8,5",
    "464,8.1,0.66,0.7,2.2,0.098,25,129,0.9972,3.08,0.53,9,5",
    "465,11.5,0.315,0.54,2.1,0.084,5,15,0.9987,2.98,0.7,9.2,6",
    "466,10,0.29,0.4,2.9,0.098,10,26,1.0006,3.48,0.91,9.7,5",
    "467,10.3,0.5,0.42,2,0.069,21,51,0.9982,3.16,0.72,11.5,6",
    "468,8.8,0.46,0.45,2.6,0.065,7,18,0.9947,3.32,0.79,14,6",
    "469,11.4,0.36,0.69,2.1,0.09,6,21,1,3.17,0.62,9.2,6",
    "470,8.7,0.82,0.02,1.2,0.07,36,48,0.9952,3.2,0.58,9.8,5",
    "471,13,0.32,0.65,2.6,0.093,15,47,0.9996,3.05,0.61,10.6,5",
    "472,9.6,0.54,0.42,2.4,0.081,25,52,0.997,3.2,0.71,11.4,6",
    "473,12.5,0.37,0.55,2.6,0.083,25,68,0.9995,3.15,0.82,10.4,6",
    "474,9.9,0.35,0.55,2.1,0.062,5,14,0.9971,3.26,0.79,10.6,5",
    "475,10.5,0.28,0.51,1.7,0.08,10,24,0.9982,3.2,0.89,9.4,6",
    "476,9.6,0.68,0.24,2.2,0.087,5,28,0.9988,3.14,0.6,10.2,5",
    "477,9.3,0.27,0.41,2,0.091,6,16,0.998,3.28,0.7,9.7,5",
    "478,10.4,0.24,0.49,1.8,0.075,6,20,0.9977,3.18,1.06,11,6",
    "479,9.6,0.68,0.24,2.2,0.087,5,28,0.9988,3.14,0.6,10.2,5",
    "480,9.4,0.685,0.11,2.7,0.077,6,31,0.9984,3.19,0.7,10.1,6",
    "481,10.6,0.28,0.39,15.5,0.069,6,23,1.0026,3.12,0.66,9.2,5",
    "482,9.4,0.3,0.56,2.8,0.08,6,17,0.9964,3.15,0.92,11.7,8",
    "483,10.6,0.36,0.59,2.2,0.152,6,18,0.9986,3.04,1.05,9.4,5",
    "484,10.6,0.36,0.6,2.2,0.152,7,18,0.9986,3.04,1.06,9.4,5",
    "485,10.6,0.44,0.68,4.1,0.114,6,24,0.997,3.06,0.66,13.4,6",
    "486,10.2,0.67,0.39,1.9,0.054,6,17,0.9976,3.17,0.47,10,5",
    "487,10.2,0.67,0.39,1.9,0.054,6,17,0.9976,3.17,0.47,10,5",
    "488,10.2,0.645,0.36,1.8,0.053,5,14,0.9982,3.17,0.42,10,6",
    "489,11.6,0.32,0.55,2.8,0.081,35,67,1.0002,3.32,0.92,10.8,7",
    "490,9.3,0.39,0.4,2.6,0.073,10,26,0.9984,3.34,0.75,10.2,6",
    "491,9.3,0.775,0.27,2.8,0.078,24,56,0.9984,3.31,0.67,10.6,6",
    "492,9.2,0.41,0.5,2.5,0.055,12,25,0.9952,3.34,0.79,13.3,7",
    "493,8.9,0.4,0.51,2.6,0.052,13,27,0.995,3.32,0.9,13.4,7",
    "494,8.7,0.69,0.31,3,0.086,23,81,1.0002,3.48,0.74,11.6,6",
    "495,6.5,0.39,0.23,8.3,0.051,28,91,0.9952,3.44,0.55,12.1,6",
    "496,10.7,0.35,0.53,2.6,0.07,5,16,0.9972,3.15,0.65,11,8",
    "497,7.8,0.52,0.25,1.9,0.081,14,38,0.9984,3.43,0.65,9,6",
    "498,7.2,0.34,0.32,2.5,0.09,43,113,0.9966,3.32,0.79,11.1,5",
    "499,10.7,0.35,0.53,2.6,0.07,5,16,0.9972,3.15,0.65,11,8",
    "500,8.7,0.69,0.31,3,0.086,23,81,1.0002,3.48,0.74,11.6,6",
    "501,7.8,0.52,0.25,1.9,0.081,14,38,0.9984,3.43,0.65,9,6",
    "502,10.4,0.44,0.73,6.55,0.074,38,76,0.999,3.17,0.85,12,7",
    "503,10.4,0.44,0.73,6.55,0.074,38,76,0.999,3.17,0.85,12,7",
    "504,10.5,0.26,0.47,1.9,0.078,6,24,0.9976,3.18,1.04,10.9,7",
    "505,10.5,0.24,0.42,1.8,0.077,6,22,0.9976,3.21,1.05,10.8,7",
    "506,10.2,0.49,0.63,2.9,0.072,10,26,0.9968,3.16,0.78,12.5,7",
    "507,10.4,0.24,0.46,1.8,0.075,6,21,0.9976,3.25,1.02,10.8,7",
    "508,11.2,0.67,0.55,2.3,0.084,6,13,1,3.17,0.71,9.5,6",
    "509,10,0.59,0.31,2.2,0.09,26,62,0.9994,3.18,0.63,10.2,6",
    "510,13.3,0.29,0.75,2.8,0.084,23,43,0.9986,3.04,0.68,11.4,7",
    "511,12.4,0.42,0.49,4.6,0.073,19,43,0.9978,3.02,0.61,9.5,5",
    "512,10,0.59,0.31,2.2,0.09,26,62,0.9994,3.18,0.63,10.2,6",
    "513,10.7,0.4,0.48,2.1,0.125,15,49,0.998,3.03,0.81,9.7,6",
    "514,10.5,0.51,0.64,2.4,0.107,6,15,0.9973,3.09,0.66,11.8,7",
    "515,10.5,0.51,0.64,2.4,0.107,6,15,0.9973,3.09,0.66,11.8,7",
    "516,8.5,0.655,0.49,6.1,0.122,34,151,1.001,3.31,1.14,9.3,5",
    "517,12.5,0.6,0.49,4.3,0.1,5,14,1.001,3.25,0.74,11.9,6",
    "518,10.4,0.61,0.49,2.1,0.2,5,16,0.9994,3.16,0.63,8.4,3",
    "519,10.9,0.21,0.49,2.8,0.088,11,32,0.9972,3.22,0.68,11.7,6",
    "520,7.3,0.365,0.49,2.5,0.088,39,106,0.9966,3.36,0.78,11,5",
    "521,9.8,0.25,0.49,2.7,0.088,15,33,0.9982,3.42,0.9,10,6",
    "522,7.6,0.41,0.49,2,0.088,16,43,0.998,3.48,0.64,9.1,5",
    "523,8.2,0.39,0.49,2.3,0.099,47,133,0.9979,3.38,0.99,9.8,5",
    "524,9.3,0.4,0.49,2.5,0.085,38,142,0.9978,3.22,0.55,9.4,5",
    "525,9.2,0.43,0.49,2.4,0.086,23,116,0.9976,3.23,0.64,9.5,5",
    "526,10.4,0.64,0.24,2.8,0.105,29,53,0.9998,3.24,0.67,9.9,5",
    "527,7.3,0.365,0.49,2.5,0.088,39,106,0.9966,3.36,0.78,11,5",
    "528,7,0.38,0.49,2.5,0.097,33,85,0.9962,3.39,0.77,11.4,6",
    "529,8.2,0.42,0.49,2.6,0.084,32,55,0.9988,3.34,0.75,8.7,6",
    "530,9.9,0.63,0.24,2.4,0.077,6,33,0.9974,3.09,0.57,9.4,5",
    "531,9.1,0.22,0.24,2.1,0.078,1,28,0.999,3.41,0.87,10.3,6",
    "532,11.9,0.38,0.49,2.7,0.098,12,42,1.0004,3.16,0.61,10.3,5",
    "533,11.9,0.38,0.49,2.7,0.098,12,42,1.0004,3.16,0.61,10.3,5",
    "534,10.3,0.27,0.24,2.1,0.072,15,33,0.9956,3.22,0.66,12.8,6",
    "535,10,0.48,0.24,2.7,0.102,13,32,1,3.28,0.56,10,6",
    "536,9.1,0.22,0.24,2.1,0.078,1,28,0.999,3.41,0.87,10.3,6",
    "537,9.9,0.63,0.24,2.4,0.077,6,33,0.9974,3.09,0.57,9.4,5",
    "538,8.1,0.825,0.24,2.1,0.084,5,13,0.9972,3.37,0.77,10.7,6",
    "539,12.9,0.35,0.49,5.8,0.066,5,35,1.0014,3.2,0.66,12,7",
    "540,11.2,0.5,0.74,5.15,0.1,5,17,0.9996,3.22,0.62,11.2,5",
    "541,9.2,0.59,0.24,3.3,0.101,20,47,0.9988,3.26,0.67,9.6,5",
    "542,9.5,0.46,0.49,6.3,0.064,5,17,0.9988,3.21,0.73,11,6",
    "543,9.3,0.715,0.24,2.1,0.07,5,20,0.9966,3.12,0.59,9.9,5",
    "544,11.2,0.66,0.24,2.5,0.085,16,53,0.9993,3.06,0.72,11,6",
    "545,14.3,0.31,0.74,1.8,0.075,6,15,1.0008,2.86,0.79,8.4,6",
    "546,9.1,0.47,0.49,2.6,0.094,38,106,0.9982,3.08,0.59,9.1,5",
    "547,7.5,0.55,0.24,2,0.078,10,28,0.9983,3.45,0.78,9.5,6",
    "548,10.6,0.31,0.49,2.5,0.067,6,21,0.9987,3.26,0.86,10.7,6",
    "549,12.4,0.35,0.49,2.6,0.079,27,69,0.9994,3.12,0.75,10.4,6",
    "550,9,0.53,0.49,1.9,0.171,6,25,0.9975,3.27,0.61,9.4,6",
    "551,6.8,0.51,0.01,2.1,0.074,9,25,0.9958,3.33,0.56,9.5,6",
    "552,9.4,0.43,0.24,2.8,0.092,14,45,0.998,3.19,0.73,10,6",
    "553,9.5,0.46,0.24,2.7,0.092,14,44,0.998,3.12,0.74,10,6",
    "554,5,1.04,0.24,1.6,0.05,32,96,0.9934,3.74,0.62,11.5,5",
    "555,15.5,0.645,0.49,4.2,0.095,10,23,1.00315,2.92,0.74,11.1,5",
    "556,15.5,0.645,0.49,4.2,0.095,10,23,1.00315,2.92,0.74,11.1,5",
    "557,10.9,0.53,0.49,4.6,0.118,10,17,1.0002,3.07,0.56,11.7,6",
    "558,15.6,0.645,0.49,4.2,0.095,10,23,1.00315,2.92,0.74,11.1,5",
    "559,10.9,0.53,0.49,4.6,0.118,10,17,1.0002,3.07,0.56,11.7,6",
    "560,13,0.47,0.49,4.3,0.085,6,47,1.0021,3.3,0.68,12.7,6",
    "561,12.7,0.6,0.49,2.8,0.075,5,19,0.9994,3.14,0.57,11.4,5",
    "562,9,0.44,0.49,2.4,0.078,26,121,0.9978,3.23,0.58,9.2,5",
    "563,9,0.54,0.49,2.9,0.094,41,110,0.9982,3.08,0.61,9.2,5",
    "564,7.6,0.29,0.49,2.7,0.092,25,60,0.9971,3.31,0.61,10.1,6",
    "565,13,0.47,0.49,4.3,0.085,6,47,1.0021,3.3,0.68,12.7,6",
    "566,12.7,0.6,0.49,2.8,0.075,5,19,0.9994,3.14,0.57,11.4,5",
    "567,8.7,0.7,0.24,2.5,0.226,5,15,0.9991,3.32,0.6,9,6",
    "568,8.7,0.7,0.24,2.5,0.226,5,15,0.9991,3.32,0.6,9,6",
    "569,9.8,0.5,0.49,2.6,0.25,5,20,0.999,3.31,0.79,10.7,6",
    "570,6.2,0.36,0.24,2.2,0.095,19,42,0.9946,3.57,0.57,11.7,6",
    "571,11.5,0.35,0.49,3.3,0.07,10,37,1.0003,3.32,0.91,11,6",
    "572,6.2,0.36,0.24,2.2,0.095,19,42,0.9946,3.57,0.57,11.7,6",
    "573,10.2,0.24,0.49,2.4,0.075,10,28,0.9978,3.14,0.61,10.4,5",
    "574,10.5,0.59,0.49,2.1,0.07,14,47,0.9991,3.3,0.56,9.6,4",
    "575,10.6,0.34,0.49,3.2,0.078,20,78,0.9992,3.19,0.7,10,6",
    "576,12.3,0.27,0.49,3.1,0.079,28,46,0.9993,3.2,0.8,10.2,6",
    "577,9.9,0.5,0.24,2.3,0.103,6,14,0.9978,3.34,0.52,10,4",
    "578,8.8,0.44,0.49,2.8,0.083,18,111,0.9982,3.3,0.6,9.5,5",
    "579,8.8,0.47,0.49,2.9,0.085,17,110,0.9982,3.29,0.6,9.8,5",
    "580,10.6,0.31,0.49,2.2,0.063,18,40,0.9976,3.14,0.51,9.8,6",
    "581,12.3,0.5,0.49,2.2,0.089,5,14,1.0002,3.19,0.44,9.6,5",
    "582,12.3,0.5,0.49,2.2,0.089,5,14,1.0002,3.19,0.44,9.6,5",
    "583,11.7,0.49,0.49,2.2,0.083,5,15,1,3.19,0.43,9.2,5",
    "584,12,0.28,0.49,1.9,0.074,10,21,0.9976,2.98,0.66,9.9,7",
    "585,11.8,0.33,0.49,3.4,0.093,54,80,1.0002,3.3,0.76,10.7,7",
    "586,7.6,0.51,0.24,2.4,0.091,8,38,0.998,3.47,0.66,9.6,6",
    "587,11.1,0.31,0.49,2.7,0.094,16,47,0.9986,3.12,1.02,10.6,7",
    "588,7.3,0.73,0.24,1.9,0.108,18,102,0.9967,3.26,0.59,9.3,5",
    "589,5,0.42,0.24,2,0.06,19,50,0.9917,3.72,0.74,14,8",
    "590,10.2,0.29,0.49,2.6,0.059,5,13,0.9976,3.05,0.74,10.5,7",
    "591,9,0.45,0.49,2.6,0.084,21,75,0.9987,3.35,0.57,9.7,5",
    "592,6.6,0.39,0.49,1.7,0.07,23,149,0.9922,3.12,0.5,11.5,6",
    "593,9,0.45,0.49,2.6,0.084,21,75,0.9987,3.35,0.57,9.7,5",
    "594,9.9,0.49,0.58,3.5,0.094,9,43,1.0004,3.29,0.58,9,5",
    "595,7.9,0.72,0.17,2.6,0.096,20,38,0.9978,3.4,0.53,9.5,5",
    "596,8.9,0.595,0.41,7.9,0.086,30,109,0.9998,3.27,0.57,9.3,5",
    "597,12.4,0.4,0.51,2,0.059,6,24,0.9994,3.04,0.6,9.3,6",
    "598,11.9,0.58,0.58,1.9,0.071,5,18,0.998,3.09,0.63,10,6",
    "599,8.5,0.585,0.18,2.1,0.078,5,30,0.9967,3.2,0.48,9.8,6",
    "600,12.7,0.59,0.45,2.3,0.082,11,22,1,3,0.7,9.3,6",
    "601,8.2,0.915,0.27,2.1,0.088,7,23,0.9962,3.26,0.47,10,4",
    "602,13.2,0.46,0.52,2.2,0.071,12,35,1.0006,3.1,0.56,9,6",
    "603,7.7,0.835,0,2.6,0.081,6,14,0.9975,3.3,0.52,9.3,5",
    "604,13.2,0.46,0.52,2.2,0.071,12,35,1.0006,3.1,0.56,9,6",
    "605,8.3,0.58,0.13,2.9,0.096,14,63,0.9984,3.17,0.62,9.1,6",
    "606,8.3,0.6,0.13,2.6,0.085,6,24,0.9984,3.31,0.59,9.2,6",
    "607,9.4,0.41,0.48,4.6,0.072,10,20,0.9973,3.34,0.79,12.2,7",
    "608,8.8,0.48,0.41,3.3,0.092,26,52,0.9982,3.31,0.53,10.5,6",
    "609,10.1,0.65,0.37,5.1,0.11,11,65,1.0026,3.32,0.64,10.4,6",
    "610,6.3,0.36,0.19,3.2,0.075,15,39,0.9956,3.56,0.52,12.7,6",
    "611,8.8,0.24,0.54,2.5,0.083,25,57,0.9983,3.39,0.54,9.2,5",
    "612,13.2,0.38,0.55,2.7,0.081,5,16,1.0006,2.98,0.54,9.4,5",
    "613,7.5,0.64,0,2.4,0.077,18,29,0.9965,3.32,0.6,10,6",
    "614,8.2,0.39,0.38,1.5,0.058,10,29,0.9962,3.26,0.74,9.8,5",
    "615,9.2,0.755,0.18,2.2,0.148,10,103,0.9969,2.87,1.36,10.2,6",
    "616,9.6,0.6,0.5,2.3,0.079,28,71,0.9997,3.5,0.57,9.7,5",
    "617,9.6,0.6,0.5,2.3,0.079,28,71,0.9997,3.5,0.57,9.7,5",
    "618,11.5,0.31,0.51,2.2,0.079,14,28,0.9982,3.03,0.93,9.8,6",
    "619,11.4,0.46,0.5,2.7,0.122,4,17,1.0006,3.13,0.7,10.2,5",
    "620,11.3,0.37,0.41,2.3,0.088,6,16,0.9988,3.09,0.8,9.3,5",
    "621,8.3,0.54,0.24,3.4,0.076,16,112,0.9976,3.27,0.61,9.4,5",
    "622,8.2,0.56,0.23,3.4,0.078,14,104,0.9976,3.28,0.62,9.4,5",
    "623,10,0.58,0.22,1.9,0.08,9,32,0.9974,3.13,0.55,9.5,5",
    "624,7.9,0.51,0.25,2.9,0.077,21,45,0.9974,3.49,0.96,12.1,6",
    "625,6.8,0.69,0,5.6,0.124,21,58,0.9997,3.46,0.72,10.2,5",
    "626,6.8,0.69,0,5.6,0.124,21,58,0.9997,3.46,0.72,10.2,5",
    "627,8.8,0.6,0.29,2.2,0.098,5,15,0.9988,3.36,0.49,9.1,5",
    "628,8.8,0.6,0.29,2.2,0.098,5,15,0.9988,3.36,0.49,9.1,5",
    "629,8.7,0.54,0.26,2.5,0.097,7,31,0.9976,3.27,0.6,9.3,6",
    "630,7.6,0.685,0.23,2.3,0.111,20,84,0.9964,3.21,0.61,9.3,5",
    "631,8.7,0.54,0.26,2.5,0.097,7,31,0.9976,3.27,0.6,9.3,6",
    "632,10.4,0.28,0.54,2.7,0.105,5,19,0.9988,3.25,0.63,9.5,5",
    "633,7.6,0.41,0.14,3,0.087,21,43,0.9964,3.32,0.57,10.5,6",
    "634,10.1,0.935,0.22,3.4,0.105,11,86,1.001,3.43,0.64,11.3,4",
    "635,7.9,0.35,0.21,1.9,0.073,46,102,0.9964,3.27,0.58,9.5,5",
    "636,8.7,0.84,0,1.4,0.065,24,33,0.9954,3.27,0.55,9.7,5",
    "637,9.6,0.88,0.28,2.4,0.086,30,147,0.9979,3.24,0.53,9.4,5",
    "638,9.5,0.885,0.27,2.3,0.084,31,145,0.9978,3.24,0.53,9.4,5",
    "639,7.7,0.915,0.12,2.2,0.143,7,23,0.9964,3.35,0.65,10.2,7",
    "640,8.9,0.29,0.35,1.9,0.067,25,57,0.997,3.18,1.36,10.3,6",
    "641,9.9,0.54,0.45,2.3,0.071,16,40,0.9991,3.39,0.62,9.4,5",
    "642,9.5,0.59,0.44,2.3,0.071,21,68,0.9992,3.46,0.63,9.5,5",
    "643,9.9,0.54,0.45,2.3,0.071,16,40,0.9991,3.39,0.62,9.4,5",
    "644,9.5,0.59,0.44,2.3,0.071,21,68,0.9992,3.46,0.63,9.5,5",
    "645,9.9,0.54,0.45,2.3,0.071,16,40,0.9991,3.39,0.62,9.4,5",
    "646,7.8,0.64,0.1,6,0.115,5,11,0.9984,3.37,0.69,10.1,7",
    "647,7.3,0.67,0.05,3.6,0.107,6,20,0.9972,3.4,0.63,10.1,5",
    "648,8.3,0.845,0.01,2.2,0.07,5,14,0.9967,3.32,0.58,11,4",
    "649,8.7,0.48,0.3,2.8,0.066,10,28,0.9964,3.33,0.67,11.2,7",
    "650,6.7,0.42,0.27,8.6,0.068,24,148,0.9948,3.16,0.57,11.3,6",
    "651,10.7,0.43,0.39,2.2,0.106,8,32,0.9986,2.89,0.5,9.6,5",
    "652,9.8,0.88,0.25,2.5,0.104,35,155,1.001,3.41,0.67,11.2,5",
    "653,15.9,0.36,0.65,7.5,0.096,22,71,0.9976,2.98,0.84,14.9,5",
    "654,9.4,0.33,0.59,2.8,0.079,9,30,0.9976,3.12,0.54,12,6",
    "655,8.6,0.47,0.47,2.4,0.074,7,29,0.9979,3.08,0.46,9.5,5",
    "656,9.7,0.55,0.17,2.9,0.087,20,53,1.0004,3.14,0.61,9.4,5",
    "657,10.7,0.43,0.39,2.2,0.106,8,32,0.9986,2.89,0.5,9.6,5",
    "658,12,0.5,0.59,1.4,0.073,23,42,0.998,2.92,0.68,10.5,7",
    "659,7.2,0.52,0.07,1.4,0.074,5,20,0.9973,3.32,0.81,9.6,6",
    "660,7.1,0.84,0.02,4.4,0.096,5,13,0.997,3.41,0.57,11,4",
    "661,7.2,0.52,0.07,1.4,0.074,5,20,0.9973,3.32,0.81,9.6,6",
    "662,7.5,0.42,0.31,1.6,0.08,15,42,0.9978,3.31,0.64,9,5",
    "663,7.2,0.57,0.06,1.6,0.076,9,27,0.9972,3.36,0.7,9.6,6",
    "664,10.1,0.28,0.46,1.8,0.05,5,13,0.9974,3.04,0.79,10.2,6",
    "665,12.1,0.4,0.52,2,0.092,15,54,1,3.03,0.66,10.2,5",
    "666,9.4,0.59,0.14,2,0.084,25,48,0.9981,3.14,0.56,9.7,5",
    "667,8.3,0.49,0.36,1.8,0.222,6,16,0.998,3.18,0.6,9.5,6",
    "668,11.3,0.34,0.45,2,0.082,6,15,0.9988,2.94,0.66,9.2,6",
    "669,10,0.73,0.43,2.3,0.059,15,31,0.9966,3.15,0.57,11,5",
    "670,11.3,0.34,0.45,2,0.082,6,15,0.9988,2.94,0.66,9.2,6",
    "671,6.9,0.4,0.24,2.5,0.083,30,45,0.9959,3.26,0.58,10,5",
    "672,8.2,0.73,0.21,1.7,0.074,5,13,0.9968,3.2,0.52,9.5,5",
    "673,9.8,1.24,0.34,2,0.079,32,151,0.998,3.15,0.53,9.5,5",
    "674,8.2,0.73,0.21,1.7,0.074,5,13,0.9968,3.2,0.52,9.5,5",
    "675,10.8,0.4,0.41,2.2,0.084,7,17,0.9984,3.08,0.67,9.3,6",
    "676,9.3,0.41,0.39,2.2,0.064,12,31,0.9984,3.26,0.65,10.2,5",
    "677,10.8,0.4,0.41,2.2,0.084,7,17,0.9984,3.08,0.67,9.3,6",
    "678,8.6,0.8,0.11,2.3,0.084,12,31,0.9979,3.4,0.48,9.9,5",
    "679,8.3,0.78,0.1,2.6,0.081,45,87,0.9983,3.48,0.53,10,5",
    "680,10.8,0.26,0.45,3.3,0.06,20,49,0.9972,3.13,0.54,9.6,5",
    "681,13.3,0.43,0.58,1.9,0.07,15,40,1.0004,3.06,0.49,9,5",
    "682,8,0.45,0.23,2.2,0.094,16,29,0.9962,3.21,0.49,10.2,6",
    "683,8.5,0.46,0.31,2.25,0.078,32,58,0.998,3.33,0.54,9.8,5",
    "684,8.1,0.78,0.23,2.6,0.059,5,15,0.997,3.37,0.56,11.3,5",
    "685,9.8,0.98,0.32,2.3,0.078,35,152,0.998,3.25,0.48,9.4,5",
    "686,8.1,0.78,0.23,2.6,0.059,5,15,0.997,3.37,0.56,11.3,5",
    "687,7.1,0.65,0.18,1.8,0.07,13,40,0.997,3.44,0.6,9.1,5",
    "688,9.1,0.64,0.23,3.1,0.095,13,38,0.9998,3.28,0.59,9.7,5",
    "689,7.7,0.66,0.04,1.6,0.039,4,9,0.9962,3.4,0.47,9.4,5",
    "690,8.1,0.38,0.48,1.8,0.157,5,17,0.9976,3.3,1.05,9.4,5",
    "691,7.4,1.185,0,4.25,0.097,5,14,0.9966,3.63,0.54,10.7,3",
    "692,9.2,0.92,0.24,2.6,0.087,12,93,0.9998,3.48,0.54,9.8,5",
    "693,8.6,0.49,0.51,2,0.422,16,62,0.9979,3.03,1.17,9,5",
    "694,9,0.48,0.32,2.8,0.084,21,122,0.9984,3.32,0.62,9.4,5",
    "695,9,0.47,0.31,2.7,0.084,24,125,0.9984,3.31,0.61,9.4,5",
    "696,5.1,0.47,0.02,1.3,0.034,18,44,0.9921,3.9,0.62,12.8,6",
    "697,7,0.65,0.02,2.1,0.066,8,25,0.9972,3.47,0.67,9.5,6",
    "698,7,0.65,0.02,2.1,0.066,8,25,0.9972,3.47,0.67,9.5,6",
    "699,9.4,0.615,0.28,3.2,0.087,18,72,1.0001,3.31,0.53,9.7,5",
    "700,11.8,0.38,0.55,2.1,0.071,5,19,0.9986,3.11,0.62,10.8,6",
    "701,10.6,1.02,0.43,2.9,0.076,26,88,0.9984,3.08,0.57,10.1,6",
    "702,7,0.65,0.02,2.1,0.066,8,25,0.9972,3.47,0.67,9.5,6",
    "703,7,0.64,0.02,2.1,0.067,9,23,0.997,3.47,0.67,9.4,6",
    "704,7.5,0.38,0.48,2.6,0.073,22,84,0.9972,3.32,0.7,9.6,4",
    "705,9.1,0.765,0.04,1.6,0.078,4,14,0.998,3.29,0.54,9.7,4",
    "706,8.4,1.035,0.15,6,0.073,11,54,0.999,3.37,0.49,9.9,5",
    "707,7,0.78,0.08,2,0.093,10,19,0.9956,3.4,0.47,10,5",
    "708,7.4,0.49,0.19,3,0.077,16,37,0.9966,3.37,0.51,10.5,5",
    "709,7.8,0.545,0.12,2.5,0.068,11,35,0.996,3.34,0.61,11.6,6",
    "710,9.7,0.31,0.47,1.6,0.062,13,33,0.9983,3.27,0.66,10,6",
    "711,10.6,1.025,0.43,2.8,0.08,21,84,0.9985,3.06,0.57,10.1,5",
    "712,8.9,0.565,0.34,3,0.093,16,112,0.9998,3.38,0.61,9.5,5",
    "713,8.7,0.69,0,3.2,0.084,13,33,0.9992,3.36,0.45,9.4,5",
    "714,8,0.43,0.36,2.3,0.075,10,48,0.9976,3.34,0.46,9.4,5",
    "715,9.9,0.74,0.28,2.6,0.078,21,77,0.998,3.28,0.51,9.8,5",
    "716,7.2,0.49,0.18,2.7,0.069,13,34,0.9967,3.29,0.48,9.2,6",
    "717,8,0.43,0.36,2.3,0.075,10,48,0.9976,3.34,0.46,9.4,5",
    "718,7.6,0.46,0.11,2.6,0.079,12,49,0.9968,3.21,0.57,10,5",
    "719,8.4,0.56,0.04,2,0.082,10,22,0.9976,3.22,0.44,9.6,5",
    "720,7.1,0.66,0,3.9,0.086,17,45,0.9976,3.46,0.54,9.5,5",
    "721,8.4,0.56,0.04,2,0.082,10,22,0.9976,3.22,0.44,9.6,5",
    "722,8.9,0.48,0.24,2.85,0.094,35,106,0.9982,3.1,0.53,9.2,5",
    "723,7.6,0.42,0.08,2.7,0.084,15,48,0.9968,3.21,0.59,10,5",
    "724,7.1,0.31,0.3,2.2,0.053,36,127,0.9965,2.94,1.62,9.5,5",
    "725,7.5,1.115,0.1,3.1,0.086,5,12,0.9958,3.54,0.6,11.2,4",
    "726,9,0.66,0.17,3,0.077,5,13,0.9976,3.29,0.55,10.4,5",
    "727,8.1,0.72,0.09,2.8,0.084,18,49,0.9994,3.43,0.72,11.1,6",
    "728,6.4,0.57,0.02,1.8,0.067,4,11,0.997,3.46,0.68,9.5,5",
    "729,6.4,0.57,0.02,1.8,0.067,4,11,0.997,3.46,0.68,9.5,5",
    "730,6.4,0.865,0.03,3.2,0.071,27,58,0.995,3.61,0.49,12.7,6",
    "731,9.5,0.55,0.66,2.3,0.387,12,37,0.9982,3.17,0.67,9.6,5",
    "732,8.9,0.875,0.13,3.45,0.088,4,14,0.9994,3.44,0.52,11.5,5",
    "733,7.3,0.835,0.03,2.1,0.092,10,19,0.9966,3.39,0.47,9.6,5",
    "734,7,0.45,0.34,2.7,0.082,16,72,0.998,3.55,0.6,9.5,5",
    "735,7.7,0.56,0.2,2,0.075,9,39,0.9987,3.48,0.62,9.3,5",
    "736,7.7,0.965,0.1,2.1,0.112,11,22,0.9963,3.26,0.5,9.5,5",
    "737,7.7,0.965,0.1,2.1,0.112,11,22,0.9963,3.26,0.5,9.5,5",
    "738,8.2,0.59,0,2.5,0.093,19,58,1.0002,3.5,0.65,9.3,6",
    "739,9,0.46,0.23,2.8,0.092,28,104,0.9983,3.1,0.56,9.2,5",
    "740,9,0.69,0,2.4,0.088,19,38,0.999,3.35,0.6,9.3,5",
    "741,8.3,0.76,0.29,4.2,0.075,12,16,0.9965,3.45,0.68,11.5,6",
    "742,9.2,0.53,0.24,2.6,0.078,28,139,0.99788,3.21,0.57,9.5,5",
    "743,6.5,0.615,0,1.9,0.065,9,18,0.9972,3.46,0.65,9.2,5",
    "744,11.6,0.41,0.58,2.8,0.096,25,101,1.00024,3.13,0.53,10,5",
    "745,11.1,0.39,0.54,2.7,0.095,21,101,1.0001,3.13,0.51,9.5,5",
    "746,7.3,0.51,0.18,2.1,0.07,12,28,0.99768,3.52,0.73,9.5,6",
    "747,8.2,0.34,0.38,2.5,0.08,12,57,0.9978,3.3,0.47,9,6",
    "748,8.6,0.33,0.4,2.6,0.083,16,68,0.99782,3.3,0.48,9.4,5",
    "749,7.2,0.5,0.18,2.1,0.071,12,31,0.99761,3.52,0.72,9.6,6",
    "750,7.3,0.51,0.18,2.1,0.07,12,28,0.99768,3.52,0.73,9.5,6",
    "751,8.3,0.65,0.1,2.9,0.089,17,40,0.99803,3.29,0.55,9.5,5",
    "752,8.3,0.65,0.1,2.9,0.089,17,40,0.99803,3.29,0.55,9.5,5",
    "753,7.6,0.54,0.13,2.5,0.097,24,66,0.99785,3.39,0.61,9.4,5",
    "754,8.3,0.65,0.1,2.9,0.089,17,40,0.99803,3.29,0.55,9.5,5",
    "755,7.8,0.48,0.68,1.7,0.415,14,32,0.99656,3.09,1.06,9.1,6",
    "756,7.8,0.91,0.07,1.9,0.058,22,47,0.99525,3.51,0.43,10.7,6",
    "757,6.3,0.98,0.01,2,0.057,15,33,0.99488,3.6,0.46,11.2,6",
    "758,8.1,0.87,0,2.2,0.084,10,31,0.99656,3.25,0.5,9.8,5",
    "759,8.1,0.87,0,2.2,0.084,10,31,0.99656,3.25,0.5,9.8,5",
    "760,8.8,0.42,0.21,2.5,0.092,33,88,0.99823,3.19,0.52,9.2,5",
    "761,9,0.58,0.25,2.8,0.075,9,104,0.99779,3.23,0.57,9.7,5",
    "762,9.3,0.655,0.26,2,0.096,5,35,0.99738,3.25,0.42,9.6,5",
    "763,8.8,0.7,0,1.7,0.069,8,19,0.99701,3.31,0.53,10,6",
    "764,9.3,0.655,0.26,2,0.096,5,35,0.99738,3.25,0.42,9.6,5",
    "765,9.1,0.68,0.11,2.8,0.093,11,44,0.99888,3.31,0.55,9.5,6",
    "766,9.2,0.67,0.1,3,0.091,12,48,0.99888,3.31,0.54,9.5,6",
    "767,8.8,0.59,0.18,2.9,0.089,12,74,0.99738,3.14,0.54,9.4,5",
    "768,7.5,0.6,0.32,2.7,0.103,13,98,0.99938,3.45,0.62,9.5,5",
    "769,7.1,0.59,0.02,2.3,0.082,24,94,0.99744,3.55,0.53,9.7,6",
    "770,7.9,0.72,0.01,1.9,0.076,7,32,0.99668,3.39,0.54,9.6,5",
    "771,7.1,0.59,0.02,2.3,0.082,24,94,0.99744,3.55,0.53,9.7,6",
    "772,9.4,0.685,0.26,2.4,0.082,23,143,0.9978,3.28,0.55,9.4,5",
    "773,9.5,0.57,0.27,2.3,0.082,23,144,0.99782,3.27,0.55,9.4,5",
    "774,7.9,0.4,0.29,1.8,0.157,1,44,0.9973,3.3,0.92,9.5,6",
    "775,7.9,0.4,0.3,1.8,0.157,2,45,0.99727,3.31,0.91,9.5,6",
    "776,7.2,1,0,3,0.102,7,16,0.99586,3.43,0.46,10,5",
    "777,6.9,0.765,0.18,2.4,0.243,5.5,48,0.99612,3.4,0.6,10.3,6",
    "778,6.9,0.635,0.17,2.4,0.241,6,18,0.9961,3.4,0.59,10.3,6",
    "779,8.3,0.43,0.3,3.4,0.079,7,34,0.99788,3.36,0.61,10.5,5",
    "780,7.1,0.52,0.03,2.6,0.076,21,92,0.99745,3.5,0.6,9.8,5",
    "781,7,0.57,0,2,0.19,12,45,0.99676,3.31,0.6,9.4,6",
    "782,6.5,0.46,0.14,2.4,0.114,9,37,0.99732,3.66,0.65,9.8,5",
    "783,9,0.82,0.05,2.4,0.081,26,96,0.99814,3.36,0.53,10,5",
    "784,6.5,0.46,0.14,2.4,0.114,9,37,0.99732,3.66,0.65,9.8,5",
    "785,7.1,0.59,0.01,2.5,0.077,20,85,0.99746,3.55,0.59,9.8,5",
    "786,9.9,0.35,0.41,2.3,0.083,11,61,0.9982,3.21,0.5,9.5,5",
    "787,9.9,0.35,0.41,2.3,0.083,11,61,0.9982,3.21,0.5,9.5,5",
    "788,10,0.56,0.24,2.2,0.079,19,58,0.9991,3.18,0.56,10.1,6",
    "789,10,0.56,0.24,2.2,0.079,19,58,0.9991,3.18,0.56,10.1,6",
    "790,8.6,0.63,0.17,2.9,0.099,21,119,0.998,3.09,0.52,9.3,5",
    "791,7.4,0.37,0.43,2.6,0.082,18,82,0.99708,3.33,0.68,9.7,6",
    "792,8.8,0.64,0.17,2.9,0.084,25,130,0.99818,3.23,0.54,9.6,5",
    "793,7.1,0.61,0.02,2.5,0.081,17,87,0.99745,3.48,0.6,9.7,6",
    "794,7.7,0.6,0,2.6,0.055,7,13,0.99639,3.38,0.56,10.8,5",
    "795,10.1,0.27,0.54,2.3,0.065,7,26,0.99531,3.17,0.53,12.5,6",
    "796,10.8,0.89,0.3,2.6,0.132,7,60,0.99786,2.99,1.18,10.2,5",
    "797,8.7,0.46,0.31,2.5,0.126,24,64,0.99746,3.1,0.74,9.6,5",
    "798,9.3,0.37,0.44,1.6,0.038,21,42,0.99526,3.24,0.81,10.8,7",
    "799,9.4,0.5,0.34,3.6,0.082,5,14,0.9987,3.29,0.52,10.7,6",
    "800,9.4,0.5,0.34,3.6,0.082,5,14,0.9987,3.29,0.52,10.7,6",
    "801,7.2,0.61,0.08,4,0.082,26,108,0.99641,3.25,0.51,9.4,5",
    "802,8.6,0.55,0.09,3.3,0.068,8,17,0.99735,3.23,0.44,10,5",
    "803,5.1,0.585,0,1.7,0.044,14,86,0.99264,3.56,0.94,12.9,7",
    "804,7.7,0.56,0.08,2.5,0.114,14,46,0.9971,3.24,0.66,9.6,6",
    "805,8.4,0.52,0.22,2.7,0.084,4,18,0.99682,3.26,0.57,9.9,6",
    "806,8.2,0.28,0.4,2.4,0.052,4,10,0.99356,3.33,0.7,12.8,7",
    "807,8.4,0.25,0.39,2,0.041,4,10,0.99386,3.27,0.71,12.5,7",
    "808,8.2,0.28,0.4,2.4,0.052,4,10,0.99356,3.33,0.7,12.8,7",
    "809,7.4,0.53,0.12,1.9,0.165,4,12,0.99702,3.26,0.86,9.2,5",
    "810,7.6,0.48,0.31,2.8,0.07,4,15,0.99693,3.22,0.55,10.3,6",
    "811,7.3,0.49,0.1,2.6,0.068,4,14,0.99562,3.3,0.47,10.5,5",
    "812,12.9,0.5,0.55,2.8,0.072,7,24,1.00012,3.09,0.68,10.9,6",
    "813,10.8,0.45,0.33,2.5,0.099,20,38,0.99818,3.24,0.71,10.8,5",
    "814,6.9,0.39,0.24,2.1,0.102,4,7,0.99462,3.44,0.58,11.4,4",
    "815,12.6,0.41,0.54,2.8,0.103,19,41,0.99939,3.21,0.76,11.3,6",
    "816,10.8,0.45,0.33,2.5,0.099,20,38,0.99818,3.24,0.71,10.8,5",
    "817,9.8,0.51,0.19,3.2,0.081,8,30,0.9984,3.23,0.58,10.5,6",
    "818,10.8,0.29,0.42,1.6,0.084,19,27,0.99545,3.28,0.73,11.9,6",
    "819,7.1,0.715,0,2.35,0.071,21,47,0.99632,3.29,0.45,9.4,5",
    "820,9.1,0.66,0.15,3.2,0.097,9,59,0.99976,3.28,0.54,9.6,5",
    "821,7,0.685,0,1.9,0.099,9,22,0.99606,3.34,0.6,9.7,5",
    "822,4.9,0.42,0,2.1,0.048,16,42,0.99154,3.71,0.74,14,7",
    "823,6.7,0.54,0.13,2,0.076,15,36,0.9973,3.61,0.64,9.8,5",
    "824,6.7,0.54,0.13,2,0.076,15,36,0.9973,3.61,0.64,9.8,5",
    "825,7.1,0.48,0.28,2.8,0.068,6,16,0.99682,3.24,0.53,10.3,5",
    "826,7.1,0.46,0.14,2.8,0.076,15,37,0.99624,3.36,0.49,10.7,5",
    "827,7.5,0.27,0.34,2.3,0.05,4,8,0.9951,3.4,0.64,11,7",
    "828,7.1,0.46,0.14,2.8,0.076,15,37,0.99624,3.36,0.49,10.7,5",
    "829,7.8,0.57,0.09,2.3,0.065,34,45,0.99417,3.46,0.74,12.7,8",
    "830,5.9,0.61,0.08,2.1,0.071,16,24,0.99376,3.56,0.77,11.1,6",
    "831,7.5,0.685,0.07,2.5,0.058,5,9,0.99632,3.38,0.55,10.9,4",
    "832,5.9,0.61,0.08,2.1,0.071,16,24,0.99376,3.56,0.77,11.1,6",
    "833,10.4,0.44,0.42,1.5,0.145,34,48,0.99832,3.38,0.86,9.9,3",
    "834,11.6,0.47,0.44,1.6,0.147,36,51,0.99836,3.38,0.86,9.9,4",
    "835,8.8,0.685,0.26,1.6,0.088,16,23,0.99694,3.32,0.47,9.4,5",
    "836,7.6,0.665,0.1,1.5,0.066,27,55,0.99655,3.39,0.51,9.3,5",
    "837,6.7,0.28,0.28,2.4,0.012,36,100,0.99064,3.26,0.39,11.7,7",
    "838,6.7,0.28,0.28,2.4,0.012,36,100,0.99064,3.26,0.39,11.7,7",
    "839,10.1,0.31,0.35,1.6,0.075,9,28,0.99672,3.24,0.83,11.2,7",
    "840,6,0.5,0.04,2.2,0.092,13,26,0.99647,3.46,0.47,10,5",
    "841,11.1,0.42,0.47,2.65,0.085,9,34,0.99736,3.24,0.77,12.1,7",
    "842,6.6,0.66,0,3,0.115,21,31,0.99629,3.45,0.63,10.3,5",
    "843,10.6,0.5,0.45,2.6,0.119,34,68,0.99708,3.23,0.72,10.9,6",
    "844,7.1,0.685,0.35,2,0.088,9,92,0.9963,3.28,0.62,9.4,5",
    "845,9.9,0.25,0.46,1.7,0.062,26,42,0.9959,3.18,0.83,10.6,6",
    "846,6.4,0.64,0.21,1.8,0.081,14,31,0.99689,3.59,0.66,9.8,5",
    "847,6.4,0.64,0.21,1.8,0.081,14,31,0.99689,3.59,0.66,9.8,5",
    "848,7.4,0.68,0.16,1.8,0.078,12,39,0.9977,3.5,0.7,9.9,6",
    "849,6.4,0.64,0.21,1.8,0.081,14,31,0.99689,3.59,0.66,9.8,5",
    "850,6.4,0.63,0.21,1.6,0.08,12,32,0.99689,3.58,0.66,9.8,5",
    "851,9.3,0.43,0.44,1.9,0.085,9,22,0.99708,3.28,0.55,9.5,5",
    "852,9.3,0.43,0.44,1.9,0.085,9,22,0.99708,3.28,0.55,9.5,5",
    "853,8,0.42,0.32,2.5,0.08,26,122,0.99801,3.22,1.07,9.7,5",
    "854,9.3,0.36,0.39,1.5,0.08,41,55,0.99652,3.47,0.73,10.9,6",
    "855,9.3,0.36,0.39,1.5,0.08,41,55,0.99652,3.47,0.73,10.9,6",
    "856,7.6,0.735,0.02,2.5,0.071,10,14,0.99538,3.51,0.71,11.7,7",
    "857,9.3,0.36,0.39,1.5,0.08,41,55,0.99652,3.47,0.73,10.9,6",
    "858,8.2,0.26,0.34,2.5,0.073,16,47,0.99594,3.4,0.78,11.3,7",
    "859,11.7,0.28,0.47,1.7,0.054,17,32,0.99686,3.15,0.67,10.6,7",
    "860,6.8,0.56,0.22,1.8,0.074,15,24,0.99438,3.4,0.82,11.2,6",
    "861,7.2,0.62,0.06,2.7,0.077,15,85,0.99746,3.51,0.54,9.5,5",
    "862,5.8,1.01,0.66,2,0.039,15,88,0.99357,3.66,0.6,11.5,6",
    "863,7.5,0.42,0.32,2.7,0.067,7,25,0.99628,3.24,0.44,10.4,5",
    "864,7.2,0.62,0.06,2.5,0.078,17,84,0.99746,3.51,0.53,9.7,5",
    "865,7.2,0.62,0.06,2.7,0.077,15,85,0.99746,3.51,0.54,9.5,5",
    "866,7.2,0.635,0.07,2.6,0.077,16,86,0.99748,3.51,0.54,9.7,5",
    "867,6.8,0.49,0.22,2.3,0.071,13,24,0.99438,3.41,0.83,11.3,6",
    "868,6.9,0.51,0.23,2,0.072,13,22,0.99438,3.4,0.84,11.2,6",
    "869,6.8,0.56,0.22,1.8,0.074,15,24,0.99438,3.4,0.82,11.2,6",
    "870,7.6,0.63,0.03,2,0.08,27,43,0.99578,3.44,0.64,10.9,6",
    "871,7.7,0.715,0.01,2.1,0.064,31,43,0.99371,3.41,0.57,11.8,6",
    "872,6.9,0.56,0.03,1.5,0.086,36,46,0.99522,3.53,0.57,10.6,5",
    "873,7.3,0.35,0.24,2,0.067,28,48,0.99576,3.43,0.54,10,4",
    "874,9.1,0.21,0.37,1.6,0.067,6,10,0.99552,3.23,0.58,11.1,7",
    "875,10.4,0.38,0.46,2.1,0.104,6,10,0.99664,3.12,0.65,11.8,7",
    "876,8.8,0.31,0.4,2.8,0.109,7,16,0.99614,3.31,0.79,11.8,7",
    "877,7.1,0.47,0,2.2,0.067,7,14,0.99517,3.4,0.58,10.9,4",
    "878,7.7,0.715,0.01,2.1,0.064,31,43,0.99371,3.41,0.57,11.8,6",
    "879,8.8,0.61,0.19,4,0.094,30,69,0.99787,3.22,0.5,10,6",
    "880,7.2,0.6,0.04,2.5,0.076,18,88,0.99745,3.53,0.55,9.5,5",
    "881,9.2,0.56,0.18,1.6,0.078,10,21,0.99576,3.15,0.49,9.9,5",
    "882,7.6,0.715,0,2.1,0.068,30,35,0.99533,3.48,0.65,11.4,6",
    "883,8.4,0.31,0.29,3.1,0.194,14,26,0.99536,3.22,0.78,12,6",
    "884,7.2,0.6,0.04,2.5,0.076,18,88,0.99745,3.53,0.55,9.5,5",
    "885,8.8,0.61,0.19,4,0.094,30,69,0.99787,3.22,0.5,10,6",
    "886,8.9,0.75,0.14,2.5,0.086,9,30,0.99824,3.34,0.64,10.5,5",
    "887,9,0.8,0.12,2.4,0.083,8,28,0.99836,3.33,0.65,10.4,6",
    "888,10.7,0.52,0.38,2.6,0.066,29,56,0.99577,3.15,0.79,12.1,7",
    "889,6.8,0.57,0,2.5,0.072,32,64,0.99491,3.43,0.56,11.2,6",
    "890,10.7,0.9,0.34,6.6,0.112,23,99,1.00289,3.22,0.68,9.3,5",
    "891,7.2,0.34,0.24,2,0.071,30,52,0.99576,3.44,0.58,10.1,5",
    "892,7.2,0.66,0.03,2.3,0.078,16,86,0.99743,3.53,0.57,9.7,5",
    "893,10.1,0.45,0.23,1.9,0.082,10,18,0.99774,3.22,0.65,9.3,6",
    "894,7.2,0.66,0.03,2.3,0.078,16,86,0.99743,3.53,0.57,9.7,5",
    "895,7.2,0.63,0.03,2.2,0.08,17,88,0.99745,3.53,0.58,9.8,6",
    "896,7.1,0.59,0.01,2.3,0.08,27,43,0.9955,3.42,0.58,10.7,6",
    "897,8.3,0.31,0.39,2.4,0.078,17,43,0.99444,3.31,0.77,12.5,7",
    "898,7.1,0.59,0.01,2.3,0.08,27,43,0.9955,3.42,0.58,10.7,6",
    "899,8.3,0.31,0.39,2.4,0.078,17,43,0.99444,3.31,0.77,12.5,7",
    "900,8.3,1.02,0.02,3.4,0.084,6,11,0.99892,3.48,0.49,11,3",
    "901,8.9,0.31,0.36,2.6,0.056,10,39,0.99562,3.4,0.69,11.8,5",
    "902,7.4,0.635,0.1,2.4,0.08,16,33,0.99736,3.58,0.69,10.8,7",
    "903,7.4,0.635,0.1,2.4,0.08,16,33,0.99736,3.58,0.69,10.8,7",
    "904,6.8,0.59,0.06,6,0.06,11,18,0.9962,3.41,0.59,10.8,7",
    "905,6.8,0.59,0.06,6,0.06,11,18,0.9962,3.41,0.59,10.8,7",
    "906,9.2,0.58,0.2,3,0.081,15,115,0.998,3.23,0.59,9.5,5",
    "907,7.2,0.54,0.27,2.6,0.084,12,78,0.9964,3.39,0.71,11,5",
    "908,6.1,0.56,0,2.2,0.079,6,9,0.9948,3.59,0.54,11.5,6",
    "909,7.4,0.52,0.13,2.4,0.078,34,61,0.99528,3.43,0.59,10.8,6",
    "910,7.3,0.305,0.39,1.2,0.059,7,11,0.99331,3.29,0.52,11.5,6",
    "911,9.3,0.38,0.48,3.8,0.132,3,11,0.99577,3.23,0.57,13.2,6",
    "912,9.1,0.28,0.46,9,0.114,3,9,0.99901,3.18,0.6,10.9,6",
    "913,10,0.46,0.44,2.9,0.065,4,8,0.99674,3.33,0.62,12.2,6",
    "914,9.4,0.395,0.46,4.6,0.094,3,10,0.99639,3.27,0.64,12.2,7",
    "915,7.3,0.305,0.39,1.2,0.059,7,11,0.99331,3.29,0.52,11.5,6",
    "916,8.6,0.315,0.4,2.2,0.079,3,6,0.99512,3.27,0.67,11.9,6",
    "917,5.3,0.715,0.19,1.5,0.161,7,62,0.99395,3.62,0.61,11,5",
    "918,6.8,0.41,0.31,8.8,0.084,26,45,0.99824,3.38,0.64,10.1,6",
    "919,8.4,0.36,0.32,2.2,0.081,32,79,0.9964,3.3,0.72,11,6",
    "920,8.4,0.62,0.12,1.8,0.072,38,46,0.99504,3.38,0.89,11.8,6",
    "921,9.6,0.41,0.37,2.3,0.091,10,23,0.99786,3.24,0.56,10.5,5",
    "922,8.4,0.36,0.32,2.2,0.081,32,79,0.9964,3.3,0.72,11,6",
    "923,8.4,0.62,0.12,1.8,0.072,38,46,0.99504,3.38,0.89,11.8,6",
    "924,6.8,0.41,0.31,8.8,0.084,26,45,0.99824,3.38,0.64,10.1,6",
    "925,8.6,0.47,0.27,2.3,0.055,14,28,0.99516,3.18,0.8,11.2,5",
    "926,8.6,0.22,0.36,1.9,0.064,53,77,0.99604,3.47,0.87,11,7",
    "927,9.4,0.24,0.33,2.3,0.061,52,73,0.99786,3.47,0.9,10.2,6",
    "928,8.4,0.67,0.19,2.2,0.093,11,75,0.99736,3.2,0.59,9.2,4",
    "929,8.6,0.47,0.27,2.3,0.055,14,28,0.99516,3.18,0.8,11.2,5",
    "930,8.7,0.33,0.38,3.3,0.063,10,19,0.99468,3.3,0.73,12,7",
    "931,6.6,0.61,0.01,1.9,0.08,8,25,0.99746,3.69,0.73,10.5,5",
    "932,7.4,0.61,0.01,2,0.074,13,38,0.99748,3.48,0.65,9.8,5",
    "933,7.6,0.4,0.29,1.9,0.078,29,66,0.9971,3.45,0.59,9.5,6",
    "934,7.4,0.61,0.01,2,0.074,13,38,0.99748,3.48,0.65,9.8,5",
    "935,6.6,0.61,0.01,1.9,0.08,8,25,0.99746,3.69,0.73,10.5,5",
    "936,8.8,0.3,0.38,2.3,0.06,19,72,0.99543,3.39,0.72,11.8,6",
    "937,8.8,0.3,0.38,2.3,0.06,19,72,0.99543,3.39,0.72,11.8,6",
    "938,12,0.63,0.5,1.4,0.071,6,26,0.99791,3.07,0.6,10.4,4",
    "939,7.2,0.38,0.38,2.8,0.068,23,42,0.99356,3.34,0.72,12.9,7",
    "940,6.2,0.46,0.17,1.6,0.073,7,11,0.99425,3.61,0.54,11.4,5",
    "941,9.6,0.33,0.52,2.2,0.074,13,25,0.99509,3.36,0.76,12.4,7",
    "942,9.9,0.27,0.49,5,0.082,9,17,0.99484,3.19,0.52,12.5,7",
    "943,10.1,0.43,0.4,2.6,0.092,13,52,0.99834,3.22,0.64,10,7",
    "944,9.8,0.5,0.34,2.3,0.094,10,45,0.99864,3.24,0.6,9.7,7",
    "945,8.3,0.3,0.49,3.8,0.09,11,24,0.99498,3.27,0.64,12.1,7",
    "946,10.2,0.44,0.42,2,0.071,7,20,0.99566,3.14,0.79,11.1,7",
    "947,10.2,0.44,0.58,4.1,0.092,11,24,0.99745,3.29,0.99,12,7",
    "948,8.3,0.28,0.48,2.1,0.093,6,12,0.99408,3.26,0.62,12.4,7",
    "949,8.9,0.12,0.45,1.8,0.075,10,21,0.99552,3.41,0.76,11.9,7",
    "950,8.9,0.12,0.45,1.8,0.075,10,21,0.99552,3.41,0.76,11.9,7",
    "951,8.9,0.12,0.45,1.8,0.075,10,21,0.99552,3.41,0.76,11.9,7",
    "952,8.3,0.28,0.48,2.1,0.093,6,12,0.99408,3.26,0.62,12.4,7",
    "953,8.2,0.31,0.4,2.2,0.058,6,10,0.99536,3.31,0.68,11.2,7",
    "954,10.2,0.34,0.48,2.1,0.052,5,9,0.99458,3.2,0.69,12.1,7",
    "955,7.6,0.43,0.4,2.7,0.082,6,11,0.99538,3.44,0.54,12.2,6",
    "956,8.5,0.21,0.52,1.9,0.09,9,23,0.99648,3.36,0.67,10.4,5",
    "957,9,0.36,0.52,2.1,0.111,5,10,0.99568,3.31,0.62,11.3,6",
    "958,9.5,0.37,0.52,2,0.088,12,51,0.99613,3.29,0.58,11.1,6",
    "959,6.4,0.57,0.12,2.3,0.12,25,36,0.99519,3.47,0.71,11.3,7",
    "960,8,0.59,0.05,2,0.089,12,32,0.99735,3.36,0.61,10,5",
    "961,8.5,0.47,0.27,1.9,0.058,18,38,0.99518,3.16,0.85,11.1,6",
    "962,7.1,0.56,0.14,1.6,0.078,7,18,0.99592,3.27,0.62,9.3,5",
    "963,6.6,0.57,0.02,2.1,0.115,6,16,0.99654,3.38,0.69,9.5,5",
    "964,8.8,0.27,0.39,2,0.1,20,27,0.99546,3.15,0.69,11.2,6",
    "965,8.5,0.47,0.27,1.9,0.058,18,38,0.99518,3.16,0.85,11.1,6",
    "966,8.3,0.34,0.4,2.4,0.065,24,48,0.99554,3.34,0.86,11,6",
    "967,9,0.38,0.41,2.4,0.103,6,10,0.99604,3.13,0.58,11.9,7",
    "968,8.5,0.66,0.2,2.1,0.097,23,113,0.99733,3.13,0.48,9.2,5",
    "969,9,0.4,0.43,2.4,0.068,29,46,0.9943,3.2,0.6,12.2,6",
    "970,6.7,0.56,0.09,2.9,0.079,7,22,0.99669,3.46,0.61,10.2,5",
    "971,10.4,0.26,0.48,1.9,0.066,6,10,0.99724,3.33,0.87,10.9,6",
    "972,10.4,0.26,0.48,1.9,0.066,6,10,0.99724,3.33,0.87,10.9,6",
    "973,10.1,0.38,0.5,2.4,0.104,6,13,0.99643,3.22,0.65,11.6,7",
    "974,8.5,0.34,0.44,1.7,0.079,6,12,0.99605,3.52,0.63,10.7,5",
    "975,8.8,0.33,0.41,5.9,0.073,7,13,0.99658,3.3,0.62,12.1,7",
    "976,7.2,0.41,0.3,2.1,0.083,35,72,0.997,3.44,0.52,9.4,5",
    "977,7.2,0.41,0.3,2.1,0.083,35,72,0.997,3.44,0.52,9.4,5",
    "978,8.4,0.59,0.29,2.6,0.109,31,119,0.99801,3.15,0.5,9.1,5",
    "979,7,0.4,0.32,3.6,0.061,9,29,0.99416,3.28,0.49,11.3,7",
    "980,12.2,0.45,0.49,1.4,0.075,3,6,0.9969,3.13,0.63,10.4,5",
    "981,9.1,0.5,0.3,1.9,0.065,8,17,0.99774,3.32,0.71,10.5,6",
    "982,9.5,0.86,0.26,1.9,0.079,13,28,0.99712,3.25,0.62,10,5",
    "983,7.3,0.52,0.32,2.1,0.07,51,70,0.99418,3.34,0.82,12.9,6",
    "984,9.1,0.5,0.3,1.9,0.065,8,17,0.99774,3.32,0.71,10.5,6",
    "985,12.2,0.45,0.49,1.4,0.075,3,6,0.9969,3.13,0.63,10.4,5",
    "986,7.4,0.58,0,2,0.064,7,11,0.99562,3.45,0.58,11.3,6",
    "987,9.8,0.34,0.39,1.4,0.066,3,7,0.9947,3.19,0.55,11.4,7",
    "988,7.1,0.36,0.3,1.6,0.08,35,70,0.99693,3.44,0.5,9.4,5",
    "989,7.7,0.39,0.12,1.7,0.097,19,27,0.99596,3.16,0.49,9.4,5",
    "990,9.7,0.295,0.4,1.5,0.073,14,21,0.99556,3.14,0.51,10.9,6",
    "991,7.7,0.39,0.12,1.7,0.097,19,27,0.99596,3.16,0.49,9.4,5",
    "992,7.1,0.34,0.28,2,0.082,31,68,0.99694,3.45,0.48,9.4,5",
    "993,6.5,0.4,0.1,2,0.076,30,47,0.99554,3.36,0.48,9.4,6",
    "994,7.1,0.34,0.28,2,0.082,31,68,0.99694,3.45,0.48,9.4,5",
    "995,10,0.35,0.45,2.5,0.092,20,88,0.99918,3.15,0.43,9.4,5",
    "996,7.7,0.6,0.06,2,0.079,19,41,0.99697,3.39,0.62,10.1,6",
    "997,5.6,0.66,0,2.2,0.087,3,11,0.99378,3.71,0.63,12.8,7",
    "998,5.6,0.66,0,2.2,0.087,3,11,0.99378,3.71,0.63,12.8,7",
    "999,8.9,0.84,0.34,1.4,0.05,4,10,0.99554,3.12,0.48,9.1,6",
    "1000,6.4,0.69,0,1.65,0.055,7,12,0.99162,3.47,0.53,12.9,6",
    "1001,7.5,0.43,0.3,2.2,0.062,6,12,0.99495,3.44,0.72,11.5,7",
    "1002,9.9,0.35,0.38,1.5,0.058,31,47,0.99676,3.26,0.82,10.6,7",
    "1003,9.1,0.29,0.33,2.05,0.063,13,27,0.99516,3.26,0.84,11.7,7",
    "1004,6.8,0.36,0.32,1.8,0.067,4,8,0.9928,3.36,0.55,12.8,7",
    "1005,8.2,0.43,0.29,1.6,0.081,27,45,0.99603,3.25,0.54,10.3,5",
    "1006,6.8,0.36,0.32,1.8,0.067,4,8,0.9928,3.36,0.55,12.8,7",
    "1007,9.1,0.29,0.33,2.05,0.063,13,27,0.99516,3.26,0.84,11.7,7",
    "1008,9.1,0.3,0.34,2,0.064,12,25,0.99516,3.26,0.84,11.7,7",
    "1009,8.9,0.35,0.4,3.6,0.11,12,24,0.99549,3.23,0.7,12,7",
    "1010,9.6,0.5,0.36,2.8,0.116,26,55,0.99722,3.18,0.68,10.9,5",
    "1011,8.9,0.28,0.45,1.7,0.067,7,12,0.99354,3.25,0.55,12.3,7",
    "1012,8.9,0.32,0.31,2,0.088,12,19,0.9957,3.17,0.55,10.4,6",
    "1013,7.7,1.005,0.15,2.1,0.102,11,32,0.99604,3.23,0.48,10,5",
    "1014,7.5,0.71,0,1.6,0.092,22,31,0.99635,3.38,0.58,10,6",
    "1015,8,0.58,0.16,2,0.12,3,7,0.99454,3.22,0.58,11.2,6",
    "1016,10.5,0.39,0.46,2.2,0.075,14,27,0.99598,3.06,0.84,11.4,6",
    "1017,8.9,0.38,0.4,2.2,0.068,12,28,0.99486,3.27,0.75,12.6,7",
    "1018,8,0.18,0.37,0.9,0.049,36,109,0.99007,2.89,0.44,12.7,6",
    "1019,8,0.18,0.37,0.9,0.049,36,109,0.99007,2.89,0.44,12.7,6",
    "1020,7,0.5,0.14,1.8,0.078,10,23,0.99636,3.53,0.61,10.4,5",
    "1021,11.3,0.36,0.66,2.4,0.123,3,8,0.99642,3.2,0.53,11.9,6",
    "1022,11.3,0.36,0.66,2.4,0.123,3,8,0.99642,3.2,0.53,11.9,6",
    "1023,7,0.51,0.09,2.1,0.062,4,9,0.99584,3.35,0.54,10.5,5",
    "1024,8.2,0.32,0.42,2.3,0.098,3,9,0.99506,3.27,0.55,12.3,6",
    "1025,7.7,0.58,0.01,1.8,0.088,12,18,0.99568,3.32,0.56,10.5,7",
    "1026,8.6,0.83,0,2.8,0.095,17,43,0.99822,3.33,0.6,10.4,6",
    "1027,7.9,0.31,0.32,1.9,0.066,14,36,0.99364,3.41,0.56,12.6,6",
    "1028,6.4,0.795,0,2.2,0.065,28,52,0.99378,3.49,0.52,11.6,5",
    "1029,7.2,0.34,0.21,2.5,0.075,41,68,0.99586,3.37,0.54,10.1,6",
    "1030,7.7,0.58,0.01,1.8,0.088,12,18,0.99568,3.32,0.56,10.5,7",
    "1031,7.1,0.59,0,2.1,0.091,9,14,0.99488,3.42,0.55,11.5,7",
    "1032,7.3,0.55,0.01,1.8,0.093,9,15,0.99514,3.35,0.58,11,7",
    "1033,8.1,0.82,0,4.1,0.095,5,14,0.99854,3.36,0.53,9.6,5",
    "1034,7.5,0.57,0.08,2.6,0.089,14,27,0.99592,3.3,0.59,10.4,6",
    "1035,8.9,0.745,0.18,2.5,0.077,15,48,0.99739,3.2,0.47,9.7,6",
    "1036,10.1,0.37,0.34,2.4,0.085,5,17,0.99683,3.17,0.65,10.6,7",
    "1037,7.6,0.31,0.34,2.5,0.082,26,35,0.99356,3.22,0.59,12.5,7",
    "1038,7.3,0.91,0.1,1.8,0.074,20,56,0.99672,3.35,0.56,9.2,5",
    "1039,8.7,0.41,0.41,6.2,0.078,25,42,0.9953,3.24,0.77,12.6,7",
    "1040,8.9,0.5,0.21,2.2,0.088,21,39,0.99692,3.33,0.83,11.1,6",
    "1041,7.4,0.965,0,2.2,0.088,16,32,0.99756,3.58,0.67,10.2,5",
    "1042,6.9,0.49,0.19,1.7,0.079,13,26,0.99547,3.38,0.64,9.8,6",
    "1043,8.9,0.5,0.21,2.2,0.088,21,39,0.99692,3.33,0.83,11.1,6",
    "1044,9.5,0.39,0.41,8.9,0.069,18,39,0.99859,3.29,0.81,10.9,7",
    "1045,6.4,0.39,0.33,3.3,0.046,12,53,0.99294,3.36,0.62,12.2,6",
    "1046,6.9,0.44,0,1.4,0.07,32,38,0.99438,3.32,0.58,11.4,6",
    "1047,7.6,0.78,0,1.7,0.076,33,45,0.99612,3.31,0.62,10.7,6",
    "1048,7.1,0.43,0.17,1.8,0.082,27,51,0.99634,3.49,0.64,10.4,5",
    "1049,9.3,0.49,0.36,1.7,0.081,3,14,0.99702,3.27,0.78,10.9,6",
    "1050,9.3,0.5,0.36,1.8,0.084,6,17,0.99704,3.27,0.77,10.8,6",
    "1051,7.1,0.43,0.17,1.8,0.082,27,51,0.99634,3.49,0.64,10.4,5",
    "1052,8.5,0.46,0.59,1.4,0.414,16,45,0.99702,3.03,1.34,9.2,5",
    "1053,5.6,0.605,0.05,2.4,0.073,19,25,0.99258,3.56,0.55,12.9,5",
    "1054,8.3,0.33,0.42,2.3,0.07,9,20,0.99426,3.38,0.77,12.7,7",
    "1055,8.2,0.64,0.27,2,0.095,5,77,0.99747,3.13,0.62,9.1,6",
    "1056,8.2,0.64,0.27,2,0.095,5,77,0.99747,3.13,0.62,9.1,6",
    "1057,8.9,0.48,0.53,4,0.101,3,10,0.99586,3.21,0.59,12.1,7",
    "1058,7.6,0.42,0.25,3.9,0.104,28,90,0.99784,3.15,0.57,9.1,5",
    "1059,9.9,0.53,0.57,2.4,0.093,30,52,0.9971,3.19,0.76,11.6,7",
    "1060,8.9,0.48,0.53,4,0.101,3,10,0.99586,3.21,0.59,12.1,7",
    "1061,11.6,0.23,0.57,1.8,0.074,3,8,0.9981,3.14,0.7,9.9,6",
    "1062,9.1,0.4,0.5,1.8,0.071,7,16,0.99462,3.21,0.69,12.5,8",
    "1063,8,0.38,0.44,1.9,0.098,6,15,0.9956,3.3,0.64,11.4,6",
    "1064,10.2,0.29,0.65,2.4,0.075,6,17,0.99565,3.22,0.63,11.8,6",
    "1065,8.2,0.74,0.09,2,0.067,5,10,0.99418,3.28,0.57,11.8,6",
    "1066,7.7,0.61,0.18,2.4,0.083,6,20,0.9963,3.29,0.6,10.2,6",
    "1067,6.6,0.52,0.08,2.4,0.07,13,26,0.99358,3.4,0.72,12.5,7",
    "1068,11.1,0.31,0.53,2.2,0.06,3,10,0.99572,3.02,0.83,10.9,7",
    "1069,11.1,0.31,0.53,2.2,0.06,3,10,0.99572,3.02,0.83,10.9,7",
    "1070,8,0.62,0.35,2.8,0.086,28,52,0.997,3.31,0.62,10.8,5",
    "1071,9.3,0.33,0.45,1.5,0.057,19,37,0.99498,3.18,0.89,11.1,7",
    "1072,7.5,0.77,0.2,8.1,0.098,30,92,0.99892,3.2,0.58,9.2,5",
    "1073,7.2,0.35,0.26,1.8,0.083,33,75,0.9968,3.4,0.58,9.5,6",
    "1074,8,0.62,0.33,2.7,0.088,16,37,0.9972,3.31,0.58,10.7,6",
    "1075,7.5,0.77,0.2,8.1,0.098,30,92,0.99892,3.2,0.58,9.2,5",
    "1076,9.1,0.25,0.34,2,0.071,45,67,0.99769,3.44,0.86,10.2,7",
    "1077,9.9,0.32,0.56,2,0.073,3,8,0.99534,3.15,0.73,11.4,6",
    "1078,8.6,0.37,0.65,6.4,0.08,3,8,0.99817,3.27,0.58,11,5",
    "1079,8.6,0.37,0.65,6.4,0.08,3,8,0.99817,3.27,0.58,11,5",
    "1080,7.9,0.3,0.68,8.3,0.05,37.5,278,0.99316,3.01,0.51,12.3,7",
    "1081,10.3,0.27,0.56,1.4,0.047,3,8,0.99471,3.16,0.51,11.8,6",
    "1082,7.9,0.3,0.68,8.3,0.05,37.5,289,0.99316,3.01,0.51,12.3,7",
    "1083,7.2,0.38,0.3,1.8,0.073,31,70,0.99685,3.42,0.59,9.5,6",
    "1084,8.7,0.42,0.45,2.4,0.072,32,59,0.99617,3.33,0.77,12,6",
    "1085,7.2,0.38,0.3,1.8,0.073,31,70,0.99685,3.42,0.59,9.5,6",
    "1086,6.8,0.48,0.08,1.8,0.074,40,64,0.99529,3.12,0.49,9.6,5",
    "1087,8.5,0.34,0.4,4.7,0.055,3,9,0.99738,3.38,0.66,11.6,7",
    "1088,7.9,0.19,0.42,1.6,0.057,18,30,0.994,3.29,0.69,11.2,6",
    "1089,11.6,0.41,0.54,1.5,0.095,22,41,0.99735,3.02,0.76,9.9,7",
    "1090,11.6,0.41,0.54,1.5,0.095,22,41,0.99735,3.02,0.76,9.9,7",
    "1091,10,0.26,0.54,1.9,0.083,42,74,0.99451,2.98,0.63,11.8,8",
    "1092,7.9,0.34,0.42,2,0.086,8,19,0.99546,3.35,0.6,11.4,6",
    "1093,7,0.54,0.09,2,0.081,10,16,0.99479,3.43,0.59,11.5,6",
    "1094,9.2,0.31,0.36,2.2,0.079,11,31,0.99615,3.33,0.86,12,7",
    "1095,6.6,0.725,0.09,5.5,0.117,9,17,0.99655,3.35,0.49,10.8,6",
    "1096,9.4,0.4,0.47,2.5,0.087,6,20,0.99772,3.15,0.5,10.5,5",
    "1097,6.6,0.725,0.09,5.5,0.117,9,17,0.99655,3.35,0.49,10.8,6",
    "1098,8.6,0.52,0.38,1.5,0.096,5,18,0.99666,3.2,0.52,9.4,5",
    "1099,8,0.31,0.45,2.1,0.216,5,16,0.99358,3.15,0.81,12.5,7",
    "1100,8.6,0.52,0.38,1.5,0.096,5,18,0.99666,3.2,0.52,9.4,5",
    "1101,8.4,0.34,0.42,2.1,0.072,23,36,0.99392,3.11,0.78,12.4,6",
    "1102,7.4,0.49,0.27,2.1,0.071,14,25,0.99388,3.35,0.63,12,6",
    "1103,6.1,0.48,0.09,1.7,0.078,18,30,0.99402,3.45,0.54,11.2,6",
    "1104,7.4,0.49,0.27,2.1,0.071,14,25,0.99388,3.35,0.63,12,6",
    "1105,8,0.48,0.34,2.2,0.073,16,25,0.9936,3.28,0.66,12.4,6",
    "1106,6.3,0.57,0.28,2.1,0.048,13,49,0.99374,3.41,0.6,12.8,5",
    "1107,8.2,0.23,0.42,1.9,0.069,9,17,0.99376,3.21,0.54,12.3,6",
    "1108,9.1,0.3,0.41,2,0.068,10,24,0.99523,3.27,0.85,11.7,7",
    "1109,8.1,0.78,0.1,3.3,0.09,4,13,0.99855,3.36,0.49,9.5,5",
    "1110,10.8,0.47,0.43,2.1,0.171,27,66,0.9982,3.17,0.76,10.8,6",
    "1111,8.3,0.53,0,1.4,0.07,6,14,0.99593,3.25,0.64,10,6",
    "1112,5.4,0.42,0.27,2,0.092,23,55,0.99471,3.78,0.64,12.3,7",
    "1113,7.9,0.33,0.41,1.5,0.056,6,35,0.99396,3.29,0.71,11,6",
    "1114,8.9,0.24,0.39,1.6,0.074,3,10,0.99698,3.12,0.59,9.5,6",
    "1115,5,0.4,0.5,4.3,0.046,29,80,0.9902,3.49,0.66,13.6,6",
    "1116,7,0.69,0.07,2.5,0.091,15,21,0.99572,3.38,0.6,11.3,6",
    "1117,7,0.69,0.07,2.5,0.091,15,21,0.99572,3.38,0.6,11.3,6",
    "1118,7,0.69,0.07,2.5,0.091,15,21,0.99572,3.38,0.6,11.3,6",
    "1119,7.1,0.39,0.12,2.1,0.065,14,24,0.99252,3.3,0.53,13.3,6",
    "1120,5.6,0.66,0,2.5,0.066,7,15,0.99256,3.52,0.58,12.9,5",
    "1121,7.9,0.54,0.34,2.5,0.076,8,17,0.99235,3.2,0.72,13.1,8",
    "1122,6.6,0.5,0,1.8,0.062,21,28,0.99352,3.44,0.55,12.3,6",
    "1123,6.3,0.47,0,1.4,0.055,27,33,0.9922,3.45,0.48,12.3,6",
    "1124,10.7,0.4,0.37,1.9,0.081,17,29,0.99674,3.12,0.65,11.2,6",
    "1125,6.5,0.58,0,2.2,0.096,3,13,0.99557,3.62,0.62,11.5,4",
    "1126,8.8,0.24,0.35,1.7,0.055,13,27,0.99394,3.14,0.59,11.3,7",
    "1127,5.8,0.29,0.26,1.7,0.063,3,11,0.9915,3.39,0.54,13.5,6",
    "1128,6.3,0.76,0,2.9,0.072,26,52,0.99379,3.51,0.6,11.5,6",
    "1129,10,0.43,0.33,2.7,0.095,28,89,0.9984,3.22,0.68,10,5",
    "1130,10.5,0.43,0.35,3.3,0.092,24,70,0.99798,3.21,0.69,10.5,6",
    "1131,9.1,0.6,0,1.9,0.058,5,10,0.9977,3.18,0.63,10.4,6",
    "1132,5.9,0.19,0.21,1.7,0.045,57,135,0.99341,3.32,0.44,9.5,5",
    "1133,7.4,0.36,0.34,1.8,0.075,18,38,0.9933,3.38,0.88,13.6,7",
    "1134,7.2,0.48,0.07,5.5,0.089,10,18,0.99684,3.37,0.68,11.2,7",
    "1135,8.5,0.28,0.35,1.7,0.061,6,15,0.99524,3.3,0.74,11.8,7",
    "1136,8,0.25,0.43,1.7,0.067,22,50,0.9946,3.38,0.6,11.9,6",
    "1137,10.4,0.52,0.45,2,0.08,6,13,0.99774,3.22,0.76,11.4,6",
    "1138,10.4,0.52,0.45,2,0.08,6,13,0.99774,3.22,0.76,11.4,6",
    "1139,7.5,0.41,0.15,3.7,0.104,29,94,0.99786,3.14,0.58,9.1,5",
    "1140,8.2,0.51,0.24,2,0.079,16,86,0.99764,3.34,0.64,9.5,6",
    "1141,7.3,0.4,0.3,1.7,0.08,33,79,0.9969,3.41,0.65,9.5,6",
    "1142,8.2,0.38,0.32,2.5,0.08,24,71,0.99624,3.27,0.85,11,6",
    "1143,6.9,0.45,0.11,2.4,0.043,6,12,0.99354,3.3,0.65,11.4,6",
    "1144,7,0.22,0.3,1.8,0.065,16,20,0.99672,3.61,0.82,10,6",
    "1145,7.3,0.32,0.23,2.3,0.066,35,70,0.99588,3.43,0.62,10.1,5",
    "1146,8.2,0.2,0.43,2.5,0.076,31,51,0.99672,3.53,0.81,10.4,6",
    "1147,7.8,0.5,0.12,1.8,0.178,6,21,0.996,3.28,0.87,9.8,6",
    "1148,10,0.41,0.45,6.2,0.071,6,14,0.99702,3.21,0.49,11.8,7",
    "1149,7.8,0.39,0.42,2,0.086,9,21,0.99526,3.39,0.66,11.6,6",
    "1150,10,0.35,0.47,2,0.061,6,11,0.99585,3.23,0.52,12,6",
    "1151,8.2,0.33,0.32,2.8,0.067,4,12,0.99473,3.3,0.76,12.8,7",
    "1152,6.1,0.58,0.23,2.5,0.044,16,70,0.99352,3.46,0.65,12.5,6",
    "1153,8.3,0.6,0.25,2.2,0.118,9,38,0.99616,3.15,0.53,9.8,5",
    "1154,9.6,0.42,0.35,2.1,0.083,17,38,0.99622,3.23,0.66,11.1,6",
    "1155,6.6,0.58,0,2.2,0.1,50,63,0.99544,3.59,0.68,11.4,6",
    "1156,8.3,0.6,0.25,2.2,0.118,9,38,0.99616,3.15,0.53,9.8,5",
    "1157,8.5,0.18,0.51,1.75,0.071,45,88,0.99524,3.33,0.76,11.8,7",
    "1158,5.1,0.51,0.18,2.1,0.042,16,101,0.9924,3.46,0.87,12.9,7",
    "1159,6.7,0.41,0.43,2.8,0.076,22,54,0.99572,3.42,1.16,10.6,6",
    "1160,10.2,0.41,0.43,2.2,0.11,11,37,0.99728,3.16,0.67,10.8,5",
    "1161,10.6,0.36,0.57,2.3,0.087,6,20,0.99676,3.14,0.72,11.1,7",
    "1162,8.8,0.45,0.43,1.4,0.076,12,21,0.99551,3.21,0.75,10.2,6",
    "1163,8.5,0.32,0.42,2.3,0.075,12,19,0.99434,3.14,0.71,11.8,7",
    "1164,9,0.785,0.24,1.7,0.078,10,21,0.99692,3.29,0.67,10,5",
    "1165,9,0.785,0.24,1.7,0.078,10,21,0.99692,3.29,0.67,10,5",
    "1166,8.5,0.44,0.5,1.9,0.369,15,38,0.99634,3.01,1.1,9.4,5",
    "1167,9.9,0.54,0.26,2,0.111,7,60,0.99709,2.94,0.98,10.2,5",
    "1168,8.2,0.33,0.39,2.5,0.074,29,48,0.99528,3.32,0.88,12.4,7",
    "1169,6.5,0.34,0.27,2.8,0.067,8,44,0.99384,3.21,0.56,12,6",
    "1170,7.6,0.5,0.29,2.3,0.086,5,14,0.99502,3.32,0.62,11.5,6",
    "1171,9.2,0.36,0.34,1.6,0.062,5,12,0.99667,3.2,0.67,10.5,6",
    "1172,7.1,0.59,0,2.2,0.078,26,44,0.99522,3.42,0.68,10.8,6",
    "1173,9.7,0.42,0.46,2.1,0.074,5,16,0.99649,3.27,0.74,12.3,6",
    "1174,7.6,0.36,0.31,1.7,0.079,26,65,0.99716,3.46,0.62,9.5,6",
    "1175,7.6,0.36,0.31,1.7,0.079,26,65,0.99716,3.46,0.62,9.5,6",
    "1176,6.5,0.61,0,2.2,0.095,48,59,0.99541,3.61,0.7,11.5,6",
    "1177,6.5,0.88,0.03,5.6,0.079,23,47,0.99572,3.58,0.5,11.2,4",
    "1178,7.1,0.66,0,2.4,0.052,6,11,0.99318,3.35,0.66,12.7,7",
    "1179,5.6,0.915,0,2.1,0.041,17,78,0.99346,3.68,0.73,11.4,5",
    "1180,8.2,0.35,0.33,2.4,0.076,11,47,0.99599,3.27,0.81,11,6",
    "1181,8.2,0.35,0.33,2.4,0.076,11,47,0.99599,3.27,0.81,11,6",
    "1182,9.8,0.39,0.43,1.65,0.068,5,11,0.99478,3.19,0.46,11.4,5",
    "1183,10.2,0.4,0.4,2.5,0.068,41,54,0.99754,3.38,0.86,10.5,6",
    "1184,6.8,0.66,0.07,1.6,0.07,16,61,0.99572,3.29,0.6,9.3,5",
    "1185,6.7,0.64,0.23,2.1,0.08,11,119,0.99538,3.36,0.7,10.9,5",
    "1186,7,0.43,0.3,2,0.085,6,39,0.99346,3.33,0.46,11.9,6",
    "1187,6.6,0.8,0.03,7.8,0.079,6,12,0.9963,3.52,0.5,12.2,5",
    "1188,7,0.43,0.3,2,0.085,6,39,0.99346,3.33,0.46,11.9,6",
    "1189,6.7,0.64,0.23,2.1,0.08,11,119,0.99538,3.36,0.7,10.9,5",
    "1190,8.8,0.955,0.05,1.8,0.075,5,19,0.99616,3.3,0.44,9.6,4",
    "1191,9.1,0.4,0.57,4.6,0.08,6,20,0.99652,3.28,0.57,12.5,6",
    "1192,6.5,0.885,0,2.3,0.166,6,12,0.99551,3.56,0.51,10.8,5",
    "1193,7.2,0.25,0.37,2.5,0.063,11,41,0.99439,3.52,0.8,12.4,7",
    "1194,6.4,0.885,0,2.3,0.166,6,12,0.99551,3.56,0.51,10.8,5",
    "1195,7,0.745,0.12,1.8,0.114,15,64,0.99588,3.22,0.59,9.5,6",
    "1196,6.2,0.43,0.22,1.8,0.078,21,56,0.99633,3.52,0.6,9.5,6",
    "1197,7.9,0.58,0.23,2.3,0.076,23,94,0.99686,3.21,0.58,9.5,6",
    "1198,7.7,0.57,0.21,1.5,0.069,4,9,0.99458,3.16,0.54,9.8,6",
    "1199,7.7,0.26,0.26,2,0.052,19,77,0.9951,3.15,0.79,10.9,6",
    "1200,7.9,0.58,0.23,2.3,0.076,23,94,0.99686,3.21,0.58,9.5,6",
    "1201,7.7,0.57,0.21,1.5,0.069,4,9,0.99458,3.16,0.54,9.8,6",
    "1202,7.9,0.34,0.36,1.9,0.065,5,10,0.99419,3.27,0.54,11.2,7",
    "1203,8.6,0.42,0.39,1.8,0.068,6,12,0.99516,3.35,0.69,11.7,8",
    "1204,9.9,0.74,0.19,5.8,0.111,33,76,0.99878,3.14,0.55,9.4,5",
    "1205,7.2,0.36,0.46,2.1,0.074,24,44,0.99534,3.4,0.85,11,7",
    "1206,7.2,0.36,0.46,2.1,0.074,24,44,0.99534,3.4,0.85,11,7",
    "1207,7.2,0.36,0.46,2.1,0.074,24,44,0.99534,3.4,0.85,11,7",
    "1208,9.9,0.72,0.55,1.7,0.136,24,52,0.99752,3.35,0.94,10,5",
    "1209,7.2,0.36,0.46,2.1,0.074,24,44,0.99534,3.4,0.85,11,7",
    "1210,6.2,0.39,0.43,2,0.071,14,24,0.99428,3.45,0.87,11.2,7",
    "1211,6.8,0.65,0.02,2.1,0.078,8,15,0.99498,3.35,0.62,10.4,6",
    "1212,6.6,0.44,0.15,2.1,0.076,22,53,0.9957,3.32,0.62,9.3,5",
    "1213,6.8,0.65,0.02,2.1,0.078,8,15,0.99498,3.35,0.62,10.4,6",
    "1214,9.6,0.38,0.42,1.9,0.071,5,13,0.99659,3.15,0.75,10.5,6",
    "1215,10.2,0.33,0.46,1.9,0.081,6,9,0.99628,3.1,0.48,10.4,6",
    "1216,8.8,0.27,0.46,2.1,0.095,20,29,0.99488,3.26,0.56,11.3,6",
    "1217,7.9,0.57,0.31,2,0.079,10,79,0.99677,3.29,0.69,9.5,6",
    "1218,8.2,0.34,0.37,1.9,0.057,43,74,0.99408,3.23,0.81,12,6",
    "1219,8.2,0.4,0.31,1.9,0.082,8,24,0.996,3.24,0.69,10.6,6",
    "1220,9,0.39,0.4,1.3,0.044,25,50,0.99478,3.2,0.83,10.9,6",
    "1221,10.9,0.32,0.52,1.8,0.132,17,44,0.99734,3.28,0.77,11.5,6",
    "1222,10.9,0.32,0.52,1.8,0.132,17,44,0.99734,3.28,0.77,11.5,6",
    "1223,8.1,0.53,0.22,2.2,0.078,33,89,0.99678,3.26,0.46,9.6,6",
    "1224,10.5,0.36,0.47,2.2,0.074,9,23,0.99638,3.23,0.76,12,6",
    "1225,12.6,0.39,0.49,2.5,0.08,8,20,0.9992,3.07,0.82,10.3,6",
    "1226,9.2,0.46,0.23,2.6,0.091,18,77,0.99922,3.15,0.51,9.4,5",
    "1227,7.5,0.58,0.03,4.1,0.08,27,46,0.99592,3.02,0.47,9.2,5",
    "1228,9,0.58,0.25,2,0.104,8,21,0.99769,3.27,0.72,9.6,5",
    "1229,5.1,0.42,0,1.8,0.044,18,88,0.99157,3.68,0.73,13.6,7",
    "1230,7.6,0.43,0.29,2.1,0.075,19,66,0.99718,3.4,0.64,9.5,5",
    "1231,7.7,0.18,0.34,2.7,0.066,15,58,0.9947,3.37,0.78,11.8,6",
    "1232,7.8,0.815,0.01,2.6,0.074,48,90,0.99621,3.38,0.62,10.8,5",
    "1233,7.6,0.43,0.29,2.1,0.075,19,66,0.99718,3.4,0.64,9.5,5",
    "1234,10.2,0.23,0.37,2.2,0.057,14,36,0.99614,3.23,0.49,9.3,4",
    "1235,7.1,0.75,0.01,2.2,0.059,11,18,0.99242,3.39,0.4,12.8,6",
    "1236,6,0.33,0.32,12.9,0.054,6,113,0.99572,3.3,0.56,11.5,4",
    "1237,7.8,0.55,0,1.7,0.07,7,17,0.99659,3.26,0.64,9.4,6",
    "1238,7.1,0.75,0.01,2.2,0.059,11,18,0.99242,3.39,0.4,12.8,6",
    "1239,8.1,0.73,0,2.5,0.081,12,24,0.99798,3.38,0.46,9.6,4",
    "1240,6.5,0.67,0,4.3,0.057,11,20,0.99488,3.45,0.56,11.8,4",
    "1241,7.5,0.61,0.2,1.7,0.076,36,60,0.99494,3.1,0.4,9.3,5",
    "1242,9.8,0.37,0.39,2.5,0.079,28,65,0.99729,3.16,0.59,9.8,5",
    "1243,9,0.4,0.41,2,0.058,15,40,0.99414,3.22,0.6,12.2,6",
    "1244,8.3,0.56,0.22,2.4,0.082,10,86,0.9983,3.37,0.62,9.5,5",
    "1245,5.9,0.29,0.25,13.4,0.067,72,160,0.99721,3.33,0.54,10.3,6",
    "1246,7.4,0.55,0.19,1.8,0.082,15,34,0.99655,3.49,0.68,10.5,5",
    "1247,7.4,0.74,0.07,1.7,0.086,15,48,0.99502,3.12,0.48,10,5",
    "1248,7.4,0.55,0.19,1.8,0.082,15,34,0.99655,3.49,0.68,10.5,5",
    "1249,6.9,0.41,0.33,2.2,0.081,22,36,0.9949,3.41,0.75,11.1,6",
    "1250,7.1,0.6,0.01,2.3,0.079,24,37,0.99514,3.4,0.61,10.9,6",
    "1251,7.1,0.6,0.01,2.3,0.079,24,37,0.99514,3.4,0.61,10.9,6",
    "1252,7.5,0.58,0.14,2.2,0.077,27,60,0.9963,3.28,0.59,9.8,5",
    "1253,7.1,0.72,0,1.8,0.123,6,14,0.99627,3.45,0.58,9.8,5",
    "1254,7.9,0.66,0,1.4,0.096,6,13,0.99569,3.43,0.58,9.5,5",
    "1255,7.8,0.7,0.06,1.9,0.079,20,35,0.99628,3.4,0.69,10.9,5",
    "1256,6.1,0.64,0.02,2.4,0.069,26,46,0.99358,3.47,0.45,11,5",
    "1257,7.5,0.59,0.22,1.8,0.082,43,60,0.99499,3.1,0.42,9.2,5",
    "1258,7,0.58,0.28,4.8,0.085,12,69,0.99633,3.32,0.7,11,6",
    "1259,6.8,0.64,0,2.7,0.123,15,33,0.99538,3.44,0.63,11.3,6",
    "1260,6.8,0.64,0,2.7,0.123,15,33,0.99538,3.44,0.63,11.3,6",
    "1261,8.6,0.635,0.68,1.8,0.403,19,56,0.99632,3.02,1.15,9.3,5",
    "1262,6.3,1.02,0,2,0.083,17,24,0.99437,3.59,0.55,11.2,4",
    "1263,9.8,0.45,0.38,2.5,0.081,34,66,0.99726,3.15,0.58,9.8,5",
    "1264,8.2,0.78,0,2.2,0.089,13,26,0.9978,3.37,0.46,9.6,4",
    "1265,8.5,0.37,0.32,1.8,0.066,26,51,0.99456,3.38,0.72,11.8,6",
    "1266,7.2,0.57,0.05,2.3,0.081,16,36,0.99564,3.38,0.6,10.3,6",
    "1267,7.2,0.57,0.05,2.3,0.081,16,36,0.99564,3.38,0.6,10.3,6",
    "1268,10.4,0.43,0.5,2.3,0.068,13,19,0.996,3.1,0.87,11.4,6",
    "1269,6.9,0.41,0.31,2,0.079,21,51,0.99668,3.47,0.55,9.5,6",
    "1270,5.5,0.49,0.03,1.8,0.044,28,87,0.9908,3.5,0.82,14,8",
    "1271,5,0.38,0.01,1.6,0.048,26,60,0.99084,3.7,0.75,14,6",
    "1272,7.3,0.44,0.2,1.6,0.049,24,64,0.9935,3.38,0.57,11.7,6",
    "1273,5.9,0.46,0,1.9,0.077,25,44,0.99385,3.5,0.53,11.2,5",
    "1274,7.5,0.58,0.2,2,0.073,34,44,0.99494,3.1,0.43,9.3,5",
    "1275,7.8,0.58,0.13,2.1,0.102,17,36,0.9944,3.24,0.53,11.2,6",
    "1276,8,0.715,0.22,2.3,0.075,13,81,0.99688,3.24,0.54,9.5,6",
    "1277,8.5,0.4,0.4,6.3,0.05,3,10,0.99566,3.28,0.56,12,4",
    "1278,7,0.69,0,1.9,0.114,3,10,0.99636,3.35,0.6,9.7,6",
    "1279,8,0.715,0.22,2.3,0.075,13,81,0.99688,3.24,0.54,9.5,6",
    "1280,9.8,0.3,0.39,1.7,0.062,3,9,0.9948,3.14,0.57,11.5,7",
    "1281,7.1,0.46,0.2,1.9,0.077,28,54,0.9956,3.37,0.64,10.4,6",
    "1282,7.1,0.46,0.2,1.9,0.077,28,54,0.9956,3.37,0.64,10.4,6",
    "1283,7.9,0.765,0,2,0.084,9,22,0.99619,3.33,0.68,10.9,6",
    "1284,8.7,0.63,0.28,2.7,0.096,17,69,0.99734,3.26,0.63,10.2,6",
    "1285,7,0.42,0.19,2.3,0.071,18,36,0.99476,3.39,0.56,10.9,5",
    "1286,11.3,0.37,0.5,1.8,0.09,20,47,0.99734,3.15,0.57,10.5,5",
    "1287,7.1,0.16,0.44,2.5,0.068,17,31,0.99328,3.35,0.54,12.4,6",
    "1288,8,0.6,0.08,2.6,0.056,3,7,0.99286,3.22,0.37,13,5",
    "1289,7,0.6,0.3,4.5,0.068,20,110,0.99914,3.3,1.17,10.2,5",
    "1290,7,0.6,0.3,4.5,0.068,20,110,0.99914,3.3,1.17,10.2,5",
    "1291,7.6,0.74,0,1.9,0.1,6,12,0.99521,3.36,0.59,11,5",
    "1292,8.2,0.635,0.1,2.1,0.073,25,60,0.99638,3.29,0.75,10.9,6",
    "1293,5.9,0.395,0.13,2.4,0.056,14,28,0.99362,3.62,0.67,12.4,6",
    "1294,7.5,0.755,0,1.9,0.084,6,12,0.99672,3.34,0.49,9.7,4",
    "1295,8.2,0.635,0.1,2.1,0.073,25,60,0.99638,3.29,0.75,10.9,6",
    "1296,6.6,0.63,0,4.3,0.093,51,77.5,0.99558,3.2,0.45,9.5,5",
    "1297,6.6,0.63,0,4.3,0.093,51,77.5,0.99558,3.2,0.45,9.5,5",
    "1298,7.2,0.53,0.14,2.1,0.064,15,29,0.99323,3.35,0.61,12.1,6",
    "1299,5.7,0.6,0,1.4,0.063,11,18,0.99191,3.45,0.56,12.2,6",
    "1300,7.6,1.58,0,2.1,0.137,5,9,0.99476,3.5,0.4,10.9,3",
    "1301,5.2,0.645,0,2.15,0.08,15,28,0.99444,3.78,0.61,12.5,6",
    "1302,6.7,0.86,0.07,2,0.1,20,57,0.99598,3.6,0.74,11.7,6",
    "1303,9.1,0.37,0.32,2.1,0.064,4,15,0.99576,3.3,0.8,11.2,6",
    "1304,8,0.28,0.44,1.8,0.081,28,68,0.99501,3.36,0.66,11.2,5",
    "1305,7.6,0.79,0.21,2.3,0.087,21,68,0.9955,3.12,0.44,9.2,5",
    "1306,7.5,0.61,0.26,1.9,0.073,24,88,0.99612,3.3,0.53,9.8,5",
    "1307,9.7,0.69,0.32,2.5,0.088,22,91,0.9979,3.29,0.62,10.1,5",
    "1308,6.8,0.68,0.09,3.9,0.068,15,29,0.99524,3.41,0.52,11.1,4",
    "1309,9.7,0.69,0.32,2.5,0.088,22,91,0.9979,3.29,0.62,10.1,5",
    "1310,7,0.62,0.1,1.4,0.071,27,63,0.996,3.28,0.61,9.2,5",
    "1311,7.5,0.61,0.26,1.9,0.073,24,88,0.99612,3.3,0.53,9.8,5",
    "1312,6.5,0.51,0.15,3,0.064,12,27,0.9929,3.33,0.59,12.8,6",
    "1313,8,1.18,0.21,1.9,0.083,14,41,0.99532,3.34,0.47,10.5,5",
    "1314,7,0.36,0.21,2.3,0.086,20,65,0.99558,3.4,0.54,10.1,6",
    "1315,7,0.36,0.21,2.4,0.086,24,69,0.99556,3.4,0.53,10.1,6",
    "1316,7.5,0.63,0.27,2,0.083,17,91,0.99616,3.26,0.58,9.8,6",
    "1317,5.4,0.74,0,1.2,0.041,16,46,0.99258,4.01,0.59,12.5,6",
    "1318,9.9,0.44,0.46,2.2,0.091,10,41,0.99638,3.18,0.69,11.9,6",
    "1319,7.5,0.63,0.27,2,0.083,17,91,0.99616,3.26,0.58,9.8,6",
    "1320,9.1,0.76,0.68,1.7,0.414,18,64,0.99652,2.9,1.33,9.1,6",
    "1321,9.7,0.66,0.34,2.6,0.094,12,88,0.99796,3.26,0.66,10.1,5",
    "1322,5,0.74,0,1.2,0.041,16,46,0.99258,4.01,0.59,12.5,6",
    "1323,9.1,0.34,0.42,1.8,0.058,9,18,0.99392,3.18,0.55,11.4,5",
    "1324,9.1,0.36,0.39,1.8,0.06,21,55,0.99495,3.18,0.82,11,7",
    "1325,6.7,0.46,0.24,1.7,0.077,18,34,0.9948,3.39,0.6,10.6,6",
    "1326,6.7,0.46,0.24,1.7,0.077,18,34,0.9948,3.39,0.6,10.6,6",
    "1327,6.7,0.46,0.24,1.7,0.077,18,34,0.9948,3.39,0.6,10.6,6",
    "1328,6.7,0.46,0.24,1.7,0.077,18,34,0.9948,3.39,0.6,10.6,6",
    "1329,6.5,0.52,0.11,1.8,0.073,13,38,0.9955,3.34,0.52,9.3,5",
    "1330,7.4,0.6,0.26,2.1,0.083,17,91,0.99616,3.29,0.56,9.8,6",
    "1331,7.4,0.6,0.26,2.1,0.083,17,91,0.99616,3.29,0.56,9.8,6",
    "1332,7.8,0.87,0.26,3.8,0.107,31,67,0.99668,3.26,0.46,9.2,5",
    "1333,8.4,0.39,0.1,1.7,0.075,6,25,0.99581,3.09,0.43,9.7,6",
    "1334,9.1,0.775,0.22,2.2,0.079,12,48,0.9976,3.18,0.51,9.6,5",
    "1335,7.2,0.835,0,2,0.166,4,11,0.99608,3.39,0.52,10,5",
    "1336,6.6,0.58,0.02,2.4,0.069,19,40,0.99387,3.38,0.66,12.6,6",
    "1337,6,0.5,0,1.4,0.057,15,26,0.99448,3.36,0.45,9.5,5",
    "1338,6,0.5,0,1.4,0.057,15,26,0.99448,3.36,0.45,9.5,5",
    "1339,6,0.5,0,1.4,0.057,15,26,0.99448,3.36,0.45,9.5,5",
    "1340,7.5,0.51,0.02,1.7,0.084,13,31,0.99538,3.36,0.54,10.5,6",
    "1341,7.5,0.51,0.02,1.7,0.084,13,31,0.99538,3.36,0.54,10.5,6",
    "1342,7.5,0.51,0.02,1.7,0.084,13,31,0.99538,3.36,0.54,10.5,6",
    "1343,7.6,0.54,0.02,1.7,0.085,17,31,0.99589,3.37,0.51,10.4,6",
    "1344,7.5,0.51,0.02,1.7,0.084,13,31,0.99538,3.36,0.54,10.5,6",
    "1345,11.5,0.42,0.48,2.6,0.077,8,20,0.99852,3.09,0.53,11,5",
    "1346,8.2,0.44,0.24,2.3,0.063,10,28,0.99613,3.25,0.53,10.2,6",
    "1347,6.1,0.59,0.01,2.1,0.056,5,13,0.99472,3.52,0.56,11.4,5",
    "1348,7.2,0.655,0.03,1.8,0.078,7,12,0.99587,3.34,0.39,9.5,5",
    "1349,7.2,0.655,0.03,1.8,0.078,7,12,0.99587,3.34,0.39,9.5,5",
    "1350,6.9,0.57,0,2.8,0.081,21,41,0.99518,3.41,0.52,10.8,5",
    "1351,9,0.6,0.29,2,0.069,32,73,0.99654,3.34,0.57,10,5",
    "1352,7.2,0.62,0.01,2.3,0.065,8,46,0.99332,3.32,0.51,11.8,6",
    "1353,7.6,0.645,0.03,1.9,0.086,14,57,0.9969,3.37,0.46,10.3,5",
    "1354,7.6,0.645,0.03,1.9,0.086,14,57,0.9969,3.37,0.46,10.3,5",
    "1355,7.2,0.58,0.03,2.3,0.077,7,28,0.99568,3.35,0.52,10,5",
    "1356,6.1,0.32,0.25,1.8,0.086,5,32,0.99464,3.36,0.44,10.1,5",
    "1357,6.1,0.34,0.25,1.8,0.084,4,28,0.99464,3.36,0.44,10.1,5",
    "1358,7.3,0.43,0.24,2.5,0.078,27,67,0.99648,3.6,0.59,11.1,6",
    "1359,7.4,0.64,0.17,5.4,0.168,52,98,0.99736,3.28,0.5,9.5,5",
    "1360,11.6,0.475,0.4,1.4,0.091,6,28,0.99704,3.07,0.65,10.03333333,6",
    "1361,9.2,0.54,0.31,2.3,0.112,11,38,0.99699,3.24,0.56,10.9,5",
    "1362,8.3,0.85,0.14,2.5,0.093,13,54,0.99724,3.36,0.54,10.1,5",
    "1363,11.6,0.475,0.4,1.4,0.091,6,28,0.99704,3.07,0.65,10.03333333,6",
    "1364,8,0.83,0.27,2,0.08,11,63,0.99652,3.29,0.48,9.8,4",
    "1365,7.2,0.605,0.02,1.9,0.096,10,31,0.995,3.46,0.53,11.8,6",
    "1366,7.8,0.5,0.09,2.2,0.115,10,42,0.9971,3.18,0.62,9.5,5",
    "1367,7.3,0.74,0.08,1.7,0.094,10,45,0.99576,3.24,0.5,9.8,5",
    "1368,6.9,0.54,0.3,2.2,0.088,9,105,0.99725,3.25,1.18,10.5,6",
    "1369,8,0.77,0.32,2.1,0.079,16,74,0.99656,3.27,0.5,9.8,6",
    "1370,6.6,0.61,0,1.6,0.069,4,8,0.99396,3.33,0.37,10.4,4",
    "1371,8.7,0.78,0.51,1.7,0.415,12,66,0.99623,3,1.17,9.2,5",
    "1372,7.5,0.58,0.56,3.1,0.153,5,14,0.99476,3.21,1.03,11.6,6",
    "1373,8.7,0.78,0.51,1.7,0.415,12,66,0.99623,3,1.17,9.2,5",
    "1374,7.7,0.75,0.27,3.8,0.11,34,89,0.99664,3.24,0.45,9.3,5",
    "1375,6.8,0.815,0,1.2,0.267,16,29,0.99471,3.32,0.51,9.8,3",
    "1376,7.2,0.56,0.26,2,0.083,13,100,0.99586,3.26,0.52,9.9,5",
    "1377,8.2,0.885,0.2,1.4,0.086,7,31,0.9946,3.11,0.46,10,5",
    "1378,5.2,0.49,0.26,2.3,0.09,23,74,0.9953,3.71,0.62,12.2,6",
    "1379,7.2,0.45,0.15,2,0.078,10,28,0.99609,3.29,0.51,9.9,6",
    "1380,7.5,0.57,0.02,2.6,0.077,11,35,0.99557,3.36,0.62,10.8,6",
    "1381,7.5,0.57,0.02,2.6,0.077,11,35,0.99557,3.36,0.62,10.8,6",
    "1382,6.8,0.83,0.09,1.8,0.074,4,25,0.99534,3.38,0.45,9.6,5",
    "1383,8,0.6,0.22,2.1,0.08,25,105,0.99613,3.3,0.49,9.9,5",
    "1384,8,0.6,0.22,2.1,0.08,25,105,0.99613,3.3,0.49,9.9,5",
    "1385,7.1,0.755,0.15,1.8,0.107,20,84,0.99593,3.19,0.5,9.5,5",
    "1386,8,0.81,0.25,3.4,0.076,34,85,0.99668,3.19,0.42,9.2,5",
    "1387,7.4,0.64,0.07,1.8,0.1,8,23,0.9961,3.3,0.58,9.6,5",
    "1388,7.4,0.64,0.07,1.8,0.1,8,23,0.9961,3.3,0.58,9.6,5",
    "1389,6.6,0.64,0.31,6.1,0.083,7,49,0.99718,3.35,0.68,10.3,5",
    "1390,6.7,0.48,0.02,2.2,0.08,36,111,0.99524,3.1,0.53,9.7,5",
    "1391,6,0.49,0,2.3,0.068,15,33,0.99292,3.58,0.59,12.5,6",
    "1392,8,0.64,0.22,2.4,0.094,5,33,0.99612,3.37,0.58,11,5",
    "1393,7.1,0.62,0.06,1.3,0.07,5,12,0.9942,3.17,0.48,9.8,5",
    "1394,8,0.52,0.25,2,0.078,19,59,0.99612,3.3,0.48,10.2,5",
    "1395,6.4,0.57,0.14,3.9,0.07,27,73,0.99669,3.32,0.48,9.2,5",
    "1396,8.6,0.685,0.1,1.6,0.092,3,12,0.99745,3.31,0.65,9.55,6",
    "1397,8.7,0.675,0.1,1.6,0.09,4,11,0.99745,3.31,0.65,9.55,5",
    "1398,7.3,0.59,0.26,2,0.08,17,104,0.99584,3.28,0.52,9.9,5",
    "1399,7,0.6,0.12,2.2,0.083,13,28,0.9966,3.52,0.62,10.2,7",
    "1400,7.2,0.67,0,2.2,0.068,10,24,0.9956,3.42,0.72,11.1,6",
    "1401,7.9,0.69,0.21,2.1,0.08,33,141,0.9962,3.25,0.51,9.9,5",
    "1402,7.9,0.69,0.21,2.1,0.08,33,141,0.9962,3.25,0.51,9.9,5",
    "1403,7.6,0.3,0.42,2,0.052,6,24,0.9963,3.44,0.82,11.9,6",
    "1404,7.2,0.33,0.33,1.7,0.061,3,13,0.996,3.23,1.1,10,8",
    "1405,8,0.5,0.39,2.6,0.082,12,46,0.9985,3.43,0.62,10.7,6",
    "1406,7.7,0.28,0.3,2,0.062,18,34,0.9952,3.28,0.9,11.3,7",
    "1407,8.2,0.24,0.34,5.1,0.062,8,22,0.9974,3.22,0.94,10.9,6",
    "1408,6,0.51,0,2.1,0.064,40,54,0.995,3.54,0.93,10.7,6",
    "1409,8.1,0.29,0.36,2.2,0.048,35,53,0.995,3.27,1.01,12.4,7",
    "1410,6,0.51,0,2.1,0.064,40,54,0.995,3.54,0.93,10.7,6",
    "1411,6.6,0.96,0,1.8,0.082,5,16,0.9936,3.5,0.44,11.9,6",
    "1412,6.4,0.47,0.4,2.4,0.071,8,19,0.9963,3.56,0.73,10.6,6",
    "1413,8.2,0.24,0.34,5.1,0.062,8,22,0.9974,3.22,0.94,10.9,6",
    "1414,9.9,0.57,0.25,2,0.104,12,89,0.9963,3.04,0.9,10.1,5",
    "1415,10,0.32,0.59,2.2,0.077,3,15,0.9994,3.2,0.78,9.6,5",
    "1416,6.2,0.58,0,1.6,0.065,8,18,0.9966,3.56,0.84,9.4,5",
    "1417,10,0.32,0.59,2.2,0.077,3,15,0.9994,3.2,0.78,9.6,5",
    "1418,7.3,0.34,0.33,2.5,0.064,21,37,0.9952,3.35,0.77,12.1,7",
    "1419,7.8,0.53,0.01,1.6,0.077,3,19,0.995,3.16,0.46,9.8,5",
    "1420,7.7,0.64,0.21,2.2,0.077,32,133,0.9956,3.27,0.45,9.9,5",
    "1421,7.8,0.53,0.01,1.6,0.077,3,19,0.995,3.16,0.46,9.8,5",
    "1422,7.5,0.4,0.18,1.6,0.079,24,58,0.9965,3.34,0.58,9.4,5",
    "1423,7,0.54,0,2.1,0.079,39,55,0.9956,3.39,0.84,11.4,6",
    "1424,6.4,0.53,0.09,3.9,0.123,14,31,0.9968,3.5,0.67,11,4",
    "1425,8.3,0.26,0.37,1.4,0.076,8,23,0.9974,3.26,0.7,9.6,6",
    "1426,8.3,0.26,0.37,1.4,0.076,8,23,0.9974,3.26,0.7,9.6,6",
    "1427,7.7,0.23,0.37,1.8,0.046,23,60,0.9971,3.41,0.71,12.1,6",
    "1428,7.6,0.41,0.33,2.5,0.078,6,23,0.9957,3.3,0.58,11.2,5",
    "1429,7.8,0.64,0,1.9,0.072,27,55,0.9962,3.31,0.63,11,5",
    "1430,7.9,0.18,0.4,2.2,0.049,38,67,0.996,3.33,0.93,11.3,5",
    "1431,7.4,0.41,0.24,1.8,0.066,18,47,0.9956,3.37,0.62,10.4,5",
    "1432,7.6,0.43,0.31,2.1,0.069,13,74,0.9958,3.26,0.54,9.9,6",
    "1433,5.9,0.44,0,1.6,0.042,3,11,0.9944,3.48,0.85,11.7,6",
    "1434,6.1,0.4,0.16,1.8,0.069,11,25,0.9955,3.42,0.74,10.1,7",
    "1435,10.2,0.54,0.37,15.4,0.214,55,95,1.00369,3.18,0.77,9,6",
    "1436,10.2,0.54,0.37,15.4,0.214,55,95,1.00369,3.18,0.77,9,6",
    "1437,10,0.38,0.38,1.6,0.169,27,90,0.99914,3.15,0.65,8.5,5",
    "1438,6.8,0.915,0.29,4.8,0.07,15,39,0.99577,3.53,0.54,11.1,5",
    "1439,7,0.59,0,1.7,0.052,3,8,0.996,3.41,0.47,10.3,5",
    "1440,7.3,0.67,0.02,2.2,0.072,31,92,0.99566,3.32,0.68,11.06666667,6",
    "1441,7.2,0.37,0.32,2,0.062,15,28,0.9947,3.23,0.73,11.3,7",
    "1442,7.4,0.785,0.19,5.2,0.094,19,98,0.99713,3.16,0.52,9.566666667,6",
    "1443,6.9,0.63,0.02,1.9,0.078,18,30,0.99712,3.4,0.75,9.8,5",
    "1444,6.9,0.58,0.2,1.75,0.058,8,22,0.99322,3.38,0.49,11.7,5",
    "1445,7.3,0.67,0.02,2.2,0.072,31,92,0.99566,3.32,0.68,11.1,6",
    "1446,7.4,0.785,0.19,5.2,0.094,19,98,0.99713,3.16,0.52,9.6,6",
    "1447,6.9,0.63,0.02,1.9,0.078,18,30,0.99712,3.4,0.75,9.8,5",
    "1448,6.8,0.67,0,1.9,0.08,22,39,0.99701,3.4,0.74,9.7,5",
    "1449,6.9,0.58,0.01,1.9,0.08,40,54,0.99683,3.4,0.73,9.7,5",
    "1450,7.2,0.38,0.31,2,0.056,15,29,0.99472,3.23,0.76,11.3,8",
    "1451,7.2,0.37,0.32,2,0.062,15,28,0.9947,3.23,0.73,11.3,7",
    "1452,7.8,0.32,0.44,2.7,0.104,8,17,0.99732,3.33,0.78,11,7",
    "1453,6.6,0.58,0.02,2,0.062,37,53,0.99374,3.35,0.76,11.6,7",
    "1454,7.6,0.49,0.33,1.9,0.074,27,85,0.99706,3.41,0.58,9,5",
    "1455,11.7,0.45,0.63,2.2,0.073,7,23,0.99974,3.21,0.69,10.9,6",
    "1456,6.5,0.9,0,1.6,0.052,9,17,0.99467,3.5,0.63,10.9,6",
    "1457,6,0.54,0.06,1.8,0.05,38,89,0.99236,3.3,0.5,10.55,6",
    "1458,7.6,0.49,0.33,1.9,0.074,27,85,0.99706,3.41,0.58,9,5",
    "1459,8.4,0.29,0.4,1.7,0.067,8,20,0.99603,3.39,0.6,10.5,5",
    "1460,7.9,0.2,0.35,1.7,0.054,7,15,0.99458,3.32,0.8,11.9,7",
    "1461,6.4,0.42,0.09,2.3,0.054,34,64,0.99724,3.41,0.68,10.4,6",
    "1462,6.2,0.785,0,2.1,0.06,6,13,0.99664,3.59,0.61,10,4",
    "1463,6.8,0.64,0.03,2.3,0.075,14,31,0.99545,3.36,0.58,10.4,6",
    "1464,6.9,0.63,0.01,2.4,0.076,14,39,0.99522,3.34,0.53,10.8,6",
    "1465,6.8,0.59,0.1,1.7,0.063,34,53,0.9958,3.41,0.67,9.7,5",
    "1466,6.8,0.59,0.1,1.7,0.063,34,53,0.9958,3.41,0.67,9.7,5",
    "1467,7.3,0.48,0.32,2.1,0.062,31,54,0.99728,3.3,0.65,10,7",
    "1468,6.7,1.04,0.08,2.3,0.067,19,32,0.99648,3.52,0.57,11,4",
    "1469,7.3,0.48,0.32,2.1,0.062,31,54,0.99728,3.3,0.65,10,7",
    "1470,7.3,0.98,0.05,2.1,0.061,20,49,0.99705,3.31,0.55,9.7,3",
    "1471,10,0.69,0.11,1.4,0.084,8,24,0.99578,2.88,0.47,9.7,5",
    "1472,6.7,0.7,0.08,3.75,0.067,8,16,0.99334,3.43,0.52,12.6,5",
    "1473,7.6,0.35,0.6,2.6,0.073,23,44,0.99656,3.38,0.79,11.1,6",
    "1474,6.1,0.6,0.08,1.8,0.071,14,45,0.99336,3.38,0.54,11,5",
    "1475,9.9,0.5,0.5,13.8,0.205,48,82,1.00242,3.16,0.75,8.8,5",
    "1476,5.3,0.47,0.11,2.2,0.048,16,89,0.99182,3.54,0.88,13.56666667,7",
    "1477,9.9,0.5,0.5,13.8,0.205,48,82,1.00242,3.16,0.75,8.8,5",
    "1478,5.3,0.47,0.11,2.2,0.048,16,89,0.99182,3.54,0.88,13.6,7",
    "1479,7.1,0.875,0.05,5.7,0.082,3,14,0.99808,3.4,0.52,10.2,3",
    "1480,8.2,0.28,0.6,3,0.104,10,22,0.99828,3.39,0.68,10.6,5",
    "1481,5.6,0.62,0.03,1.5,0.08,6,13,0.99498,3.66,0.62,10.1,4",
    "1482,8.2,0.28,0.6,3,0.104,10,22,0.99828,3.39,0.68,10.6,5",
    "1483,7.2,0.58,0.54,2.1,0.114,3,9,0.99719,3.33,0.57,10.3,4",
    "1484,8.1,0.33,0.44,1.5,0.042,6,12,0.99542,3.35,0.61,10.7,5",
    "1485,6.8,0.91,0.06,2,0.06,4,11,0.99592,3.53,0.64,10.9,4",
    "1486,7,0.655,0.16,2.1,0.074,8,25,0.99606,3.37,0.55,9.7,5",
    "1487,6.8,0.68,0.21,2.1,0.07,9,23,0.99546,3.38,0.6,10.3,5",
    "1488,6,0.64,0.05,1.9,0.066,9,17,0.99496,3.52,0.78,10.6,5",
    "1489,5.6,0.54,0.04,1.7,0.049,5,13,0.9942,3.72,0.58,11.4,5",
    "1490,6.2,0.57,0.1,2.1,0.048,4,11,0.99448,3.44,0.76,10.8,6",
    "1491,7.1,0.22,0.49,1.8,0.039,8,18,0.99344,3.39,0.56,12.4,6",
    "1492,5.6,0.54,0.04,1.7,0.049,5,13,0.9942,3.72,0.58,11.4,5",
    "1493,6.2,0.65,0.06,1.6,0.05,6,18,0.99348,3.57,0.54,11.95,5",
    "1494,7.7,0.54,0.26,1.9,0.089,23,147,0.99636,3.26,0.59,9.7,5",
    "1495,6.4,0.31,0.09,1.4,0.066,15,28,0.99459,3.42,0.7,10,7",
    "1496,7,0.43,0.02,1.9,0.08,15,28,0.99492,3.35,0.81,10.6,6",
    "1497,7.7,0.54,0.26,1.9,0.089,23,147,0.99636,3.26,0.59,9.7,5",
    "1498,6.9,0.74,0.03,2.3,0.054,7,16,0.99508,3.45,0.63,11.5,6",
    "1499,6.6,0.895,0.04,2.3,0.068,7,13,0.99582,3.53,0.58,10.8,6",
    "1500,6.9,0.74,0.03,2.3,0.054,7,16,0.99508,3.45,0.63,11.5,6",
    "1501,7.5,0.725,0.04,1.5,0.076,8,15,0.99508,3.26,0.53,9.6,5",
    "1502,7.8,0.82,0.29,4.3,0.083,21,64,0.99642,3.16,0.53,9.4,5",
    "1503,7.3,0.585,0.18,2.4,0.078,15,60,0.99638,3.31,0.54,9.8,5",
    "1504,6.2,0.44,0.39,2.5,0.077,6,14,0.99555,3.51,0.69,11,6",
    "1505,7.5,0.38,0.57,2.3,0.106,5,12,0.99605,3.36,0.55,11.4,6",
    "1506,6.7,0.76,0.02,1.8,0.078,6,12,0.996,3.55,0.63,9.95,3",
    "1507,6.8,0.81,0.05,2,0.07,6,14,0.99562,3.51,0.66,10.8,6",
    "1508,7.5,0.38,0.57,2.3,0.106,5,12,0.99605,3.36,0.55,11.4,6",
    "1509,7.1,0.27,0.6,2.1,0.074,17,25,0.99814,3.38,0.72,10.6,6",
    "1510,7.9,0.18,0.4,1.8,0.062,7,20,0.9941,3.28,0.7,11.1,5",
    "1511,6.4,0.36,0.21,2.2,0.047,26,48,0.99661,3.47,0.77,9.7,6",
    "1512,7.1,0.69,0.04,2.1,0.068,19,27,0.99712,3.44,0.67,9.8,5",
    "1513,6.4,0.79,0.04,2.2,0.061,11,17,0.99588,3.53,0.65,10.4,6",
    "1514,6.4,0.56,0.15,1.8,0.078,17,65,0.99294,3.33,0.6,10.5,6",
    "1515,6.9,0.84,0.21,4.1,0.074,16,65,0.99842,3.53,0.72,9.233333333,6",
    "1516,6.9,0.84,0.21,4.1,0.074,16,65,0.99842,3.53,0.72,9.25,6",
    "1517,6.1,0.32,0.25,2.3,0.071,23,58,0.99633,3.42,0.97,10.6,5",
    "1518,6.5,0.53,0.06,2,0.063,29,44,0.99489,3.38,0.83,10.3,6",
    "1519,7.4,0.47,0.46,2.2,0.114,7,20,0.99647,3.32,0.63,10.5,5",
    "1520,6.6,0.7,0.08,2.6,0.106,14,27,0.99665,3.44,0.58,10.2,5",
    "1521,6.5,0.53,0.06,2,0.063,29,44,0.99489,3.38,0.83,10.3,6",
    "1522,6.9,0.48,0.2,1.9,0.082,9,23,0.99585,3.39,0.43,9.05,4",
    "1523,6.1,0.32,0.25,2.3,0.071,23,58,0.99633,3.42,0.97,10.6,5",
    "1524,6.8,0.48,0.25,2,0.076,29,61,0.9953,3.34,0.6,10.4,5",
    "1525,6,0.42,0.19,2,0.075,22,47,0.99522,3.39,0.78,10,6",
    "1526,6.7,0.48,0.08,2.1,0.064,18,34,0.99552,3.33,0.64,9.7,5",
    "1527,6.8,0.47,0.08,2.2,0.064,18,38,0.99553,3.3,0.65,9.6,6",
    "1528,7.1,0.53,0.07,1.7,0.071,15,24,0.9951,3.29,0.66,10.8,6",
    "1529,7.9,0.29,0.49,2.2,0.096,21,59,0.99714,3.31,0.67,10.1,6",
    "1530,7.1,0.69,0.08,2.1,0.063,42,52,0.99608,3.42,0.6,10.2,6",
    "1531,6.6,0.44,0.09,2.2,0.063,9,18,0.99444,3.42,0.69,11.3,6",
    "1532,6.1,0.705,0.1,2.8,0.081,13,28,0.99631,3.6,0.66,10.2,5",
    "1533,7.2,0.53,0.13,2,0.058,18,22,0.99573,3.21,0.68,9.9,6",
    "1534,8,0.39,0.3,1.9,0.074,32,84,0.99717,3.39,0.61,9,5",
    "1535,6.6,0.56,0.14,2.4,0.064,13,29,0.99397,3.42,0.62,11.7,7",
    "1536,7,0.55,0.13,2.2,0.075,15,35,0.9959,3.36,0.59,9.7,6",
    "1537,6.1,0.53,0.08,1.9,0.077,24,45,0.99528,3.6,0.68,10.3,6",
    "1538,5.4,0.58,0.08,1.9,0.059,20,31,0.99484,3.5,0.64,10.2,6",
    "1539,6.2,0.64,0.09,2.5,0.081,15,26,0.99538,3.57,0.63,12,5",
    "1540,7.2,0.39,0.32,1.8,0.065,34,60,0.99714,3.46,0.78,9.9,5",
    "1541,6.2,0.52,0.08,4.4,0.071,11,32,0.99646,3.56,0.63,11.6,6",
    "1542,7.4,0.25,0.29,2.2,0.054,19,49,0.99666,3.4,0.76,10.9,7",
    "1543,6.7,0.855,0.02,1.9,0.064,29,38,0.99472,3.3,0.56,10.75,6",
    "1544,11.1,0.44,0.42,2.2,0.064,14,19,0.99758,3.25,0.57,10.4,6",
    "1545,8.4,0.37,0.43,2.3,0.063,12,19,0.9955,3.17,0.81,11.2,7",
    "1546,6.5,0.63,0.33,1.8,0.059,16,28,0.99531,3.36,0.64,10.1,6",
    "1547,7,0.57,0.02,2,0.072,17,26,0.99575,3.36,0.61,10.2,5",
    "1548,6.3,0.6,0.1,1.6,0.048,12,26,0.99306,3.55,0.51,12.1,5",
    "1549,11.2,0.4,0.5,2,0.099,19,50,0.99783,3.1,0.58,10.4,5",
    "1550,7.4,0.36,0.3,1.8,0.074,17,24,0.99419,3.24,0.7,11.4,8",
    "1551,7.1,0.68,0,2.3,0.087,17,26,0.99783,3.45,0.53,9.5,5",
    "1552,7.1,0.67,0,2.3,0.083,18,27,0.99768,3.44,0.54,9.4,5",
    "1553,6.3,0.68,0.01,3.7,0.103,32,54,0.99586,3.51,0.66,11.3,6",
    "1554,7.3,0.735,0,2.2,0.08,18,28,0.99765,3.41,0.6,9.4,5",
    "1555,6.6,0.855,0.02,2.4,0.062,15,23,0.99627,3.54,0.6,11,6",
    "1556,7,0.56,0.17,1.7,0.065,15,24,0.99514,3.44,0.68,10.55,7",
    "1557,6.6,0.88,0.04,2.2,0.066,12,20,0.99636,3.53,0.56,9.9,5",
    "1558,6.6,0.855,0.02,2.4,0.062,15,23,0.99627,3.54,0.6,11,6",
    "1559,6.9,0.63,0.33,6.7,0.235,66,115,0.99787,3.22,0.56,9.5,5",
    "1560,7.8,0.6,0.26,2,0.08,31,131,0.99622,3.21,0.52,9.9,5",
    "1561,7.8,0.6,0.26,2,0.08,31,131,0.99622,3.21,0.52,9.9,5",
    "1562,7.8,0.6,0.26,2,0.08,31,131,0.99622,3.21,0.52,9.9,5",
    "1563,7.2,0.695,0.13,2,0.076,12,20,0.99546,3.29,0.54,10.1,5",
    "1564,7.2,0.695,0.13,2,0.076,12,20,0.99546,3.29,0.54,10.1,5",
    "1565,7.2,0.695,0.13,2,0.076,12,20,0.99546,3.29,0.54,10.1,5",
    "1566,6.7,0.67,0.02,1.9,0.061,26,42,0.99489,3.39,0.82,10.9,6",
    "1567,6.7,0.16,0.64,2.1,0.059,24,52,0.99494,3.34,0.71,11.2,6",
    "1568,7.2,0.695,0.13,2,0.076,12,20,0.99546,3.29,0.54,10.1,5",
    "1569,7,0.56,0.13,1.6,0.077,25,42,0.99629,3.34,0.59,9.2,5",
    "1570,6.2,0.51,0.14,1.9,0.056,15,34,0.99396,3.48,0.57,11.5,6",
    "1571,6.4,0.36,0.53,2.2,0.23,19,35,0.9934,3.37,0.93,12.4,6",
    "1572,6.4,0.38,0.14,2.2,0.038,15,25,0.99514,3.44,0.65,11.1,6",
    "1573,7.3,0.69,0.32,2.2,0.069,35,104,0.99632,3.33,0.51,9.5,5",
    "1574,6,0.58,0.2,2.4,0.075,15,50,0.99467,3.58,0.67,12.5,6",
    "1575,5.6,0.31,0.78,13.9,0.074,23,92,0.99677,3.39,0.48,10.5,6",
    "1576,7.5,0.52,0.4,2.2,0.06,12,20,0.99474,3.26,0.64,11.8,6",
    "1577,8,0.3,0.63,1.6,0.081,16,29,0.99588,3.3,0.78,10.8,6",
    "1578,6.2,0.7,0.15,5.1,0.076,13,27,0.99622,3.54,0.6,11.9,6",
    "1579,6.8,0.67,0.15,1.8,0.118,13,20,0.9954,3.42,0.67,11.3,6",
    "1580,6.2,0.56,0.09,1.7,0.053,24,32,0.99402,3.54,0.6,11.3,5",
    "1581,7.4,0.35,0.33,2.4,0.068,9,26,0.9947,3.36,0.6,11.9,6",
    "1582,6.2,0.56,0.09,1.7,0.053,24,32,0.99402,3.54,0.6,11.3,5",
    "1583,6.1,0.715,0.1,2.6,0.053,13,27,0.99362,3.57,0.5,11.9,5",
    "1584,6.2,0.46,0.29,2.1,0.074,32,98,0.99578,3.33,0.62,9.8,5",
    "1585,6.7,0.32,0.44,2.4,0.061,24,34,0.99484,3.29,0.8,11.6,7",
    "1586,7.2,0.39,0.44,2.6,0.066,22,48,0.99494,3.3,0.84,11.5,6",
    "1587,7.5,0.31,0.41,2.4,0.065,34,60,0.99492,3.34,0.85,11.4,6",
    "1588,5.8,0.61,0.11,1.8,0.066,18,28,0.99483,3.55,0.66,10.9,6",
    "1589,7.2,0.66,0.33,2.5,0.068,34,102,0.99414,3.27,0.78,12.8,6",
    "1590,6.6,0.725,0.2,7.8,0.073,29,79,0.9977,3.29,0.54,9.2,5",
    "1591,6.3,0.55,0.15,1.8,0.077,26,35,0.99314,3.32,0.82,11.6,6",
    "1592,5.4,0.74,0.09,1.7,0.089,16,26,0.99402,3.67,0.56,11.6,6",
    "1593,6.3,0.51,0.13,2.3,0.076,29,40,0.99574,3.42,0.75,11,6",
    "1594,6.8,0.62,0.08,1.9,0.068,28,38,0.99651,3.42,0.82,9.5,6",
    "1595,6.2,0.6,0.08,2,0.09,32,44,0.9949,3.45,0.58,10.5,5",
    "1596,5.9,0.55,0.1,2.2,0.062,39,51,0.99512,3.52,0.76,11.2,6",
    "1597,6.3,0.51,0.13,2.3,0.076,29,40,0.99574,3.42,0.75,11,6",
    "1598,5.9,0.645,0.12,2,0.075,32,44,0.99547,3.57,0.71,10.2,5",
    "1599,6,0.31,0.47,3.6,0.067,18,42,0.99549,3.39,0.66,11,6",
)

Pima_indian_Diabetes_Dataset = (
"id_int,pregnant-times_int,2hr-glucose-concentration_int,Diastolic-blood-pressure_int,Triceps-skinfold-thickness_int,2-Hour-serum-insulin_int,Body-mass-index_float,Diabetes-pedigree-function_float,age_int,diagnosis-result_int",
"1,6,148,72,35,0,33.6,0.627,50,1",
"2,1,85,66,29,0,26.6,0.351,31,0",
"3,8,183,64,0,0,23.3,0.672,32,1",
"4,1,89,66,23,94,28.1,0.167,21,0",
"5,0,137,40,35,168,43.1,2.288,33,1",
"6,5,116,74,0,0,25.6,0.201,30,0",
"7,3,78,50,32,88,31.0,0.248,26,1",
"8,10,115,0,0,0,35.3,0.134,29,0",
"9,2,197,70,45,543,30.5,0.158,53,1",
"10,8,125,96,0,0,0.0,0.232,54,1",
"11,4,110,92,0,0,37.6,0.191,30,0",
"12,10,168,74,0,0,38.0,0.537,34,1",
"13,10,139,80,0,0,27.1,1.441,57,0",
"14,1,189,60,23,846,30.1,0.398,59,1",
"15,5,166,72,19,175,25.8,0.587,51,1",
"16,7,100,0,0,0,30.0,0.484,32,1",
"17,0,118,84,47,230,45.8,0.551,31,1",
"18,7,107,74,0,0,29.6,0.254,31,1",
"19,1,103,30,38,83,43.3,0.183,33,0",
"20,1,115,70,30,96,34.6,0.529,32,1",
"21,3,126,88,41,235,39.3,0.704,27,0",
"22,8,99,84,0,0,35.4,0.388,50,0",
"23,7,196,90,0,0,39.8,0.451,41,1",
"24,9,119,80,35,0,29.0,0.263,29,1",
"25,11,143,94,33,146,36.6,0.254,51,1",
"26,10,125,70,26,115,31.1,0.205,41,1",
"27,7,147,76,0,0,39.4,0.257,43,1",
"28,1,97,66,15,140,23.2,0.487,22,0",
"29,13,145,82,19,110,22.2,0.245,57,0",
"30,5,117,92,0,0,34.1,0.337,38,0",
"31,5,109,75,26,0,36.0,0.546,60,0",
"32,3,158,76,36,245,31.6,0.851,28,1",
"33,3,88,58,11,54,24.8,0.267,22,0",
"34,6,92,92,0,0,19.9,0.188,28,0",
"35,10,122,78,31,0,27.6,0.512,45,0",
"36,4,103,60,33,192,24.0,0.966,33,0",
"37,11,138,76,0,0,33.2,0.420,35,0",
"38,9,102,76,37,0,32.9,0.665,46,1",
"39,2,90,68,42,0,38.2,0.503,27,1",
"40,4,111,72,47,207,37.1,1.390,56,1",
"41,3,180,64,25,70,34.0,0.271,26,0",
"42,7,133,84,0,0,40.2,0.696,37,0",
"43,7,106,92,18,0,22.7,0.235,48,0",
"44,9,171,110,24,240,45.4,0.721,54,1",
"45,7,159,64,0,0,27.4,0.294,40,0",
"46,0,180,66,39,0,42.0,1.893,25,1",
"47,1,146,56,0,0,29.7,0.564,29,0",
"48,2,71,70,27,0,28.0,0.586,22,0",
"49,7,103,66,32,0,39.1,0.344,31,1",
"50,7,105,0,0,0,0.0,0.305,24,0",
"51,1,103,80,11,82,19.4,0.491,22,0",
"52,1,101,50,15,36,24.2,0.526,26,0",
"53,5,88,66,21,23,24.4,0.342,30,0",
"54,8,176,90,34,300,33.7,0.467,58,1",
"55,7,150,66,42,342,34.7,0.718,42,0",
"56,1,73,50,10,0,23.0,0.248,21,0",
"57,7,187,68,39,304,37.7,0.254,41,1",
"58,0,100,88,60,110,46.8,0.962,31,0",
"59,0,146,82,0,0,40.5,1.781,44,0",
"60,0,105,64,41,142,41.5,0.173,22,0",
"61,2,84,0,0,0,0.0,0.304,21,0",
"62,8,133,72,0,0,32.9,0.270,39,1",
"63,5,44,62,0,0,25.0,0.587,36,0",
"64,2,141,58,34,128,25.4,0.699,24,0",
"65,7,114,66,0,0,32.8,0.258,42,1",
"66,5,99,74,27,0,29.0,0.203,32,0",
"67,0,109,88,30,0,32.5,0.855,38,1",
"68,2,109,92,0,0,42.7,0.845,54,0",
"69,1,95,66,13,38,19.6,0.334,25,0",
"70,4,146,85,27,100,28.9,0.189,27,0",
"71,2,100,66,20,90,32.9,0.867,28,1",
"72,5,139,64,35,140,28.6,0.411,26,0",
"73,13,126,90,0,0,43.4,0.583,42,1",
"74,4,129,86,20,270,35.1,0.231,23,0",
"75,1,79,75,30,0,32.0,0.396,22,0",
"76,1,0,48,20,0,24.7,0.140,22,0",
"77,7,62,78,0,0,32.6,0.391,41,0",
"78,5,95,72,33,0,37.7,0.370,27,0",
"79,0,131,0,0,0,43.2,0.270,26,1",
"80,2,112,66,22,0,25.0,0.307,24,0",
"81,3,113,44,13,0,22.4,0.140,22,0",
"82,2,74,0,0,0,0.0,0.102,22,0",
"83,7,83,78,26,71,29.3,0.767,36,0",
"84,0,101,65,28,0,24.6,0.237,22,0",
"85,5,137,108,0,0,48.8,0.227,37,1",
"86,2,110,74,29,125,32.4,0.698,27,0",
"87,13,106,72,54,0,36.6,0.178,45,0",
"88,2,100,68,25,71,38.5,0.324,26,0",
"89,15,136,70,32,110,37.1,0.153,43,1",
"90,1,107,68,19,0,26.5,0.165,24,0",
"91,1,80,55,0,0,19.1,0.258,21,0",
"92,4,123,80,15,176,32.0,0.443,34,0",
"93,7,81,78,40,48,46.7,0.261,42,0",
"94,4,134,72,0,0,23.8,0.277,60,1",
"95,2,142,82,18,64,24.7,0.761,21,0",
"96,6,144,72,27,228,33.9,0.255,40,0",
"97,2,92,62,28,0,31.6,0.130,24,0",
"98,1,71,48,18,76,20.4,0.323,22,0",
"99,6,93,50,30,64,28.7,0.356,23,0",
"100,1,122,90,51,220,49.7,0.325,31,1",
"101,1,163,72,0,0,39.0,1.222,33,1",
"102,1,151,60,0,0,26.1,0.179,22,0",
"103,0,125,96,0,0,22.5,0.262,21,0",
"104,1,81,72,18,40,26.6,0.283,24,0",
"105,2,85,65,0,0,39.6,0.930,27,0",
"106,1,126,56,29,152,28.7,0.801,21,0",
"107,1,96,122,0,0,22.4,0.207,27,0",
"108,4,144,58,28,140,29.5,0.287,37,0",
"109,3,83,58,31,18,34.3,0.336,25,0",
"110,0,95,85,25,36,37.4,0.247,24,1",
"111,3,171,72,33,135,33.3,0.199,24,1",
"112,8,155,62,26,495,34.0,0.543,46,1",
"113,1,89,76,34,37,31.2,0.192,23,0",
"114,4,76,62,0,0,34.0,0.391,25,0",
"115,7,160,54,32,175,30.5,0.588,39,1",
"116,4,146,92,0,0,31.2,0.539,61,1",
"117,5,124,74,0,0,34.0,0.220,38,1",
"118,5,78,48,0,0,33.7,0.654,25,0",
"119,4,97,60,23,0,28.2,0.443,22,0",
"120,4,99,76,15,51,23.2,0.223,21,0",
"121,0,162,76,56,100,53.2,0.759,25,1",
"122,6,111,64,39,0,34.2,0.260,24,0",
"123,2,107,74,30,100,33.6,0.404,23,0",
"124,5,132,80,0,0,26.8,0.186,69,0",
"125,0,113,76,0,0,33.3,0.278,23,1",
"126,1,88,30,42,99,55.0,0.496,26,1",
"127,3,120,70,30,135,42.9,0.452,30,0",
"128,1,118,58,36,94,33.3,0.261,23,0",
"129,1,117,88,24,145,34.5,0.403,40,1",
"130,0,105,84,0,0,27.9,0.741,62,1",
"131,4,173,70,14,168,29.7,0.361,33,1",
"132,9,122,56,0,0,33.3,1.114,33,1",
"133,3,170,64,37,225,34.5,0.356,30,1",
"134,8,84,74,31,0,38.3,0.457,39,0",
"135,2,96,68,13,49,21.1,0.647,26,0",
"136,2,125,60,20,140,33.8,0.088,31,0",
"137,0,100,70,26,50,30.8,0.597,21,0",
"138,0,93,60,25,92,28.7,0.532,22,0",
"139,0,129,80,0,0,31.2,0.703,29,0",
"140,5,105,72,29,325,36.9,0.159,28,0",
"141,3,128,78,0,0,21.1,0.268,55,0",
"142,5,106,82,30,0,39.5,0.286,38,0",
"143,2,108,52,26,63,32.5,0.318,22,0",
"144,10,108,66,0,0,32.4,0.272,42,1",
"145,4,154,62,31,284,32.8,0.237,23,0",
"146,0,102,75,23,0,0.0,0.572,21,0",
"147,9,57,80,37,0,32.8,0.096,41,0",
"148,2,106,64,35,119,30.5,1.400,34,0",
"149,5,147,78,0,0,33.7,0.218,65,0",
"150,2,90,70,17,0,27.3,0.085,22,0",
"151,1,136,74,50,204,37.4,0.399,24,0",
"152,4,114,65,0,0,21.9,0.432,37,0",
"153,9,156,86,28,155,34.3,1.189,42,1",
"154,1,153,82,42,485,40.6,0.687,23,0",
"155,8,188,78,0,0,47.9,0.137,43,1",
"156,7,152,88,44,0,50.0,0.337,36,1",
"157,2,99,52,15,94,24.6,0.637,21,0",
"158,1,109,56,21,135,25.2,0.833,23,0",
"159,2,88,74,19,53,29.0,0.229,22,0",
"160,17,163,72,41,114,40.9,0.817,47,1",
"161,4,151,90,38,0,29.7,0.294,36,0",
"162,7,102,74,40,105,37.2,0.204,45,0",
"163,0,114,80,34,285,44.2,0.167,27,0",
"164,2,100,64,23,0,29.7,0.368,21,0",
"165,0,131,88,0,0,31.6,0.743,32,1",
"166,6,104,74,18,156,29.9,0.722,41,1",
"167,3,148,66,25,0,32.5,0.256,22,0",
"168,4,120,68,0,0,29.6,0.709,34,0",
"169,4,110,66,0,0,31.9,0.471,29,0",
"170,3,111,90,12,78,28.4,0.495,29,0",
"171,6,102,82,0,0,30.8,0.180,36,1",
"172,6,134,70,23,130,35.4,0.542,29,1",
"173,2,87,0,23,0,28.9,0.773,25,0",
"174,1,79,60,42,48,43.5,0.678,23,0",
"175,2,75,64,24,55,29.7,0.370,33,0",
"176,8,179,72,42,130,32.7,0.719,36,1",
"177,6,85,78,0,0,31.2,0.382,42,0",
"178,0,129,110,46,130,67.1,0.319,26,1",
"179,5,143,78,0,0,45.0,0.190,47,0",
"180,5,130,82,0,0,39.1,0.956,37,1",
"181,6,87,80,0,0,23.2,0.084,32,0",
"182,0,119,64,18,92,34.9,0.725,23,0",
"183,1,0,74,20,23,27.7,0.299,21,0",
"184,5,73,60,0,0,26.8,0.268,27,0",
"185,4,141,74,0,0,27.6,0.244,40,0",
"186,7,194,68,28,0,35.9,0.745,41,1",
"187,8,181,68,36,495,30.1,0.615,60,1",
"188,1,128,98,41,58,32.0,1.321,33,1",
"189,8,109,76,39,114,27.9,0.640,31,1",
"190,5,139,80,35,160,31.6,0.361,25,1",
"191,3,111,62,0,0,22.6,0.142,21,0",
"192,9,123,70,44,94,33.1,0.374,40,0",
"193,7,159,66,0,0,30.4,0.383,36,1",
"194,11,135,0,0,0,52.3,0.578,40,1",
"195,8,85,55,20,0,24.4,0.136,42,0",
"196,5,158,84,41,210,39.4,0.395,29,1",
"197,1,105,58,0,0,24.3,0.187,21,0",
"198,3,107,62,13,48,22.9,0.678,23,1",
"199,4,109,64,44,99,34.8,0.905,26,1",
"200,4,148,60,27,318,30.9,0.150,29,1",
"201,0,113,80,16,0,31.0,0.874,21,0",
"202,1,138,82,0,0,40.1,0.236,28,0",
"203,0,108,68,20,0,27.3,0.787,32,0",
"204,2,99,70,16,44,20.4,0.235,27,0",
"205,6,103,72,32,190,37.7,0.324,55,0",
"206,5,111,72,28,0,23.9,0.407,27,0",
"207,8,196,76,29,280,37.5,0.605,57,1",
"208,5,162,104,0,0,37.7,0.151,52,1",
"209,1,96,64,27,87,33.2,0.289,21,0",
"210,7,184,84,33,0,35.5,0.355,41,1",
"211,2,81,60,22,0,27.7,0.290,25,0",
"212,0,147,85,54,0,42.8,0.375,24,0",
"213,7,179,95,31,0,34.2,0.164,60,0",
"214,0,140,65,26,130,42.6,0.431,24,1",
"215,9,112,82,32,175,34.2,0.260,36,1",
"216,12,151,70,40,271,41.8,0.742,38,1",
"217,5,109,62,41,129,35.8,0.514,25,1",
"218,6,125,68,30,120,30.0,0.464,32,0",
"219,5,85,74,22,0,29.0,1.224,32,1",
"220,5,112,66,0,0,37.8,0.261,41,1",
"221,0,177,60,29,478,34.6,1.072,21,1",
"222,2,158,90,0,0,31.6,0.805,66,1",
"223,7,119,0,0,0,25.2,0.209,37,0",
"224,7,142,60,33,190,28.8,0.687,61,0",
"225,1,100,66,15,56,23.6,0.666,26,0",
"226,1,87,78,27,32,34.6,0.101,22,0",
"227,0,101,76,0,0,35.7,0.198,26,0",
"228,3,162,52,38,0,37.2,0.652,24,1",
"229,4,197,70,39,744,36.7,2.329,31,0",
"230,0,117,80,31,53,45.2,0.089,24,0",
"231,4,142,86,0,0,44.0,0.645,22,1",
"232,6,134,80,37,370,46.2,0.238,46,1",
"233,1,79,80,25,37,25.4,0.583,22,0",
"234,4,122,68,0,0,35.0,0.394,29,0",
"235,3,74,68,28,45,29.7,0.293,23,0",
"236,4,171,72,0,0,43.6,0.479,26,1",
"237,7,181,84,21,192,35.9,0.586,51,1",
"238,0,179,90,27,0,44.1,0.686,23,1",
"239,9,164,84,21,0,30.8,0.831,32,1",
"240,0,104,76,0,0,18.4,0.582,27,0",
"241,1,91,64,24,0,29.2,0.192,21,0",
"242,4,91,70,32,88,33.1,0.446,22,0",
"243,3,139,54,0,0,25.6,0.402,22,1",
"244,6,119,50,22,176,27.1,1.318,33,1",
"245,2,146,76,35,194,38.2,0.329,29,0",
"246,9,184,85,15,0,30.0,1.213,49,1",
"247,10,122,68,0,0,31.2,0.258,41,0",
"248,0,165,90,33,680,52.3,0.427,23,0",
"249,9,124,70,33,402,35.4,0.282,34,0",
"250,1,111,86,19,0,30.1,0.143,23,0",
"251,9,106,52,0,0,31.2,0.380,42,0",
"252,2,129,84,0,0,28.0,0.284,27,0",
"253,2,90,80,14,55,24.4,0.249,24,0",
"254,0,86,68,32,0,35.8,0.238,25,0",
"255,12,92,62,7,258,27.6,0.926,44,1",
"256,1,113,64,35,0,33.6,0.543,21,1",
"257,3,111,56,39,0,30.1,0.557,30,0",
"258,2,114,68,22,0,28.7,0.092,25,0",
"259,1,193,50,16,375,25.9,0.655,24,0",
"260,11,155,76,28,150,33.3,1.353,51,1",
"261,3,191,68,15,130,30.9,0.299,34,0",
"262,3,141,0,0,0,30.0,0.761,27,1",
"263,4,95,70,32,0,32.1,0.612,24,0",
"264,3,142,80,15,0,32.4,0.200,63,0",
"265,4,123,62,0,0,32.0,0.226,35,1",
"266,5,96,74,18,67,33.6,0.997,43,0",
"267,0,138,0,0,0,36.3,0.933,25,1",
"268,2,128,64,42,0,40.0,1.101,24,0",
"269,0,102,52,0,0,25.1,0.078,21,0",
"270,2,146,0,0,0,27.5,0.240,28,1",
"271,10,101,86,37,0,45.6,1.136,38,1",
"272,2,108,62,32,56,25.2,0.128,21,0",
"273,3,122,78,0,0,23.0,0.254,40,0",
"274,1,71,78,50,45,33.2,0.422,21,0",
"275,13,106,70,0,0,34.2,0.251,52,0",
"276,2,100,70,52,57,40.5,0.677,25,0",
"277,7,106,60,24,0,26.5,0.296,29,1",
"278,0,104,64,23,116,27.8,0.454,23,0",
"279,5,114,74,0,0,24.9,0.744,57,0",
"280,2,108,62,10,278,25.3,0.881,22,0",
"281,0,146,70,0,0,37.9,0.334,28,1",
"282,10,129,76,28,122,35.9,0.280,39,0",
"283,7,133,88,15,155,32.4,0.262,37,0",
"284,7,161,86,0,0,30.4,0.165,47,1",
"285,2,108,80,0,0,27.0,0.259,52,1",
"286,7,136,74,26,135,26.0,0.647,51,0",
"287,5,155,84,44,545,38.7,0.619,34,0",
"288,1,119,86,39,220,45.6,0.808,29,1",
"289,4,96,56,17,49,20.8,0.340,26,0",
"290,5,108,72,43,75,36.1,0.263,33,0",
"291,0,78,88,29,40,36.9,0.434,21,0",
"292,0,107,62,30,74,36.6,0.757,25,1",
"293,2,128,78,37,182,43.3,1.224,31,1",
"294,1,128,48,45,194,40.5,0.613,24,1",
"295,0,161,50,0,0,21.9,0.254,65,0",
"296,6,151,62,31,120,35.5,0.692,28,0",
"297,2,146,70,38,360,28.0,0.337,29,1",
"298,0,126,84,29,215,30.7,0.520,24,0",
"299,14,100,78,25,184,36.6,0.412,46,1",
"300,8,112,72,0,0,23.6,0.840,58,0",
"301,0,167,0,0,0,32.3,0.839,30,1",
"302,2,144,58,33,135,31.6,0.422,25,1",
"303,5,77,82,41,42,35.8,0.156,35,0",
"304,5,115,98,0,0,52.9,0.209,28,1",
"305,3,150,76,0,0,21.0,0.207,37,0",
"306,2,120,76,37,105,39.7,0.215,29,0",
"307,10,161,68,23,132,25.5,0.326,47,1",
"308,0,137,68,14,148,24.8,0.143,21,0",
"309,0,128,68,19,180,30.5,1.391,25,1",
"310,2,124,68,28,205,32.9,0.875,30,1",
"311,6,80,66,30,0,26.2,0.313,41,0",
"312,0,106,70,37,148,39.4,0.605,22,0",
"313,2,155,74,17,96,26.6,0.433,27,1",
"314,3,113,50,10,85,29.5,0.626,25,0",
"315,7,109,80,31,0,35.9,1.127,43,1",
"316,2,112,68,22,94,34.1,0.315,26,0",
"317,3,99,80,11,64,19.3,0.284,30,0",
"318,3,182,74,0,0,30.5,0.345,29,1",
"319,3,115,66,39,140,38.1,0.150,28,0",
"320,6,194,78,0,0,23.5,0.129,59,1",
"321,4,129,60,12,231,27.5,0.527,31,0",
"322,3,112,74,30,0,31.6,0.197,25,1",
"323,0,124,70,20,0,27.4,0.254,36,1",
"324,13,152,90,33,29,26.8,0.731,43,1",
"325,2,112,75,32,0,35.7,0.148,21,0",
"326,1,157,72,21,168,25.6,0.123,24,0",
"327,1,122,64,32,156,35.1,0.692,30,1",
"328,10,179,70,0,0,35.1,0.200,37,0",
"329,2,102,86,36,120,45.5,0.127,23,1",
"330,6,105,70,32,68,30.8,0.122,37,0",
"331,8,118,72,19,0,23.1,1.476,46,0",
"332,2,87,58,16,52,32.7,0.166,25,0",
"333,1,180,0,0,0,43.3,0.282,41,1",
"334,12,106,80,0,0,23.6,0.137,44,0",
"335,1,95,60,18,58,23.9,0.260,22,0",
"336,0,165,76,43,255,47.9,0.259,26,0",
"337,0,117,0,0,0,33.8,0.932,44,0",
"338,5,115,76,0,0,31.2,0.343,44,1",
"339,9,152,78,34,171,34.2,0.893,33,1",
"340,7,178,84,0,0,39.9,0.331,41,1",
"341,1,130,70,13,105,25.9,0.472,22,0",
"342,1,95,74,21,73,25.9,0.673,36,0",
"343,1,0,68,35,0,32.0,0.389,22,0",
"344,5,122,86,0,0,34.7,0.290,33,0",
"345,8,95,72,0,0,36.8,0.485,57,0",
"346,8,126,88,36,108,38.5,0.349,49,0",
"347,1,139,46,19,83,28.7,0.654,22,0",
"348,3,116,0,0,0,23.5,0.187,23,0",
"349,3,99,62,19,74,21.8,0.279,26,0",
"350,5,0,80,32,0,41.0,0.346,37,1",
"351,4,92,80,0,0,42.2,0.237,29,0",
"352,4,137,84,0,0,31.2,0.252,30,0",
"353,3,61,82,28,0,34.4,0.243,46,0",
"354,1,90,62,12,43,27.2,0.580,24,0",
"355,3,90,78,0,0,42.7,0.559,21,0",
"356,9,165,88,0,0,30.4,0.302,49,1",
"357,1,125,50,40,167,33.3,0.962,28,1",
"358,13,129,0,30,0,39.9,0.569,44,1",
"359,12,88,74,40,54,35.3,0.378,48,0",
"360,1,196,76,36,249,36.5,0.875,29,1",
"361,5,189,64,33,325,31.2,0.583,29,1",
"362,5,158,70,0,0,29.8,0.207,63,0",
"363,5,103,108,37,0,39.2,0.305,65,0",
"364,4,146,78,0,0,38.5,0.520,67,1",
"365,4,147,74,25,293,34.9,0.385,30,0",
"366,5,99,54,28,83,34.0,0.499,30,0",
"367,6,124,72,0,0,27.6,0.368,29,1",
"368,0,101,64,17,0,21.0,0.252,21,0",
"369,3,81,86,16,66,27.5,0.306,22,0",
"370,1,133,102,28,140,32.8,0.234,45,1",
"371,3,173,82,48,465,38.4,2.137,25,1",
"372,0,118,64,23,89,0.0,1.731,21,0",
"373,0,84,64,22,66,35.8,0.545,21,0",
"374,2,105,58,40,94,34.9,0.225,25,0",
"375,2,122,52,43,158,36.2,0.816,28,0",
"376,12,140,82,43,325,39.2,0.528,58,1",
"377,0,98,82,15,84,25.2,0.299,22,0",
"378,1,87,60,37,75,37.2,0.509,22,0",
"379,4,156,75,0,0,48.3,0.238,32,1",
"380,0,93,100,39,72,43.4,1.021,35,0",
"381,1,107,72,30,82,30.8,0.821,24,0",
"382,0,105,68,22,0,20.0,0.236,22,0",
"383,1,109,60,8,182,25.4,0.947,21,0",
"384,1,90,62,18,59,25.1,1.268,25,0",
"385,1,125,70,24,110,24.3,0.221,25,0",
"386,1,119,54,13,50,22.3,0.205,24,0",
"387,5,116,74,29,0,32.3,0.660,35,1",
"388,8,105,100,36,0,43.3,0.239,45,1",
"389,5,144,82,26,285,32.0,0.452,58,1",
"390,3,100,68,23,81,31.6,0.949,28,0",
"391,1,100,66,29,196,32.0,0.444,42,0",
"392,5,166,76,0,0,45.7,0.340,27,1",
"393,1,131,64,14,415,23.7,0.389,21,0",
"394,4,116,72,12,87,22.1,0.463,37,0",
"395,4,158,78,0,0,32.9,0.803,31,1",
"396,2,127,58,24,275,27.7,1.600,25,0",
"397,3,96,56,34,115,24.7,0.944,39,0",
"398,0,131,66,40,0,34.3,0.196,22,1",
"399,3,82,70,0,0,21.1,0.389,25,0",
"400,3,193,70,31,0,34.9,0.241,25,1",
"401,4,95,64,0,0,32.0,0.161,31,1",
"402,6,137,61,0,0,24.2,0.151,55,0",
"403,5,136,84,41,88,35.0,0.286,35,1",
"404,9,72,78,25,0,31.6,0.280,38,0",
"405,5,168,64,0,0,32.9,0.135,41,1",
"406,2,123,48,32,165,42.1,0.520,26,0",
"407,4,115,72,0,0,28.9,0.376,46,1",
"408,0,101,62,0,0,21.9,0.336,25,0",
"409,8,197,74,0,0,25.9,1.191,39,1",
"410,1,172,68,49,579,42.4,0.702,28,1",
"411,6,102,90,39,0,35.7,0.674,28,0",
"412,1,112,72,30,176,34.4,0.528,25,0",
"413,1,143,84,23,310,42.4,1.076,22,0",
"414,1,143,74,22,61,26.2,0.256,21,0",
"415,0,138,60,35,167,34.6,0.534,21,1",
"416,3,173,84,33,474,35.7,0.258,22,1",
"417,1,97,68,21,0,27.2,1.095,22,0",
"418,4,144,82,32,0,38.5,0.554,37,1",
"419,1,83,68,0,0,18.2,0.624,27,0",
"420,3,129,64,29,115,26.4,0.219,28,1",
"421,1,119,88,41,170,45.3,0.507,26,0",
"422,2,94,68,18,76,26.0,0.561,21,0",
"423,0,102,64,46,78,40.6,0.496,21,0",
"424,2,115,64,22,0,30.8,0.421,21,0",
"425,8,151,78,32,210,42.9,0.516,36,1",
"426,4,184,78,39,277,37.0,0.264,31,1",
"427,0,94,0,0,0,0.0,0.256,25,0",
"428,1,181,64,30,180,34.1,0.328,38,1",
"429,0,135,94,46,145,40.6,0.284,26,0",
"430,1,95,82,25,180,35.0,0.233,43,1",
"431,2,99,0,0,0,22.2,0.108,23,0",
"432,3,89,74,16,85,30.4,0.551,38,0",
"433,1,80,74,11,60,30.0,0.527,22,0",
"434,2,139,75,0,0,25.6,0.167,29,0",
"435,1,90,68,8,0,24.5,1.138,36,0",
"436,0,141,0,0,0,42.4,0.205,29,1",
"437,12,140,85,33,0,37.4,0.244,41,0",
"438,5,147,75,0,0,29.9,0.434,28,0",
"439,1,97,70,15,0,18.2,0.147,21,0",
"440,6,107,88,0,0,36.8,0.727,31,0",
"441,0,189,104,25,0,34.3,0.435,41,1",
"442,2,83,66,23,50,32.2,0.497,22,0",
"443,4,117,64,27,120,33.2,0.230,24,0",
"444,8,108,70,0,0,30.5,0.955,33,1",
"445,4,117,62,12,0,29.7,0.380,30,1",
"446,0,180,78,63,14,59.4,2.420,25,1",
"447,1,100,72,12,70,25.3,0.658,28,0",
"448,0,95,80,45,92,36.5,0.330,26,0",
"449,0,104,64,37,64,33.6,0.510,22,1",
"450,0,120,74,18,63,30.5,0.285,26,0",
"451,1,82,64,13,95,21.2,0.415,23,0",
"452,2,134,70,0,0,28.9,0.542,23,1",
"453,0,91,68,32,210,39.9,0.381,25,0",
"454,2,119,0,0,0,19.6,0.832,72,0",
"455,2,100,54,28,105,37.8,0.498,24,0",
"456,14,175,62,30,0,33.6,0.212,38,1",
"457,1,135,54,0,0,26.7,0.687,62,0",
"458,5,86,68,28,71,30.2,0.364,24,0",
"459,10,148,84,48,237,37.6,1.001,51,1",
"460,9,134,74,33,60,25.9,0.460,81,0",
"461,9,120,72,22,56,20.8,0.733,48,0",
"462,1,71,62,0,0,21.8,0.416,26,0",
"463,8,74,70,40,49,35.3,0.705,39,0",
"464,5,88,78,30,0,27.6,0.258,37,0",
"465,10,115,98,0,0,24.0,1.022,34,0",
"466,0,124,56,13,105,21.8,0.452,21,0",
"467,0,74,52,10,36,27.8,0.269,22,0",
"468,0,97,64,36,100,36.8,0.600,25,0",
"469,8,120,0,0,0,30.0,0.183,38,1",
"470,6,154,78,41,140,46.1,0.571,27,0",
"471,1,144,82,40,0,41.3,0.607,28,0",
"472,0,137,70,38,0,33.2,0.170,22,0",
"473,0,119,66,27,0,38.8,0.259,22,0",
"474,7,136,90,0,0,29.9,0.210,50,0",
"475,4,114,64,0,0,28.9,0.126,24,0",
"476,0,137,84,27,0,27.3,0.231,59,0",
"477,2,105,80,45,191,33.7,0.711,29,1",
"478,7,114,76,17,110,23.8,0.466,31,0",
"479,8,126,74,38,75,25.9,0.162,39,0",
"480,4,132,86,31,0,28.0,0.419,63,0",
"481,3,158,70,30,328,35.5,0.344,35,1",
"482,0,123,88,37,0,35.2,0.197,29,0",
"483,4,85,58,22,49,27.8,0.306,28,0",
"484,0,84,82,31,125,38.2,0.233,23,0",
"485,0,145,0,0,0,44.2,0.630,31,1",
"486,0,135,68,42,250,42.3,0.365,24,1",
"487,1,139,62,41,480,40.7,0.536,21,0",
"488,0,173,78,32,265,46.5,1.159,58,0",
"489,4,99,72,17,0,25.6,0.294,28,0",
"490,8,194,80,0,0,26.1,0.551,67,0",
"491,2,83,65,28,66,36.8,0.629,24,0",
"492,2,89,90,30,0,33.5,0.292,42,0",
"493,4,99,68,38,0,32.8,0.145,33,0",
"494,4,125,70,18,122,28.9,1.144,45,1",
"495,3,80,0,0,0,0.0,0.174,22,0",
"496,6,166,74,0,0,26.6,0.304,66,0",
"497,5,110,68,0,0,26.0,0.292,30,0",
"498,2,81,72,15,76,30.1,0.547,25,0",
"499,7,195,70,33,145,25.1,0.163,55,1",
"500,6,154,74,32,193,29.3,0.839,39,0",
"501,2,117,90,19,71,25.2,0.313,21,0",
"502,3,84,72,32,0,37.2,0.267,28,0",
"503,6,0,68,41,0,39.0,0.727,41,1",
"504,7,94,64,25,79,33.3,0.738,41,0",
"505,3,96,78,39,0,37.3,0.238,40,0",
"506,10,75,82,0,0,33.3,0.263,38,0",
"507,0,180,90,26,90,36.5,0.314,35,1",
"508,1,130,60,23,170,28.6,0.692,21,0",
"509,2,84,50,23,76,30.4,0.968,21,0",
"510,8,120,78,0,0,25.0,0.409,64,0",
"511,12,84,72,31,0,29.7,0.297,46,1",
"512,0,139,62,17,210,22.1,0.207,21,0",
"513,9,91,68,0,0,24.2,0.200,58,0",
"514,2,91,62,0,0,27.3,0.525,22,0",
"515,3,99,54,19,86,25.6,0.154,24,0",
"516,3,163,70,18,105,31.6,0.268,28,1",
"517,9,145,88,34,165,30.3,0.771,53,1",
"518,7,125,86,0,0,37.6,0.304,51,0",
"519,13,76,60,0,0,32.8,0.180,41,0",
"520,6,129,90,7,326,19.6,0.582,60,0",
"521,2,68,70,32,66,25.0,0.187,25,0",
"522,3,124,80,33,130,33.2,0.305,26,0",
"523,6,114,0,0,0,0.0,0.189,26,0",
"524,9,130,70,0,0,34.2,0.652,45,1",
"525,3,125,58,0,0,31.6,0.151,24,0",
"526,3,87,60,18,0,21.8,0.444,21,0",
"527,1,97,64,19,82,18.2,0.299,21,0",
"528,3,116,74,15,105,26.3,0.107,24,0",
"529,0,117,66,31,188,30.8,0.493,22,0",
"530,0,111,65,0,0,24.6,0.660,31,0",
"531,2,122,60,18,106,29.8,0.717,22,0",
"532,0,107,76,0,0,45.3,0.686,24,0",
"533,1,86,66,52,65,41.3,0.917,29,0",
"534,6,91,0,0,0,29.8,0.501,31,0",
"535,1,77,56,30,56,33.3,1.251,24,0",
"536,4,132,0,0,0,32.9,0.302,23,1",
"537,0,105,90,0,0,29.6,0.197,46,0",
"538,0,57,60,0,0,21.7,0.735,67,0",
"539,0,127,80,37,210,36.3,0.804,23,0",
"540,3,129,92,49,155,36.4,0.968,32,1",
"541,8,100,74,40,215,39.4,0.661,43,1",
"542,3,128,72,25,190,32.4,0.549,27,1",
"543,10,90,85,32,0,34.9,0.825,56,1",
"544,4,84,90,23,56,39.5,0.159,25,0",
"545,1,88,78,29,76,32.0,0.365,29,0",
"546,8,186,90,35,225,34.5,0.423,37,1",
"547,5,187,76,27,207,43.6,1.034,53,1",
"548,4,131,68,21,166,33.1,0.160,28,0",
"549,1,164,82,43,67,32.8,0.341,50,0",
"550,4,189,110,31,0,28.5,0.680,37,0",
"551,1,116,70,28,0,27.4,0.204,21,0",
"552,3,84,68,30,106,31.9,0.591,25,0",
"553,6,114,88,0,0,27.8,0.247,66,0",
"554,1,88,62,24,44,29.9,0.422,23,0",
"555,1,84,64,23,115,36.9,0.471,28,0",
"556,7,124,70,33,215,25.5,0.161,37,0",
"557,1,97,70,40,0,38.1,0.218,30,0",
"558,8,110,76,0,0,27.8,0.237,58,0",
"559,11,103,68,40,0,46.2,0.126,42,0",
"560,11,85,74,0,0,30.1,0.300,35,0",
"561,6,125,76,0,0,33.8,0.121,54,1",
"562,0,198,66,32,274,41.3,0.502,28,1",
"563,1,87,68,34,77,37.6,0.401,24,0",
"564,6,99,60,19,54,26.9,0.497,32,0",
"565,0,91,80,0,0,32.4,0.601,27,0",
"566,2,95,54,14,88,26.1,0.748,22,0",
"567,1,99,72,30,18,38.6,0.412,21,0",
"568,6,92,62,32,126,32.0,0.085,46,0",
"569,4,154,72,29,126,31.3,0.338,37,0",
"570,0,121,66,30,165,34.3,0.203,33,1",
"571,3,78,70,0,0,32.5,0.270,39,0",
"572,2,130,96,0,0,22.6,0.268,21,0",
"573,3,111,58,31,44,29.5,0.430,22,0",
"574,2,98,60,17,120,34.7,0.198,22,0",
"575,1,143,86,30,330,30.1,0.892,23,0",
"576,1,119,44,47,63,35.5,0.280,25,0",
"577,6,108,44,20,130,24.0,0.813,35,0",
"578,2,118,80,0,0,42.9,0.693,21,1",
"579,10,133,68,0,0,27.0,0.245,36,0",
"580,2,197,70,99,0,34.7,0.575,62,1",
"581,0,151,90,46,0,42.1,0.371,21,1",
"582,6,109,60,27,0,25.0,0.206,27,0",
"583,12,121,78,17,0,26.5,0.259,62,0",
"584,8,100,76,0,0,38.7,0.190,42,0",
"585,8,124,76,24,600,28.7,0.687,52,1",
"586,1,93,56,11,0,22.5,0.417,22,0",
"587,8,143,66,0,0,34.9,0.129,41,1",
"588,6,103,66,0,0,24.3,0.249,29,0",
"589,3,176,86,27,156,33.3,1.154,52,1",
"590,0,73,0,0,0,21.1,0.342,25,0",
"591,11,111,84,40,0,46.8,0.925,45,1",
"592,2,112,78,50,140,39.4,0.175,24,0",
"593,3,132,80,0,0,34.4,0.402,44,1",
"594,2,82,52,22,115,28.5,1.699,25,0",
"595,6,123,72,45,230,33.6,0.733,34,0",
"596,0,188,82,14,185,32.0,0.682,22,1",
"597,0,67,76,0,0,45.3,0.194,46,0",
"598,1,89,24,19,25,27.8,0.559,21,0",
"599,1,173,74,0,0,36.8,0.088,38,1",
"600,1,109,38,18,120,23.1,0.407,26,0",
"601,1,108,88,19,0,27.1,0.400,24,0",
"602,6,96,0,0,0,23.7,0.190,28,0",
"603,1,124,74,36,0,27.8,0.100,30,0",
"604,7,150,78,29,126,35.2,0.692,54,1",
"605,4,183,0,0,0,28.4,0.212,36,1",
"606,1,124,60,32,0,35.8,0.514,21,0",
"607,1,181,78,42,293,40.0,1.258,22,1",
"608,1,92,62,25,41,19.5,0.482,25,0",
"609,0,152,82,39,272,41.5,0.270,27,0",
"610,1,111,62,13,182,24.0,0.138,23,0",
"611,3,106,54,21,158,30.9,0.292,24,0",
"612,3,174,58,22,194,32.9,0.593,36,1",
"613,7,168,88,42,321,38.2,0.787,40,1",
"614,6,105,80,28,0,32.5,0.878,26,0",
"615,11,138,74,26,144,36.1,0.557,50,1",
"616,3,106,72,0,0,25.8,0.207,27,0",
"617,6,117,96,0,0,28.7,0.157,30,0",
"618,2,68,62,13,15,20.1,0.257,23,0",
"619,9,112,82,24,0,28.2,1.282,50,1",
"620,0,119,0,0,0,32.4,0.141,24,1",
"621,2,112,86,42,160,38.4,0.246,28,0",
"622,2,92,76,20,0,24.2,1.698,28,0",
"623,6,183,94,0,0,40.8,1.461,45,0",
"624,0,94,70,27,115,43.5,0.347,21,0",
"625,2,108,64,0,0,30.8,0.158,21,0",
"626,4,90,88,47,54,37.7,0.362,29,0",
"627,0,125,68,0,0,24.7,0.206,21,0",
"628,0,132,78,0,0,32.4,0.393,21,0",
"629,5,128,80,0,0,34.6,0.144,45,0",
"630,4,94,65,22,0,24.7,0.148,21,0",
"631,7,114,64,0,0,27.4,0.732,34,1",
"632,0,102,78,40,90,34.5,0.238,24,0",
"633,2,111,60,0,0,26.2,0.343,23,0",
"634,1,128,82,17,183,27.5,0.115,22,0",
"635,10,92,62,0,0,25.9,0.167,31,0",
"636,13,104,72,0,0,31.2,0.465,38,1",
"637,5,104,74,0,0,28.8,0.153,48,0",
"638,2,94,76,18,66,31.6,0.649,23,0",
"639,7,97,76,32,91,40.9,0.871,32,1",
"640,1,100,74,12,46,19.5,0.149,28,0",
"641,0,102,86,17,105,29.3,0.695,27,0",
"642,4,128,70,0,0,34.3,0.303,24,0",
"643,6,147,80,0,0,29.5,0.178,50,1",
"644,4,90,0,0,0,28.0,0.610,31,0",
"645,3,103,72,30,152,27.6,0.730,27,0",
"646,2,157,74,35,440,39.4,0.134,30,0",
"647,1,167,74,17,144,23.4,0.447,33,1",
"648,0,179,50,36,159,37.8,0.455,22,1",
"649,11,136,84,35,130,28.3,0.260,42,1",
"650,0,107,60,25,0,26.4,0.133,23,0",
"651,1,91,54,25,100,25.2,0.234,23,0",
"652,1,117,60,23,106,33.8,0.466,27,0",
"653,5,123,74,40,77,34.1,0.269,28,0",
"654,2,120,54,0,0,26.8,0.455,27,0",
"655,1,106,70,28,135,34.2,0.142,22,0",
"656,2,155,52,27,540,38.7,0.240,25,1",
"657,2,101,58,35,90,21.8,0.155,22,0",
"658,1,120,80,48,200,38.9,1.162,41,0",
"659,11,127,106,0,0,39.0,0.190,51,0",
"660,3,80,82,31,70,34.2,1.292,27,1",
"661,10,162,84,0,0,27.7,0.182,54,0",
"662,1,199,76,43,0,42.9,1.394,22,1",
"663,8,167,106,46,231,37.6,0.165,43,1",
"664,9,145,80,46,130,37.9,0.637,40,1",
"665,6,115,60,39,0,33.7,0.245,40,1",
"666,1,112,80,45,132,34.8,0.217,24,0",
"667,4,145,82,18,0,32.5,0.235,70,1",
"668,10,111,70,27,0,27.5,0.141,40,1",
"669,6,98,58,33,190,34.0,0.430,43,0",
"670,9,154,78,30,100,30.9,0.164,45,0",
"671,6,165,68,26,168,33.6,0.631,49,0",
"672,1,99,58,10,0,25.4,0.551,21,0",
"673,10,68,106,23,49,35.5,0.285,47,0",
"674,3,123,100,35,240,57.3,0.880,22,0",
"675,8,91,82,0,0,35.6,0.587,68,0",
"676,6,195,70,0,0,30.9,0.328,31,1",
"677,9,156,86,0,0,24.8,0.230,53,1",
"678,0,93,60,0,0,35.3,0.263,25,0",
"679,3,121,52,0,0,36.0,0.127,25,1",
"680,2,101,58,17,265,24.2,0.614,23,0",
"681,2,56,56,28,45,24.2,0.332,22,0",
"682,0,162,76,36,0,49.6,0.364,26,1",
"683,0,95,64,39,105,44.6,0.366,22,0",
"684,4,125,80,0,0,32.3,0.536,27,1",
"685,5,136,82,0,0,0.0,0.640,69,0",
"686,2,129,74,26,205,33.2,0.591,25,0",
"687,3,130,64,0,0,23.1,0.314,22,0",
"688,1,107,50,19,0,28.3,0.181,29,0",
"689,1,140,74,26,180,24.1,0.828,23,0",
"690,1,144,82,46,180,46.1,0.335,46,1",
"691,8,107,80,0,0,24.6,0.856,34,0",
"692,13,158,114,0,0,42.3,0.257,44,1",
"693,2,121,70,32,95,39.1,0.886,23,0",
"694,7,129,68,49,125,38.5,0.439,43,1",
"695,2,90,60,0,0,23.5,0.191,25,0",
"696,7,142,90,24,480,30.4,0.128,43,1",
"697,3,169,74,19,125,29.9,0.268,31,1",
"698,0,99,0,0,0,25.0,0.253,22,0",
"699,4,127,88,11,155,34.5,0.598,28,0",
"700,4,118,70,0,0,44.5,0.904,26,0",
"701,2,122,76,27,200,35.9,0.483,26,0",
"702,6,125,78,31,0,27.6,0.565,49,1",
"703,1,168,88,29,0,35.0,0.905,52,1",
"704,2,129,0,0,0,38.5,0.304,41,0",
"705,4,110,76,20,100,28.4,0.118,27,0",
"706,6,80,80,36,0,39.8,0.177,28,0",
"707,10,115,0,0,0,0.0,0.261,30,1",
"708,2,127,46,21,335,34.4,0.176,22,0",
"709,9,164,78,0,0,32.8,0.148,45,1",
"710,2,93,64,32,160,38.0,0.674,23,1",
"711,3,158,64,13,387,31.2,0.295,24,0",
"712,5,126,78,27,22,29.6,0.439,40,0",
"713,10,129,62,36,0,41.2,0.441,38,1",
"714,0,134,58,20,291,26.4,0.352,21,0",
"715,3,102,74,0,0,29.5,0.121,32,0",
"716,7,187,50,33,392,33.9,0.826,34,1",
"717,3,173,78,39,185,33.8,0.970,31,1",
"718,10,94,72,18,0,23.1,0.595,56,0",
"719,1,108,60,46,178,35.5,0.415,24,0",
"720,5,97,76,27,0,35.6,0.378,52,1",
"721,4,83,86,19,0,29.3,0.317,34,0",
"722,1,114,66,36,200,38.1,0.289,21,0",
"723,1,149,68,29,127,29.3,0.349,42,1",
"724,5,117,86,30,105,39.1,0.251,42,0",
"725,1,111,94,0,0,32.8,0.265,45,0",
"726,4,112,78,40,0,39.4,0.236,38,0",
"727,1,116,78,29,180,36.1,0.496,25,0",
"728,0,141,84,26,0,32.4,0.433,22,0",
"729,2,175,88,0,0,22.9,0.326,22,0",
"730,2,92,52,0,0,30.1,0.141,22,0",
"731,3,130,78,23,79,28.4,0.323,34,1",
"732,8,120,86,0,0,28.4,0.259,22,1",
"733,2,174,88,37,120,44.5,0.646,24,1",
"734,2,106,56,27,165,29.0,0.426,22,0",
"735,2,105,75,0,0,23.3,0.560,53,0",
"736,4,95,60,32,0,35.4,0.284,28,0",
"737,0,126,86,27,120,27.4,0.515,21,0",
"738,8,65,72,23,0,32.0,0.600,42,0",
"739,2,99,60,17,160,36.6,0.453,21,0",
"740,1,102,74,0,0,39.5,0.293,42,1",
"741,11,120,80,37,150,42.3,0.785,48,1",
"742,3,102,44,20,94,30.8,0.400,26,0",
"743,1,109,58,18,116,28.5,0.219,22,0",
"744,9,140,94,0,0,32.7,0.734,45,1",
"745,13,153,88,37,140,40.6,1.174,39,0",
"746,12,100,84,33,105,30.0,0.488,46,0",
"747,1,147,94,41,0,49.3,0.358,27,1",
"748,1,81,74,41,57,46.3,1.096,32,0",
"749,3,187,70,22,200,36.4,0.408,36,1",
"750,6,162,62,0,0,24.3,0.178,50,1",
"751,4,136,70,0,0,31.2,1.182,22,1",
"752,1,121,78,39,74,39.0,0.261,28,0",
"753,3,108,62,24,0,26.0,0.223,25,0",
"754,0,181,88,44,510,43.3,0.222,26,1",
"755,8,154,78,32,0,32.4,0.443,45,1",
"756,1,128,88,39,110,36.5,1.057,37,1",
"757,7,137,90,41,0,32.0,0.391,39,0",
"758,0,123,72,0,0,36.3,0.258,52,1",
"759,1,106,76,0,0,37.5,0.197,26,0",
"760,6,190,92,0,0,35.5,0.278,66,1",
"761,2,88,58,26,16,28.4,0.766,22,0",
"762,9,170,74,31,0,44.0,0.403,43,1",
"763,9,89,62,0,0,22.5,0.142,33,0",
"764,10,101,76,48,180,32.9,0.171,63,0",
"765,2,122,70,27,0,36.8,0.340,27,0",
"766,5,121,72,23,112,26.2,0.245,30,0",
"767,1,126,60,0,0,30.1,0.349,47,1",
"768,1,93,70,31,0,30.4,0.315,23,0",
)

Banknote_Dataset = (
"Id_int,Variance of Wavelet Transformed image_float,Skewness of Wavelet Transformed image_float,Kurtosis of Wavelet Transformed image_float,Entropy of image_float,Class_int",
"1,3.6216,8.6661,-2.8073,-0.44699,0",
"2,4.5459,8.1674,-2.4586,-1.4621,0",
"3,3.866,-2.6383,1.9242,0.10645,0",
"4,3.4566,9.5228,-4.0112,-3.5944,0",
"5,0.32924,-4.4552,4.5718,-0.9888,0",
"6,4.3684,9.6718,-3.9606,-3.1625,0",
"7,3.5912,3.0129,0.72888,0.56421,0",
"8,2.0922,-6.81,8.4636,-0.60216,0",
"9,3.2032,5.7588,-0.75345,-0.61251,0",
"10,1.5356,9.1772,-2.2718,-0.73535,0",
"11,1.2247,8.7779,-2.2135,-0.80647,0",
"12,3.9899,-2.7066,2.3946,0.86291,0",
"13,1.8993,7.6625,0.15394,-3.1108,0",
"14,-1.5768,10.843,2.5462,-2.9362,0",
"15,3.404,8.7261,-2.9915,-0.57242,0",
"16,4.6765,-3.3895,3.4896,1.4771,0",
"17,2.6719,3.0646,0.37158,0.58619,0",
"18,0.80355,2.8473,4.3439,0.6017,0",
"19,1.4479,-4.8794,8.3428,-2.1086,0",
"20,5.2423,11.0272,-4.353,-4.1013,0",
"21,5.7867,7.8902,-2.6196,-0.48708,0",
"22,0.3292,-4.4552,4.5718,-0.9888,0",
"23,3.9362,10.1622,-3.8235,-4.0172,0",
"24,0.93584,8.8855,-1.6831,-1.6599,0",
"25,4.4338,9.887,-4.6795,-3.7483,0",
"26,0.7057,-5.4981,8.3368,-2.8715,0",
"27,1.1432,-3.7413,5.5777,-0.63578,0",
"28,-0.38214,8.3909,2.1624,-3.7405,0",
"29,6.5633,9.8187,-4.4113,-3.2258,0",
"30,4.8906,-3.3584,3.4202,1.0905,0",
"31,-0.24811,-0.17797,4.9068,0.15429,0",
"32,1.4884,3.6274,3.308,0.48921,0",
"33,4.2969,7.617,-2.3874,-0.96164,0",
"34,-0.96511,9.4111,1.7305,-4.8629,0",
"35,-1.6162,0.80908,8.1628,0.60817,0",
"36,2.4391,6.4417,-0.80743,-0.69139,0",
"37,2.6881,6.0195,-0.46641,-0.69268,0",
"38,3.6289,0.81322,1.6277,0.77627,0",
"39,4.5679,3.1929,-2.1055,0.29653,0",
"40,3.4805,9.7008,-3.7541,-3.4379,0",
"41,4.1711,8.722,-3.0224,-0.59699,0",
"42,-0.2062,9.2207,-3.7044,-6.8103,0",
"43,-0.0068919,9.2931,-0.41243,-1.9638,0",
"44,0.96441,5.8395,2.3235,0.066365,0",
"45,2.8561,6.9176,-0.79372,0.48403,0",
"46,-0.7869,9.5663,-3.7867,-7.5034,0",
"47,2.0843,6.6258,0.48382,-2.2134,0",
"48,-0.7869,9.5663,-3.7867,-7.5034,0",
"49,3.9102,6.065,-2.4534,-0.68234,0",
"50,1.6349,3.286,2.8753,0.087054,0",
"51,4.3239,-4.8835,3.4356,-0.5776,0",
"52,5.262,3.9834,-1.5572,1.0103,0",
"53,3.1452,5.825,-0.51439,-1.4944,0",
"54,2.549,6.1499,-1.1605,-1.2371,0",
"55,4.9264,5.496,-2.4774,-0.50648,0",
"56,4.8265,0.80287,1.6371,1.1875,0",
"57,2.5635,6.7769,-0.61979,0.38576,0",
"58,5.807,5.0097,-2.2384,0.43878,0",
"59,3.1377,-4.1096,4.5701,0.98963,0",
"60,-0.78289,11.3603,-0.37644,-7.0495,0",
"61,2.888,0.44696,4.5907,-0.24398,0",
"62,0.49665,5.527,1.7785,-0.47156,0",
"63,4.2586,11.2962,-4.0943,-4.3457,0",
"64,1.7939,-1.1174,1.5454,-0.26079,0",
"65,5.4021,3.1039,-1.1536,1.5651,0",
"66,2.5367,2.599,2.0938,0.20085,0",
"67,4.6054,-4.0765,2.7587,0.31981,0",
"68,2.4235,9.5332,-3.0789,-2.7746,0",
"69,1.0009,7.7846,-0.28219,-2.6608,0",
"70,0.12326,8.9848,-0.9351,-2.4332,0",
"71,3.9529,-2.3548,2.3792,0.48274,0",
"72,4.1373,0.49248,1.093,1.8276,0",
"73,4.7181,10.0153,-3.9486,-3.8582,0",
"74,4.1654,-3.4495,3.643,1.0879,0",
"75,4.4069,10.9072,-4.5775,-4.4271,0",
"76,2.3066,3.5364,0.57551,0.41938,0",
"77,3.7935,7.9853,-2.5477,-1.872,0",
"78,0.049175,6.1437,1.7828,-0.72113,0",
"79,0.24835,7.6439,0.9885,-0.87371,0",
"80,1.1317,3.9647,3.3979,0.84351,0",
"81,2.8033,9.0862,-3.3668,-1.0224,0",
"82,4.4682,2.2907,0.95766,0.83058,0",
"83,5.0185,8.5978,-2.9375,-1.281,0",
"84,1.8664,7.7763,-0.23849,-2.9634,0",
"85,3.245,6.63,-0.63435,0.86937,0",
"86,4.0296,2.6756,0.80685,0.71679,0",
"87,-1.1313,1.9037,7.5339,1.022,0",
"88,0.87603,6.8141,0.84198,-0.17156,0",
"89,4.1197,-2.7956,2.0707,0.67412,0",
"90,3.8027,0.81529,2.1041,1.0245,0",
"91,1.4806,7.6377,-2.7876,-1.0341,0",
"92,4.0632,3.584,0.72545,0.39481,0",
"93,4.3064,8.2068,-2.7824,-1.4336,0",
"94,2.4486,-6.3175,7.9632,0.20602,0",
"95,3.2718,1.7837,2.1161,0.61334,0",
"96,-0.64472,-4.6062,8.347,-2.7099,0",
"97,2.9543,1.076,0.64577,0.89394,0",
"98,2.1616,-6.8804,8.1517,-0.081048,0",
"99,3.82,10.9279,-4.0112,-5.0284,0",
"100,-2.7419,11.4038,2.5394,-5.5793,0",
"101,3.3669,-5.1856,3.6935,-1.1427,0",
"102,4.5597,-2.4211,2.6413,1.6168,0",
"103,5.1129,-0.49871,0.62863,1.1189,0",
"104,3.3397,-4.6145,3.9823,-0.23751,0",
"105,4.2027,0.22761,0.96108,0.97282,0",
"106,3.5438,1.2395,1.997,2.1547,0",
"107,2.3136,10.6651,-3.5288,-4.7672,0",
"108,-1.8584,7.886,-1.6643,-1.8384,0",
"109,3.106,9.5414,-4.2536,-4.003,0",
"110,2.9163,10.8306,-3.3437,-4.122,0",
"111,3.9922,-4.4676,3.7304,-0.1095,0",
"112,1.518,5.6946,0.094818,-0.026738,0",
"113,3.2351,9.647,-3.2074,-2.5948,0",
"114,4.2188,6.8162,-1.2804,0.76076,0",
"115,1.7819,6.9176,-1.2744,-1.5759,0",
"116,2.5331,2.9135,-0.822,-0.12243,0",
"117,3.8969,7.4163,-1.8245,0.14007,0",
"118,2.108,6.7955,-0.1708,0.4905,0",
"119,2.8969,0.70768,2.29,1.8663,0",
"120,0.9297,-3.7971,4.6429,-0.2957,0",
"121,3.4642,10.6878,-3.4071,-4.109,0",
"122,4.0713,10.4023,-4.1722,-4.7582,0",
"123,-1.4572,9.1214,1.7425,-5.1241,0",
"124,-1.5075,1.9224,7.1466,0.89136,0",
"125,-0.91718,9.9884,1.1804,-5.2263,0",
"126,2.994,7.2011,-1.2153,0.3211,0",
"127,-2.343,12.9516,3.3285,-5.9426,0",
"128,3.7818,-2.8846,2.2558,-0.15734,0",
"129,4.6689,1.3098,0.055404,1.909,0",
"130,3.4663,1.1112,1.7425,1.3388,0",
"131,3.2697,-4.3414,3.6884,-0.29829,0",
"132,5.1302,8.6703,-2.8913,-1.5086,0",
"133,2.0139,6.1416,0.37929,0.56938,0",
"134,0.4339,5.5395,2.033,-0.40432,0",
"135,-1.0401,9.3987,0.85998,-5.3336,0",
"136,4.1605,11.2196,-3.6136,-4.0819,0",
"137,5.438,9.4669,-4.9417,-3.9202,0",
"138,5.032,8.2026,-2.6256,-1.0341,0",
"139,5.2418,10.5388,-4.1174,-4.2797,0",
"140,-0.2062,9.2207,-3.7044,-6.8103,0",
"141,2.0911,0.94358,4.5512,1.234,0",
"142,1.7317,-0.34765,4.1905,-0.99138,0",
"143,4.1736,3.3336,-1.4244,0.60429,0",
"144,3.9232,-3.2467,3.4579,0.83705,0",
"145,3.8481,10.1539,-3.8561,-4.2228,0",
"146,0.5195,-3.2633,3.0895,-0.9849,0",
"147,3.8584,0.78425,1.1033,1.7008,0",
"148,1.7496,-0.1759,5.1827,1.2922,0",
"149,3.6277,0.9829,0.68861,0.63403,0",
"150,2.7391,7.4018,0.071684,-2.5302,0",
"151,4.5447,8.2274,-2.4166,-1.5875,0",
"152,-1.7599,11.9211,2.6756,-3.3241,0",
"153,5.0691,0.21313,0.20278,1.2095,0",
"154,3.4591,11.112,-4.2039,-5.0931,0",
"155,1.9358,8.1654,-0.023425,-2.2586,0",
"156,2.486,-0.99533,5.3404,-0.15475,0",
"157,2.4226,-4.5752,5.947,0.21507,0",
"158,3.9479,-3.7723,2.883,0.019813,0",
"159,2.2634,-4.4862,3.6558,-0.61251,0",
"160,1.3566,4.2358,2.1341,0.3211,0",
"161,5.0452,3.8964,-1.4304,0.86291,0",
"162,3.5499,8.6165,-3.2794,-1.2009,0",
"163,0.17346,7.8695,0.26876,-3.7883,0",
"164,2.4008,9.3593,-3.3565,-3.3526,0",
"165,4.8851,1.5995,-0.00029081,1.6401,0",
"166,4.1927,-3.2674,2.5839,0.21766,0",
"167,1.1166,8.6496,-0.96252,-1.8112,0",
"168,1.0235,6.901,-2.0062,-2.7125,0",
"169,-1.803,11.8818,2.0458,-5.2728,0",
"170,0.11739,6.2761,-1.5495,-2.4746,0",
"171,0.5706,-0.0248,1.2421,-0.5621,0",
"172,4.0552,-2.4583,2.2806,1.0323,0",
"173,-1.6952,1.0657,8.8294,0.94955,0",
"174,-1.1193,10.7271,2.0938,-5.6504,0",
"175,1.8799,2.4707,2.4931,0.37671,0",
"176,3.583,-3.7971,3.4391,-0.12501,0",
"177,0.19081,9.1297,-3.725,-5.8224,0",
"178,3.6582,5.6864,-1.7157,-0.23751,0",
"179,-0.13144,-1.7775,8.3316,0.35214,0",
"180,2.3925,9.798,-3.0361,-2.8224,0",
"181,1.6426,3.0149,0.22849,-0.147,0",
"182,-0.11783,-1.5789,8.03,-0.028031,0",
"183,-0.69572,8.6165,1.8419,-4.3289,0",
"184,2.9421,7.4101,-0.97709,-0.88406,0",
"185,-1.7559,11.9459,3.0946,-4.8978,0",
"186,-1.2537,10.8803,1.931,-4.3237,0",
"187,3.2585,-4.4614,3.8024,-0.15087,0",
"188,1.8314,6.3672,-0.036278,0.049554,0",
"189,4.5645,-3.6275,2.8684,0.27714,0",
"190,2.7365,-5.0325,6.6608,-0.57889,0",
"191,0.9297,-3.7971,4.6429,-0.2957,0",
"192,3.9663,10.1684,-4.1131,-4.6056,0",
"193,1.4578,-0.08485,4.1785,0.59136,0",
"194,4.8272,3.0687,0.68604,0.80731,0",
"195,-2.341,12.3784,0.70403,-7.5836,0",
"196,-1.8584,7.886,-1.6643,-1.8384,0",
"197,4.1454,7.257,-1.9153,-0.86078,0",
"198,1.9157,6.0816,0.23705,-2.0116,0",
"199,4.0215,-2.1914,2.4648,1.1409,0",
"200,5.8862,5.8747,-2.8167,-0.30087,0",
"201,-2.0897,10.8265,2.3603,-3.4198,0",
"202,4.0026,-3.5943,3.5573,0.26809,0",
"203,-0.78689,9.5663,-3.7867,-7.5034,0",
"204,4.1757,10.2615,-3.8552,-4.3056,0",
"205,0.83292,7.5404,0.65005,-0.92544,0",
"206,4.8077,2.2327,-0.26334,1.5534,0",
"207,5.3063,5.2684,-2.8904,-0.52716,0",
"208,2.5605,9.2683,-3.5913,-1.356,0",
"209,2.1059,7.6046,-0.47755,-1.8461,0",
"210,2.1721,-0.73874,5.4672,-0.72371,0",
"211,4.2899,9.1814,-4.6067,-4.3263,0",
"212,3.5156,10.1891,-4.2759,-4.978,0",
"213,2.614,8.0081,-3.7258,-1.3069,0",
"214,0.68087,2.3259,4.9085,0.54998,0",
"215,4.1962,0.74493,0.83256,0.753,0",
"216,6.0919,2.9673,-1.3267,1.4551,0",
"217,1.3234,3.2964,0.2362,-0.11984,0",
"218,1.3264,1.0326,5.6566,-0.41337,0",
"219,-0.16735,7.6274,1.2061,-3.6241,0",
"220,-1.3,10.2678,-2.953,-5.8638,0",
"221,-2.2261,12.5398,2.9438,-3.5258,0",
"222,2.4196,6.4665,-0.75688,0.228,0",
"223,1.0987,0.6394,5.989,-0.58277,0",
"224,4.6464,10.5326,-4.5852,-4.206,0",
"225,-0.36038,4.1158,3.1143,-0.37199,0",
"226,1.3562,3.2136,4.3465,0.78662,0",
"227,0.5706,-0.0248,1.2421,-0.5621,0",
"228,-2.6479,10.1374,-1.331,-5.4707,0",
"229,3.1219,-3.137,1.9259,-0.37458,0",
"230,5.4944,1.5478,0.041694,1.9284,0",
"231,-1.3389,1.552,7.0806,1.031,0",
"232,-2.3361,11.9604,3.0835,-5.4435,0",
"233,2.2596,-0.033118,4.7355,-0.2776,0",
"234,0.46901,-0.63321,7.3848,0.36507,0",
"235,2.7296,2.8701,0.51124,0.5099,0",
"236,2.0466,2.03,2.1761,-0.083634,0",
"237,-1.3274,9.498,2.4408,-5.2689,0",
"238,3.8905,-2.1521,2.6302,1.1047,0",
"239,3.9994,0.90427,1.1693,1.6892,0",
"240,2.3952,9.5083,-3.1783,-3.0086,0",
"241,3.2704,6.9321,-1.0456,0.23447,0",
"242,-1.3931,1.5664,7.5382,0.78403,0",
"243,1.6406,3.5488,1.3964,-0.36424,0",
"244,2.7744,6.8576,-1.0671,0.075416,0",
"245,2.4287,9.3821,-3.2477,-1.4543,0",
"246,4.2134,-2.806,2.0116,0.67412,0",
"247,1.6472,0.48213,4.7449,1.225,0",
"248,2.0597,-0.99326,5.2119,-0.29312,0",
"249,0.3798,0.7098,0.7572,-0.4444,0",
"250,1.0135,8.4551,-1.672,-2.0815,0",
"251,4.5691,-4.4552,3.1769,0.0042961,0",
"252,0.57461,10.1105,-1.6917,-4.3922,0",
"253,0.5734,9.1938,-0.9094,-1.872,0",
"254,5.2868,3.257,-1.3721,1.1668,0",
"255,4.0102,10.6568,-4.1388,-5.0646,0",
"256,4.1425,-3.6792,3.8281,1.6297,0",
"257,3.0934,-2.9177,2.2232,0.22283,0",
"258,2.2034,5.9947,0.53009,0.84998,0",
"259,3.744,0.79459,0.95851,1.0077,0",
"260,3.0329,2.2948,2.1135,0.35084,0",
"261,3.7731,7.2073,-1.6814,-0.94742,0",
"262,3.1557,2.8908,0.59693,0.79825,0",
"263,1.8114,7.6067,-0.9788,-2.4668,0",
"264,4.988,7.2052,-3.2846,-1.1608,0",
"265,2.483,6.6155,-0.79287,-0.90863,0",
"266,1.594,4.7055,1.3758,0.081882,0",
"267,-0.016103,9.7484,0.15394,-1.6134,0",
"268,3.8496,9.7939,-4.1508,-4.4582,0",
"269,0.9297,-3.7971,4.6429,-0.2957,0",
"270,4.9342,2.4107,-0.17594,1.6245,0",
"271,3.8417,10.0215,-4.2699,-4.9159,0",
"272,5.3915,9.9946,-3.8081,-3.3642,0",
"273,4.4072,-0.070365,2.0416,1.1319,0",
"274,2.6946,6.7976,-0.40301,0.44912,0",
"275,5.2756,0.13863,0.12138,1.1435,0",
"276,3.4312,6.2637,-1.9513,-0.36165,0",
"277,4.052,-0.16555,0.45383,0.51248,0",
"278,1.3638,-4.7759,8.4182,-1.8836,0",
"279,0.89566,7.7763,-2.7473,-1.9353,0",
"280,1.9265,7.7557,-0.16823,-3.0771,0",
"281,0.20977,-0.46146,7.7267,0.90946,0",
"282,4.068,-2.9363,2.1992,0.50084,0",
"283,2.877,-4.0599,3.6259,-0.32544,0",
"284,0.3223,-0.89808,8.0883,0.69222,0",
"285,-1.3,10.2678,-2.953,-5.8638,0",
"286,1.7747,-6.4334,8.15,-0.89828,0",
"287,1.3419,-4.4221,8.09,-1.7349,0",
"288,0.89606,10.5471,-1.4175,-4.0327,0",
"289,0.44125,2.9487,4.3225,0.7155,0",
"290,3.2422,6.2265,0.12224,-1.4466,0",
"291,2.5678,3.5136,0.61406,-0.40691,0",
"292,-2.2153,11.9625,0.078538,-7.7853,0",
"293,4.1349,6.1189,-2.4294,-0.19613,0",
"294,1.934,-9.2828e-06,4.816,-0.33967,0",
"295,2.5068,1.1588,3.9249,0.12585,0",
"296,2.1464,6.0795,-0.5778,-2.2302,0",
"297,0.051979,7.0521,-2.0541,-3.1508,0",
"298,1.2706,8.035,-0.19651,-2.1888,0",
"299,1.143,0.83391,5.4552,-0.56984,0",
"300,2.2928,9.0386,-3.2417,-1.2991,0",
"301,0.3292,-4.4552,4.5718,-0.9888,0",
"302,2.9719,6.8369,-0.2702,0.71291,0",
"303,1.6849,8.7489,-1.2641,-1.3858,0",
"304,-1.9177,11.6894,2.5454,-3.2763,0",
"305,2.3729,10.4726,-3.0087,-3.2013,0",
"306,1.0284,9.767,-1.3687,-1.7853,0",
"307,0.27451,9.2186,-3.2863,-4.8448,0",
"308,1.6032,-4.7863,8.5193,-2.1203,0",
"309,4.616,10.1788,-4.2185,-4.4245,0",
"310,4.2478,7.6956,-2.7696,-1.0767,0",
"311,4.0215,-2.7004,2.4957,0.36636,0",
"312,5.0297,-4.9704,3.5025,-0.23751,0",
"313,1.5902,2.2948,3.2403,0.18404,0",
"314,2.1274,5.1939,-1.7971,-1.1763,0",
"315,1.1811,8.3847,-2.0567,-0.90345,0",
"316,0.3292,-4.4552,4.5718,-0.9888,0",
"317,5.7353,5.2808,-2.2598,0.075416,0",
"318,2.6718,5.6574,0.72974,-1.4892,0",
"319,1.5799,-4.7076,7.9186,-1.5487,0",
"320,2.9499,2.2493,1.3458,-0.037083,0",
"321,0.5195,-3.2633,3.0895,-0.9849,0",
"322,3.7352,9.5911,-3.9032,-3.3487,0",
"323,-1.7344,2.0175,7.7618,0.93532,0",
"324,3.884,10.0277,-3.9298,-4.0819,0",
"325,3.5257,1.2829,1.9276,1.7991,0",
"326,4.4549,2.4976,1.0313,0.96894,0",
"327,-0.16108,-6.4624,8.3573,-1.5216,0",
"328,4.2164,9.4607,-4.9288,-5.2366,0",
"329,3.5152,6.8224,-0.67377,-0.46898,0",
"330,1.6988,2.9094,2.9044,0.11033,0",
"331,1.0607,2.4542,2.5188,-0.17027,0",
"332,2.0421,1.2436,4.2171,0.90429,0",
"333,3.5594,1.3078,1.291,1.6556,0",
"334,3.0009,5.8126,-2.2306,-0.66553,0",
"335,3.9294,1.4112,1.8076,0.89782,0",
"336,3.4667,-4.0724,4.2882,1.5418,0",
"337,3.966,3.9213,0.70574,0.33662,0",
"338,1.0191,2.33,4.9334,0.82929,0",
"339,0.96414,5.616,2.2138,-0.12501,0",
"340,1.8205,6.7562,0.0099913,0.39481,0",
"341,4.9923,7.8653,-2.3515,-0.71984,0",
"342,-1.1804,11.5093,0.15565,-6.8194,0",
"343,4.0329,0.23175,0.89082,1.1823,0",
"344,0.66018,10.3878,-1.4029,-3.9151,0",
"345,3.5982,7.1307,-1.3035,0.21248,0",
"346,-1.8584,7.886,-1.6643,-1.8384,0",
"347,4.0972,0.46972,1.6671,0.91593,0",
"348,3.3299,0.91254,1.5806,0.39352,0",
"349,3.1088,3.1122,0.80857,0.4336,0",
"350,-4.2859,8.5234,3.1392,-0.91639,0",
"351,-1.2528,10.2036,2.1787,-5.6038,0",
"352,0.5195,-3.2633,3.0895,-0.9849,0",
"353,0.3292,-4.4552,4.5718,-0.9888,0",
"354,0.88872,5.3449,2.045,-0.19355,0",
"355,3.5458,9.3718,-4.0351,-3.9564,0",
"356,-0.21661,8.0329,1.8848,-3.8853,0",
"357,2.7206,9.0821,-3.3111,-0.96811,0",
"358,3.2051,8.6889,-2.9033,-0.7819,0",
"359,2.6917,10.8161,-3.3,-4.2888,0",
"360,-2.3242,11.5176,1.8231,-5.375,0",
"361,2.7161,-4.2006,4.1914,0.16981,0",
"362,3.3848,3.2674,0.90967,0.25128,0",
"363,1.7452,4.8028,2.0878,0.62627,0",
"364,2.805,0.57732,1.3424,1.2133,0",
"365,5.7823,5.5788,-2.4089,-0.056479,0",
"366,3.8999,1.734,1.6011,0.96765,0",
"367,3.5189,6.332,-1.7791,-0.020273,0",
"368,3.2294,7.7391,-0.37816,-2.5405,0",
"369,3.4985,3.1639,0.22677,-0.1651,0",
"370,2.1948,1.3781,1.1582,0.85774,0",
"371,2.2526,9.9636,-3.1749,-2.9944,0",
"372,4.1529,-3.9358,2.8633,-0.017686,0",
"373,0.74307,11.17,-1.3824,-4.0728,0",
"374,1.9105,8.871,-2.3386,-0.75604,0",
"375,-1.5055,0.070346,6.8681,-0.50648,0",
"376,0.58836,10.7727,-1.3884,-4.3276,0",
"377,3.2303,7.8384,-3.5348,-1.2151,0",
"378,-1.9922,11.6542,2.6542,-5.2107,0",
"379,2.8523,9.0096,-3.761,-3.3371,0",
"380,4.2772,2.4955,0.48554,0.36119,0",
"381,1.5099,0.039307,6.2332,-0.30346,0",
"382,5.4188,10.1457,-4.084,-3.6991,0",
"383,0.86202,2.6963,4.2908,0.54739,0",
"384,3.8117,10.1457,-4.0463,-4.5629,0",
"385,0.54777,10.3754,-1.5435,-4.1633,0",
"386,2.3718,7.4908,0.015989,-1.7414,0",
"387,-2.4953,11.1472,1.9353,-3.4638,0",
"388,4.6361,-2.6611,2.8358,1.1991,0",
"389,-2.2527,11.5321,2.5899,-3.2737,0",
"390,3.7982,10.423,-4.1602,-4.9728,0",
"391,-0.36279,8.2895,-1.9213,-3.3332,0",
"392,2.1265,6.8783,0.44784,-2.2224,0",
"393,0.86736,5.5643,1.6765,-0.16769,0",
"394,3.7831,10.0526,-3.8869,-3.7366,0",
"395,-2.2623,12.1177,0.28846,-7.7581,0",
"396,1.2616,4.4303,-1.3335,-1.7517,0",
"397,2.6799,3.1349,0.34073,0.58489,0",
"398,-0.39816,5.9781,1.3912,-1.1621,0",
"399,4.3937,0.35798,2.0416,1.2004,0",
"400,2.9695,5.6222,0.27561,-1.1556,0",
"401,1.3049,-0.15521,6.4911,-0.75346,0",
"402,2.2123,-5.8395,7.7687,-0.85302,0",
"403,1.9647,6.9383,0.57722,0.66377,0",
"404,3.0864,-2.5845,2.2309,0.30947,0",
"405,0.3798,0.7098,0.7572,-0.4444,0",
"406,0.58982,7.4266,1.2353,-2.9595,0",
"407,0.14783,7.946,1.0742,-3.3409,0",
"408,-0.062025,6.1975,1.099,-1.131,0",
"409,4.223,1.1319,0.72202,0.96118,0",
"410,0.64295,7.1018,0.3493,-0.41337,0",
"411,1.941,0.46351,4.6472,1.0879,0",
"412,4.0047,0.45937,1.3621,1.6181,0",
"413,3.7767,9.7794,-3.9075,-3.5323,0",
"414,3.4769,-0.15314,2.53,2.4495,0",
"415,1.9818,9.2621,-3.521,-1.872,0",
"416,3.8023,-3.8696,4.044,0.95343,0",
"417,4.3483,11.1079,-4.0857,-4.2539,0",
"418,1.1518,1.3864,5.2727,-0.43536,0",
"419,-1.2576,1.5892,7.0078,0.42455,0",
"420,1.9572,-5.1153,8.6127,-1.4297,0",
"421,-2.484,12.1611,2.8204,-3.7418,0",
"422,-1.1497,1.2954,7.701,0.62627,0",
"423,4.8368,10.0132,-4.3239,-4.3276,0",
"424,-0.12196,8.8068,0.94566,-4.2267,0",
"425,1.9429,6.3961,0.092248,0.58102,0",
"426,1.742,-4.809,8.2142,-2.0659,0",
"427,-1.5222,10.8409,2.7827,-4.0974,0",
"428,-1.3,10.2678,-2.953,-5.8638,0",
"429,3.4246,-0.14693,0.80342,0.29136,0",
"430,2.5503,-4.9518,6.3729,-0.41596,0",
"431,1.5691,6.3465,-0.1828,-2.4099,0",
"432,1.3087,4.9228,2.0013,0.22024,0",
"433,5.1776,8.2316,-3.2511,-1.5694,0",
"434,2.229,9.6325,-3.1123,-2.7164,0",
"435,5.6272,10.0857,-4.2931,-3.8142,0",
"436,1.2138,8.7986,-2.1672,-0.74182,0",
"437,0.3798,0.7098,0.7572,-0.4444,0",
"438,0.5415,6.0319,1.6825,-0.46122,0",
"439,4.0524,5.6802,-1.9693,0.026279,0",
"440,4.7285,2.1065,-0.28305,1.5625,0",
"441,3.4359,0.66216,2.1041,1.8922,0",
"442,0.86816,10.2429,-1.4912,-4.0082,0",
"443,3.359,9.8022,-3.8209,-3.7133,0",
"444,3.6702,2.9942,0.85141,0.30688,0",
"445,1.3349,6.1189,0.46497,0.49826,0",
"446,3.1887,-3.4143,2.7742,-0.2026,0",
"447,2.4527,2.9653,0.20021,-0.056479,0",
"448,3.9121,2.9735,0.92852,0.60558,0",
"449,3.9364,10.5885,-3.725,-4.3133,0",
"450,3.9414,-3.2902,3.1674,1.0866,0",
"451,3.6922,-3.9585,4.3439,1.3517,0",
"452,5.681,7.795,-2.6848,-0.92544,0",
"453,0.77124,9.0862,-1.2281,-1.4996,0",
"454,3.5761,9.7753,-3.9795,-3.4638,0",
"455,1.602,6.1251,0.52924,0.47886,0",
"456,2.6682,10.216,-3.4414,-4.0069,0",
"457,2.0007,1.8644,2.6491,0.47369,0",
"458,0.64215,3.1287,4.2933,0.64696,0",
"459,4.3848,-3.0729,3.0423,1.2741,0",
"460,0.77445,9.0552,-2.4089,-1.3884,0",
"461,0.96574,8.393,-1.361,-1.4659,0",
"462,3.0948,8.7324,-2.9007,-0.96682,0",
"463,4.9362,7.6046,-2.3429,-0.85302,0",
"464,-1.9458,11.2217,1.9079,-3.4405,0",
"465,5.7403,-0.44284,0.38015,1.3763,0",
"466,-2.6989,12.1984,0.67661,-8.5482,0",
"467,1.1472,3.5985,1.9387,-0.43406,0",
"468,2.9742,8.96,-2.9024,-1.0379,0",
"469,4.5707,7.2094,-3.2794,-1.4944,0",
"470,0.1848,6.5079,2.0133,-0.87242,0",
"471,0.87256,9.2931,-0.7843,-2.1978,0",
"472,0.39559,6.8866,1.0588,-0.67587,0",
"473,3.8384,6.1851,-2.0439,-0.033204,0",
"474,2.8209,7.3108,-0.81857,-1.8784,0",
"475,2.5817,9.7546,-3.1749,-2.9957,0",
"476,3.8213,0.23175,2.0133,2.0564,0",
"477,0.3798,0.7098,0.7572,-0.4444,0",
"478,3.4893,6.69,-1.2042,-0.38751,0",
"479,-1.7781,0.8546,7.1303,0.027572,0",
"480,2.0962,2.4769,1.9379,-0.040962,0",
"481,0.94732,-0.57113,7.1903,-0.67587,0",
"482,2.8261,9.4007,-3.3034,-1.0509,0",
"483,0.0071249,8.3661,0.50781,-3.8155,0",
"484,0.96788,7.1907,1.2798,-2.4565,0",
"485,4.7432,2.1086,0.1368,1.6543,0",
"486,3.6575,7.2797,-2.2692,-1.144,0",
"487,3.8832,6.4023,-2.432,-0.98363,0",
"488,3.4776,8.811,-3.1886,-0.92285,0",
"489,1.1315,7.9212,1.093,-2.8444,0",
"490,2.8237,2.8597,0.19678,0.57196,0",
"491,1.9321,6.0423,0.26019,-2.053,0",
"492,3.0632,-3.3315,5.1305,0.8267,0",
"493,-1.8411,10.8306,2.769,-3.0901,0",
"494,2.8084,11.3045,-3.3394,-4.4194,0",
"495,2.5698,-4.4076,5.9856,0.078002,0",
"496,-0.12624,10.3216,-3.7121,-6.1185,0",
"497,3.3756,-4.0951,4.367,1.0698,0",
"498,-0.048008,-1.6037,8.4756,0.75558,0",
"499,0.5706,-0.0248,1.2421,-0.5621,0",
"500,0.88444,6.5906,0.55837,-0.44182,0",
"501,3.8644,3.7061,0.70403,0.35214,0",
"502,1.2999,2.5762,2.0107,-0.18967,0",
"503,2.0051,-6.8638,8.132,-0.2401,0",
"504,4.9294,0.27727,0.20792,0.33662,0",
"505,2.8297,6.3485,-0.73546,-0.58665,0",
"506,2.565,8.633,-2.9941,-1.3082,0",
"507,2.093,8.3061,0.022844,-3.2724,0",
"508,4.6014,5.6264,-2.1235,0.19309,0",
"509,5.0617,-0.35799,0.44698,0.99868,0",
"510,-0.2951,9.0489,-0.52725,-2.0789,0",
"511,3.577,2.4004,1.8908,0.73231,0",
"512,3.9433,2.5017,1.5215,0.903,0",
"513,2.6648,10.754,-3.3994,-4.1685,0",
"514,5.9374,6.1664,-2.5905,-0.36553,0",
"515,2.0153,1.8479,3.1375,0.42843,0",
"516,5.8782,5.9409,-2.8544,-0.60863,0",
"517,-2.3983,12.606,2.9464,-5.7888,0",
"518,1.762,4.3682,2.1384,0.75429,0",
"519,4.2406,-2.4852,1.608,0.7155,0",
"520,3.4669,6.87,-1.0568,-0.73147,0",
"521,3.1896,5.7526,-0.18537,-0.30087,0",
"522,0.81356,9.1566,-2.1492,-4.1814,0",
"523,0.52855,0.96427,4.0243,-1.0483,0",
"524,2.1319,-2.0403,2.5574,-0.061652,0",
"525,0.33111,4.5731,2.057,-0.18967,0",
"526,1.2746,8.8172,-1.5323,-1.7957,0",
"527,2.2091,7.4556,-1.3284,-3.3021,0",
"528,2.5328,7.528,-0.41929,-2.6478,0",
"529,3.6244,1.4609,1.3501,1.9284,0",
"530,-1.3885,12.5026,0.69118,-7.5487,0",
"531,5.7227,5.8312,-2.4097,-0.24527,0",
"532,3.3583,10.3567,-3.7301,-3.6991,0",
"533,2.5227,2.2369,2.7236,0.79438,0",
"534,0.045304,6.7334,1.0708,-0.9332,0",
"535,4.8278,7.7598,-2.4491,-1.2216,0",
"536,1.9476,-4.7738,8.527,-1.8668,0",
"537,2.7659,0.66216,4.1494,-0.28406,0",
"538,-0.10648,-0.76771,7.7575,0.64179,0",
"539,0.72252,-0.053811,5.6703,-1.3509,0",
"540,4.2475,1.4816,-0.48355,0.95343,0",
"541,3.9772,0.33521,2.2566,2.1625,0",
"542,3.6667,4.302,0.55923,0.33791,0",
"543,2.8232,10.8513,-3.1466,-3.9784,0",
"544,-1.4217,11.6542,-0.057699,-7.1025,0",
"545,4.2458,1.1981,0.66633,0.94696,0",
"546,4.1038,-4.8069,3.3491,-0.49225,0",
"547,1.4507,8.7903,-2.2324,-0.65259,0",
"548,3.4647,-3.9172,3.9746,0.36119,0",
"549,1.8533,6.1458,1.0176,-2.0401,0",
"550,3.5288,0.71596,1.9507,1.9375,0",
"551,3.9719,1.0367,0.75973,1.0013,0",
"552,3.534,9.3614,-3.6316,-1.2461,0",
"553,3.6894,9.887,-4.0788,-4.3664,0",
"554,3.0672,-4.4117,3.8238,-0.81682,0",
"555,2.6463,-4.8152,6.3549,0.003003,0",
"556,2.2893,3.733,0.6312,-0.39786,0",
"557,1.5673,7.9274,-0.056842,-2.1694,0",
"558,4.0405,0.51524,1.0279,1.106,0",
"559,4.3846,-4.8794,3.3662,-0.029324,0",
"560,2.0165,-0.25246,5.1707,1.0763,0",
"561,4.0446,11.1741,-4.3582,-4.7401,0",
"562,-0.33729,-0.64976,7.6659,0.72326,0",
"563,-2.4604,12.7302,0.91738,-7.6418,0",
"564,4.1195,10.9258,-3.8929,-4.1802,0",
"565,2.0193,0.82356,4.6369,1.4202,0",
"566,1.5701,7.9129,0.29018,-2.1953,0",
"567,2.6415,7.586,-0.28562,-1.6677,0",
"568,5.0214,8.0764,-3.0515,-1.7155,0",
"569,4.3435,3.3295,0.83598,0.64955,0",
"570,1.8238,-6.7748,8.3873,-0.54139,0",
"571,3.9382,0.9291,0.78543,0.6767,0",
"572,2.2517,-5.1422,4.2916,-1.2487,0",
"573,5.504,10.3671,-4.413,-4.0211,0",
"574,2.8521,9.171,-3.6461,-1.2047,0",
"575,1.1676,9.1566,-2.0867,-0.80647,0",
"576,2.6104,8.0081,-0.23592,-1.7608,0",
"577,0.32444,10.067,-1.1982,-4.1284,0",
"578,3.8962,-4.7904,3.3954,-0.53751,0",
"579,2.1752,-0.8091,5.1022,-0.67975,0",
"580,1.1588,8.9331,-2.0807,-1.1272,0",
"581,4.7072,8.2957,-2.5605,-1.4905,0",
"582,-1.9667,11.8052,-0.40472,-7.8719,0",
"583,4.0552,0.40143,1.4563,0.65343,0",
"584,2.3678,-6.839,8.4207,-0.44829,0",
"585,0.33565,6.8369,0.69718,-0.55691,0",
"586,4.3398,-5.3036,3.8803,-0.70432,0",
"587,1.5456,8.5482,0.4187,-2.1784,0",
"588,1.4276,8.3847,-2.0995,-1.9677,0",
"589,-0.27802,8.1881,-3.1338,-2.5276,0",
"590,0.93611,8.6413,-1.6351,-1.3043,0",
"591,4.6352,-3.0087,2.6773,1.212,0",
"592,1.5268,-5.5871,8.6564,-1.722,0",
"593,0.95626,2.4728,4.4578,0.21636,0",
"594,-2.7914,1.7734,6.7756,-0.39915,0",
"595,5.2032,3.5116,-1.2538,1.0129,0",
"596,3.1836,7.2321,-1.0713,-2.5909,0",
"597,0.65497,5.1815,1.0673,-0.42113,0",
"598,5.6084,10.3009,-4.8003,-4.3534,0",
"599,1.105,7.4432,0.41099,-3.0332,0",
"600,3.9292,-2.9156,2.2129,0.30817,0",
"601,1.1558,6.4003,1.5506,0.6961,0",
"602,2.5581,2.6218,1.8513,0.40257,0",
"603,2.7831,10.9796,-3.557,-4.4039,0",
"604,3.7635,2.7811,0.66119,0.34179,0",
"605,-2.6479,10.1374,-1.331,-5.4707,0",
"606,1.0652,8.3682,-1.4004,-1.6509,0",
"607,-1.4275,11.8797,0.41613,-6.9978,0",
"608,5.7456,10.1808,-4.7857,-4.3366,0",
"609,5.086,3.2798,-1.2701,1.1189,0",
"610,3.4092,5.4049,-2.5228,-0.89958,0",
"611,-0.2361,9.3221,2.1307,-4.3793,0",
"612,3.8197,8.9951,-4.383,-4.0327,0",
"613,-1.1391,1.8127,6.9144,0.70127,0",
"614,4.9249,0.68906,0.77344,1.2095,0",
"615,2.5089,6.841,-0.029423,0.44912,0",
"616,-0.2062,9.2207,-3.7044,-6.8103,0",
"617,3.946,6.8514,-1.5443,-0.5582,0",
"618,-0.278,8.1881,-3.1338,-2.5276,0",
"619,1.8592,3.2074,-0.15966,-0.26208,0",
"620,0.56953,7.6294,1.5754,-3.2233,0",
"621,3.4626,-4.449,3.5427,0.15429,0",
"622,3.3951,1.1484,2.1401,2.0862,0",
"623,5.0429,-0.52974,0.50439,1.106,0",
"624,3.7758,7.1783,-1.5195,0.40128,0",
"625,4.6562,7.6398,-2.4243,-1.2384,0",
"626,4.0948,-2.9674,2.3689,0.75429,0",
"627,1.8384,6.063,0.54723,0.51248,0",
"628,2.0153,0.43661,4.5864,-0.3151,0",
"629,3.5251,0.7201,1.6928,0.64438,0",
"630,3.757,-5.4236,3.8255,-1.2526,0",
"631,2.5989,3.5178,0.7623,0.81119,0",
"632,1.8994,0.97462,4.2265,0.81377,0",
"633,3.6941,-3.9482,4.2625,1.1577,0",
"634,4.4295,-2.3507,1.7048,0.90946,0",
"635,6.8248,5.2187,-2.5425,0.5461,0",
"636,1.8967,-2.5163,2.8093,-0.79742,0",
"637,2.1526,-6.1665,8.0831,-0.34355,0",
"638,3.3004,7.0811,-1.3258,0.22283,0",
"639,2.7213,7.05,-0.58808,0.41809,0",
"640,3.8846,-3.0336,2.5334,0.20214,0",
"641,4.1665,-0.4449,0.23448,0.27843,0",
"642,0.94225,5.8561,1.8762,-0.32544,0",
"643,5.1321,-0.031048,0.32616,1.1151,0",
"644,0.38251,6.8121,1.8128,-0.61251,0",
"645,3.0333,-2.5928,2.3183,0.303,0",
"646,2.9233,6.0464,-0.11168,-0.58665,0",
"647,1.162,10.2926,-1.2821,-4.0392,0",
"648,3.7791,2.5762,1.3098,0.5655,0",
"649,0.77765,5.9781,1.1941,-0.3526,0",
"650,-0.38388,-1.0471,8.0514,0.49567,0",
"651,0.21084,9.4359,-0.094543,-1.859,0",
"652,2.9571,-4.5938,5.9068,0.57196,0",
"653,4.6439,-3.3729,2.5976,0.55257,0",
"654,3.3577,-4.3062,6.0241,0.18274,0",
"655,3.5127,2.9073,1.0579,0.40774,0",
"656,2.6562,10.7044,-3.3085,-4.0767,0",
"657,-1.3612,10.694,1.7022,-2.9026,0",
"658,-0.278,8.1881,-3.1338,-2.5276,0",
"659,1.04,-6.9321,8.2888,-1.2991,0",
"660,2.1881,2.7356,1.3278,-0.1832,0",
"661,4.2756,-2.6528,2.1375,0.94437,0",
"662,-0.11996,6.8741,0.91995,-0.6694,0",
"663,2.9736,8.7944,-3.6359,-1.3754,0",
"664,3.7798,-3.3109,2.6491,0.066365,0",
"665,5.3586,3.7557,-1.7345,1.0789,0",
"666,1.8373,6.1292,0.84027,0.55257,0",
"667,1.2262,0.89599,5.7568,-0.11596,0",
"668,-0.048008,-0.56078,7.7215,0.453,0",
"669,0.5706,-0.024841,1.2421,-0.56208,0",
"670,4.3634,0.46351,1.4281,2.0202,0",
"671,3.482,-4.1634,3.5008,-0.078462,0",
"672,0.51947,-3.2633,3.0895,-0.98492,0",
"673,2.3164,-2.628,3.1529,-0.08622,0",
"674,-1.8348,11.0334,3.1863,-4.8888,0",
"675,1.3754,8.8793,-1.9136,-0.53751,0",
"676,-0.16682,5.8974,0.49839,-0.70044,0",
"677,0.29961,7.1328,-0.31475,-1.1828,0",
"678,0.25035,9.3262,-3.6873,-6.2543,0",
"679,2.4673,1.3926,1.7125,0.41421,0",
"680,0.77805,6.6424,-1.1425,-1.0573,0",
"681,3.4465,2.9508,1.0271,0.5461,0",
"682,2.2429,-4.1427,5.2333,-0.40173,0",
"683,3.7321,-3.884,3.3577,-0.0060486,0",
"684,4.3365,-3.584,3.6884,0.74912,0",
"685,-2.0759,10.8223,2.6439,-4.837,0",
"686,4.0715,7.6398,-2.0824,-1.1698,0",
"687,0.76163,5.8209,1.1959,-0.64613,0",
"688,-0.53966,7.3273,0.46583,-1.4543,0",
"689,2.6213,5.7919,0.065686,-1.5759,0",
"690,3.0242,-3.3378,2.5865,-0.54785,0",
"691,5.8519,5.3905,-2.4037,-0.061652,0",
"692,0.5706,-0.0248,1.2421,-0.5621,0",
"693,3.9771,11.1513,-3.9272,-4.3444,0",
"694,1.5478,9.1814,-1.6326,-1.7375,0",
"695,0.74054,0.36625,2.1992,0.48403,0",
"696,0.49571,10.2243,-1.097,-4.0159,0",
"697,1.645,7.8612,-0.87598,-3.5569,0",
"698,3.6077,6.8576,-1.1622,0.28231,0",
"699,3.2403,-3.7082,5.2804,0.41291,0",
"700,3.9166,10.2491,-4.0926,-4.4659,0",
"701,3.9262,6.0299,-2.0156,-0.065531,0",
"702,5.591,10.4643,-4.3839,-4.3379,0",
"703,3.7522,-3.6978,3.9943,1.3051,0",
"704,1.3114,4.5462,2.2935,0.22541,0",
"705,3.7022,6.9942,-1.8511,-0.12889,0",
"706,4.364,-3.1039,2.3757,0.78532,0",
"707,3.5829,1.4423,1.0219,1.4008,0",
"708,4.65,-4.8297,3.4553,-0.25174,0",
"709,5.1731,3.9606,-1.983,0.40774,0",
"710,3.2692,3.4184,0.20706,-0.066824,0",
"711,2.4012,1.6223,3.0312,0.71679,0",
"712,1.7257,-4.4697,8.2219,-1.8073,0",
"713,4.7965,6.9859,-1.9967,-0.35001,0",
"714,4.0962,10.1891,-3.9323,-4.1827,0",
"715,2.5559,3.3605,2.0321,0.26809,0",
"716,3.4916,8.5709,-3.0326,-0.59182,0",
"717,0.5195,-3.2633,3.0895,-0.9849,0",
"718,2.9856,7.2673,-0.409,-2.2431,0",
"719,4.0932,5.4132,-1.8219,0.23576,0",
"720,1.7748,-0.76978,5.5854,1.3039,0",
"721,5.2012,0.32694,0.17965,1.1797,0",
"722,-0.45062,-1.3678,7.0858,-0.40303,0",
"723,4.8451,8.1116,-2.9512,-1.4724,0",
"724,0.74841,7.2756,1.1504,-0.5388,0",
"725,5.1213,8.5565,-3.3917,-1.5474,0",
"726,3.6181,-3.7454,2.8273,-0.71208,0",
"727,0.040498,8.5234,1.4461,-3.9306,0",
"728,-2.6479,10.1374,-1.331,-5.4707,0",
"729,0.37984,0.70975,0.75716,-0.44441,0",
"730,-0.95923,0.091039,6.2204,-1.4828,0",
"731,2.8672,10.0008,-3.2049,-3.1095,0",
"732,1.0182,9.109,-0.62064,-1.7129,0",
"733,-2.7143,11.4535,2.1092,-3.9629,0",
"734,3.8244,-3.1081,2.4537,0.52024,0",
"735,2.7961,2.121,1.8385,0.38317,0",
"736,3.5358,6.7086,-0.81857,0.47886,0",
"737,-0.7056,8.7241,2.2215,-4.5965,0",
"738,4.1542,7.2756,-2.4766,-1.2099,0",
"739,0.92703,9.4318,-0.66263,-1.6728,0",
"740,1.8216,-6.4748,8.0514,-0.41855,0",
"741,-2.4473,12.6247,0.73573,-7.6612,0",
"742,3.5862,-3.0957,2.8093,0.24481,0",
"743,0.66191,9.6594,-0.28819,-1.6638,0",
"744,4.7926,1.7071,-0.051701,1.4926,0",
"745,4.9852,8.3516,-2.5425,-1.2823,0",
"746,0.75736,3.0294,2.9164,-0.068117,0",
"747,4.6499,7.6336,-1.9427,-0.37458,0",
"748,-0.023579,7.1742,0.78457,-0.75734,0",
"749,0.85574,0.0082678,6.6042,-0.53104,0",
"750,0.88298,0.66009,6.0096,-0.43277,0",
"751,4.0422,-4.391,4.7466,1.137,0",
"752,2.2546,8.0992,-0.24877,-3.2698,0",
"753,0.38478,6.5989,-0.3336,-0.56466,0",
"754,3.1541,-5.1711,6.5991,0.57455,0",
"755,2.3969,0.23589,4.8477,1.437,0",
"756,4.7114,2.0755,-0.2702,1.2379,0",
"757,4.0127,10.1477,-3.9366,-4.0728,0",
"758,2.6606,3.1681,1.9619,0.18662,0",
"759,3.931,1.8541,-0.023425,1.2314,0",
"760,0.01727,8.693,1.3989,-3.9668,0",
"761,3.2414,0.40971,1.4015,1.1952,0",
"762,2.2504,3.5757,0.35273,0.2836,0",
"763,-1.3971,3.3191,-1.3927,-1.9948,1",
"764,0.39012,-0.14279,-0.031994,0.35084,1",
"765,-1.6677,-7.1535,7.8929,0.96765,1",
"766,-3.8483,-12.8047,15.6824,-1.281,1",
"767,-3.5681,-8.213,10.083,0.96765,1",
"768,-2.2804,-0.30626,1.3347,1.3763,1",
"769,-1.7582,2.7397,-2.5323,-2.234,1",
"770,-0.89409,3.1991,-1.8219,-2.9452,1",
"771,0.3434,0.12415,-0.28733,0.14654,1",
"772,-0.9854,-6.661,5.8245,0.5461,1",
"773,-2.4115,-9.1359,9.3444,-0.65259,1",
"774,-1.5252,-6.2534,5.3524,0.59912,1",
"775,-0.61442,-0.091058,-0.31818,0.50214,1",
"776,-0.36506,2.8928,-3.6461,-3.0603,1",
"777,-5.9034,6.5679,0.67661,-6.6797,1",
"778,-1.8215,2.7521,-0.72261,-2.353,1",
"779,-0.77461,-1.8768,2.4023,1.1319,1",
"780,-1.8187,-9.0366,9.0162,-0.12243,1",
"781,-3.5801,-12.9309,13.1779,-2.5677,1",
"782,-1.8219,-6.8824,5.4681,0.057313,1",
"783,-0.3481,-0.38696,-0.47841,0.62627,1",
"784,0.47368,3.3605,-4.5064,-4.0431,1",
"785,-3.4083,4.8587,-0.76888,-4.8668,1",
"786,-1.6662,-0.30005,1.4238,0.024986,1",
"787,-2.0962,-7.1059,6.6188,-0.33708,1",
"788,-2.6685,-10.4519,9.1139,-1.7323,1",
"789,-0.47465,-4.3496,1.9901,0.7517,1",
"790,1.0552,1.1857,-2.6411,0.11033,1",
"791,1.1644,3.8095,-4.9408,-4.0909,1",
"792,-4.4779,7.3708,-0.31218,-6.7754,1",
"793,-2.7338,0.45523,2.4391,0.21766,1",
"794,-2.286,-5.4484,5.8039,0.88231,1",
"795,-1.6244,-6.3444,4.6575,0.16981,1",
"796,0.50813,0.47799,-1.9804,0.57714,1",
"797,1.6408,4.2503,-4.9023,-2.6621,1",
"798,0.81583,4.84,-5.2613,-6.0823,1",
"799,-5.4901,9.1048,-0.38758,-5.9763,1",
"800,-3.2238,2.7935,0.32274,-0.86078,1",
"801,-2.0631,-1.5147,1.219,0.44524,1",
"802,-0.91318,-2.0113,-0.19565,0.066365,1",
"803,0.6005,1.9327,-3.2888,-0.32415,1",
"804,0.91315,3.3377,-4.0557,-1.6741,1",
"805,-0.28015,3.0729,-3.3857,-2.9155,1",
"806,-3.6085,3.3253,-0.51954,-3.5737,1",
"807,-6.2003,8.6806,0.0091344,-3.703,1",
"808,-4.2932,3.3419,0.77258,-0.99785,1",
"809,-3.0265,-0.062088,0.68604,-0.055186,1",
"810,-1.7015,-0.010356,-0.99337,-0.53104,1",
"811,-0.64326,2.4748,-2.9452,-1.0276,1",
"812,-0.86339,1.9348,-2.3729,-1.0897,1",
"813,-2.0659,1.0512,-0.46298,-1.0974,1",
"814,-2.1333,1.5685,-0.084261,-1.7453,1",
"815,-1.2568,-1.4733,2.8718,0.44653,1",
"816,-3.1128,-6.841,10.7402,-1.0172,1",
"817,-4.8554,-5.9037,10.9818,-0.82199,1",
"818,-2.588,3.8654,-0.3336,-1.2797,1",
"819,0.24394,1.4733,-1.4192,-0.58535,1",
"820,-1.5322,-5.0966,6.6779,0.17498,1",
"821,-4.0025,-13.4979,17.6772,-3.3202,1",
"822,-4.0173,-8.3123,12.4547,-1.4375,1",
"823,-3.0731,-0.53181,2.3877,0.77627,1",
"824,-1.979,3.2301,-1.3575,-2.5819,1",
"825,-0.4294,-0.14693,0.044265,-0.15605,1",
"826,-2.234,-7.0314,7.4936,0.61334,1",
"827,-4.211,-12.4736,14.9704,-1.3884,1",
"828,-3.8073,-8.0971,10.1772,0.65084,1",
"829,-2.5912,-0.10554,1.2798,1.0414,1",
"830,-2.2482,3.0915,-2.3969,-2.6711,1",
"831,-1.4427,3.2922,-1.9702,-3.4392,1",
"832,-0.39416,-0.020702,-0.066267,-0.44699,1",
"833,-1.522,-6.6383,5.7491,-0.10691,1",
"834,-2.8267,-9.0407,9.0694,-0.98233,1",
"835,-1.7263,-6.0237,5.2419,0.29524,1",
"836,-0.94255,0.039307,-0.24192,0.31593,1",
"837,-0.89569,3.0025,-3.6067,-3.4457,1",
"838,-6.2815,6.6651,0.52581,-7.0107,1",
"839,-2.3211,3.166,-1.0002,-2.7151,1",
"840,-1.3414,-2.0776,2.8093,0.60688,1",
"841,-2.258,-9.3263,9.3727,-0.85949,1",
"842,-3.8858,-12.8461,12.7957,-3.1353,1",
"843,-1.8969,-6.7893,5.2761,-0.32544,1",
"844,-0.52645,-0.24832,-0.45613,0.41938,1",
"845,0.0096613,3.5612,-4.407,-4.4103,1",
"846,-3.8826,4.898,-0.92311,-5.0801,1",
"847,-2.1405,-0.16762,1.321,-0.20906,1",
"848,-2.4824,-7.3046,6.839,-0.59053,1",
"849,-2.9098,-10.0712,8.4156,-1.9948,1",
"850,-0.60975,-4.002,1.8471,0.6017,1",
"851,0.83625,1.1071,-2.4706,-0.062945,1",
"852,0.60731,3.9544,-4.772,-4.4853,1",
"853,-4.8861,7.0542,-0.17252,-6.959,1",
"854,-3.1366,0.42212,2.6225,-0.064238,1",
"855,-2.5754,-5.6574,6.103,0.65214,1",
"856,-1.8782,-6.5865,4.8486,-0.021566,1",
"857,0.24261,0.57318,-1.9402,0.44007,1",
"858,1.296,4.2855,-4.8457,-2.9013,1",
"859,0.25943,5.0097,-5.0394,-6.3862,1",
"860,-5.873,9.1752,-0.27448,-6.0422,1",
"861,-3.4605,2.6901,0.16165,-1.0224,1",
"862,-2.3797,-1.4402,1.1273,0.16076,1",
"863,-1.2424,-1.7175,-0.52553,-0.21036,1",
"864,0.20216,1.9182,-3.2828,-0.61768,1",
"865,0.59823,3.5012,-3.9795,-1.7841,1",
"866,-0.77995,3.2322,-3.282,-3.1004,1",
"867,-4.1409,3.4619,-0.47841,-3.8879,1",
"868,-6.5084,8.7696,0.23191,-3.937,1",
"869,-4.4996,3.4288,0.56265,-1.1672,1",
"870,-3.3125,0.10139,0.55323,-0.2957,1",
"871,-1.9423,0.3766,-1.2898,-0.82458,1",
"872,-0.75793,2.5349,-3.0464,-1.2629,1",
"873,-0.95403,1.9824,-2.3163,-1.1957,1",
"874,-2.2173,1.4671,-0.72689,-1.1724,1",
"875,-2.799,1.9679,-0.42357,-2.1125,1",
"876,-1.8629,-0.84841,2.5377,0.097399,1",
"877,-3.5916,-6.2285,10.2389,-1.1543,1",
"878,-5.1216,-5.3118,10.3846,-1.0612,1",
"879,-3.2854,4.0372,-0.45356,-1.8228,1",
"880,-0.56877,1.4174,-1.4252,-1.1246,1",
"881,-2.3518,-4.8359,6.6479,-0.060358,1",
"882,-4.4861,-13.2889,17.3087,-3.2194,1",
"883,-4.3876,-7.7267,11.9655,-1.4543,1",
"884,-3.3604,-0.32696,2.1324,0.6017,1",
"885,-1.0112,2.9984,-1.1664,-1.6185,1",
"886,0.030219,-1.0512,1.4024,0.77369,1",
"887,-1.6514,-8.4985,9.1122,1.2379,1",
"888,-3.2692,-12.7406,15.5573,-0.14182,1",
"889,-2.5701,-6.8452,8.9999,2.1353,1",
"890,-1.3066,0.25244,0.7623,1.7758,1",
"891,-1.6637,3.2881,-2.2701,-2.2224,1",
"892,-0.55008,2.8659,-1.6488,-2.4319,1",
"893,0.21431,-0.69529,0.87711,0.29653,1",
"894,-0.77288,-7.4473,6.492,0.36119,1",
"895,-1.8391,-9.0883,9.2416,-0.10432,1",
"896,-0.63298,-5.1277,4.5624,1.4797,1",
"897,0.0040545,0.62905,-0.64121,0.75817,1",
"898,-0.28696,3.1784,-3.5767,-3.1896,1",
"899,-5.2406,6.6258,-0.19908,-6.8607,1",
"900,-1.4446,2.1438,-0.47241,-1.6677,1",
"901,-0.65767,-2.8018,3.7115,0.99739,1",
"902,-1.5449,-10.1498,9.6152,-1.2332,1",
"903,-2.8957,-12.0205,11.9149,-2.7552,1",
"904,-0.81479,-5.7381,4.3919,0.3211,1",
"905,0.50225,0.65388,-1.1793,0.39998,1",
"906,0.74521,3.6357,-4.4044,-4.1414,1",
"907,-2.9146,4.0537,-0.45699,-4.0327,1",
"908,-1.3907,-1.3781,2.3055,-0.021566,1",
"909,-1.786,-8.1157,7.0858,-1.2112,1",
"910,-1.7322,-9.2828,7.719,-1.7168,1",
"911,0.55298,-3.4619,1.7048,1.1008,1",
"912,2.031,1.852,-3.0121,0.003003,1",
"913,1.2279,4.0309,-4.6435,-3.9125,1",
"914,-4.2249,6.2699,0.15822,-5.5457,1",
"915,-2.5346,-0.77392,3.3602,0.00171,1",
"916,-1.749,-6.332,6.0987,0.14266,1",
"917,-0.539,-5.167,3.4399,0.052141,1",
"918,1.5631,0.89599,-1.9702,0.65472,1",
"919,2.3917,4.5565,-4.9888,-2.8987,1",
"920,0.89512,4.7738,-4.8431,-5.5909,1",
"921,-5.4808,8.1819,0.27818,-5.0323,1",
"922,-2.8833,1.7713,0.68946,-0.4638,1",
"923,-1.4174,-2.2535,1.518,0.61981,1",
"924,0.4283,-0.94981,-1.0731,0.3211,1",
"925,1.5904,2.2121,-3.1183,-0.11725,1",
"926,1.7425,3.6833,-4.0129,-1.7207,1",
"927,-0.23356,3.2405,-3.0669,-2.7784,1",
"928,-3.6227,3.9958,-0.35845,-3.9047,1",
"929,-6.1536,7.9295,0.61663,-3.2646,1",
"930,-3.9172,2.6652,0.78886,-0.7819,1",
"931,-2.2214,-0.23798,0.56008,0.05602,1",
"932,-0.49241,0.89392,-1.6283,-0.56854,1",
"933,0.26517,2.4066,-2.8416,-0.59958,1",
"934,-0.10234,1.8189,-2.2169,-0.56725,1",
"935,-1.6176,1.0926,-0.35502,-0.59958,1",
"936,-1.8448,1.254,0.27218,-1.0728,1",
"937,-1.2786,-2.4087,4.5735,0.47627,1",
"938,-2.902,-7.6563,11.8318,-0.84268,1",
"939,-4.3773,-5.5167,10.939,-0.4082,1",
"940,-2.0529,3.8385,-0.79544,-1.2138,1",
"941,0.18868,0.70148,-0.51182,0.0055892,1",
"942,-1.7279,-6.841,8.9494,0.68058,1",
"943,-3.3793,-13.7731,17.9274,-2.0323,1",
"944,-3.1273,-7.1121,11.3897,-0.083634,1",
"945,-2.121,-0.05588,1.949,1.353,1",
"946,-1.7697,3.4329,-1.2144,-2.3789,1",
"947,-0.0012852,0.13863,-0.19651,0.0081754,1",
"948,-1.682,-6.8121,7.1398,1.3323,1",
"949,-3.4917,-12.1736,14.3689,-0.61639,1",
"950,-3.1158,-8.6289,10.4403,0.97153,1",
"951,-2.0891,-0.48422,1.704,1.7435,1",
"952,-1.6936,2.7852,-2.1835,-1.9276,1",
"953,-1.2846,3.2715,-1.7671,-3.2608,1",
"954,-0.092194,0.39315,-0.32846,-0.13794,1",
"955,-1.0292,-6.3879,5.5255,0.79955,1",
"956,-2.2083,-9.1069,8.9991,-0.28406,1",
"957,-1.0744,-6.3113,5.355,0.80472,1",
"958,-0.51003,-0.23591,0.020273,0.76334,1",
"959,-0.36372,3.0439,-3.4816,-2.7836,1",
"960,-6.3979,6.4479,1.0836,-6.6176,1",
"961,-2.2501,3.3129,-0.88369,-2.8974,1",
"962,-1.1859,-1.2519,2.2635,0.77239,1",
"963,-1.8076,-8.8131,8.7086,-0.21682,1",
"964,-3.3863,-12.9889,13.0545,-2.7202,1",
"965,-1.4106,-7.108,5.6454,0.31335,1",
"966,-0.21394,-0.68287,0.096532,1.1965,1",
"967,0.48797,3.5674,-4.3882,-3.8116,1",
"968,-3.8167,5.1401,-0.65063,-5.4306,1",
"969,-1.9555,0.20692,1.2473,-0.3707,1",
"970,-2.1786,-6.4479,6.0344,-0.20777,1",
"971,-2.3299,-9.9532,8.4756,-1.8733,1",
"972,0.0031201,-4.0061,1.7956,0.91722,1",
"973,1.3518,1.0595,-2.3437,0.39998,1",
"974,1.2309,3.8923,-4.8277,-4.0069,1",
"975,-5.0301,7.5032,-0.13396,-7.5034,1",
"976,-3.0799,0.60836,2.7039,-0.23751,1",
"977,-2.2987,-5.227,5.63,0.91722,1",
"978,-1.239,-6.541,4.8151,-0.033204,1",
"979,0.75896,0.29176,-1.6506,0.83834,1",
"980,1.6799,4.2068,-4.5398,-2.3931,1",
"981,0.63655,5.2022,-5.2159,-6.1211,1",
"982,-6.0598,9.2952,-0.43642,-6.3694,1",
"983,-3.518,2.8763,0.1548,-1.2086,1",
"984,-2.0336,-1.4092,1.1582,0.36507,1",
"985,-0.69745,-1.7672,-0.34474,-0.12372,1",
"986,0.75108,1.9161,-3.1098,-0.20518,1",
"987,0.84546,3.4826,-3.6307,-1.3961,1",
"988,-0.55648,3.2136,-3.3085,-2.7965,1",
"989,-3.6817,3.2239,-0.69347,-3.4004,1",
"990,-6.7526,8.8172,-0.061983,-3.725,1",
"991,-4.577,3.4515,0.66719,-0.94742,1",
"992,-2.9883,0.31245,0.45041,0.068951,1",
"993,-1.4781,0.14277,-1.1622,-0.48579,1",
"994,-0.46651,2.3383,-2.9812,-1.0431,1",
"995,-0.8734,1.6533,-2.1964,-0.78061,1",
"996,-2.1234,1.1815,-0.55552,-0.81165,1",
"997,-2.3142,2.0838,-0.46813,-1.6767,1",
"998,-1.4233,-0.98912,2.3586,0.39481,1",
"999,-3.0866,-6.6362,10.5405,-0.89182,1",
"1000,-4.7331,-6.1789,11.388,-1.0741,1",
"1001,-2.8829,3.8964,-0.1888,-1.1672,1",
"1002,-0.036127,1.525,-1.4089,-0.76121,1",
"1003,-1.7104,-4.778,6.2109,0.3974,1",
"1004,-3.8203,-13.0551,16.9583,-2.3052,1",
"1005,-3.7181,-8.5089,12.363,-0.95518,1",
"1006,-2.899,-0.60424,2.6045,1.3776,1",
"1007,-0.98193,2.7956,-1.2341,-1.5668,1",
"1008,-0.17296,-1.1816,1.3818,0.7336,1",
"1009,-1.9409,-8.6848,9.155,0.94049,1",
"1010,-3.5713,-12.4922,14.8881,-0.47027,1",
"1011,-2.9915,-6.6258,8.6521,1.8198,1",
"1012,-1.8483,0.31038,0.77344,1.4189,1",
"1013,-2.2677,3.2964,-2.2563,-2.4642,1",
"1014,-0.50816,2.868,-1.8108,-2.2612,1",
"1015,0.14329,-1.0885,1.0039,0.48791,1",
"1016,-0.90784,-7.9026,6.7807,0.34179,1",
"1017,-2.0042,-9.3676,9.3333,-0.10303,1",
"1018,-0.93587,-5.1008,4.5367,1.3866,1",
"1019,-0.40804,0.54214,-0.52725,0.6586,1",
"1020,-0.8172,3.3812,-3.6684,-3.456,1",
"1021,-4.8392,6.6755,-0.24278,-6.5775,1",
"1022,-1.2792,2.1376,-0.47584,-1.3974,1",
"1023,-0.66008,-3.226,3.8058,1.1836,1",
"1024,-1.7713,-10.7665,10.2184,-1.0043,1",
"1025,-3.0061,-12.2377,11.9552,-2.1603,1",
"1026,-1.1022,-5.8395,4.5641,0.68705,1",
"1027,0.11806,0.39108,-0.98223,0.42843,1",
"1028,0.11686,3.735,-4.4379,-4.3741,1",
"1029,-2.7264,3.9213,-0.49212,-3.6371,1",
"1030,-1.2369,-1.6906,2.518,0.51636,1",
"1031,-1.8439,-8.6475,7.6796,-0.66682,1",
"1032,-1.8554,-9.6035,7.7764,-0.97716,1",
"1033,0.16358,-3.3584,1.3749,1.3569,1",
"1034,1.5077,1.9596,-3.0584,-0.12243,1",
"1035,0.67886,4.1199,-4.569,-4.1414,1",
"1036,-3.9934,5.8333,0.54723,-4.9379,1",
"1037,-2.3898,-0.78427,3.0141,0.76205,1",
"1038,-1.7976,-6.7686,6.6753,0.89912,1",
"1039,-0.70867,-5.5602,4.0483,0.903,1",
"1040,1.0194,1.1029,-2.3,0.59395,1",
"1041,1.7875,4.78,-5.1362,-3.2362,1",
"1042,0.27331,4.8773,-4.9194,-5.8198,1",
"1043,-5.1661,8.0433,0.044265,-4.4983,1",
"1044,-2.7028,1.6327,0.83598,-0.091393,1",
"1045,-1.4904,-2.2183,1.6054,0.89394,1",
"1046,-0.014902,-1.0243,-0.94024,0.64955,1",
"1047,0.88992,2.2638,-3.1046,-0.11855,1",
"1048,1.0637,3.6957,-4.1594,-1.9379,1",
"1049,-0.8471,3.1329,-3.0112,-2.9388,1",
"1050,-3.9594,4.0289,-0.35845,-3.8957,1",
"1051,-5.8818,7.6584,0.5558,-2.9155,1",
"1052,-3.7747,2.5162,0.83341,-0.30993,1",
"1053,-2.4198,-0.24418,0.70146,0.41809,1",
"1054,-0.83535,0.80494,-1.6411,-0.19225,1",
"1055,-0.30432,2.6528,-2.7756,-0.65647,1",
"1056,-0.60254,1.7237,-2.1501,-0.77027,1",
"1057,-2.1059,1.1815,-0.53324,-0.82716,1",
"1058,-2.0441,1.2271,0.18564,-1.091,1",
"1059,-1.5621,-2.2121,4.2591,0.27972,1",
"1060,-3.2305,-7.2135,11.6433,-0.94613,1",
"1061,-4.8426,-4.9932,10.4052,-0.53104,1",
"1062,-2.3147,3.6668,-0.6969,-1.2474,1",
"1063,-0.11716,0.60422,-0.38587,-0.059065,1",
"1064,-2.0066,-6.719,9.0162,0.099985,1",
"1065,-3.6961,-13.6779,17.5795,-2.6181,1",
"1066,-3.6012,-6.5389,10.5234,-0.48967,1",
"1067,-2.6286,0.18002,1.7956,0.97282,1",
"1068,-0.82601,2.9611,-1.2864,-1.4647,1",
"1069,0.31803,-0.99326,1.0947,0.88619,1",
"1070,-1.4454,-8.4385,8.8483,0.96894,1",
"1071,-3.1423,-13.0365,15.6773,-0.66165,1",
"1072,-2.5373,-6.959,8.8054,1.5289,1",
"1073,-1.366,0.18416,0.90539,1.5806,1",
"1074,-1.7064,3.3088,-2.2829,-2.1978,1",
"1075,-0.41965,2.9094,-1.7859,-2.2069,1",
"1076,0.37637,-0.82358,0.78543,0.74524,1",
"1077,-0.55355,-7.9233,6.7156,0.74394,1",
"1078,-1.6001,-9.5828,9.4044,0.081882,1",
"1079,-0.37013,-5.554,4.7749,1.547,1",
"1080,0.12126,0.22347,-0.47327,0.97024,1",
"1081,-0.27068,3.2674,-3.5562,-3.0888,1",
"1082,-5.119,6.6486,-0.049987,-6.5206,1",
"1083,-1.3946,2.3134,-0.44499,-1.4905,1",
"1084,-0.69879,-3.3771,4.1211,1.5043,1",
"1085,-1.48,-10.5244,9.9176,-0.5026,1",
"1086,-2.6649,-12.813,12.6689,-1.9082,1",
"1087,-0.62684,-6.301,4.7843,1.106,1",
"1088,0.518,0.25865,-0.84085,0.96118,1",
"1089,0.64376,3.764,-4.4738,-4.0483,1",
"1090,-2.9821,4.1986,-0.5898,-3.9642,1",
"1091,-1.4628,-1.5706,2.4357,0.49826,1",
"1092,-1.7101,-8.7903,7.9735,-0.45475,1",
"1093,-1.5572,-9.8808,8.1088,-1.0806,1",
"1094,0.74428,-3.7723,1.6131,1.5754,1",
"1095,2.0177,1.7982,-2.9581,0.2099,1",
"1096,1.164,3.913,-4.5544,-3.8672,1",
"1097,-4.3667,6.0692,0.57208,-5.4668,1",
"1098,-2.5919,-1.0553,3.8949,0.77757,1",
"1099,-1.8046,-6.8141,6.7019,1.1681,1",
"1100,-0.71868,-5.7154,3.8298,1.0233,1",
"1101,1.4378,0.66837,-2.0267,1.0271,1",
"1102,2.1943,4.5503,-4.976,-2.7254,1",
"1103,0.7376,4.8525,-4.7986,-5.6659,1",
"1104,-5.637,8.1261,0.13081,-5.0142,1",
"1105,-3.0193,1.7775,0.73745,-0.45346,1",
"1106,-1.6706,-2.09,1.584,0.71162,1",
"1107,-0.1269,-1.1505,-0.95138,0.57843,1",
"1108,1.2198,2.0982,-3.1954,0.12843,1",
"1109,1.4501,3.6067,-4.0557,-1.5966,1",
"1110,-0.40857,3.0977,-2.9607,-2.6892,1",
"1111,-3.8952,3.8157,-0.31304,-3.8194,1",
"1112,-6.3679,8.0102,0.4247,-3.2207,1",
"1113,-4.1429,2.7749,0.68261,-0.71984,1",
"1114,-2.6864,-0.097265,0.61663,0.061192,1",
"1115,-1.0555,0.79459,-1.6968,-0.46768,1",
"1116,-0.29858,2.4769,-2.9512,-0.66165,1",
"1117,-0.49948,1.7734,-2.2469,-0.68104,1",
"1118,-1.9881,0.99945,-0.28562,-0.70044,1",
"1119,-1.9389,1.5706,0.045979,-1.122,1",
"1120,-1.4375,-1.8624,4.026,0.55127,1",
"1121,-3.1875,-7.5756,11.8678,-0.57889,1",
"1122,-4.6765,-5.6636,10.969,-0.33449,1",
"1123,-2.0285,3.8468,-0.63435,-1.175,1",
"1124,0.26637,0.73252,-0.67891,0.03533,1",
"1125,-1.7589,-6.4624,8.4773,0.31981,1",
"1126,-3.5985,-13.6593,17.6052,-2.4927,1",
"1127,-3.3582,-7.2404,11.4419,-0.57113,1",
"1128,-2.3629,-0.10554,1.9336,1.1358,1",
"1129,-2.1802,3.3791,-1.2256,-2.6621,1",
"1130,-0.40951,-0.15521,0.060545,-0.088807,1",
"1131,-2.2918,-7.257,7.9597,0.9211,1",
"1132,-4.0214,-12.8006,15.6199,-0.95647,1",
"1133,-3.3884,-8.215,10.3315,0.98187,1",
"1134,-2.0046,-0.49457,1.333,1.6543,1",
"1135,-1.7063,2.7956,-2.378,-2.3491,1",
"1136,-1.6386,3.3584,-1.7302,-3.5646,1",
"1137,-0.41645,0.32487,-0.33617,-0.36036,1",
"1138,-1.5877,-6.6072,5.8022,0.31593,1",
"1139,-2.5961,-9.349,9.7942,-0.28018,1",
"1140,-1.5228,-6.4789,5.7568,0.87325,1",
"1141,-0.53072,-0.097265,-0.21793,1.0426,1",
"1142,-0.49081,2.8452,-3.6436,-3.1004,1",
"1143,-6.5773,6.8017,0.85483,-7.5344,1",
"1144,-2.4621,2.7645,-0.62578,-2.8573,1",
"1145,-1.3995,-1.9162,2.5154,0.59912,1",
"1146,-2.3221,-9.3304,9.233,-0.79871,1",
"1147,-3.73,-12.9723,12.9817,-2.684,1",
"1148,-1.6988,-7.1163,5.7902,0.16723,1",
"1149,-0.26654,-0.64562,-0.42014,0.89136,1",
"1150,0.33325,3.3108,-4.5081,-4.012,1",
"1151,-4.2091,4.7283,-0.49126,-5.2159,1",
"1152,-2.3142,-0.68494,1.9833,-0.44829,1",
"1153,-2.4835,-7.4494,6.8964,-0.64484,1",
"1154,-2.7611,-10.5099,9.0239,-1.9547,1",
"1155,-0.36025,-4.449,2.1067,0.94308,1",
"1156,1.0117,0.9022,-2.3506,0.42714,1",
"1157,0.96708,3.8426,-4.9314,-4.1323,1",
"1158,-5.2049,7.259,0.070827,-7.3004,1",
"1159,-3.3203,-0.02691,2.9618,-0.44958,1",
"1160,-2.565,-5.7899,6.0122,0.046968,1",
"1161,-1.5951,-6.572,4.7689,-0.94354,1",
"1162,0.7049,0.17174,-1.7859,0.36119,1",
"1163,1.7331,3.9544,-4.7412,-2.5017,1",
"1164,0.6818,4.8504,-5.2133,-6.1043,1",
"1165,-6.3364,9.2848,0.014275,-6.7844,1",
"1166,-3.8053,2.4273,0.6809,-1.0871,1",
"1167,-2.1979,-2.1252,1.7151,0.45171,1",
"1168,-0.87874,-2.2121,-0.051701,0.099985,1",
"1169,0.74067,1.7299,-3.1963,-0.1457,1",
"1170,0.98296,3.4226,-3.9692,-1.7116,1",
"1171,-0.3489,3.1929,-3.4054,-3.1832,1",
"1172,-3.8552,3.5219,-0.38415,-3.8608,1",
"1173,-6.9599,8.9931,0.2182,-4.572,1",
"1174,-4.7462,3.1205,1.075,-1.2966,1",
"1175,-3.2051,-0.14279,0.97565,0.045675,1",
"1176,-1.7549,-0.080711,-0.75774,-0.3707,1",
"1177,-0.59587,2.4811,-2.8673,-0.89828,1",
"1178,-0.89542,2.0279,-2.3652,-1.2746,1",
"1179,-2.0754,1.2767,-0.64206,-1.2642,1",
"1180,-3.2778,1.8023,0.1805,-2.3931,1",
"1181,-2.2183,-1.254,2.9986,0.36378,1",
"1182,-3.5895,-6.572,10.5251,-0.16381,1",
"1183,-5.0477,-5.8023,11.244,-0.3901,1",
"1184,-3.5741,3.944,-0.07912,-2.1203,1",
"1185,-0.7351,1.7361,-1.4938,-1.1582,1",
"1186,-2.2617,-4.7428,6.3489,0.11162,1",
"1187,-4.244,-13.0634,17.1116,-2.8017,1",
"1188,-4.0218,-8.304,12.555,-1.5099,1",
"1189,-3.0201,-0.67253,2.7056,0.85774,1",
"1190,-2.4941,3.5447,-1.3721,-2.8483,1",
"1191,-0.83121,0.039307,0.05369,-0.23105,1",
"1192,-2.5665,-6.8824,7.5416,0.70774,1",
"1193,-4.4018,-12.9371,15.6559,-1.6806,1",
"1194,-3.7573,-8.2916,10.3032,0.38059,1",
"1195,-2.4725,-0.40145,1.4855,1.1189,1",
"1196,-1.9725,2.8825,-2.3086,-2.3724,1",
"1197,-2.0149,3.6874,-1.9385,-3.8918,1",
"1198,-0.82053,0.65181,-0.48869,-0.52716,1",
"1199,-1.7886,-6.3486,5.6154,0.42584,1",
"1200,-2.9138,-9.4711,9.7668,-0.60216,1",
"1201,-1.8343,-6.5907,5.6429,0.54998,1",
"1202,-0.8734,-0.033118,-0.20165,0.55774,1",
"1203,-0.70346,2.957,-3.5947,-3.1457,1",
"1204,-6.7387,6.9879,0.67833,-7.5887,1",
"1205,-2.7723,3.2777,-0.9351,-3.1457,1",
"1206,-1.6641,-1.3678,1.997,0.52283,1",
"1207,-2.4349,-9.2497,8.9922,-0.50001,1",
"1208,-3.793,-12.7095,12.7957,-2.825,1",
"1209,-1.9551,-6.9756,5.5383,-0.12889,1",
"1210,-0.69078,-0.50077,-0.35417,0.47498,1",
"1211,0.025013,3.3998,-4.4327,-4.2655,1",
"1212,-4.3967,4.9601,-0.64892,-5.4719,1",
"1213,-2.456,-0.24418,1.4041,-0.45863,1",
"1214,-2.62,-6.8555,6.2169,-0.62285,1",
"1215,-2.9662,-10.3257,8.784,-2.1138,1",
"1216,-0.71494,-4.4448,2.2241,0.49826,1",
"1217,0.6005,0.99945,-2.2126,0.097399,1",
"1218,0.61652,3.8944,-4.7275,-4.3948,1",
"1219,-5.4414,7.2363,0.10938,-7.5642,1",
"1220,-3.5798,0.45937,2.3457,-0.45734,1",
"1221,-2.7769,-5.6967,5.9179,0.37671,1",
"1222,-1.8356,-6.7562,5.0585,-0.55044,1",
"1223,0.30081,0.17381,-1.7542,0.48921,1",
"1224,1.3403,4.1323,-4.7018,-2.5987,1",
"1225,0.26877,4.987,-5.1508,-6.3913,1",
"1226,-6.5235,9.6014,-0.25392,-6.9642,1",
"1227,-4.0679,2.4955,0.79571,-1.1039,1",
"1228,-2.564,-1.7051,1.5026,0.32757,1",
"1229,-1.3414,-1.9162,-0.15538,-0.11984,1",
"1230,0.23874,2.0879,-3.3522,-0.66553,1",
"1231,0.6212,3.6771,-4.0771,-2.0711,1",
"1232,-0.77848,3.4019,-3.4859,-3.5569,1",
"1233,-4.1244,3.7909,-0.6532,-4.1802,1",
"1234,-7.0421,9.2,0.25933,-4.6832,1",
"1235,-4.9462,3.5716,0.82742,-1.4957,1",
"1236,-3.5359,0.30417,0.6569,-0.2957,1",
"1237,-2.0662,0.16967,-1.0054,-0.82975,1",
"1238,-0.88728,2.808,-3.1432,-1.2035,1",
"1239,-1.0941,2.3072,-2.5237,-1.4453,1",
"1240,-2.4458,1.6285,-0.88541,-1.4802,1",
"1241,-3.551,1.8955,0.1865,-2.4409,1",
"1242,-2.2811,-0.85669,2.7185,0.044382,1",
"1243,-3.6053,-5.974,10.0916,-0.82846,1",
"1244,-5.0676,-5.1877,10.4266,-0.86725,1",
"1245,-3.9204,4.0723,-0.23678,-2.1151,1",
"1246,-1.1306,1.8458,-1.3575,-1.3806,1",
"1247,-2.4561,-4.5566,6.4534,-0.056479,1",
"1248,-4.4775,-13.0303,17.0834,-3.0345,1",
"1249,-4.1958,-8.1819,12.1291,-1.6017,1",
"1250,-3.38,-0.7077,2.5325,0.71808,1",
"1251,-2.4365,3.6026,-1.4166,-2.8948,1",
"1252,-0.77688,0.13036,-0.031137,-0.35389,1",
"1253,-2.7083,-6.8266,7.5339,0.59007,1",
"1254,-4.5531,-12.5854,15.4417,-1.4983,1",
"1255,-3.8894,-7.8322,9.8208,0.47498,1",
"1256,-2.5084,-0.22763,1.488,1.2069,1",
"1257,-2.1652,3.0211,-2.4132,-2.4241,1",
"1258,-1.8974,3.5074,-1.7842,-3.8491,1",
"1259,-0.62043,0.5587,-0.38587,-0.66423,1",
"1260,-1.8387,-6.301,5.6506,0.19567,1",
"1261,-3,-9.1566,9.5766,-0.73018,1",
"1262,-1.9116,-6.1603,5.606,0.48533,1",
"1263,-1.005,0.084831,-0.2462,0.45688,1",
"1264,-0.87834,3.257,-3.6778,-3.2944,1",
"1265,-6.651,6.7934,0.68604,-7.5887,1",
"1266,-2.5463,3.1101,-0.83228,-3.0358,1",
"1267,-1.4377,-1.432,2.1144,0.42067,1",
"1268,-2.4554,-9.0407,8.862,-0.86983,1",
"1269,-3.9411,-12.8792,13.0597,-3.3125,1",
"1270,-2.1241,-6.8969,5.5992,-0.47156,1",
"1271,-0.74324,-0.32902,-0.42785,0.23317,1",
"1272,-0.071503,3.7412,-4.5415,-4.2526,1",
"1273,-4.2333,4.9166,-0.49212,-5.3207,1",
"1274,-2.3675,-0.43663,1.692,-0.43018,1",
"1275,-2.5526,-7.3625,6.9255,-0.66811,1",
"1276,-3.0986,-10.4602,8.9717,-2.3427,1",
"1277,-0.89809,-4.4862,2.2009,0.50731,1",
"1278,0.56232,1.0015,-2.2726,-0.0060486,1",
"1279,0.53936,3.8944,-4.8166,-4.3418,1",
"1280,-5.3012,7.3915,0.029699,-7.3987,1",
"1281,-3.3553,0.35591,2.6473,-0.37846,1",
"1282,-2.7908,-5.7133,5.953,0.45946,1",
"1283,-1.9983,-6.6072,4.8254,-0.41984,1",
"1284,0.15423,0.11794,-1.6823,0.59524,1",
"1285,1.208,4.0744,-4.7635,-2.6129,1",
"1286,0.2952,4.8856,-5.149,-6.2323,1",
"1287,-6.4247,9.5311,0.022844,-6.8517,1",
"1288,-3.9933,2.6218,0.62863,-1.1595,1",
"1289,-2.659,-1.6058,1.3647,0.16464,1",
"1290,-1.4094,-2.1252,-0.10397,-0.19225,1",
"1291,0.11032,1.9741,-3.3668,-0.65259,1",
"1292,0.52374,3.644,-4.0746,-1.9909,1",
"1293,-0.76794,3.4598,-3.4405,-3.4276,1",
"1294,-3.9698,3.6812,-0.60008,-4.0133,1",
"1295,-7.0364,9.2931,0.16594,-4.5396,1",
"1296,-4.9447,3.3005,1.063,-1.444,1",
"1297,-3.5933,0.22968,0.7126,-0.3332,1",
"1298,-2.1674,0.12415,-1.0465,-0.86208,1",
"1299,-0.9607,2.6963,-3.1226,-1.3121,1",
"1300,-1.0802,2.1996,-2.5862,-1.2759,1",
"1301,-2.3277,1.4381,-0.82114,-1.2862,1",
"1302,-3.7244,1.9037,-0.035421,-2.5095,1",
"1303,-2.5724,-0.95602,2.7073,-0.16639,1",
"1304,-3.9297,-6.0816,10.0958,-1.0147,1",
"1305,-5.2943,-5.1463,10.3332,-1.1181,1",
"1306,-3.8953,4.0392,-0.3019,-2.1836,1",
"1307,-1.2244,1.7485,-1.4801,-1.4181,1",
"1308,-2.6406,-4.4159,5.983,-0.13924,1",
"1309,-4.6338,-12.7509,16.7166,-3.2168,1",
"1310,-4.2887,-7.8633,11.8387,-1.8978,1",
"1311,-3.3458,-0.50491,2.6328,0.53705,1",
"1312,-1.1188,3.3357,-1.3455,-1.9573,1",
"1313,0.55939,-0.3104,0.18307,0.44653,1",
"1314,-1.5078,-7.3191,7.8981,1.2289,1",
"1315,-3.506,-12.5667,15.1606,-0.75216,1",
"1316,-2.9498,-8.273,10.2646,1.1629,1",
"1317,-1.6029,-0.38903,1.62,1.9103,1",
"1318,-1.2667,2.8183,-2.426,-1.8862,1",
"1319,-0.49281,3.0605,-1.8356,-2.834,1",
"1320,0.66365,-0.045533,-0.18794,0.23447,1",
"1321,-0.72068,-6.7583,5.8408,0.62369,1",
"1322,-1.9966,-9.5001,9.682,-0.12889,1",
"1323,-0.97325,-6.4168,5.6026,1.0323,1",
"1324,-0.025314,-0.17383,-0.11339,1.2198,1",
"1325,0.062525,2.9301,-3.5467,-2.6737,1",
"1326,-5.525,6.3258,0.89768,-6.6241,1",
"1327,-1.2943,2.6735,-0.84085,-2.0323,1",
"1328,-0.24037,-1.7837,2.135,1.2418,1",
"1329,-1.3968,-9.6698,9.4652,-0.34872,1",
"1330,-2.9672,-13.2869,13.4727,-2.6271,1",
"1331,-1.1005,-7.2508,6.0139,0.36895,1",
"1332,0.22432,-0.52147,-0.40386,1.2017,1",
"1333,0.90407,3.3708,-4.4987,-3.6965,1",
"1334,-2.8619,4.5193,-0.58123,-4.2629,1",
"1335,-1.0833,-0.31247,1.2815,0.41291,1",
"1336,-1.5681,-7.2446,6.5537,-0.1276,1",
"1337,-2.0545,-10.8679,9.4926,-1.4116,1",
"1338,0.2346,-4.5152,2.1195,1.4448,1",
"1339,1.581,0.86909,-2.3138,0.82412,1",
"1340,1.5514,3.8013,-4.9143,-3.7483,1",
"1341,-4.1479,7.1225,-0.083404,-6.4172,1",
"1342,-2.2625,-0.099335,2.8127,0.48662,1",
"1343,-1.7479,-5.823,5.8699,1.212,1",
"1344,-0.95923,-6.7128,4.9857,0.32886,1",
"1345,1.3451,0.23589,-1.8785,1.3258,1",
"1346,2.2279,4.0951,-4.8037,-2.1112,1",
"1347,1.2572,4.8731,-5.2861,-5.8741,1",
"1348,-5.3857,9.1214,-0.41929,-5.9181,1",
"1349,-2.9786,2.3445,0.52667,-0.40173,1",
"1350,-1.5851,-2.1562,1.7082,0.9017,1",
"1351,-0.21888,-2.2038,-0.0954,0.56421,1",
"1352,1.3183,1.9017,-3.3111,0.065071,1",
"1353,1.4896,3.4288,-4.0309,-1.4259,1",
"1354,0.11592,3.2219,-3.4302,-2.8457,1",
"1355,-3.3924,3.3564,-0.72004,-3.5233,1",
"1356,-6.1632,8.7096,-0.21621,-3.6345,1",
"1357,-4.0786,2.9239,0.87026,-0.65389,1",
"1358,-2.5899,-0.3911,0.93452,0.42972,1",
"1359,-1.0116,-0.19038,-0.90597,0.003003,1",
"1360,0.066129,2.4914,-2.9401,-0.62156,1",
"1361,-0.24745,1.9368,-2.4697,-0.80518,1",
"1362,-1.5732,1.0636,-0.71232,-0.8388,1",
"1363,-2.1668,1.5933,0.045122,-1.678,1",
"1364,-1.1667,-1.4237,2.9241,0.66119,1",
"1365,-2.8391,-6.63,10.4849,-0.42113,1",
"1366,-4.5046,-5.8126,10.8867,-0.52846,1",
"1367,-2.41,3.7433,-0.40215,-1.2953,1",
"1368,0.40614,1.3492,-1.4501,-0.55949,1",
"1369,-1.3887,-4.8773,6.4774,0.34179,1",
"1370,-3.7503,-13.4586,17.5932,-2.7771,1",
"1371,-3.5637,-8.3827,12.393,-1.2823,1",
"1372,-2.5419,-0.65804,2.6842,1.1952,1",
)

Wheat_Seeds_Dataset = (
"Id_int,Area_float,Perimeter_float,Compactness_float,Length of kernel_float,Width of kernel_float,Asymmetry coefficient_float,Length of kernel groove_float,Class_int",
"1,15.26,14.84,0.871,5.763,3.312,2.221,5.22,1",
"2,14.88,14.57,0.8811,5.554,3.333,1.018,4.956,1",
"3,14.29,14.09,0.905,5.291,3.337,2.699,4.825,1",
"4,13.84,13.94,0.8955,5.324,3.379,2.259,4.805,1",
"5,16.14,14.99,0.9034,5.658,3.562,1.355,5.175,1",
"6,14.38,14.21,0.8951,5.386,3.312,2.462,4.956,1",
"7,14.69,14.49,0.8799,5.563,3.259,3.586,5.219,1",
"8,14.11,14.1,0.8911,5.42,3.302,2.7,5,1",
"9,16.63,15.46,0.8747,6.053,3.465,2.04,5.877,1",
"10,16.44,15.25,0.888,5.884,3.505,1.969,5.533,1",
"11,15.26,14.85,0.8696,5.714,3.242,4.543,5.314,1",
"12,14.03,14.16,0.8796,5.438,3.201,1.717,5.001,1",
"13,13.89,14.02,0.888,5.439,3.199,3.986,4.738,1",
"14,13.78,14.06,0.8759,5.479,3.156,3.136,4.872,1",
"15,13.74,14.05,0.8744,5.482,3.114,2.932,4.825,1",
"16,14.59,14.28,0.8993,5.351,3.333,4.185,4.781,1",
"17,13.99,13.83,0.9183,5.119,3.383,5.234,4.781,1",
"18,15.69,14.75,0.9058,5.527,3.514,1.599,5.046,1",
"19,14.7,14.21,0.9153,5.205,3.466,1.767,4.649,1",
"20,12.72,13.57,0.8686,5.226,3.049,4.102,4.914,1",
"21,14.16,14.4,0.8584,5.658,3.129,3.072,5.176,1",
"22,14.11,14.26,0.8722,5.52,3.168,2.688,5.219,1",
"23,15.88,14.9,0.8988,5.618,3.507,0.7651,5.091,1",
"24,12.08,13.23,0.8664,5.099,2.936,1.415,4.961,1",
"25,15.01,14.76,0.8657,5.789,3.245,1.791,5.001,1",
"26,16.19,15.16,0.8849,5.833,3.421,0.903,5.307,1",
"27,13.02,13.76,0.8641,5.395,3.026,3.373,4.825,1",
"28,12.74,13.67,0.8564,5.395,2.956,2.504,4.869,1",
"29,14.11,14.18,0.882,5.541,3.221,2.754,5.038,1",
"30,13.45,14.02,0.8604,5.516,3.065,3.531,5.097,1",
"31,13.16,13.82,0.8662,5.454,2.975,0.8551,5.056,1",
"32,15.49,14.94,0.8724,5.757,3.371,3.412,5.228,1",
"33,14.09,14.41,0.8529,5.717,3.186,3.92,5.299,1",
"34,13.94,14.17,0.8728,5.585,3.15,2.124,5.012,1",
"35,15.05,14.68,0.8779,5.712,3.328,2.129,5.36,1",
"36,16.12,15,0.9,5.709,3.485,2.27,5.443,1",
"37,16.2,15.27,0.8734,5.826,3.464,2.823,5.527,1",
"38,17.08,15.38,0.9079,5.832,3.683,2.956,5.484,1",
"39,14.8,14.52,0.8823,5.656,3.288,3.112,5.309,1",
"40,14.28,14.17,0.8944,5.397,3.298,6.685,5.001,1",
"41,13.54,13.85,0.8871,5.348,3.156,2.587,5.178,1",
"42,13.5,13.85,0.8852,5.351,3.158,2.249,5.176,1",
"43,13.16,13.55,0.9009,5.138,3.201,2.461,4.783,1",
"44,15.5,14.86,0.882,5.877,3.396,4.711,5.528,1",
"45,15.11,14.54,0.8986,5.579,3.462,3.128,5.18,1",
"46,13.8,14.04,0.8794,5.376,3.155,1.56,4.961,1",
"47,15.36,14.76,0.8861,5.701,3.393,1.367,5.132,1",
"48,14.99,14.56,0.8883,5.57,3.377,2.958,5.175,1",
"49,14.79,14.52,0.8819,5.545,3.291,2.704,5.111,1",
"50,14.86,14.67,0.8676,5.678,3.258,2.129,5.351,1",
"51,14.43,14.4,0.8751,5.585,3.272,3.975,5.144,1",
"52,15.78,14.91,0.8923,5.674,3.434,5.593,5.136,1",
"53,14.49,14.61,0.8538,5.715,3.113,4.116,5.396,1",
"54,14.33,14.28,0.8831,5.504,3.199,3.328,5.224,1",
"55,14.52,14.6,0.8557,5.741,3.113,1.481,5.487,1",
"56,15.03,14.77,0.8658,5.702,3.212,1.933,5.439,1",
"57,14.46,14.35,0.8818,5.388,3.377,2.802,5.044,1",
"58,14.92,14.43,0.9006,5.384,3.412,1.142,5.088,1",
"59,15.38,14.77,0.8857,5.662,3.419,1.999,5.222,1",
"60,12.11,13.47,0.8392,5.159,3.032,1.502,4.519,1",
"61,11.42,12.86,0.8683,5.008,2.85,2.7,4.607,1",
"62,11.23,12.63,0.884,4.902,2.879,2.269,4.703,1",
"63,12.36,13.19,0.8923,5.076,3.042,3.22,4.605,1",
"64,13.22,13.84,0.868,5.395,3.07,4.157,5.088,1",
"65,12.78,13.57,0.8716,5.262,3.026,1.176,4.782,1",
"66,12.88,13.5,0.8879,5.139,3.119,2.352,4.607,1",
"67,14.34,14.37,0.8726,5.63,3.19,1.313,5.15,1",
"68,14.01,14.29,0.8625,5.609,3.158,2.217,5.132,1",
"69,14.37,14.39,0.8726,5.569,3.153,1.464,5.3,1",
"70,12.73,13.75,0.8458,5.412,2.882,3.533,5.067,1",
"71,17.63,15.98,0.8673,6.191,3.561,4.076,6.06,2",
"72,16.84,15.67,0.8623,5.998,3.484,4.675,5.877,2",
"73,17.26,15.73,0.8763,5.978,3.594,4.539,5.791,2",
"74,19.11,16.26,0.9081,6.154,3.93,2.936,6.079,2",
"75,16.82,15.51,0.8786,6.017,3.486,4.004,5.841,2",
"76,16.77,15.62,0.8638,5.927,3.438,4.92,5.795,2",
"77,17.32,15.91,0.8599,6.064,3.403,3.824,5.922,2",
"78,20.71,17.23,0.8763,6.579,3.814,4.451,6.451,2",
"79,18.94,16.49,0.875,6.445,3.639,5.064,6.362,2",
"80,17.12,15.55,0.8892,5.85,3.566,2.858,5.746,2",
"81,16.53,15.34,0.8823,5.875,3.467,5.532,5.88,2",
"82,18.72,16.19,0.8977,6.006,3.857,5.324,5.879,2",
"83,20.2,16.89,0.8894,6.285,3.864,5.173,6.187,2",
"84,19.57,16.74,0.8779,6.384,3.772,1.472,6.273,2",
"85,19.51,16.71,0.878,6.366,3.801,2.962,6.185,2",
"86,18.27,16.09,0.887,6.173,3.651,2.443,6.197,2",
"87,18.88,16.26,0.8969,6.084,3.764,1.649,6.109,2",
"88,18.98,16.66,0.859,6.549,3.67,3.691,6.498,2",
"89,21.18,17.21,0.8989,6.573,4.033,5.78,6.231,2",
"90,20.88,17.05,0.9031,6.45,4.032,5.016,6.321,2",
"91,20.1,16.99,0.8746,6.581,3.785,1.955,6.449,2",
"92,18.76,16.2,0.8984,6.172,3.796,3.12,6.053,2",
"93,18.81,16.29,0.8906,6.272,3.693,3.237,6.053,2",
"94,18.59,16.05,0.9066,6.037,3.86,6.001,5.877,2",
"95,18.36,16.52,0.8452,6.666,3.485,4.933,6.448,2",
"96,16.87,15.65,0.8648,6.139,3.463,3.696,5.967,2",
"97,19.31,16.59,0.8815,6.341,3.81,3.477,6.238,2",
"98,18.98,16.57,0.8687,6.449,3.552,2.144,6.453,2",
"99,18.17,16.26,0.8637,6.271,3.512,2.853,6.273,2",
"100,18.72,16.34,0.881,6.219,3.684,2.188,6.097,2",
"101,16.41,15.25,0.8866,5.718,3.525,4.217,5.618,2",
"102,17.99,15.86,0.8992,5.89,3.694,2.068,5.837,2",
"103,19.46,16.5,0.8985,6.113,3.892,4.308,6.009,2",
"104,19.18,16.63,0.8717,6.369,3.681,3.357,6.229,2",
"105,18.95,16.42,0.8829,6.248,3.755,3.368,6.148,2",
"106,18.83,16.29,0.8917,6.037,3.786,2.553,5.879,2",
"107,18.85,16.17,0.9056,6.152,3.806,2.843,6.2,2",
"108,17.63,15.86,0.88,6.033,3.573,3.747,5.929,2",
"109,19.94,16.92,0.8752,6.675,3.763,3.252,6.55,2",
"110,18.55,16.22,0.8865,6.153,3.674,1.738,5.894,2",
"111,18.45,16.12,0.8921,6.107,3.769,2.235,5.794,2",
"112,19.38,16.72,0.8716,6.303,3.791,3.678,5.965,2",
"113,19.13,16.31,0.9035,6.183,3.902,2.109,5.924,2",
"114,19.14,16.61,0.8722,6.259,3.737,6.682,6.053,2",
"115,20.97,17.25,0.8859,6.563,3.991,4.677,6.316,2",
"116,19.06,16.45,0.8854,6.416,3.719,2.248,6.163,2",
"117,18.96,16.2,0.9077,6.051,3.897,4.334,5.75,2",
"118,19.15,16.45,0.889,6.245,3.815,3.084,6.185,2",
"119,18.89,16.23,0.9008,6.227,3.769,3.639,5.966,2",
"120,20.03,16.9,0.8811,6.493,3.857,3.063,6.32,2",
"121,20.24,16.91,0.8897,6.315,3.962,5.901,6.188,2",
"122,18.14,16.12,0.8772,6.059,3.563,3.619,6.011,2",
"123,16.17,15.38,0.8588,5.762,3.387,4.286,5.703,2",
"124,18.43,15.97,0.9077,5.98,3.771,2.984,5.905,2",
"125,15.99,14.89,0.9064,5.363,3.582,3.336,5.144,2",
"126,18.75,16.18,0.8999,6.111,3.869,4.188,5.992,2",
"127,18.65,16.41,0.8698,6.285,3.594,4.391,6.102,2",
"128,17.98,15.85,0.8993,5.979,3.687,2.257,5.919,2",
"129,20.16,17.03,0.8735,6.513,3.773,1.91,6.185,2",
"130,17.55,15.66,0.8991,5.791,3.69,5.366,5.661,2",
"131,18.3,15.89,0.9108,5.979,3.755,2.837,5.962,2",
"132,18.94,16.32,0.8942,6.144,3.825,2.908,5.949,2",
"133,15.38,14.9,0.8706,5.884,3.268,4.462,5.795,2",
"134,16.16,15.33,0.8644,5.845,3.395,4.266,5.795,2",
"135,15.56,14.89,0.8823,5.776,3.408,4.972,5.847,2",
"136,15.38,14.66,0.899,5.477,3.465,3.6,5.439,2",
"137,17.36,15.76,0.8785,6.145,3.574,3.526,5.971,2",
"138,15.57,15.15,0.8527,5.92,3.231,2.64,5.879,2",
"139,15.6,15.11,0.858,5.832,3.286,2.725,5.752,2",
"140,16.23,15.18,0.885,5.872,3.472,3.769,5.922,2",
"141,13.07,13.92,0.848,5.472,2.994,5.304,5.395,3",
"142,13.32,13.94,0.8613,5.541,3.073,7.035,5.44,3",
"143,13.34,13.95,0.862,5.389,3.074,5.995,5.307,3",
"144,12.22,13.32,0.8652,5.224,2.967,5.469,5.221,3",
"145,11.82,13.4,0.8274,5.314,2.777,4.471,5.178,3",
"146,11.21,13.13,0.8167,5.279,2.687,6.169,5.275,3",
"147,11.43,13.13,0.8335,5.176,2.719,2.221,5.132,3",
"148,12.49,13.46,0.8658,5.267,2.967,4.421,5.002,3",
"149,12.7,13.71,0.8491,5.386,2.911,3.26,5.316,3",
"150,10.79,12.93,0.8107,5.317,2.648,5.462,5.194,3",
"151,11.83,13.23,0.8496,5.263,2.84,5.195,5.307,3",
"152,12.01,13.52,0.8249,5.405,2.776,6.992,5.27,3",
"153,12.26,13.6,0.8333,5.408,2.833,4.756,5.36,3",
"154,11.18,13.04,0.8266,5.22,2.693,3.332,5.001,3",
"155,11.36,13.05,0.8382,5.175,2.755,4.048,5.263,3",
"156,11.19,13.05,0.8253,5.25,2.675,5.813,5.219,3",
"157,11.34,12.87,0.8596,5.053,2.849,3.347,5.003,3",
"158,12.13,13.73,0.8081,5.394,2.745,4.825,5.22,3",
"159,11.75,13.52,0.8082,5.444,2.678,4.378,5.31,3",
"160,11.49,13.22,0.8263,5.304,2.695,5.388,5.31,3",
"161,12.54,13.67,0.8425,5.451,2.879,3.082,5.491,3",
"162,12.02,13.33,0.8503,5.35,2.81,4.271,5.308,3",
"163,12.05,13.41,0.8416,5.267,2.847,4.988,5.046,3",
"164,12.55,13.57,0.8558,5.333,2.968,4.419,5.176,3",
"165,11.14,12.79,0.8558,5.011,2.794,6.388,5.049,3",
"166,12.1,13.15,0.8793,5.105,2.941,2.201,5.056,3",
"167,12.44,13.59,0.8462,5.319,2.897,4.924,5.27,3",
"168,12.15,13.45,0.8443,5.417,2.837,3.638,5.338,3",
"169,11.35,13.12,0.8291,5.176,2.668,4.337,5.132,3",
"170,11.24,13,0.8359,5.09,2.715,3.521,5.088,3",
"171,11.02,13,0.8189,5.325,2.701,6.735,5.163,3",
"172,11.55,13.1,0.8455,5.167,2.845,6.715,4.956,3",
"173,11.27,12.97,0.8419,5.088,2.763,4.309,5,3",
"174,11.4,13.08,0.8375,5.136,2.763,5.588,5.089,3",
"175,10.83,12.96,0.8099,5.278,2.641,5.182,5.185,3",
"176,10.8,12.57,0.859,4.981,2.821,4.773,5.063,3",
"177,11.26,13.01,0.8355,5.186,2.71,5.335,5.092,3",
"178,10.74,12.73,0.8329,5.145,2.642,4.702,4.963,3",
"179,11.48,13.05,0.8473,5.18,2.758,5.876,5.002,3",
"180,12.21,13.47,0.8453,5.357,2.893,1.661,5.178,3",
"181,11.41,12.95,0.856,5.09,2.775,4.957,4.825,3",
"182,12.46,13.41,0.8706,5.236,3.017,4.987,5.147,3",
"183,12.19,13.36,0.8579,5.24,2.909,4.857,5.158,3",
"184,11.65,13.07,0.8575,5.108,2.85,5.209,5.135,3",
"185,12.89,13.77,0.8541,5.495,3.026,6.185,5.316,3",
"186,11.56,13.31,0.8198,5.363,2.683,4.062,5.182,3",
"187,11.81,13.45,0.8198,5.413,2.716,4.898,5.352,3",
"188,10.91,12.8,0.8372,5.088,2.675,4.179,4.956,3",
"189,11.23,12.82,0.8594,5.089,2.821,7.524,4.957,3",
"190,10.59,12.41,0.8648,4.899,2.787,4.975,4.794,3",
"191,10.93,12.8,0.839,5.046,2.717,5.398,5.045,3",
"192,11.27,12.86,0.8563,5.091,2.804,3.985,5.001,3",
"193,11.87,13.02,0.8795,5.132,2.953,3.597,5.132,3",
"194,10.82,12.83,0.8256,5.18,2.63,4.853,5.089,3",
"195,12.11,13.27,0.8639,5.236,2.975,4.132,5.012,3",
"196,12.8,13.47,0.886,5.16,3.126,4.873,4.914,3",
"197,12.79,13.53,0.8786,5.224,3.054,5.483,4.958,3",
"198,13.37,13.78,0.8849,5.32,3.128,4.67,5.091,3",
"199,12.62,13.67,0.8481,5.41,2.911,3.306,5.231,3",
"200,12.76,13.38,0.8964,5.073,3.155,2.828,4.83,3",
"201,12.38,13.44,0.8609,5.219,2.989,5.472,5.045,3",
"202,12.67,13.32,0.8977,4.984,3.135,2.3,4.745,3",
"203,11.18,12.72,0.868,5.009,2.81,4.051,4.828,3",
"204,12.7,13.41,0.8874,5.183,3.091,8.456,5,3",
"205,12.37,13.47,0.8567,5.204,2.96,3.919,5.001,3",
"206,12.19,13.2,0.8783,5.137,2.981,3.631,4.87,3",
"207,11.23,12.88,0.8511,5.14,2.795,4.325,5.003,3",
"208,13.2,13.66,0.8883,5.236,3.232,8.315,5.056,3",
"209,11.84,13.21,0.8521,5.175,2.836,3.598,5.044,3",
"210,12.3,13.34,0.8684,5.243,2.974,5.637,5.063,3",
)


class iris(tabular_dataloader):
    def __init__(self, name='iris_dataset', data_contents=Iris_Dataset, *args, **kwargs):
        super().__init__(name=name, data_contents=data_contents, *args, **kwargs)


class diabetes(tabular_dataloader):
    def __init__(self, name='pima_indians_diabetes_dataset', data_contents=Pima_indian_Diabetes_Dataset, *args, **kwargs):
        super().__init__(name=name, data_contents=data_contents, *args, **kwargs)


class banknote(tabular_dataloader):
    def __init__(self, name='banknote_dataset', data_contents=Banknote_Dataset, *args, **kwargs):
        super().__init__(name=name, data_contents=data_contents, *args, **kwargs)


class wheat(tabular_dataloader):
    def __init__(self, name='wheat_seeds_dataset', data_contents=Wheat_Seeds_Dataset, *args, **kwargs):
        super().__init__(name=name, data_contents=data_contents, *args, **kwargs)


if __name__ == '__main__':
    loader = wheat()
    data = loader.load(split_type='cross_validation')
    print(data['profile'])

    train_loader = data['train_loader']
    test_loader = data['test_loader']
    for fold in train_loader:
        for x, y in train_loader[fold]:
            print(fold, x, y)

    for fold in train_loader:
        for x, y in test_loader[fold]:
            print(fold, x, y)

